#include "MParser/CharStreamException.h"
#include "MParser/CharStreamIOException.h"
#include "MParser/NoViableAltForCharException.h"
#include "MParser/TokenStreamIOException.h"
#include "MParser/TokenStreamRecognitionException.h"
#include "MParser/CommonToken.h"
#include "MParser/StringUtils.h"
#include "MParser/ByteBuffer.h"
#include "MParser/CharBuffer.h"
#include "MGrammar/ANTLRLexer.h"

static int tohexnum (int c)

{
    c = toupper(c);
    switch (c)
     {
     case '0':
     case '1':
     case '2':
     case '3':
     case '4':
     case '5':
     case '6':
     case '7':
     case '8':
     case '9':
	  return (c - '0');
     case 'A':
     case 'B':
     case 'C':
     case 'D':
     case 'E':
     case 'F':
	  return (c - 'A' + 10);
     default:
	  return -1;
     }
}


/**Convert 'c' to an integer char value. */
int ANTLRLexer::escapeCharValue(const MLANG_String& cs) 

{
   //System.out.println("escapeCharValue("+cs+")");
   if ( cs.charAt(1)!='\\' ) return 0;
   switch ( cs.charAt(2) ) {
   case 'b' : return '\b';
   case 'r' : return '\r';
   case 't' : return '\t';
   case 'n' : return '\n';
   case 'f' : return '\f';
   case '"' : return '\"';
   case '\'' :return '\'';
   case '\\' :return '\\';

   case 'u' :
      // Unicode char
      if (cs.length() != 8) {
      	return 0;
      }
      else {
      	return
         tohexnum(cs.charAt(3)) * 16 * 16 * 16 +
         tohexnum(cs.charAt(4)) * 16 * 16 +
         tohexnum(cs.charAt(5)) * 16 +
         tohexnum(cs.charAt(6));
      }

   case '0' :
   case '1' :
   case '2' :
   case '3' :
      if ( cs.length()>5 && isdigit(cs.charAt(4)) ) {
      	return (cs.charAt(2)-'0')*8*8 + (cs.charAt(3)-'0')*8 + (cs.charAt(4)-'0');
      }
      if ( cs.length()>4 && isdigit(cs.charAt(3)) ) {
      	return (cs.charAt(2)-'0')*8 + (cs.charAt(3)-'0');
      }
      return cs.charAt(2)-'0';

   case '4' :
   case '5' :
   case '6' :
   case '7' :
      if ( cs.length()>4 && isdigit(cs.charAt(3)) ) {
      	return (cs.charAt(2)-'0')*8 + (cs.charAt(3)-'0');
      }
      return cs.charAt(2)-'0';

   default :
      return 0;
   }
}

int ANTLRLexer::tokenTypeForCharLiteral(const MLANG_String& lit)

{
   if ( lit.length()>3 ) {  // does char contain escape?
      return escapeCharValue(lit);
    }
   else
    {
      return lit.charAt(1);
    }
}

ANTLRLexer::ANTLRLexer (const MIO_InputStream_var& in)
  : super(new MParser_ByteBuffer(in))
{
    caseSensitiveLiterals_ = true;
    this->init();
}

ANTLRLexer::ANTLRLexer(const MIO_Reader_var& in)
  : super(new MParser_CharBuffer(in))
{
    caseSensitiveLiterals_ = true;
    this->init();
}

ANTLRLexer::ANTLRLexer (const MParser_InputBuffer_var& ib)
  : super(new MParser_LexerSharedInputState(ib))
{
    caseSensitiveLiterals_ = true;
    this->init();
}

void ANTLRLexer::init()

{
   setCaseSensitive(true);
   literals_.put("public", 31);
   literals_.put("class", 10);
   literals_.put("header", 5);
   literals_.put("throws", 37);
   literals_.put("lexclass", 9);
   literals_.put("catch", 40);
   literals_.put("private", 32);
   literals_.put("options", 51);
   literals_.put("extends", 11);
   literals_.put("protected", 30);
   literals_.put("TreeParser", 13);
   literals_.put("Parser", 29);
   literals_.put("Lexer", 12);
   literals_.put("returns", 35);
   literals_.put("charVocabulary", 18);
   literals_.put("tokens", 4);
   literals_.put("exception", 39);
}

ANTLRLexer::ANTLRLexer(const MParser_LexerSharedInputState_var& state)
  : super(state)

{
    caseSensitiveLiterals_ = true;
    this->init();
}

MParser_Token_var ANTLRLexer::nextToken()

{
   MParser_Token_var theRetToken;
tryAgain:
	for (;;) {
   MParser_Token_var _token;
   int _ttype = MParser_Token::INVALID_TYPE;
   resetText();
   try {   // for char stream error handling
      try {   // for lexical error handling
      	switch ( LA(1)) {
      	case '\t':  case '\n':  case '\r':  case ' ':
      	{
         mWS(true);
         theRetToken=returnToken_;
         break;
      }
      	case '/':
      	{
         mCOMMENT(true);
         theRetToken=returnToken_;
         break;
      }
      	case '<':
      	{
         mOPEN_ELEMENT_OPTION(true);
         theRetToken=returnToken_;
         break;
      }
      	case '>':
      	{
         mCLOSE_ELEMENT_OPTION(true);
         theRetToken=returnToken_;
         break;
      }
      	case ',':
      	{
         mCOMMA(true);
         theRetToken=returnToken_;
         break;
      }
      	case '?':
      	{
         mQUESTION(true);
         theRetToken=returnToken_;
         break;
      }
      	case '#':
      	{
         mTREE_BEGIN(true);
         theRetToken=returnToken_;
         break;
      }
      	case '(':
      	{
         mLPAREN(true);
         theRetToken=returnToken_;
         break;
      }
      	case ')':
      	{
         mRPAREN(true);
         theRetToken=returnToken_;
         break;
      }
      	case ':':
      	{
         mCOLON(true);
         theRetToken=returnToken_;
         break;
      }
      	case '*':
      	{
         mSTAR(true);
         theRetToken=returnToken_;
         break;
      }
      	case '+':
      	{
         mPLUS(true);
         theRetToken=returnToken_;
         break;
      }
      	case ';':
      	{
         mSEMI(true);
         theRetToken=returnToken_;
         break;
      }
      	case '^':
      	{
         mCARET(true);
         theRetToken=returnToken_;
         break;
      }
      	case '!':
      	{
         mBANG(true);
         theRetToken=returnToken_;
         break;
      }
      	case '|':
      	{
         mOR(true);
         theRetToken=returnToken_;
         break;
      }
      	case '~':
      	{
         mNOT_OP(true);
         theRetToken=returnToken_;
         break;
      }
      	case '}':
      	{
         mRCURLY(true);
         theRetToken=returnToken_;
         break;
      }
      	case '\'':
      	{
         mCHAR_LITERAL(true);
         theRetToken=returnToken_;
         break;
      }
      	case '"':
      	{
         mSTRING_LITERAL(true);
         theRetToken=returnToken_;
         break;
      }
      	case '0':  case '1':  case '2':  case '3':
      	case '4':  case '5':  case '6':  case '7':
      	case '8':  case '9':
      	{
         mINT(true);
         theRetToken=returnToken_;
         break;
      }
      	case '[':
      	{
         mARG_ACTION(true);
         theRetToken=returnToken_;
         break;
      }
      	case '{':
      	{
         mACTION(true);
         theRetToken=returnToken_;
         break;
      }
      	case 'A':  case 'B':  case 'C':  case 'D':
      	case 'E':  case 'F':  case 'G':  case 'H':
      	case 'I':  case 'J':  case 'K':  case 'L':
      	case 'M':  case 'N':  case 'O':  case 'P':
      	case 'Q':  case 'R':  case 'S':  case 'T':
      	case 'U':  case 'V':  case 'W':  case 'X':
      	case 'Y':  case 'Z':
      	{
         mTOKEN_REF(true);
         theRetToken=returnToken_;
         break;
      }
      	case 'a':  case 'b':  case 'c':  case 'd':
      	case 'e':  case 'f':  case 'g':  case 'h':
      	case 'i':  case 'j':  case 'k':  case 'l':
      	case 'm':  case 'n':  case 'o':  case 'p':
      	case 'q':  case 'r':  case 's':  case 't':
      	case 'u':  case 'v':  case 'w':  case 'x':
      	case 'y':  case 'z':
      	{
         mRULE_REF(true);
         theRetToken=returnToken_;
         break;
      }
      	default:
         if ((LA(1)=='=') && (LA(2)=='>')) {
            mIMPLIES(true);
            theRetToken=returnToken_;
         }
         else if ((LA(1)=='.') && (LA(2)=='.')) {
            mRANGE(true);
            theRetToken=returnToken_;
         }
         else if ((LA(1)=='=') && (true)) {
            mASSIGN(true);
            theRetToken=returnToken_;
         }
         else if ((LA(1)=='.') && (true)) {
            mWILDCARD(true);
            theRetToken=returnToken_;
         }
      	else {
         if (LA(1)==EOF_CHAR) {uponEOF(); returnToken_ = makeToken(MParser_Token::EOF_TYPE);}
      	else {throw MParser_NoViableAltForCharException((char)LA(1), getFilename(), getLine(), getColumn());}
      }
      }
      	if ( MParser_Token::is_nil(returnToken_) ) goto tryAgain; // found SKIP token
      	_ttype = returnToken_->getType();
      	returnToken_->setType(_ttype);
      	return returnToken_;
      }
      catch (MParser_RecognitionException& e) {
      	throw MParser_TokenStreamRecognitionException(e);
      }
   }
   catch (MParser_CharStreamException& cse) {
      if ( ACE_dynamic_cast (MParser_CharStreamIOException*, &cse)) {
      	throw MParser_TokenStreamIOException(((MParser_CharStreamIOException&)cse).io);
      }
      else {
      	throw MParser_TokenStreamException(cse.getMessage().c_str());
      }
   }
}
}

void ANTLRLexer::mWS(bool _createToken)

{
   int _ttype; MParser_Token_var _token; int _begin=text_.length();
   _ttype = WS;
   int _saveIndex;
   
   {
   switch ( LA(1)) {
   case ' ':
   {
      match(' ');
      break;
   }
   case '\t':
   {
      match('\t');
      break;
   }
   case '\n':
   {
      match('\n');
      newline();
      break;
   }
   default:
      if ((LA(1)=='\r') && (LA(2)=='\n')) {
      	match('\r');
      	match('\n');
      	newline();
      }
      else if ((LA(1)=='\r') && (true)) {
      	match('\r');
      	newline();
      }
   else {
      throw MParser_NoViableAltForCharException((char)LA(1), getFilename(), getLine(), getColumn());
   }
   }
   }
   _ttype = MParser_Token::SKIP;
   if ( _createToken && MParser_Token::is_nil(_token) && _ttype!=MParser_Token::SKIP ) {
      _token = makeToken(_ttype);
      _token->setText(MLANG_String(text_.getBuffer() + _begin, text_.length()-_begin));
   }
   returnToken_ = _token;
}
	
void ANTLRLexer::mCOMMENT(bool _createToken)

{
   int _ttype; MParser_Token_var _token; int _begin=text_.length();
   _ttype = COMMENT;
   int _saveIndex;
   MParser_Token_var t;
   
   {
   if ((LA(1)=='/') && (LA(2)=='/')) {
      mSL_COMMENT(false);
   }
   else if ((LA(1)=='/') && (LA(2)=='*')) {
      mML_COMMENT(true);
      t=returnToken_;
      _ttype = t->getType();
   }
   else {
      throw MParser_NoViableAltForCharException((char)LA(1), getFilename(), getLine(), getColumn());
   }
   
   }
   if ( _ttype != DOC_COMMENT ) _ttype = MParser_Token::SKIP;
   if ( _createToken && MParser_Token::is_nil(_token) && _ttype!=MParser_Token::SKIP ) {
      _token = makeToken(_ttype);
      _token->setText(MLANG_String(text_.getBuffer() + _begin, text_.length()-_begin));
   }
   returnToken_ = _token;
}
	
void ANTLRLexer::mSL_COMMENT(bool _createToken)

{
   int _ttype; MParser_Token_var _token; int _begin=text_.length();
   _ttype = SL_COMMENT;
   int _saveIndex;
   
   match("//");
   {
   LABEL(_loop153)
   do {
      if ((_tokenSet_0.member(LA(1)))) {
      	{
      	match(_tokenSet_0);
      }
      }
      else {
      	BREAK(_loop153);
      }
      
   } while (true);
   }
   {
   if ((LA(1)=='\r') && (LA(2)=='\n')) {
      match('\r');
      match('\n');
   }
   else if ((LA(1)=='\r') && (true)) {
      match('\r');
   }
   else if ((LA(1)=='\n')) {
      match('\n');
   }
   else {
      throw MParser_NoViableAltForCharException((char)LA(1), getFilename(), getLine(), getColumn());
   }
   
   }
   newline();
   if ( _createToken && MParser_Token::is_nil(_token) && _ttype!=MParser_Token::SKIP ) {
      _token = makeToken(_ttype);
      _token->setText(MLANG_String(text_.getBuffer() + _begin, text_.length()-_begin));
   }
   returnToken_ = _token;
}
	
void ANTLRLexer::mML_COMMENT(bool _createToken)

{
   int _ttype; MParser_Token_var _token; int _begin=text_.length();
   _ttype = ML_COMMENT;
   int _saveIndex;
   
   match("/*");
   {
   if (((LA(1)=='*') && ((LA(2) >= char(3) && LA(2) <= int(0xff)))&&( LA(2)!='/' ))) {
      match('*');
      _ttype = DOC_COMMENT;
   }
   else if (((LA(1) >= char(3) && LA(1) <= int(0xff)) && ((LA(2) >= char(3) && LA(2) <= int(0xff))))) {
   }
   else {
      throw MParser_NoViableAltForCharException((char)LA(1), getFilename(), getLine(), getColumn());
   }
   
   }
   {
   LABEL(_loop159)
   do {
      // nongreedy exit test
      if ((LA(1)=='*') && (LA(2)=='/')) BREAK(_loop159);
      if ((LA(1)=='\r') && (LA(2)=='\n')) {
      	match('\r');
      	match('\n');
      	newline();
      }
      else if ((LA(1)=='\r') && ((LA(2) >= char(3) && LA(2) <= int(0xff)))) {
      	match('\r');
      	newline();
      }
      else if ((_tokenSet_0.member(LA(1))) && ((LA(2) >= char(3) && LA(2) <= int(0xff)))) {
      	{
      	match(_tokenSet_0);
      }
      }
      else if ((LA(1)=='\n')) {
      	match('\n');
      	newline();
      }
      else {
      	BREAK(_loop159);
      }
      
   } while (true);
   }
   match("*/");
   if ( _createToken && MParser_Token::is_nil(_token) && _ttype!=MParser_Token::SKIP ) {
      _token = makeToken(_ttype);
      _token->setText(MLANG_String(text_.getBuffer() + _begin, text_.length()-_begin));
   }
   returnToken_ = _token;
}
	
void ANTLRLexer::mOPEN_ELEMENT_OPTION(bool _createToken)

{
   int _ttype; MParser_Token_var _token; int _begin=text_.length();
   _ttype = OPEN_ELEMENT_OPTION;
   int _saveIndex;
   
   match('<');
   if ( _createToken && MParser_Token::is_nil(_token) && _ttype!=MParser_Token::SKIP ) {
      _token = makeToken(_ttype);
      _token->setText(MLANG_String(text_.getBuffer() + _begin, text_.length()-_begin));
   }
   returnToken_ = _token;
}
	
void ANTLRLexer::mCLOSE_ELEMENT_OPTION(bool _createToken)

{
   int _ttype; MParser_Token_var _token; int _begin=text_.length();
   _ttype = CLOSE_ELEMENT_OPTION;
   int _saveIndex;
   
   match('>');
   if ( _createToken && MParser_Token::is_nil(_token) && _ttype!=MParser_Token::SKIP ) {
      _token = makeToken(_ttype);
      _token->setText(MLANG_String(text_.getBuffer() + _begin, text_.length()-_begin));
   }
   returnToken_ = _token;
}
	
void ANTLRLexer::mCOMMA(bool _createToken)

{
   int _ttype; MParser_Token_var _token; int _begin=text_.length();
   _ttype = COMMA;
   int _saveIndex;
   
   match(',');
   if ( _createToken && MParser_Token::is_nil(_token) && _ttype!=MParser_Token::SKIP ) {
      _token = makeToken(_ttype);
      _token->setText(MLANG_String(text_.getBuffer() + _begin, text_.length()-_begin));
   }
   returnToken_ = _token;
}
	
void ANTLRLexer::mQUESTION(bool _createToken)

{
   int _ttype; MParser_Token_var _token; int _begin=text_.length();
   _ttype = QUESTION;
   int _saveIndex;
   
   match('?');
   if ( _createToken && MParser_Token::is_nil(_token) && _ttype!=MParser_Token::SKIP ) {
      _token = makeToken(_ttype);
      _token->setText(MLANG_String(text_.getBuffer() + _begin, text_.length()-_begin));
   }
   returnToken_ = _token;
}
	
void ANTLRLexer::mTREE_BEGIN(bool _createToken)

{
   int _ttype; MParser_Token_var _token; int _begin=text_.length();
   _ttype = TREE_BEGIN;
   int _saveIndex;
   
   match("#(");
   if ( _createToken && MParser_Token::is_nil(_token) && _ttype!=MParser_Token::SKIP ) {
      _token = makeToken(_ttype);
      _token->setText(MLANG_String(text_.getBuffer() + _begin, text_.length()-_begin));
   }
   returnToken_ = _token;
}
	
void ANTLRLexer::mLPAREN(bool _createToken)

{
   int _ttype; MParser_Token_var _token; int _begin=text_.length();
   _ttype = LPAREN;
   int _saveIndex;
   
   match('(');
   if ( _createToken && MParser_Token::is_nil(_token) && _ttype!=MParser_Token::SKIP ) {
      _token = makeToken(_ttype);
      _token->setText(MLANG_String(text_.getBuffer() + _begin, text_.length()-_begin));
   }
   returnToken_ = _token;
}
	
void ANTLRLexer::mRPAREN(bool _createToken)

{
   int _ttype; MParser_Token_var _token; int _begin=text_.length();
   _ttype = RPAREN;
   int _saveIndex;
   
   match(')');
   if ( _createToken && MParser_Token::is_nil(_token) && _ttype!=MParser_Token::SKIP ) {
      _token = makeToken(_ttype);
      _token->setText(MLANG_String(text_.getBuffer() + _begin, text_.length()-_begin));
   }
   returnToken_ = _token;
}
	
void ANTLRLexer::mCOLON(bool _createToken)

{
   int _ttype; MParser_Token_var _token; int _begin=text_.length();
   _ttype = COLON;
   int _saveIndex;
   
   match(':');
   if ( _createToken && MParser_Token::is_nil(_token) && _ttype!=MParser_Token::SKIP ) {
      _token = makeToken(_ttype);
      _token->setText(MLANG_String(text_.getBuffer() + _begin, text_.length()-_begin));
   }
   returnToken_ = _token;
}
	
void ANTLRLexer::mSTAR(bool _createToken)

{
   int _ttype; MParser_Token_var _token; int _begin=text_.length();
   _ttype = STAR;
   int _saveIndex;
   
   match('*');
   if ( _createToken && MParser_Token::is_nil(_token) && _ttype!=MParser_Token::SKIP ) {
      _token = makeToken(_ttype);
      _token->setText(MLANG_String(text_.getBuffer() + _begin, text_.length()-_begin));
   }
   returnToken_ = _token;
}
	
void ANTLRLexer::mPLUS(bool _createToken)

{
   int _ttype; MParser_Token_var _token; int _begin=text_.length();
   _ttype = PLUS;
   int _saveIndex;
   
   match('+');
   if ( _createToken && MParser_Token::is_nil(_token) && _ttype!=MParser_Token::SKIP ) {
      _token = makeToken(_ttype);
      _token->setText(MLANG_String(text_.getBuffer() + _begin, text_.length()-_begin));
   }
   returnToken_ = _token;
}
	
void ANTLRLexer::mASSIGN(bool _createToken)

{
   int _ttype; MParser_Token_var _token; int _begin=text_.length();
   _ttype = ASSIGN;
   int _saveIndex;
   
   match('=');
   if ( _createToken && MParser_Token::is_nil(_token) && _ttype!=MParser_Token::SKIP ) {
      _token = makeToken(_ttype);
      _token->setText(MLANG_String(text_.getBuffer() + _begin, text_.length()-_begin));
   }
   returnToken_ = _token;
}
	
void ANTLRLexer::mIMPLIES(bool _createToken)

{
   int _ttype; MParser_Token_var _token; int _begin=text_.length();
   _ttype = IMPLIES;
   int _saveIndex;
   
   match("=>");
   if ( _createToken && MParser_Token::is_nil(_token) && _ttype!=MParser_Token::SKIP ) {
      _token = makeToken(_ttype);
      _token->setText(MLANG_String(text_.getBuffer() + _begin, text_.length()-_begin));
   }
   returnToken_ = _token;
}
	
void ANTLRLexer::mSEMI(bool _createToken)

{
   int _ttype; MParser_Token_var _token; int _begin=text_.length();
   _ttype = SEMI;
   int _saveIndex;
   
   match(';');
   if ( _createToken && MParser_Token::is_nil(_token) && _ttype!=MParser_Token::SKIP ) {
      _token = makeToken(_ttype);
      _token->setText(MLANG_String(text_.getBuffer() + _begin, text_.length()-_begin));
   }
   returnToken_ = _token;
}
	
void ANTLRLexer::mCARET(bool _createToken)

{
   int _ttype; MParser_Token_var _token; int _begin=text_.length();
   _ttype = CARET;
   int _saveIndex;
   
   match('^');
   if ( _createToken && MParser_Token::is_nil(_token) && _ttype!=MParser_Token::SKIP ) {
      _token = makeToken(_ttype);
      _token->setText(MLANG_String(text_.getBuffer() + _begin, text_.length()-_begin));
   }
   returnToken_ = _token;
}
	
void ANTLRLexer::mBANG(bool _createToken)

{
   int _ttype; MParser_Token_var _token; int _begin=text_.length();
   _ttype = BANG;
   int _saveIndex;
   
   match('!');
   if ( _createToken && MParser_Token::is_nil(_token) && _ttype!=MParser_Token::SKIP ) {
      _token = makeToken(_ttype);
      _token->setText(MLANG_String(text_.getBuffer() + _begin, text_.length()-_begin));
   }
   returnToken_ = _token;
}
	
void ANTLRLexer::mOR(bool _createToken)

{
   int _ttype; MParser_Token_var _token; int _begin=text_.length();
   _ttype = OR;
   int _saveIndex;
   
   match('|');
   if ( _createToken && MParser_Token::is_nil(_token) && _ttype!=MParser_Token::SKIP ) {
      _token = makeToken(_ttype);
      _token->setText(MLANG_String(text_.getBuffer() + _begin, text_.length()-_begin));
   }
   returnToken_ = _token;
}
	
void ANTLRLexer::mWILDCARD(bool _createToken)

{
   int _ttype; MParser_Token_var _token; int _begin=text_.length();
   _ttype = WILDCARD;
   int _saveIndex;
   
   match('.');
   if ( _createToken && MParser_Token::is_nil(_token) && _ttype!=MParser_Token::SKIP ) {
      _token = makeToken(_ttype);
      _token->setText(MLANG_String(text_.getBuffer() + _begin, text_.length()-_begin));
   }
   returnToken_ = _token;
}
	
void ANTLRLexer::mRANGE(bool _createToken)

{
   int _ttype; MParser_Token_var _token; int _begin=text_.length();
   _ttype = RANGE;
   int _saveIndex;
   
   match("..");
   if ( _createToken && MParser_Token::is_nil(_token) && _ttype!=MParser_Token::SKIP ) {
      _token = makeToken(_ttype);
      _token->setText(MLANG_String(text_.getBuffer() + _begin, text_.length()-_begin));
   }
   returnToken_ = _token;
}
	
void ANTLRLexer::mNOT_OP(bool _createToken)

{
   int _ttype; MParser_Token_var _token; int _begin=text_.length();
   _ttype = NOT_OP;
   int _saveIndex;
   
   match('~');
   if ( _createToken && MParser_Token::is_nil(_token) && _ttype!=MParser_Token::SKIP ) {
      _token = makeToken(_ttype);
      _token->setText(MLANG_String(text_.getBuffer() + _begin, text_.length()-_begin));
   }
   returnToken_ = _token;
}
	
void ANTLRLexer::mRCURLY(bool _createToken)

{
   int _ttype; MParser_Token_var _token; int _begin=text_.length();
   _ttype = RCURLY;
   int _saveIndex;
   
   match('}');
   if ( _createToken && MParser_Token::is_nil(_token) && _ttype!=MParser_Token::SKIP ) {
      _token = makeToken(_ttype);
      _token->setText(MLANG_String(text_.getBuffer() + _begin, text_.length()-_begin));
   }
   returnToken_ = _token;
}
	
void ANTLRLexer::mCHAR_LITERAL(bool _createToken)

{
   int _ttype; MParser_Token_var _token; int _begin=text_.length();
   _ttype = CHAR_LITERAL;
   int _saveIndex;
   
   match('\'');
   {
   if ((LA(1)=='\\')) {
      mESC(false);
   }
   else if ((_tokenSet_1.member(LA(1)))) {
      matchNot('\'');
   }
   else {
      throw MParser_NoViableAltForCharException((char)LA(1), getFilename(), getLine(), getColumn());
   }
   
   }
   match('\'');
   if ( _createToken && MParser_Token::is_nil(_token) && _ttype!=MParser_Token::SKIP ) {
      _token = makeToken(_ttype);
      _token->setText(MLANG_String(text_.getBuffer() + _begin, text_.length()-_begin));
   }
   returnToken_ = _token;
}
	
void ANTLRLexer::mESC(bool _createToken)

{
   int _ttype; MParser_Token_var _token; int _begin=text_.length();
   _ttype = ESC;
   int _saveIndex;
   
   match('\\');
   {
   switch ( LA(1)) {
   case 'n':
   {
      match('n');
      break;
   }
   case 'r':
   {
      match('r');
      break;
   }
   case 't':
   {
      match('t');
      break;
   }
   case 'b':
   {
      match('b');
      break;
   }
   case 'f':
   {
      match('f');
      break;
   }
   case 'w':
   {
      match('w');
      break;
   }
   case 'a':
   {
      match('a');
      break;
   }
   case '"':
   {
      match('"');
      break;
   }
   case '\'':
   {
      match('\'');
      break;
   }
   case '\\':
   {
      match('\\');
      break;
   }
   case '0':  case '1':  case '2':  case '3':
   {
      {
      matchRange('0','3');
      }
      {
      if (((LA(1) >= '0' && LA(1) <= '7')) && ((LA(2) >= char(3) && LA(2) <= int(0xff)))) {
      	matchRange('0','7');
      	{
      	if (((LA(1) >= '0' && LA(1) <= '7')) && ((LA(2) >= char(3) && LA(2) <= int(0xff)))) {
         matchRange('0','7');
      }
      	else if (((LA(1) >= char(3) && LA(1) <= int(0xff))) && (true)) {
      }
      	else {
         throw MParser_NoViableAltForCharException((char)LA(1), getFilename(), getLine(), getColumn());
      }
      	
      }
      }
      else if (((LA(1) >= char(3) && LA(1) <= int(0xff))) && (true)) {
      }
      else {
      	throw MParser_NoViableAltForCharException((char)LA(1), getFilename(), getLine(), getColumn());
      }
      
      }
      break;
   }
   case '4':  case '5':  case '6':  case '7':
   {
      {
      matchRange('4','7');
      }
      {
      if (((LA(1) >= '0' && LA(1) <= '7')) && ((LA(2) >= char(3) && LA(2) <= (int)0xff))) {
      	matchRange('0','7');
      }
      else if (((LA(1) >= char(3) && LA(1) <= int(0xff))) && (true)) {
      }
      else {
      	throw MParser_NoViableAltForCharException((char)LA(1), getFilename(), getLine(), getColumn());
      }
      
      }
      break;
   }
   case 'u':
   {
      match('u');
      mXDIGIT(false);
      mXDIGIT(false);
      mXDIGIT(false);
      mXDIGIT(false);
      break;
   }
   default:
   {
      throw MParser_NoViableAltForCharException((char)LA(1), getFilename(), getLine(), getColumn());
   }
   }
   }
   if ( _createToken && MParser_Token::is_nil(_token) && _ttype!=MParser_Token::SKIP ) {
      _token = makeToken(_ttype);
      _token->setText(MLANG_String(text_.getBuffer() + _begin, text_.length()-_begin));
   }
   returnToken_ = _token;
}
	
void ANTLRLexer::mSTRING_LITERAL(bool _createToken)

{
   int _ttype; MParser_Token_var _token; int _begin=text_.length();
   _ttype = STRING_LITERAL;
   int _saveIndex;
   
   match('"');
   {
   LABEL(_loop184)
   do {
      if ((LA(1)=='\\')) {
      	mESC(false);
      }
      else if ((_tokenSet_2.member(LA(1)))) {
      	matchNot('"');
      }
      else {
      	BREAK(_loop184);
      }
      
   } while (true);
   }
   match('"');
   if ( _createToken && MParser_Token::is_nil(_token) && _ttype!=MParser_Token::SKIP ) {
      _token = makeToken(_ttype);
      _token->setText(MLANG_String(text_.getBuffer() + _begin, text_.length()-_begin));
   }
   returnToken_ = _token;
}
	
void ANTLRLexer::mXDIGIT(bool _createToken)

{
   int _ttype; MParser_Token_var _token; int _begin=text_.length();
   _ttype = XDIGIT;
   int _saveIndex;
   
   switch ( LA(1)) {
   case '0':  case '1':  case '2':  case '3':
   case '4':  case '5':  case '6':  case '7':
   case '8':  case '9':
   {
      matchRange('0','9');
      break;
   }
   case 'a':  case 'b':  case 'c':  case 'd':
   case 'e':  case 'f':
   {
      matchRange('a','f');
      break;
   }
   case 'A':  case 'B':  case 'C':  case 'D':
   case 'E':  case 'F':
   {
      matchRange('A','F');
      break;
   }
   default:
   {
      throw MParser_NoViableAltForCharException((char)LA(1), getFilename(), getLine(), getColumn());
   }
   }
   if ( _createToken && MParser_Token::is_nil(_token) && _ttype!=MParser_Token::SKIP ) {
      _token = makeToken(_ttype);
      _token->setText(MLANG_String(text_.getBuffer() + _begin, text_.length()-_begin));
   }
   returnToken_ = _token;
}
	
void ANTLRLexer::mDIGIT(bool _createToken)

{
   int _ttype; MParser_Token_var _token; int _begin=text_.length();
   _ttype = DIGIT;
   int _saveIndex;
   
   matchRange('0','9');
   if ( _createToken && MParser_Token::is_nil(_token) && _ttype!=MParser_Token::SKIP ) {
      _token = makeToken(_ttype);
      _token->setText(MLANG_String(text_.getBuffer() + _begin, text_.length()-_begin));
   }
   returnToken_ = _token;
}
	
void ANTLRLexer::mINT(bool _createToken)

{
   int _ttype; MParser_Token_var _token; int _begin=text_.length();
   _ttype = INT;
   int _saveIndex;
   
   {
   int _cnt196=0;
   LABEL(_loop196)
   do {
      if (((LA(1) >= '0' && LA(1) <= '9'))) {
      	matchRange('0','9');
      }
      else {
      	if ( _cnt196>=1 ) { BREAK(_loop196); } else {throw MParser_NoViableAltForCharException((char)LA(1), getFilename(), getLine(), getColumn());}
      }
      
      _cnt196++;
   } while (true);
   }
   if ( _createToken && MParser_Token::is_nil(_token) && _ttype!=MParser_Token::SKIP ) {
      _token = makeToken(_ttype);
      _token->setText(MLANG_String(text_.getBuffer() + _begin, text_.length()-_begin));
   }
   returnToken_ = _token;
}
	
void ANTLRLexer::mARG_ACTION(bool _createToken)

{
   int _ttype; MParser_Token_var _token; int _begin=text_.length();
   _ttype = ARG_ACTION;
   int _saveIndex;
   
   mNESTED_ARG_ACTION(false);
   setText(MParser_StringUtils::stripFrontBack(getText(), "[", "]"));
   if ( _createToken && MParser_Token::is_nil(_token) && _ttype!=MParser_Token::SKIP ) {
      _token = makeToken(_ttype);
      _token->setText(MLANG_String(text_.getBuffer() + _begin, text_.length()-_begin));
   }
   returnToken_ = _token;
}
	
void ANTLRLexer::mNESTED_ARG_ACTION(bool _createToken)

{
   int _ttype; MParser_Token_var _token; int _begin=text_.length();
   _ttype = NESTED_ARG_ACTION;
   int _saveIndex;
   
   match('[');
   {
   LABEL(_loop200)
   do {
      switch ( LA(1)) {
      case '[':
      {
      	mNESTED_ARG_ACTION(false);
      	break;
      }
      case '\n':
      {
      	match('\n');
      	newline();
      	break;
      }
      case '\'':
      {
      	mCHAR_LITERAL(false);
      	break;
      }
      case '"':
      {
      	mSTRING_LITERAL(false);
      	break;
      }
      default:
      	if ((LA(1)=='\r') && (LA(2)=='\n')) {
         match('\r');
         match('\n');
         newline();
      }
      	else if ((LA(1)=='\r') && ((LA(2) >= char(3) && LA(2) <= int(0xff)))) {
         match('\r');
         newline();
      }
      	else if ((_tokenSet_3.member(LA(1)))) {
         matchNot(']');
      }
      else {
      	BREAK(_loop200);
      }
      }
   } while (true);
   }
   match(']');
   if ( _createToken && MParser_Token::is_nil(_token) && _ttype!=MParser_Token::SKIP ) {
      _token = makeToken(_ttype);
      _token->setText(MLANG_String(text_.getBuffer() + _begin, text_.length()-_begin));
   }
   returnToken_ = _token;
}
	
void ANTLRLexer::mACTION(bool _createToken)

{
   int _ttype; MParser_Token_var _token; int _begin=text_.length();
   _ttype = ACTION;
   int _saveIndex;
   int actionLine=getLine(); int actionColumn = getColumn();
   
   mNESTED_ACTION(false);
   {
   if ((LA(1)=='?')) {
      match('?');
      _ttype = SEMPRED;
   }
   else {
   }
   
   }
   
         if ( _ttype==ACTION ) {
            setText(MParser_StringUtils::stripFrontBack(getText(), "{", "}"));
         }
         else {
            setText(MParser_StringUtils::stripFrontBack(getText(), "{", "}?"));
         }
         MParser_CommonToken_var t = new MParser_CommonToken(_ttype,MLANG_String(text_.getBuffer() + _begin,text_.length()-_begin));
         t->setLine(actionLine);      // set action line to start
         t->setColumn(actionColumn);
         _token = t;
      	
   if ( _createToken && MParser_Token::is_nil(_token) && _ttype!=MParser_Token::SKIP ) {
      _token = makeToken(_ttype);
      _token->setText(MLANG_String(text_.getBuffer() + _begin, text_.length()-_begin));
   }
   returnToken_ = _token;
}
	
void ANTLRLexer::mNESTED_ACTION(bool _createToken)

{
   int _ttype; MParser_Token_var _token; int _begin=text_.length();
   _ttype = NESTED_ACTION;
   int _saveIndex;
   
   match('{');
   {
   LABEL(_loop206)
   do {
      // nongreedy exit test
      if ((LA(1)=='}') && (true)) BREAK(_loop206);
      if ((LA(1)=='\n'||LA(1)=='\r') && ((LA(2) >= char(3) && LA(2) <= int(0xff)))) {
      	{
      	if ((LA(1)=='\r') && (LA(2)=='\n')) {
         match('\r');
         match('\n');
         newline();
      }
      	else if ((LA(1)=='\r') && ((LA(2) >= char(3) && LA(2) <= int(0xff)))) {
         match('\r');
         newline();
      }
      	else if ((LA(1)=='\n')) {
         match('\n');
         newline();
      }
      	else {
         throw MParser_NoViableAltForCharException((char)LA(1), getFilename(), getLine(), getColumn());
      }
      	
      }
      }
      else if ((LA(1)=='{') && ((LA(2) >= char(3) && LA(2) <= int(0xff)))) {
      	mNESTED_ACTION(false);
      }
      else if ((LA(1)=='\'') && (_tokenSet_4.member(LA(2)))) {
      	mCHAR_LITERAL(false);
      }
      else if ((LA(1)=='/') && (LA(2)=='*'||LA(2)=='/')) {
      	mCOMMENT(false);
      }
      else if ((LA(1)=='"') && ((LA(2) >= char(3) && LA(2) <= int(0xff)))) {
      	mSTRING_LITERAL(false);
      }
      else if (((LA(1) >= char(3) && LA(1) <= int(0xff))) && ((LA(2) >= char(3) && LA(2) <= int(0xff)))) {
      	matchNot(EOF_CHAR);
      }
      else {
      	BREAK(_loop206);
      }
      
   } while (true);
   }
   match('}');
   if ( _createToken && MParser_Token::is_nil(_token) && _ttype!=MParser_Token::SKIP ) {
      _token = makeToken(_ttype);
      _token->setText(MLANG_String(text_.getBuffer() + _begin, text_.length()-_begin));
   }
   returnToken_ = _token;
}
	
void ANTLRLexer::mTOKEN_REF(bool _createToken)

{
   int _ttype; MParser_Token_var _token; int _begin=text_.length();
   _ttype = TOKEN_REF;
   int _saveIndex;
   
   matchRange('A','Z');
   {
   LABEL(_loop209)
   do {
      switch ( LA(1)) {
      case 'a':  case 'b':  case 'c':  case 'd':
      case 'e':  case 'f':  case 'g':  case 'h':
      case 'i':  case 'j':  case 'k':  case 'l':
      case 'm':  case 'n':  case 'o':  case 'p':
      case 'q':  case 'r':  case 's':  case 't':
      case 'u':  case 'v':  case 'w':  case 'x':
      case 'y':  case 'z':
      {
      	matchRange('a','z');
      	break;
      }
      case 'A':  case 'B':  case 'C':  case 'D':
      case 'E':  case 'F':  case 'G':  case 'H':
      case 'I':  case 'J':  case 'K':  case 'L':
      case 'M':  case 'N':  case 'O':  case 'P':
      case 'Q':  case 'R':  case 'S':  case 'T':
      case 'U':  case 'V':  case 'W':  case 'X':
      case 'Y':  case 'Z':
      {
      	matchRange('A','Z');
      	break;
      }
      case '_':
      {
      	match('_');
      	break;
      }
      case '0':  case '1':  case '2':  case '3':
      case '4':  case '5':  case '6':  case '7':
      case '8':  case '9':
      {
      	matchRange('0','9');
      	break;
      }
      default:
      {
      	BREAK(_loop209);
      }
      }
   } while (true);
   }
   _ttype = testLiteralsTable(_ttype);
   if ( _createToken && MParser_Token::is_nil(_token) && _ttype!=MParser_Token::SKIP ) {
      _token = makeToken(_ttype);
      _token->setText(MLANG_String(text_.getBuffer() + _begin, text_.length()-_begin));
   }
   returnToken_ = _token;
}
	
void ANTLRLexer::mRULE_REF(bool _createToken)

{
   int _ttype; MParser_Token_var _token; int _begin=text_.length();
   _ttype = RULE_REF;
   int _saveIndex;
   
      int t=0;
   
   
   t=mINTERNAL_RULE_REF(false);
   _ttype=t;
   {
   if (( true )&&(t==LITERAL_options)) {
      mWS_LOOP(false);
      {
      if ((LA(1)=='{')) {
      	match('{');
      	_ttype = OPTIONS;
      }
      else {
      }
      
      }
   }
   else if (( true )&&(t==LITERAL_tokens)) {
      mWS_LOOP(false);
      {
      if ((LA(1)=='{')) {
      	match('{');
      	_ttype = TOKENS;
      }
      else {
      }
      
      }
   }
   else {
   }
   
   }
   if ( _createToken && MParser_Token::is_nil(_token) && _ttype!=MParser_Token::SKIP ) {
      _token = makeToken(_ttype);
      _token->setText(MLANG_String(text_.getBuffer() + _begin, text_.length()-_begin));
   }
   returnToken_ = _token;
}
	
int ANTLRLexer::mINTERNAL_RULE_REF(bool _createToken)

{
   int t;
   int _ttype; MParser_Token_var _token; int _begin=text_.length();
   _ttype = INTERNAL_RULE_REF;
   int _saveIndex;
   
      t = RULE_REF;
   
   
   matchRange('a','z');
   {
   LABEL(_loop219)
   do {
      switch ( LA(1)) {
      case 'a':  case 'b':  case 'c':  case 'd':
      case 'e':  case 'f':  case 'g':  case 'h':
      case 'i':  case 'j':  case 'k':  case 'l':
      case 'm':  case 'n':  case 'o':  case 'p':
      case 'q':  case 'r':  case 's':  case 't':
      case 'u':  case 'v':  case 'w':  case 'x':
      case 'y':  case 'z':
      {
      	matchRange('a','z');
      	break;
      }
      case 'A':  case 'B':  case 'C':  case 'D':
      case 'E':  case 'F':  case 'G':  case 'H':
      case 'I':  case 'J':  case 'K':  case 'L':
      case 'M':  case 'N':  case 'O':  case 'P':
      case 'Q':  case 'R':  case 'S':  case 'T':
      case 'U':  case 'V':  case 'W':  case 'X':
      case 'Y':  case 'Z':
      {
      	matchRange('A','Z');
      	break;
      }
      case '_':
      {
      	match('_');
      	break;
      }
      case '0':  case '1':  case '2':  case '3':
      case '4':  case '5':  case '6':  case '7':
      case '8':  case '9':
      {
      	matchRange('0','9');
      	break;
      }
      default:
      {
      	BREAK(_loop219);
      }
      }
   } while (true);
   }
   t = testLiteralsTable(t);
   if ( _createToken && MParser_Token::is_nil(_token) && _ttype!=MParser_Token::SKIP ) {
      _token = makeToken(_ttype);
      _token->setText(MLANG_String(text_.getBuffer() + _begin, text_.length()-_begin));
   }
   returnToken_ = _token;
   return t;
}
	
void ANTLRLexer::mWS_LOOP(bool _createToken)

{
   int _ttype; MParser_Token_var _token; int _begin=text_.length();
   _ttype = WS_LOOP;
   int _saveIndex;
   
   {
   LABEL(_loop216)
   do {
      switch ( LA(1)) {
      case '\t':  case '\n':  case '\r':  case ' ':
      {
      	mWS(false);
      	break;
      }
      case '/':
      {
      	mCOMMENT(false);
      	break;
      }
      default:
      {
      	BREAK(_loop216);
      }
      }
   } while (true);
   }
   if ( _createToken && MParser_Token::is_nil(_token) && _ttype!=MParser_Token::SKIP ) {
      _token = makeToken(_ttype);
      _token->setText(MLANG_String(text_.getBuffer() + _begin, text_.length()-_begin));
   }
   returnToken_ = _token;
}
	
void ANTLRLexer::mWS_OPT(bool _createToken)

{
   int _ttype; MParser_Token_var _token; int _begin=text_.length();
   _ttype = WS_OPT;
   int _saveIndex;
   
   {
   if ((_tokenSet_5.member(LA(1)))) {
      mWS(false);
   }
   else {
   }
   
   }
   if ( _createToken && MParser_Token::is_nil(_token) && _ttype!=MParser_Token::SKIP ) {
      _token = makeToken(_ttype);
      _token->setText(MLANG_String(text_.getBuffer() + _begin, text_.length()-_begin));
   }
   returnToken_ = _token;
}
	
	
longArray ANTLRLexer::mk_tokenSet_0() {
   static ACE_UINT64 data[8];
   data[0]=-9224L;
   for (int i = 1; i<=3; i++) { data[i]=-1L; }
   return longArray(data,sizeof(data)/sizeof(data[0]));
}
MParser_BitSet ANTLRLexer::_tokenSet_0(ANTLRLexer::mk_tokenSet_0());
longArray ANTLRLexer::mk_tokenSet_1() {
   static ACE_UINT64 data[8];
   data[0]=-549755813896L;
   data[1]=-268435457L;
   for (int i = 2; i<=3; i++) { data[i]=-1L; }
   return longArray(data,sizeof(data)/sizeof(data[0]));
}
MParser_BitSet ANTLRLexer::_tokenSet_1(ANTLRLexer::mk_tokenSet_1());
longArray ANTLRLexer::mk_tokenSet_2() {
   static ACE_UINT64 data[8];
   data[0]=-17179869192L;
   data[1]=-268435457L;
   for (int i = 2; i<=3; i++) { data[i]=-1L; }
   return longArray(data,sizeof(data)/sizeof(data[0]));
}
MParser_BitSet ANTLRLexer::_tokenSet_2(ANTLRLexer::mk_tokenSet_2());
longArray ANTLRLexer::mk_tokenSet_3() {
   static ACE_UINT64 data[8];
   data[0]=-566935692296L;
   data[1]=-671088641L;
   for (int i = 2; i<=3; i++) { data[i]=-1L; }
   return longArray(data,sizeof(data)/sizeof(data[0]));
}
MParser_BitSet ANTLRLexer::_tokenSet_3(ANTLRLexer::mk_tokenSet_3());
longArray ANTLRLexer::mk_tokenSet_4() {
   static ACE_UINT64 data[8];
   data[0]=-549755813896L;
   for (int i = 1; i<=3; i++) { data[i]=-1L; }
   return longArray(data,sizeof(data)/sizeof(data[0]));
}
MParser_BitSet ANTLRLexer::_tokenSet_4(ANTLRLexer::mk_tokenSet_4());
longArray ANTLRLexer::mk_tokenSet_5() {
   ACE_UINT64 data[] = { 4294977024L, 0L, 0L, 0L, 0L};
   return longArray(data,sizeof(data)/sizeof(data[0]));
}
MParser_BitSet ANTLRLexer::_tokenSet_5(ANTLRLexer::mk_tokenSet_5());

ANTLRLexer::~ANTLRLexer ()

{
}
