#include "MGrammar/Grammar.h"
#include "MGrammar/CodeGenerator.h"
#include "MGrammar/GrammarAnalyzer.h"
#include "MGrammar/Tool.h"
#include "MGrammar/Alternative.h"
#include "MGrammar/AlternativeBlock.h"

/* ANTLR Translator Generator
 * Project led by Terence Parr at http://www.jGuru.com
 * Software rights: http://www.antlr.org/license.html
 *
 * $Id: //depot/code/org.antlr/release/antlr-2.7.4/antlr/AlternativeBlock.java#1 $
 */


int MGrammar_AlternativeBlock::nblks_ = 0; // how many blocks have we allocated?

MGrammar_AlternativeBlock::MGrammar_AlternativeBlock (
   MGrammar_Grammar_ptr g
) : MGrammar_AlternativeElement (g),
    hasAnAction_(false),
    hasASynPred_(false),
    ID_ (0),
    no_(false),
    greedy_(true),
    greedySet_(false),
    doAutoGen_(true),
    warnWhenFollowAmbig_(true),
    generateAmbigWarnings_(true)

{
   nblks_++;
   ID_ = nblks_;
}

MGrammar_AlternativeBlock::MGrammar_AlternativeBlock (
   MGrammar_Grammar_ptr g,
   const MParser_Token_var& start,
   bool no
) : MGrammar_AlternativeElement (g, start),
    hasAnAction_(false),
    hasASynPred_(false),
    ID_ (0),
    no_(no),
    greedy_(true),
    greedySet_(false),
    doAutoGen_(true),
    warnWhenFollowAmbig_(true),
    generateAmbigWarnings_(true)

{
    nblks_++;
    ID_ = nblks_;
}

void MGrammar_AlternativeBlock::addAlternative(MGrammar_Alternative_ptr alt)

{
   alternatives_.push_back(alt);
}

void MGrammar_AlternativeBlock::generate()

{
   grammar_->generator_->gen(this);
}

MGrammar_Lookahead MGrammar_AlternativeBlock::look(int k) 

{
   return grammar_->theLLkAnalyzer_->look(k, this);
}

void MGrammar_AlternativeBlock::prepareForAnalysis()

{
   for (size_t i = 0; i < alternatives_.size(); i++) 
    {
       // deterministic() uses an alternative cache and sets lookahead depth
       MGrammar_Alternative_var a = alternatives_[i];
       a->cache_.resize(0, MGrammar_Lookahead());
       a->cache_.resize(grammar_->maxk_ + 1, MGrammar_Lookahead());
       for (int j=0; j<grammar_->maxk_ + 1; j++)
           a->cache_[j].set_null();
       a->lookaheadDepth_ = MGrammar_GrammarAnalyzer::LOOKAHEAD_DEPTH_INIT;
     }
}

/**Walk the syntactic predicate and, for a rule ref R, remove
 * the ref from the list of FOLLOW references for R (stored
 * in the symbol table.
 */
void MGrammar_AlternativeBlock::removeTrackingOfRuleRefs(MGrammar_Grammar_ptr g)

{
   for (size_t i = 0; i < alternatives_.size(); i++)
    {
       MGrammar_Alternative_var alt = getAlternativeAt(i);
       MGrammar_AlternativeElement_var elem = alt->head_;
       while (!MGrammar_AlternativeElement::is_nil(elem))
        {
           MGrammar_RuleRefElement_var rr = MGrammar_RuleRefElement::_narrow(elem);
           if (!MGrammar_RuleRefElement::is_nil(rr))
            {
              MGrammar_RuleSymbol_var rs = 
                  MGrammar_RuleSymbol::_narrow(g->getSymbol(rr->targetRule()));
              if (MGrammar_RuleSymbol::is_nil(rs))
               {
                 grammar_->antlrTool_->error("rule " + rr->targetRule() + " referenced in (...)=>, but not defined");
               }
              else
               {
                 rs->removeReference(rr);
               }
            }
           else 
            {// recurse into subrules
               MGrammar_AlternativeBlock_var ab = MGrammar_AlternativeBlock::_narrow(elem);
               if (!MGrammar_AlternativeBlock::is_nil(ab))
                {
                  ab->removeTrackingOfRuleRefs(g);
                }
            }
           elem = elem->next_;
        }
    }
}

void MGrammar_AlternativeBlock::setOption (
   const MParser_Token_var& key,
   const MParser_Token_var& value
)

{
   if (key->getText().equals("warnWhenFollowAmbig"))
    {
      if (value->getText().equals("true"))
       {
         warnWhenFollowAmbig_ = true;
       }
      else if (value->getText().equals("false"))
       {
         warnWhenFollowAmbig_ = false;
       }
      else
       {
         grammar_->antlrTool_->error("Value for warnWhenFollowAmbig must be true or false", grammar_->getFilename(), key->getLine(), key->getColumn());
       }
    }
   else if (key->getText().equals("generateAmbigWarnings"))
    {
      if (value->getText().equals("true"))
       {
         generateAmbigWarnings_ = true;
       }
      else if (value->getText().equals("false")) 
       {
         generateAmbigWarnings_ = false;
       }
      else 
       {
         grammar_->antlrTool_->error("Value for generateAmbigWarnings must be true or false", grammar_->getFilename(), key->getLine(), key->getColumn());
       }
    }
   else if (key->getText().equals("greedy"))
    {
       if (value->getText().equals("true"))
        {
          greedy_ = true;
          greedySet_ = true;
        }
       else if (value->getText().equals("false"))
        {
          greedy_ = false;
          greedySet_ = true;
        }
       else 
        {
          grammar_->antlrTool_->error("Value for greedy must be true or false", grammar_->getFilename(), key->getLine(), key->getColumn());
        }
    }
   else
    {
       grammar_->antlrTool_->error("Invalid subrule option: " + key->getText(), grammar_->getFilename(), key->getLine(), key->getColumn());
    }
}

MLANG_String MGrammar_AlternativeBlock::toString() const

{
    MLANG_String s = " (";
    if (initAction_.length() > 0)
     {
       s += initAction_;
     }
    for (size_t i = 0; i < alternatives_.size(); i++)
     {
        MGrammar_Alternative_var alt = getAlternativeAt(i);
        ACE_Vector<MGrammar_Lookahead> cache = alt->cache_;
        int k = alt->lookaheadDepth_;
        // dump lookahead set
        if (k == MGrammar_GrammarAnalyzer::LOOKAHEAD_DEPTH_INIT)
         {
         }
        else if (k == MGrammar_GrammarAnalyzer::NONDETERMINISTIC)
         {
           s += "{?}:";
         }
        else 
         {
           s += " {";
           for (int j = 1; j <= k; j++)
            {
               s += cache[j].toString(",", grammar_->tokenManager_->getVocabulary());
               if (j < k && !cache[j + 1].nil()) s += ";";
            }
           s += "}:";
         }
        // dump alternative including pred (if any)
        MGrammar_AlternativeElement_var p = alt->head_;
        MLANG_String pred = alt->semPred_;
        if (pred.length())
         {
           s += pred;
         }
        while (!MGrammar_AlternativeElement::is_nil(p))
         {
           s += p->toString();
           p = p->next_;
         }
        if (i < (alternatives_.size() - 1))
         {
           s += " |";
         }
    }
   s += " )";
   return s;
}

MGrammar_AlternativeBlock::~MGrammar_AlternativeBlock ()

{
}
