#include "MParser/NumberFormatException.h"
#include "MParser/System.h"
#include "MParser/Token.h"
#include "MGrammar/StringLiteralElement.h"
#include "MGrammar/StringLiteralSymbol.h"
#include "MGrammar/RuleSymbol.h"
#include "MIO/PrintWriter.h"
#include "MGrammar/Tool.h"
#include "MGrammar/TokenManager.h"
#include "MGrammar/Grammar.h"
#include "MGrammar/CodeGenerator.h"

/* ANTLR Translator Generator
 * Project led by Terence Parr at http://www.jGuru.com
 * Software rights: http://www.antlr.org/license.html
 *
 * $Id: //depot/code/org.antlr/release/antlr-2.7.4/antlr/CodeGenerator.java#1 $
 */

/**A generic ANTLR code generator.  All code generators
 * Derive from this class.
 *
 * <p>
 * A CodeGenerator knows about a Grammar data structure and
 * a grammar analyzer.  The Grammar is walked to generate the
 * appropriate code for both a parser and lexer (if present).
 * This interface may change slightly so that the lexer is
 * itself living inside of a Grammar object (in which case,
 * this class generates only one recognizer).  The main method
 * to call is <tt>gen()</tt>, which initiates all code gen.
 *
 * <p>
 * The interaction of the code generator with the analyzer is
 * simple: each subrule block calls deterministic() before generating
 * code for the block.  Method deterministic() sets lookahead caches
 * in each Alternative object.  Technically, a code generator
 * doesn't need the grammar analyzer if all lookahead analysis
 * is done at runtime, but this would result in a slower parser.
 *
 * <p>
 * This class provides a set of support utilities to handle argument
 * list parsing and so on.
 *
 * @author  Terence Parr, John Lilley
 * @version 2.00a
 * @see     antlr.JavaCodeGenerator
 * @see     antlr.DiagnosticCodeGenerator
 * @see     antlr.LLkAnalyzer
 * @see     antlr.Grammar
 * @see     antlr.AlternativeElement
 * @see     antlr.Lookahead
 */

DEFINE_CLASS_INFO(MGrammar_CodeGenerator)

/** If there are more than 8 long words to init in a bitset,
 *  try to optimize it; e.g., detect runs of -1L and 0L.
 */
MLANG_String MGrammar_CodeGenerator::TokenTypesFileSuffix_ = "TokenTypes";
MLANG_String MGrammar_CodeGenerator::TokenTypesFileExt_ = ".txt";

/** Construct code generator base class */
MGrammar_CodeGenerator::MGrammar_CodeGenerator()
  : tabs_(0),
    grammar_(0),
    DEBUG_CODE_GENERATOR(false),
    makeSwitchThreshold_(DEFAULT_MAKE_SWITCH_THRESHOLD),
    bitsetTestThreshold_(DEFAULT_BITSET_TEST_THRESHOLD)

{
}

/** Output a String to the currentOutput stream.
 * Ignored if string is null.
 * @param s The string to output
 */
void MGrammar_CodeGenerator::_print (const MLANG_String& s) 

{
   if (s.length() != 0)
    {
      currentOutput_->print(s);
    }
}

/** Print an action without leading tabs, attempting to
 * preserve the current indentation level for multi-line actions
 * Ignored if string is null.
 * @param s The action string to output
 */
void MGrammar_CodeGenerator::_printAction(const MLANG_String& s)

{
    if (s.length() == 0)
     {
       return;
     }

    // Skip leading newlines, tabs and spaces
    size_t start = 0;
    while (start < s.length() && isspace(s.charAt(start)))
     {
       start++;
     }

    // Skip leading newlines, tabs and spaces
    int end = s.length() - 1;
    while (end > (int) start && isspace(s.charAt(end)))
     {
       end--;
     }

    char c = 0;
    for (int i = start; i <= end;)
     {
        c = s.charAt(i);
        i++;
        bool newline = false;
        switch (c)
         {
         case '\n':
              newline = true;
              break;
         case '\r':
              if (i <= end && s.charAt(i) == '\n')
               {
                 i++;
               }
              newline = true;
              break;
         default:
              currentOutput_->print(c);
              break;
         }
        if (newline)
         {
           currentOutput_->println();
           printTabs();
           // Absorb leading whitespace
           while (i <= end && isspace(s.charAt(i)))
            {
               i++;
            }
           newline = false;
         }
     }
    currentOutput_->println();
}

/** Output a String followed by newline, to the currentOutput stream.
 * Ignored if string is null.
 * @param s The string to output
 */
void MGrammar_CodeGenerator::_println (const MLANG_String& s)

{
   if (s.length() != 0)
    {
      currentOutput_->println(s);
    }
}

/** Test if a set element array represents a contiguous range.
 * @param elems The array of elements representing the set, usually from BitSet.toArray().
 * @return true if the elements are a contiguous range (with two or more).
 */
bool MGrammar_CodeGenerator::elementsAreRange (const MLANG_Vector<size_t>& elems)

{
   if (elems.size() == 0)
    {
      return false;
    }
   int begin = elems[0];
   int end = elems[elems.size() - 1];
   if (elems.size() <= 2)
    {
      // Not enough elements for a range expression
      return false;
    }
   if (end - begin + 1 > (int)elems.size())
    {
      // The set does not represent a contiguous range
      return false;
    }
   int v = begin + 1;
   for (size_t i = 1; i < elems.size() - 1; i++)
    {
       if (v != (int)elems[i])
        {
          // The set does not represent a contiguous range
          return false;
        }
       v++;
    }
   return true;
}

/** Get the identifier portion of an argument-action token.
 * The ID of an action is assumed to be a trailing identifier.
 * Specific code-generators may want to override this
 * if the language has unusual declaration syntax.
 * @param t The action token
 * @return A string containing the text of the identifier
 */
MLANG_String MGrammar_CodeGenerator::extractIdOfAction(MParser_Token_ptr t)

{
   return extractIdOfAction(t->getText(), t->getLine(), t->getColumn());
}

/** Get the identifier portion of an argument-action.
 * The ID of an action is assumed to be a trailing identifier.
 * Specific code-generators may want to override this
 * if the language has unusual declaration syntax.
 * @param s The action text
 * @param line Line used for error reporting.
 * @param column Line used for error reporting.
 * @return A string containing the text of the identifier
 */
MLANG_String MGrammar_CodeGenerator::extractIdOfAction (
   const MLANG_String& s,
   int line,
   int column
)

{
   MLANG_String r = removeAssignmentFromDeclaration(s);
   // Search back from the end for a non alphanumeric.  That marks the
   // beginning of the identifier
   for (int i = r.length() - 2; i >= 0; i--)
    {
       // TODO: make this work for language-independent identifiers?
       if (!isalnum(r.charAt(i)) && r.charAt(i) != '_')
        {
          // Found end of type part
          return r.substring(i + 1);
        }
    }
   // Something is bogus, but we cannot parse the language-specific
   // actions any better.  The compiler will have to catch the problem.
   antlrTool_->warning("Ill-formed action", grammar_->getFilename(), line, column);
   return "";
}

/** Get the type string out of an argument-action token.
 * The type of an action is assumed to precede a trailing identifier
 * Specific code-generators may want to override this
 * if the language has unusual declaration syntax.
 * @param t The action token
 * @return A string containing the text of the type
 */
MLANG_String MGrammar_CodeGenerator::extractTypeOfAction(MParser_Token_ptr t)

{
   return extractTypeOfAction(t->getText(), t->getLine(), t->getColumn());
}

/** Get the type portion of an argument-action.
 * The type of an action is assumed to precede a trailing identifier
 * Specific code-generators may want to override this
 * if the language has unusual declaration syntax.
 * @param s The action text
 * @param line Line used for error reporting.
 * @return A string containing the text of the type
 */
MLANG_String MGrammar_CodeGenerator::extractTypeOfAction (
   const MLANG_String& s,
   int line,
   int column
)

{

   MLANG_String r = removeAssignmentFromDeclaration(s);
   // Search back from the end for a non alphanumeric.  That marks the
   // beginning of the identifier
   for (int i = r.length() - 2; i >= 0; i--)
    {
       // TODO: make this work for language-independent identifiers?
       if (!isalnum(r.charAt(i)) && r.charAt(i) != '_')
        {
          // Found end of type part
          return r.substring(0, i + 1);
        }
    }
   // Something is bogus, but we cannot parse the language-specific
   // actions any better.  The compiler will have to catch the problem.
   antlrTool_->warning("Ill-formed action", grammar_->getFilename(), line, column);
   return "";
}

/** Generate the token types as a text file for persistence across shared lexer/parser */
void MGrammar_CodeGenerator::genTokenInterchange(MGrammar_TokenManager_ptr tm)

{
   // Open the token output Java file and set the currentOutput stream
   MLANG_String fName = tm->getName() + TokenTypesFileSuffix_ + TokenTypesFileExt_;
   currentOutput_ = 
      antlrTool_->openOutputFile(fName);

   println("// $ANTLR " + antlrTool_->version() + ": " +
           antlrTool_->fileMinusPath(antlrTool_->grammarFile()) +
           " -> " +
           fName +
           "$");

   tabs_ = 0;

   // Header
   println(tm->getName() + "    // output token vocab name");

   // Generate a definition for each token type
   MGrammar_TokenManager::VOCABULARY v = tm->getVocabulary();
   for (int i = MParser_Token::MIN_USER_TYPE; i < (int) v.size(); i++)
    {
      MLANG_String s = v[i];
      if (DEBUG_CODE_GENERATOR)
       {
         SYSTEM->out()->println("gen persistence file entry for: " + s);
       }
      if (s.length() != 0 && !s.startsWith("<"))
       {
         // if literal, find label
         if (s.startsWith("\""))
          {
            MGrammar_StringLiteralSymbol_var sl = 
                MGrammar_StringLiteralSymbol::_narrow(tm->getTokenSymbol(s));
            if (!MGrammar_StringLiteralSymbol::is_nil(sl) && sl->getLabel().length() != 0)
             {
               print(sl->getLabel() + "=");
             }
            println(s + "=" +MLANG_String::valueOf(i));
          }
         else
          {
            print(s);
            // check for a paraphrase
            MGrammar_TokenSymbol_var ts = 
                 MGrammar_TokenSymbol::_duplicate(tm->getTokenSymbol(s));
            if (MGrammar_TokenSymbol::is_nil(ts))
             {
               antlrTool_->warning("undefined token symbol: " + s);
             }
            else
             {
               if (ts->getParaphrase().length() != 0)
                {
                  print("(" + ts->getParaphrase() + ")");
                }
             }
            println("=" +MLANG_String::valueOf(i));
          }
       }
    }

   // Close the tokens output file
   currentOutput_->close();
   currentOutput_ = 0;
}

/** Given the index of a bitset in the bitset list, generate a unique name.
 * Specific code-generators may want to override this
 * if the language does not allow '_' or numerals in identifiers.
 * @param index  The index of the bitset in the bitset list.
 */
MLANG_String MGrammar_CodeGenerator::getBitsetName(size_t index) const

{
   return "_tokenSet_" + MLANG_String::valueOf(index);
}

MLANG_String MGrammar_CodeGenerator::encodeLexerRuleName(const MLANG_String& id)

{
   return "m" + id;
}

MLANG_String MGrammar_CodeGenerator::decodeLexerRuleName(const MLANG_String& id)

{
   if ( id.length()==0)
    {
      return "";
    }
   return id.substring(1,id.length());
}

/** Add a bitset to the list of bitsets to be generated.
 * if the bitset is already in the list, ignore the request.
 * Always adds the bitset to the end of the list, so the
 * caller can rely on the position of bitsets in the list.
 * The returned position can be used to format the bitset
 * name, since it is invariant.
 * @param p Bit set to mark for code generation
 * @param forParser true if the bitset is used for the parser, false for the lexer
 * @return The position of the bitset in the list.
 */
int MGrammar_CodeGenerator::markBitsetForGen(const MParser_BitSet& p)

{
    // Is the bitset (or an identical one) already marked for gen?
    for (size_t i = 0; i < bitsetsUsed_.size(); i++)
     {
        MParser_BitSet set = bitsetsUsed_[i];
        if (p.equals(set))
         {
           // Use the identical one already stored
           return i;
         }
     }

    // Add the new bitset
    bitsetsUsed_.push_back(p);
    return bitsetsUsed_.size() - 1;
}

/** Output tab indent followed by a String, to the currentOutput stream.
 * Ignored if string is null.
 * @param s The string to output.
 */
void MGrammar_CodeGenerator::print(const MLANG_String& s)

{
#if 0
   if (s.length() != 0)
#endif
    {
      printTabs();
      currentOutput_->print(s);
    }
}

/** Print an action with leading tabs, attempting to
 * preserve the current indentation level for multi-line actions
 * Ignored if string is null.
 * @param s The action string to output
 */
void MGrammar_CodeGenerator::printAction(const MLANG_String& s)

{
#if 0
   if (s.length() != 0)
#endif
    {
      printTabs();
      _printAction(s);
    }
}

/** Output tab indent followed by a String followed by newline,
 * to the currentOutput stream.  Ignored if string is null.
 * @param s The string to output
 */
void MGrammar_CodeGenerator::println(const MLANG_String& s)

{
#if 0
   if (s.length() != 0) 
#endif
    {
      printTabs();
      currentOutput_->println(s);
    }
}

/** Output the current tab indentation.  This outputs the number of tabs
 * indicated by the "tabs" variable to the currentOutput stream.
 */
void MGrammar_CodeGenerator::printTabs()

{
   for (int i = 1; i <= tabs_; i++) 
    {
       currentOutput_->print(' ');
    }
}

MLANG_String MGrammar_CodeGenerator::getFOLLOWBitSet (
   const MLANG_String& ruleName,
   int k
)

{
   MGrammar_RuleSymbol_var rs = Dynamic_Cast<MGrammar_RuleSymbol>(grammar_->getSymbol(ruleName));
   if (MGrammar_RuleSymbol::is_nil(rs))
    {
      return "";
    }
   MGrammar_RuleBlock_var blk = rs->getBlock();
   MGrammar_Lookahead follow = grammar_->theLLkAnalyzer_->FOLLOW(k, blk->endNode());
   MLANG_String followSetName = getBitsetName(markBitsetForGen(follow.fset_));
   return followSetName;
}

MLANG_String MGrammar_CodeGenerator::getFIRSTBitSet (
   const MLANG_String& ruleName,
   int k
)

{
   MGrammar_RuleSymbol_var rs = 
       MGrammar_RuleSymbol::_narrow(grammar_->getSymbol(ruleName));
   if (MGrammar_RuleSymbol::is_nil(rs))
    {
      return "";
    }
   MGrammar_RuleBlock_var blk = 
       MGrammar_RuleBlock::_duplicate(rs->getBlock());
   MGrammar_Lookahead first = grammar_->theLLkAnalyzer_->look(k, blk);
   MLANG_String firstSetName = getBitsetName(markBitsetForGen(first.fset_));
   return firstSetName;
 }

/**
 * Remove the assignment portion of a declaration, if any.
 * @param d the declaration
 * @return the declaration without any assignment portion
 */
MLANG_String MGrammar_CodeGenerator::removeAssignmentFromDeclaration (
   const MLANG_String& d
)

{
   MLANG_String r = d;
   // If d contains an equal sign, then it's a declaration
   // with an initialization.  Strip off the initialization part.
   if (r.indexOf('=') >= 0) r = r.substring(0, r.indexOf('=')).trim();
   return r;
}

/** Set all fields back like one just created */
void MGrammar_CodeGenerator::reset()

{
   tabs_ = 0;
   // Allocate list of bitsets tagged for code generation
   bitsetsUsed_ = MLANG_Vector<MParser_BitSet>();
   currentOutput_ = 0;
   grammar_ = 0;
   DEBUG_CODE_GENERATOR = false;
   makeSwitchThreshold_ = DEFAULT_MAKE_SWITCH_THRESHOLD;
   bitsetTestThreshold_ = DEFAULT_BITSET_TEST_THRESHOLD;
}

MLANG_String MGrammar_CodeGenerator::reverseLexerRuleName (const MLANG_String& id)

{
   return id.substring(1, id.length());
}

void MGrammar_CodeGenerator::setAnalyzer (
   const MGrammar_LLkGrammarAnalyzer_var& analyzer
)

{
   analyzer_ = analyzer;
}

void MGrammar_CodeGenerator::setBehavior (
   const MGrammar_DefineGrammarSymbols_var& behavior
)

{
   behavior_ = behavior;
}

/** Set a grammar for the code generator to use */
void MGrammar_CodeGenerator::setGrammar(MGrammar_Grammar_ptr g)

{
   reset();
   grammar_ = g;
   // Lookup make-switch threshold in the grammar generic options
   if (grammar_->hasOption("codeGenMakeSwitchThreshold"))
    {
      try
       {
          makeSwitchThreshold_ = grammar_->getIntegerOption("codeGenMakeSwitchThreshold");
                //SYSTEM->out()->println("setting codeGenMakeSwitchThreshold to " + makeSwitchThreshold);
       }
      catch (MParser_NumberFormatException& e)
       {
          MParser_Token_var tok = 
              MParser_Token::_duplicate(grammar_->getOption("codeGenMakeSwitchThreshold"));
          antlrTool_->error (
                "option 'codeGenMakeSwitchThreshold' must be an integer",
                grammar_->getClassName(),
                tok->getLine(), tok->getColumn()
            );
       }
    }

   // Lookup bitset-test threshold in the grammar generic options
   if (grammar_->hasOption("codeGenBitsetTestThreshold"))
    {
      try
       {
          bitsetTestThreshold_ = grammar_->getIntegerOption("codeGenBitsetTestThreshold");
          //SYSTEM->out()->println("setting codeGenBitsetTestThreshold to " + bitsetTestThreshold);
       }
      catch (MParser_NumberFormatException& e)
       {
          MParser_Token_var tok = grammar_->getOption("codeGenBitsetTestThreshold");
          antlrTool_->error (
               "option 'codeGenBitsetTestThreshold' must be an integer",
               grammar_->getClassName(),
               tok->getLine(), tok->getColumn()
           );
       }
    }

   // Lookup debug code-gen in the grammar generic options
   if (grammar_->hasOption("codeGenDebug"))
    {
      MParser_Token_var t = 
           MParser_Token::_duplicate(grammar_->getOption("codeGenDebug"));
      if (t->getText().equals("true"))
       {
          //SYSTEM->out()->println("setting code-generation debug ON");
          DEBUG_CODE_GENERATOR = true;
       }
      else if (t->getText().equals("false"))
       {
          //SYSTEM->out()->println("setting code-generation debug OFF");
          DEBUG_CODE_GENERATOR = false;
       }
      else
       {
          antlrTool_->error("option 'codeGenDebug' must be true or false", grammar_->getClassName(), t->getLine(), t->getColumn());
       }
    }
}

void MGrammar_CodeGenerator::setTool (MGrammar_Tool_ptr tool)

{
   antlrTool_ = tool;
}

MGrammar_CodeGenerator::~MGrammar_CodeGenerator()

{
}
