#include "MLANG/String.h"
#include "MLANG/StringBuffer.h"
#include "MGrammar/JavaCharFormatter.h"
#include "MGrammar/Lookahead.h"
#include "MGrammar/Alternative.h"
#include "MGrammar/Tool.h"
#include "MGrammar/Grammar.h"
#include "MGrammar/CodeGenerator.h"
#include "MGrammar/DefaultToolErrorHandler.h"

/* ANTLR Translator Generator
 * Project led by Terence Parr at http://www.jGuru.com
 * Software rights: http://www.antlr.org/license.html
 *
 * $Id: //depot/code/org.antlr/release/antlr-2.7.4/antlr/DefaultToolErrorHandler.java#1 $
 */

MGrammar_DefaultToolErrorHandler::MGrammar_DefaultToolErrorHandler (
   MGrammar_Tool_ptr tool
) : antlrTool_(tool),
    javaCharFormatter_(new MGrammar_JavaCharFormatter())

{
}


/** Dump token/character sets to a string array suitable for
 * {@link antlr.Tool.warning(String[], String, int, int)
 * @param output The array that will contain the token/character set dump,
 *               one element per k (lookahead) value
 * @param outputStartIndex The index into <code>output</code> that the
 *                         dump should start at.
 * @param lexicalAnalysis  true for lexical rule
 * @param depth  The depth of the ambiguity
 * @param sets  An array of bitsets containing the ambiguities
 */
void MGrammar_DefaultToolErrorHandler::dumpSets (
   ACE_Vector<MLANG_String>& output,
   int outputStartIndex,
   MGrammar_Grammar_ptr grammar,
   bool lexicalAnalysis,
   int depth,
   const ACE_Vector<MGrammar_Lookahead>& sets
)

{
   MLANG_StringBuffer line;
   for (int i = 1; i <= depth; i++)
    {
       line.append("k==").append(i).append(':');
       if (lexicalAnalysis)
	{
          MLANG_String bits = sets[i].fset_.toStringWithRanges(",", *javaCharFormatter_.in());
          if (sets[i].containsEpsilon())
	   {
             line.append("<end-of-token>");
             if (bits.length() > 0)
              {
                line.append(',');
              }
           }
          line.append(bits);
        }
       else
	{
          line.append(sets[i].fset_.toString(",", grammar->tokenManager_->getVocabulary()));
        }
       output[outputStartIndex++] = line.toString();
       line.setLength(0);
   }
}

/** Issue a warning about ambiguity between a alternates
 * @param blk  The block being analyzed
 * @param lexicalAnalysis  true for lexical rule
 * @param depth  The depth of the ambiguity
 * @param sets  An array of bitsets containing the ambiguities
 * @param altIdx1  The zero-based index of the first ambiguous alternative
 * @param altIdx2  The zero-based index of the second ambiguous alternative
 */
void MGrammar_DefaultToolErrorHandler::warnAltAmbiguity (
   MGrammar_Grammar_ptr grammar,
   MGrammar_AlternativeBlock_ptr blk,
   bool lexicalAnalysis,
   int depth,
   const ACE_Vector<MGrammar_Lookahead>& sets,
   int altIdx1,
   int altIdx2
)

{
   MLANG_StringBuffer line;
   if (ACE_dynamic_cast(MGrammar_RuleBlock_ptr, blk)!=0
   &&  ACE_dynamic_cast(MGrammar_RuleBlock_ptr, blk)->isLexerAutoGenRule())
    {
       MGrammar_Alternative_ptr ai = blk->getAlternativeAt(altIdx1);
       MGrammar_Alternative_ptr aj = blk->getAlternativeAt(altIdx2);
       MGrammar_RuleRefElement_var rri = MGrammar_RuleRefElement::_narrow(ai->head_.in());
       MGrammar_RuleRefElement_var rrj = MGrammar_RuleRefElement::_narrow(aj->head_.in());
       MLANG_String ri = MGrammar_CodeGenerator::reverseLexerRuleName(rri->targetRule());
       MLANG_String rj = MGrammar_CodeGenerator::reverseLexerRuleName(rrj->targetRule());
       line.append("lexical nondeterminism between rules ");
       line.append(ri).append(" and ").append(rj).append(" upon");
    }
   else
    {
       if (lexicalAnalysis)
	{
          line.append("lexical ");
        }
       line.append("nondeterminism between alts ");
       line.append(altIdx1 + 1).append(" and ");
       line.append(altIdx2 + 1).append(" of block upon");
    }
   ACE_Vector<MLANG_String> output;
   output.resize (depth + 1, MLANG_String());
   output[0] = line.toString();
   dumpSets(output, 1, grammar, lexicalAnalysis, depth, sets);
   antlrTool_->warning(output, grammar->getFilename(), blk->getLine(), blk->getColumn());

}

/** Issue a warning about ambiguity between an alternate and exit path.
 * @param blk  The block being analyzed
 * @param lexicalAnalysis  true for lexical rule
 * @param depth  The depth of the ambiguity
 * @param sets  An array of bitsets containing the ambiguities
 * @param altIdx  The zero-based index of the ambiguous alternative
 */
void MGrammar_DefaultToolErrorHandler::warnAltExitAmbiguity (
   MGrammar_Grammar_ptr grammar,
   MGrammar_BlockWithImpliedExitPath_ptr blk,
   bool lexicalAnalysis,
   int depth,
   const ACE_Vector<MGrammar_Lookahead>& sets,
   int altIdx
)

{
   ACE_Vector<MLANG_String> output;
   output.resize (depth + 2, MLANG_String());
   output[0] = MLANG_String(lexicalAnalysis ? "lexical " : "") + "nondeterminism upon";
   dumpSets(output, 1, grammar, lexicalAnalysis, depth, sets);
   output[depth + 1] = "between alt " +MLANG_String::valueOf(altIdx + 1) + " and exit branch of block";
   antlrTool_->warning(output, grammar->getFilename(), blk->getLine(), blk->getColumn());
}

MGrammar_DefaultToolErrorHandler::~MGrammar_DefaultToolErrorHandler ()

{
}
