#include "MLANG/String.h"
#include "MLANG/StringBuffer.h"
#include "MParser/TokenTypes.h"
#include "MParser/NumberFormatException.h"
#include "MParser/CommonToken.h"
#include "MGrammar/Tool.h"
#include "MGrammar/LLkAnalyzer.h"
#include "MGrammar/CodeGenerator.h"
#include "MGrammar/Grammar.h"

/* ANTLR Translator Generator
 * Project led by Terence Parr at http://www.jGuru.com
 * Software rights: http://www.antlr.org/license.html
 *
 * $Id: //depot/code/org.antlr/release/antlr-2.7.4/antlr/Grammar.java#1 $
 */


MGrammar_Grammar::MGrammar_Grammar (
   const MLANG_String& className, 
   MGrammar_Tool_ptr tool,
   const MLANG_String& superClass
) : antlrTool_(tool),
    buildAST_(false),
    analyzerDebug_(false),
    interactive_(false),
    superClass_(superClass),
    preambleAction_(new MParser_CommonToken(MParser_Token::INVALID_TYPE, "")),
    className_(className),
    classMemberAction_(new MParser_CommonToken(MParser_Token::INVALID_TYPE, "")),
    hasSyntacticPredicate_(false),
    hasUserErrorHandling_(false),
    maxk_(1),
    traceRules_(false),
    debuggingOutput_(false),
    defaultErrorHandler_(true)

{
}

/** Define a rule */
void MGrammar_Grammar::define(const MGrammar_RuleSymbol_var& rs)

{
    rules_.push_back(rs);
    // add the symbol to the rules hash table
    symbols_[rs->getId()] = rs;
}

/** Get an integer option.  Given the name of the option find its
 * associated integer value.  If the associated value is not an integer or
 * is not in the table, then throw an exception of type NumberFormatException.
 * @param key The name of the option
 * @return The value associated with the key.
 */
int MGrammar_Grammar::getIntegerOption(const MLANG_String& key)

{
   OPTIONS::const_iterator iter =
            options_.find (key);
   if (iter == options_.end() 
   ||  iter->second->getType() != MParser_TokenTypes::INT)
    {
      throw MParser_NumberFormatException();
    }
   else
    {
      return ACE_OS::atoi (iter->second->getText().c_str());
    }
}

/** Get an option.  Given the name of the option find its associated value.
 * @param key The name of the option
 * @return The value associated with the key, or null if the key has not been set.
 */
MParser_Token_var MGrammar_Grammar::getOption(const MLANG_String& key)

{
   MParser_Token_var token;
   OPTIONS::const_iterator iter =
            options_.find (key);
   if (iter == options_.end())
    {
      return token;
    }
   else
    {
      return iter->second;
    }
}

MGrammar_GrammarSymbol_var MGrammar_Grammar::getSymbol(const MLANG_String& s)

{
   MGrammar_GrammarSymbol_var sym;
   SYMBOLS::const_iterator iter = symbols_.find (s);
   if (iter != symbols_.end())
    {
      sym = iter->second;
    }
   return sym;
}

SYMBOLS::iterator MGrammar_Grammar::getSymbols() 

{
   return symbols_.begin();
}

/** Check the existence of an option in the table
 * @param key The name of the option
 * @return true if the option is in the table
 */
bool MGrammar_Grammar::hasOption(const MLANG_String& key) const

{
    return options_.count(key);
}

/** Is a rule symbol defined? (not used for tokens) */
bool MGrammar_Grammar::isDefined(const MLANG_String& s) const

{
  return symbols_.count(s)!=0;
}

/** Set a generic option.
 * This associates a generic option key with a Token value.
 * No validation is performed by this method, although users of the value
 * (code generation and/or analysis) may require certain formats.
 * The value is stored as a token so that the location of an error
 * can be reported.
 * @param key The name of the option.
 * @param value The value to associate with the key.
 * @return true if the option was a valid generic grammar option, false o/w
 */
bool MGrammar_Grammar::setOption (
   const MLANG_String& key, 
   const MParser_Token_var& value
)

{
   options_[key] = value;
   MLANG_String s = value->getText();
   int i;
   if (key.equals("k"))
    {
      try
       {
         maxk_ = getIntegerOption("k");
         if ( maxk_<=0 )
          {
            antlrTool_->error("option 'k' must be greater than 0 (was " +
			       value->getText() + ")",
				getFilename(),
				value->getLine(),
				value->getColumn());
            maxk_ = 1;
          }
       }
      catch (MParser_NumberFormatException& e)
       {
         antlrTool_->error("option 'k' must be an integer (was " + 
                           value->getText() + ")", 
                           getFilename(), 
                           value->getLine(), 
                           value->getColumn());
       }
      return true;
    }
   if (key.equals("codeGenMakeSwitchThreshold"))
    {
      try
       {
          i = getIntegerOption("codeGenMakeSwitchThreshold");
       }
      catch (MParser_NumberFormatException& e)
       {
          antlrTool_->error("option 'codeGenMakeSwitchThreshold' must be an integer", getFilename(), value->getLine(), value->getColumn());
       }
      return true;
    }
   if (key.equals("codeGenBitsetTestThreshold"))
    {
      try
       {
         i = getIntegerOption("codeGenBitsetTestThreshold");
       }
      catch (MParser_NumberFormatException& e)
       {
         antlrTool_->error("option 'codeGenBitsetTestThreshold' must be an integer", getFilename(), value->getLine(), value->getColumn());
       }
      return true;
    }
   if (key.equals("defaultErrorHandler"))
    {
      if (s.equals("true"))
       {
         defaultErrorHandler_ = true;
       }
      else if (s.equals("false"))
       {
         defaultErrorHandler_ = false;
       }
      else 
       {
         antlrTool_->error("Value for defaultErrorHandler must be true or false", getFilename(), value->getLine(), value->getColumn());
       }
      return true;
    }
   if (key.equals("analyzerDebug"))
    {
      if (s.equals("true"))
       {
         analyzerDebug_ = true;
       }
      else if (s.equals("false"))
       {
         analyzerDebug_ = false;
       }
      else
       {
         antlrTool_->error("option 'analyzerDebug' must be true or false", getFilename(), value->getLine(), value->getColumn());
       }
      return true;
    }
   if (key.equals("codeGenDebug"))
    {
      if (s.equals("true"))
       {
         analyzerDebug_ = true;
       }
      else if (s.equals("false"))
       {
         analyzerDebug_ = false;
       }
      else
       {
         antlrTool_->error("option 'codeGenDebug' must be true or false", getFilename(), value->getLine(), value->getColumn());
       }
      return true;
    }
   if (key.equals("classHeaderSuffix")) 
    {
      return true;
    }
   if (key.equals("classHeaderPrefix")) 
    {
      return true;
    }
   if (key.equals("namespaceAntlr")) 
    {
      return true;
    }
   if (key.equals("namespaceStd")) 
    {
      return true;
    }
   if (key.equals("genHashLines")) 
    {
      return true;
    }
   if (key.equals("noConstructors")) 
    {
      return true;
    }
   return false;
}

/** Print out the grammar without actions */
MLANG_String MGrammar_Grammar::toString() const

{
   MLANG_StringBuffer buf;
   for (size_t i=0; i<rules_.size(); i++)
    {
      MGrammar_RuleSymbol_var rs = rules_[i];
      if (!rs->getId().equals("mnextToken"))
       {
         buf.append(rs->getBlock()->toString());
         buf.append("\n\n");
       }
    }
   return buf.toString();
}

MGrammar_Grammar::~MGrammar_Grammar ()

{
}
