#ifndef __MGRAMMAR_GRAMMAR_H__
#define __MGRAMMAR_GRAMMAR_H__
#include "ace/ARGV.h"
#include "MLANG/Refcountable.h"
#include "MLANG/SmartPtr.h"
#include "MLANG/HashMap.h"
#include "MParser/BitSet.h"
#include "MParser/Token.h"
#include "MGrammar/RuleSymbol.h"
#include "MGrammar/LLkAnalyzer.h"

class MGrammar_Grammar;
typedef class MGrammar_Grammar *MGrammar_Grammar_ptr;
typedef MLANG_SmartPtr<MGrammar_Grammar> MGrammar_Grammar_var;

class MGrammar_Tool;
typedef class MGrammar_Tool *MGrammar_Tool_ptr;
typedef class MGrammar_Tool *MGrammar_Tool_var;

//#include "MGrammar/CodeGenerator.h"
class MGrammar_CodeGenerator;
typedef MGrammar_CodeGenerator *MGrammar_CodeGenerator_ptr;
typedef MGrammar_CodeGenerator *MGrammar_CodeGenerator_var;
//typedef MLANG_SmartPtr<MGrammar_CodeGenerator> MGrammar_CodeGenerator_var;

/* ANTLR Translator Generator
 * Project led by Terence Parr at http://www.jGuru.com
 * Software rights: http://www.antlr.org/license.html
 *
 * $Id: //depot/code/org.antlr/release/antlr-2.7.4/antlr/Grammar.java#1 $
 */

typedef MLANG_HashMap<MLANG_String, MGrammar_RuleSymbol_var> SYMBOLS;
typedef MLANG_HashMap<MLANG_String, MParser_Token_var> OPTIONS;

/**A Grammar holds a set of rules (which are stored
 * in a symbol table).  Most of the time a grammar
 * needs a code generator and an LLkAnalyzer too.
 */
class MGrammar_Grammar : public MLANG_Refcountable<ACE_SYNCH_MUTEX>

{
public:
    typedef MLANG_Vector<MGrammar_RuleSymbol_var> RULES;
public:
    MGrammar_Tool_ptr antlrTool_;
    MGrammar_CodeGenerator_var generator_;
    MGrammar_LLkGrammarAnalyzer_var theLLkAnalyzer_;
    SYMBOLS symbols_;
    bool buildAST_;
    bool analyzerDebug_;
    bool interactive_;
    MLANG_String superClass_;

    /** The token manager associated with the grammar, if any.
     // The token manager is responsible for maintaining the set of valid tokens, and
     // is conceptually shared between the lexer and parser.  This may be either a
     // LexerGrammar or a ImportVocabTokenManager.
     */
    MGrammar_TokenManager_var tokenManager_;

    /** The name of the export vocabulary...used to generate the output
     *  token types interchange file.
     */
    MLANG_String exportVocab_;

    /** The name of the import vocabulary.  "Initial conditions"
     */
    MLANG_String importVocab_;

    // Mapping from String keys to Token option values
    OPTIONS options_;
    // Vector of RuleSymbol entries
    RULES rules_;

    MParser_Token_var preambleAction_;
    MLANG_String className_;
    MLANG_String fileName_;
    MParser_Token_var classMemberAction_;
    bool hasSyntacticPredicate_;
    bool hasUserErrorHandling_;

    // max lookahead that can be attempted for this parser.
    int maxk_;

    // options
    bool traceRules_;
    bool debuggingOutput_;
    bool defaultErrorHandler_;

    MLANG_String comment_;

public:

    MGrammar_Grammar (
       const MLANG_String& className,
       MGrammar_Tool_ptr tool,
       const MLANG_String& superClass
     );
    virtual ~MGrammar_Grammar();

public:

   static bool is_nil (MGrammar_Grammar_ptr g)
   { return g == 0; }

   static MGrammar_Grammar_ptr _nil()
   { return 0; }
   
   static MGrammar_Grammar_ptr _duplicate(MGrammar_Grammar_ptr g)
   { return MLANG_acquire (g); }

public:

    int maxk() const
    { return maxk_; }

    const MLANG_String& superClass() const
    { return superClass_; }

    const MParser_Token_var& classMemberAction() const
    { return classMemberAction_; }

    const OPTIONS& options() const
    { return options_; }

    bool buildAST() const
    { return buildAST_; }

    virtual bool isLexer() const { return false; }

    virtual bool isParser() const { return false; }

    virtual bool isTreeWalker() const { return false; }

    bool traceRules() const
    { return traceRules_; }

    bool debuggingOutput() const
    { return debuggingOutput_; }

    bool hasSyntacticPredicate() const
    { return hasSyntacticPredicate_; }

    const RULES& rules() const
    { return rules_; }

    /** Define a rule */
    virtual void define(const MGrammar_RuleSymbol_var& rs);

    /** Top-level call to generate the code for this grammar */
    virtual void generate() = 0;

    const MLANG_String& exportVocab() const
    { return exportVocab_; }

    void exportVocab(const MLANG_String& value) 
    { exportVocab_ = value; }

    const MLANG_String& importVocab() const
    { return importVocab_; }

    void importVocab(const MLANG_String& value)
    { importVocab_ = value; }

    const MLANG_String getClassName() const
    {
        return className_;
    }

    /* Does this grammar have a default error handler? */
    bool getDefaultErrorHandler() const
    {
        return defaultErrorHandler_;
    }

    const MLANG_String& getFilename() const
    {
        return fileName_;
    }

    /** Get an integer option.  Given the name of the option find its
     * associated integer value.  If the associated value is not an integer or
     * is not in the table, then throw an exception of type NumberFormatException.
     * @param key The name of the option
     * @return The value associated with the key.
     */
    virtual int getIntegerOption(const MLANG_String& key);

    /** Get an option.  Given the name of the option find its associated value.
     * @param key The name of the option
     * @return The value associated with the key, or null if the key has not been set.
     */
    virtual MParser_Token_var getOption(const MLANG_String& key);

    // Get name of class from which generated parser/lexer inherits
    virtual const MLANG_String getSuperClass() const = 0;

    virtual MGrammar_GrammarSymbol_var getSymbol(const MLANG_String& s);

    virtual SYMBOLS::iterator getSymbols();

    /** Check the existence of an option in the table
     * @param key The name of the option
     * @return true if the option is in the table
     */
    virtual bool hasOption(const MLANG_String& key) const;

    /** Is a rule symbol defined? (not used for tokens) */
    virtual bool isDefined(const MLANG_String& s) const;

    /**Process command line arguments.  Implemented in subclasses */
    virtual void processArguments(ACE_ARGV& args) = 0;

    virtual void setCodeGenerator(const MGrammar_CodeGenerator_var& gen)
    {
        generator_ = gen;
    }

    virtual void setFilename(const MLANG_String& s)
    {
        fileName_ = s;
    }

    virtual void setGrammarAnalyzer(const MGrammar_LLkGrammarAnalyzer_var& a)
    {
        theLLkAnalyzer_ = a;
    }

    virtual const MParser_Token_var& preambleAction() const
    { return preambleAction_; }

    virtual MGrammar_Grammar& preambleAction(const MParser_Token_var& preamble)
    { preambleAction_ = preamble; return *this; }

    /** Set a generic option.
     * This associates a generic option key with a Token value.
     * No validation is performed by this method, although users of the value
     * (code generation and/or analysis) may require certain formats.
     * The value is stored as a token so that the location of an error
     * can be reported.
     * @param key The name of the option.
     * @param value The value to associate with the key.
     * @return true if the option was a valid generic grammar option, false o/w
     */
    virtual bool setOption(const MLANG_String& key, const MParser_Token_var& value);

    virtual void setTokenManager(const MGrammar_TokenManager_var& tokenManager)
    {
        tokenManager_ = tokenManager;
    }

    const MLANG_String& comment() const
    { return comment_; }

    MGrammar_Grammar& comment(const MLANG_String& value)
    { comment_ = value; return *this; }

    /** Print out the grammar without actions */
    MLANG_String toString() const;
};


#endif /* __MGRAMMAR_GRAMMAR_H__ */
