#include "MParser/System.h"
#include "MGrammar/JavaCharFormatter.h"
#include "MGrammar/Alternative.h"
#include "MGrammar/WildcardElement.h"
#include "MGrammar/CharLiteralElement.h"
#include "MGrammar/StringLiteralElement.h"
#include "MGrammar/CharRangeElement.h"
#include "MGrammar/TokenRangeElement.h"
#include "MGrammar/BlockEndElement.h"
#include "MGrammar/AlternativeElement.h"
#include "MGrammar/TreeElement.h"
#include "MGrammar/ActionElement.h"
#include "MGrammar/OneOrMoreBlock.h"
#include "MGrammar/ZeroOrMoreBlock.h"
#include "MGrammar/SynPredBlock.h"
#include "MGrammar/Grammar.h"
#include "MGrammar/LexerGrammar.h"
#include "MGrammar/CodeGenerator.h"
#include "MGrammar/LLkAnalyzer.h"

/* ANTLR Translator Generator
 * Project led by Terence Parr at http://www.jGuru.com
 * Software rights: http://www.antlr.org/license.html
 *
 * $Id: //depot/code/org.antlr/release/antlr-2.7.4/antlr/LLkAnalyzer.java#1 $
 */

/**A linear-approximate LL(k) grammar analzyer.
 *
 * All lookahead elements are sets of token types.
 *
 * @author  Terence Parr, John Lilley
 * @see     antlr.Grammar
 * @see     antlr.Lookahead
 */
    // Set "analyzerDebug" to true

/** Create an LLk analyzer */
MGrammar_LLkAnalyzer::MGrammar_LLkAnalyzer(MGrammar_Tool_ptr tool)
   : DEBUG_ANALYZER(false),
     tool_(tool),
     grammar_(0),
     lexicalAnalysis_(false),
     charFormatter_(new MGrammar_JavaCharFormatter)

{
}

/** Return true if someone used the '.' wildcard default idiom.
 *  Either #(. children) or '.' as an alt by itself.
 */
bool MGrammar_LLkAnalyzer::altUsesWildcardDefault(MGrammar_Alternative_ptr alt)

{
   MGrammar_AlternativeElement_var head = alt->head_;
   // if element is #(. blah) then check to see if el is root
   if (instanceof<MGrammar_TreeElement>(head) != 0
   &&  instanceof<MGrammar_WildcardElement>
          (Dynamic_Cast<MGrammar_TreeElement>(head)->root_))
    {
      return true;
    }
   if (instanceof<MGrammar_WildcardElement>(head)
   &&  instanceof<MGrammar_BlockEndElement>(head->next_))
    {
      return true;
    }
   return false;
}

/**Is this block of alternatives LL(k)?  Fill in alternative cache for this block.
 * @return true if the block is deterministic
 */
bool MGrammar_LLkAnalyzer::deterministic(MGrammar_AlternativeBlock_ptr blk)

{
   /** The lookahead depth for this decision */
   int k = 1;	// start at k=1
   if (DEBUG_ANALYZER) SYSTEM->out()->println("deterministic(" + blk->toString() + ")");
   bool det = true;
   int nalts = blk->alternatives_.size();
   MGrammar_AlternativeBlock_var saveCurrentBlock = currentBlock_;
   MGrammar_Alternative_var wildcardAlt = 0;
   currentBlock_ = MGrammar_AlternativeBlock::_duplicate(blk);

   /* don't allow nongreedy (...) blocks */
   if (!blk->greedy_
    && !(instanceof<MGrammar_OneOrMoreBlock>(blk))
    && !(instanceof<MGrammar_ZeroOrMoreBlock>(blk)))
    {
      tool_->warning("Being nongreedy only makes sense for (...)+ and (...)*", grammar_->getFilename(), blk->getLine(), blk->getColumn());
    }

   // SPECIAL CASE: only one alternative.  We don't need to check the
   // determinism, but other code expects the lookahead cache to be
   // set for the single alt.
   if (nalts == 1)
    {
      MGrammar_AlternativeElement_var e = blk->getAlternativeAt(0)->head_;
      currentBlock_->alti = 0;
      blk->getAlternativeAt(0)->cache_[1] = e->look(1);
      blk->getAlternativeAt(0)->lookaheadDepth_ = 1;	// set lookahead to LL(1)
      currentBlock_ = saveCurrentBlock;
      return true;	// always deterministic for one alt
    }

   for (int i = 0; i < nalts - 1; i++)
    {
       currentBlock_->alti = i;
       currentBlock_->analysisAlt = i;	// which alt are we analyzing?
       currentBlock_->altj = i + 1;		// reset this alt.  Haven't computed yet,
       // but we need the alt number.
       // compare against other alternatives with lookahead depth k
       for (int j = i + 1; j < nalts; j++)
        {
           currentBlock_->altj = j;
           if (DEBUG_ANALYZER) SYSTEM->out()->println("comparing " +MLANG_String::valueOf(i) + " against alt " +MLANG_String::valueOf(j));
           currentBlock_->analysisAlt = j;	// which alt are we analyzing?
           k = 1;	// always attempt minimum lookahead possible.

           // check to see if there is a lookahead depth that distinguishes
           // between alternatives i and j.
           ACE_Vector<MGrammar_Lookahead> r;
           r.resize(grammar_->maxk_ + 1, MGrammar_Lookahead());
           bool haveAmbiguity;
           do
            {
              haveAmbiguity = false;
              if (DEBUG_ANALYZER) SYSTEM->out()->println("checking depth " +MLANG_String::valueOf(k) + "<=" +MLANG_String::valueOf(grammar_->maxk_));
              MGrammar_Lookahead p,q;
              p = getAltLookahead(blk, i, k);
              q = getAltLookahead(blk, j, k);

              // compare LOOK(alt i) with LOOK(alt j).  Is there an intersection?
              // Lookahead must be disjoint.
              if (DEBUG_ANALYZER) SYSTEM->out()->println("p is " + p.toString(",", *charFormatter_.in(), grammar_));
              if (DEBUG_ANALYZER) SYSTEM->out()->println("q is " + q.toString(",", *charFormatter_.in(), grammar_));
              // r[i] = p.fset.and(q.fset);
              r[k] = p.intersection(q);
              if (DEBUG_ANALYZER) SYSTEM->out()->println("intersection at depth " +MLANG_String::valueOf(k) + " is " + r[k].toString());
              if (!r[k].nil())
               {
                 haveAmbiguity = true;
                 k++;
               }
              // go until no more lookahead to use or no intersection
            }
           while (haveAmbiguity && k <= grammar_->maxk_);

           MGrammar_Alternative_ptr ai = blk->getAlternativeAt(i);
           MGrammar_Alternative_ptr aj = blk->getAlternativeAt(j);
           if (haveAmbiguity)
            {
              det = false;
              ai->lookaheadDepth_ = NONDETERMINISTIC;
              aj->lookaheadDepth_ = NONDETERMINISTIC;

              /* if ith alt starts with a syntactic predicate, computing the
               * lookahead is still done for code generation, but messages
               * should not be generated when comparing against alt j.
               * Alternatives with syn preds that are unnecessary do
               * not result in syn pred try-blocks.
               */
              if (!MGrammar_SynPredBlock::is_nil(ai->synPred_))
               {
                 if (DEBUG_ANALYZER)
                  {
                    SYSTEM->out()->println("alt " +MLANG_String::valueOf(i) + " has a syn pred");
                  }
                 // The alt with the (...)=> block is nondeterministic for sure.
                 // If the (...)=> conflicts with alt j, j is nondeterministic.
                 // This prevents alt j from being in any switch statements.
                 // move on to next alternative=>no possible ambiguity!
                 //						continue inner;
               }

              /* if ith alt starts with a semantic predicate, computing the
               * lookahead is still done for code generation, but messages
               * should not be generated when comparing against alt j.
               */
              else if (ai->semPred_.length() != 0)
               {
                 if (DEBUG_ANALYZER)
                  {
                    SYSTEM->out()->println("alt " +MLANG_String::valueOf(i) + " has a sem pred");
                  }
               }

              /* if jth alt is exactly the wildcard or wildcard root of tree,
               * then remove elements from alt i lookahead from alt j's lookahead.
               * Don't do an ambiguity warning.
               */
              else if (altUsesWildcardDefault(aj))
               {
                 // SYSTEM->out()->println("removing pred sets");
                 // removeCompetingPredictionSetsFromWildcard(aj.cache, aj.head, grammar.maxk);
                 wildcardAlt = aj;
               }

              /* If the user specified warnWhenFollowAmbig=false, then we
               * can turn off this warning IFF one of the alts is empty;
               * that is, it points immediately at the end block.
               */
              else if (!blk->warnWhenFollowAmbig_ 
                   &&  (instanceof<MGrammar_BlockEndElement>(ai->head_)
                   ||  instanceof<MGrammar_BlockEndElement>(aj->head_)))
               {
                  // SYSTEM->out()->println("ai.head pts to "+ai.head.getClass());
                  // SYSTEM->out()->println("aj.head pts to "+aj.head.getClass());
               }

              /* If they have the generateAmbigWarnings option off for the block
               * then don't generate a warning.
               */
              else if (!blk->generateAmbigWarnings_)
               {
               }

              /* If greedy=true and *one* empty alt shut off warning. */
              else if (blk->greedySet_ && blk->greedy_
                   &&  ((instanceof<MGrammar_BlockEndElement>(ai->head_)
                   &&   !(instanceof<MGrammar_BlockEndElement>(aj->head_)))
                   ||   (instanceof<MGrammar_BlockEndElement>(aj->head_) &&
                        !(instanceof<MGrammar_BlockEndElement>(ai->head_)))))
               {
                  // SYSTEM->out()->println("greedy set to true; one alt empty");
               }


              /* We have no choice, but to report a nondetermism */
              else
               {
                 tool_->errorHandler_->warnAltAmbiguity(
                              grammar_,
                              blk, // the block
                              lexicalAnalysis_, // true if lexical
                              grammar_->maxk_, // depth of ambiguity
                              r, // set of linear ambiguities
                              i, // first ambiguous alternative
                              j				// second ambiguous alternative
                         );
               }
            }
           else
            {
              // a lookahead depth, k, was found where i and j do not conflict
              ai->lookaheadDepth_ = MAX(ai->lookaheadDepth_, k);
              aj->lookaheadDepth_ = MAX(aj->lookaheadDepth_, k);
            }
         }
       }

   // finished with block.

   // If had wildcard default clause idiom, remove competing lookahead
   /*
      if ( wildcardAlt!=null )
       {
         removeCompetingPredictionSetsFromWildcard(wildcardAlt.cache, wildcardAlt.head, grammar.maxk);
       }
    */

   currentBlock_ = saveCurrentBlock;
   return det;
}

/**Is (...)+ block LL(1)?  Fill in alternative cache for this block.
 * @return true if the block is deterministic
 */
bool MGrammar_LLkAnalyzer::deterministic(MGrammar_OneOrMoreBlock_ptr blk)

{
   if (DEBUG_ANALYZER) SYSTEM->out()->println("deterministic(...)+(" + blk->toString() + ")");
   MGrammar_AlternativeBlock_var saveCurrentBlock = currentBlock_;
   currentBlock_ = MGrammar_OneOrMoreBlock::_duplicate(blk);
   bool blkOk = deterministic((MGrammar_AlternativeBlock_ptr)blk);
   // block has been checked, now check that what follows does not conflict
   // with the lookahead of the (...)+ block.
   bool det = deterministicImpliedPath(blk);
   currentBlock_ = saveCurrentBlock;
   return det && blkOk;
}

/**Is (...)* block LL(1)?  Fill in alternative cache for this block.
 * @return true if the block is deterministic
 */
bool MGrammar_LLkAnalyzer::deterministic(MGrammar_ZeroOrMoreBlock_ptr blk)

{
   if (DEBUG_ANALYZER) SYSTEM->out()->println("deterministic(...)*(" + blk->toString() + ")");
   MGrammar_AlternativeBlock_var saveCurrentBlock = currentBlock_;
   currentBlock_ = MGrammar_ZeroOrMoreBlock::_duplicate(blk);
   bool blkOk = deterministic(Dynamic_Cast<MGrammar_AlternativeBlock>(blk));
   // block has been checked, now check that what follows does not conflict
   // with the lookahead of the (...)* block.
   bool det = deterministicImpliedPath(blk);
   currentBlock_ = saveCurrentBlock;
   return det && blkOk;
}

/**Is this (...)* or (...)+ block LL(k)?
 * @return true if the block is deterministic
 */
bool MGrammar_LLkAnalyzer::deterministicImpliedPath(MGrammar_BlockWithImpliedExitPath_ptr blk)

{
   /** The lookahead depth for this decision considering implied exit path */
   int k;
   bool det = true;
   ALTVEC alts = blk->getAlternatives();
   int nalts = alts.size();
   currentBlock_->altj = -1;	// comparing against implicit optional/exit alt

   if (DEBUG_ANALYZER) SYSTEM->out()->println("deterministicImpliedPath");
   for (int i = 0; i < nalts; i++) 
    {		// check follow against all alts
       MGrammar_Alternative_ptr alt = blk->getAlternativeAt(i);

       if (instanceof<MGrammar_BlockEndElement>(alt->head_))
        {
          tool_->warning("empty alternative makes no sense in (...)* or (...)+", grammar_->getFilename(), blk->getLine(), blk->getColumn());
        }

       k = 1;							// assume eac alt is LL(1) with exit branch
       // check to see if there is a lookahead depth that distinguishes
       // between alternative i and the exit branch.
       ACE_Vector<MGrammar_Lookahead> r;
       r.resize (grammar_->maxk_ + 1, MGrammar_Lookahead());
       bool haveAmbiguity;
       do
        {
          haveAmbiguity = false;
          if (DEBUG_ANALYZER) SYSTEM->out()->println("checking depth " +MLANG_String::valueOf(k) + "<=" +MLANG_String::valueOf(grammar_->maxk_));
          MGrammar_Lookahead p;
          MGrammar_Lookahead follow = blk->next_->look(k);
          blk->exitCache_[k] = follow;
          currentBlock_->alti = i;
          p = getAltLookahead(blk, i, k);

          if (DEBUG_ANALYZER) SYSTEM->out()->println("follow is " + follow.toString(",", *charFormatter_.in(), grammar_));
          if (DEBUG_ANALYZER) SYSTEM->out()->println("p is " + p.toString(",", *charFormatter_.in(), grammar_));
          //r[k] = follow.fset.and(p.fset);
          r[k] = follow.intersection(p);
          if (DEBUG_ANALYZER) SYSTEM->out()->println("intersection at depth " +MLANG_String::valueOf(k) + " is " + r[k].toString());
          if (!r[k].nil())
           {
             haveAmbiguity = true;
             k++;
           }
          // go until no more lookahead to use or no intersection
        }
       while (haveAmbiguity && k <= grammar_->maxk_);

       if (haveAmbiguity)
        {
          det = false;
          alt->lookaheadDepth_ = NONDETERMINISTIC;
          blk->exitLookaheadDepth_ = NONDETERMINISTIC;
          MGrammar_Alternative_ptr ambigAlt = blk->getAlternativeAt(currentBlock_->alti);

          /* If the user specified warnWhenFollowAmbig=false, then we
           * can turn off this warning.
           */
          if (!blk->warnWhenFollowAmbig_)
           {
           }

          /* If they have the generateAmbigWarnings option off for the block
           * then don't generate a warning.
           */
          else if (!blk->generateAmbigWarnings_)
           {
           }

          /* If greedy=true and alt not empty, shut off warning */
          else if (blk->greedy_
               &&  blk->greedySet_
               &&  !(instanceof<MGrammar_BlockEndElement>(ambigAlt->head_)))
           {
             if (DEBUG_ANALYZER) SYSTEM->out()->println("greedy loop");
           }

          /* If greedy=false then shut off warning...will have
           * to add "if FOLLOW break"
           * block during code gen to compensate for removal of warning.
           */
          else if (!blk->greedy_ &&
                    !(instanceof<MGrammar_BlockEndElement>(ambigAlt->head_)))
           {
             if (DEBUG_ANALYZER) SYSTEM->out()->println("nongreedy loop");
             // if FOLLOW not single k-string (|set[k]| can
             // be > 1 actually) then must warn them that
             // loop may terminate incorrectly.
             // For example, ('a'..'d')+ ("ad"|"cb")
             if (!lookaheadEquivForApproxAndFullAnalysis(blk->exitCache_, grammar_->maxk_))
              {
                ACE_Vector<MLANG_String> lines;
                lines.push_back("nongreedy block may exit incorrectly due");
                lines.push_back("\tto limitations of linear approximate lookahead (first k-1 sets");
                lines.push_back("\tin lookahead not singleton).");
                tool_->warning(lines,
                               grammar_->getFilename(), blk->getLine(), blk->getColumn());
                    }
              }

             // no choice but to generate a warning
             else
              {
                tool_->errorHandler_->warnAltExitAmbiguity(
                        grammar_,
                        blk, // the block
                        lexicalAnalysis_, // true if lexical
                        grammar_->maxk_, // depth of ambiguity
                        r, // set of linear ambiguities
                        i		// ambiguous alternative
                    );
              }
           }
          else
           {
             alt->lookaheadDepth_ = MAX(alt->lookaheadDepth_, k);
             blk->exitLookaheadDepth_ = MAX(blk->exitLookaheadDepth_, k);
           }
    }
   return det;
}

/**Compute the lookahead set of whatever follows references to
 * the rule associated witht the FOLLOW block.
 */
MGrammar_Lookahead MGrammar_LLkAnalyzer::FOLLOW(int k, MGrammar_RuleEndElement_ptr end)

{
   // what rule are we trying to compute FOLLOW of?
   MGrammar_RuleBlock_var rb = Dynamic_Cast<MGrammar_RuleBlock>(end->block_);
   // rule name is different in lexer
   MLANG_String rule;
   if (lexicalAnalysis_)
    {
      rule = MGrammar_CodeGenerator::encodeLexerRuleName(rb->getRuleName());
    }
   else
    {
      rule = rb->getRuleName();
    }

   if (DEBUG_ANALYZER) SYSTEM->out()->println("FOLLOW(" +MLANG_String::valueOf(k) + "," + rule + ")");

   // are we in the midst of computing this FOLLOW already?
   if (end->lock_[k])
    {
      if (DEBUG_ANALYZER) SYSTEM->out()->println("FOLLOW cycle to " + rule);
      return MGrammar_Lookahead(rule);
    }

   // Check to see if there is cached value
   if (!end->cache_[k].is_null())
    {
      if (DEBUG_ANALYZER)
       {
         SYSTEM->out()->println("cache entry FOLLOW(" +MLANG_String::valueOf(k) + ") for " + rule + ": " + end->cache_[k].toString(",", *charFormatter_.in(), grammar_));
       }
      // if the cache is a complete computation then simply return entry
      if (end->cache_[k].cycle().length() == 0)
       {
         return end->cache_[k];
       }
      // A cache entry exists, but it is a reference to a cyclic computation.
      MGrammar_RuleSymbol_var rs = 
           MGrammar_RuleSymbol::_narrow(grammar_->getSymbol(end->cache_[k].cycle_));
      MGrammar_RuleEndElement_var re = rs->getBlock()->endNode_;
      // The other entry may not exist because it is still being
      // computed when this cycle cache entry was found here.
      if (re->cache_[k].is_null())
       {
         // return the cycle...that's all we can do at the moment.
         return end->cache_[k];
       }
      else
       {
         if (DEBUG_ANALYZER)
          {
            SYSTEM->out()->println("combining FOLLOW(" +MLANG_String::valueOf(k) + ") for " +
                                rule + ": from "+
                                end->cache_[k].toString(",", *charFormatter_.in(), grammar_) + " with FOLLOW for "+
                                ((MGrammar_RuleBlock_ptr)re->block_.in())->getRuleName()+": "+
                                re->cache_[k].toString(",", *charFormatter_.in(), grammar_));
          }
         // combine results from other rule's FOLLOW
         if ( re->cache_[k].cycle_.length()==0 )
          {
            // current rule depends on another rule's FOLLOW and
            // it is complete with no cycle; just kill our cycle and
            // combine full result from other rule's FOLLOW
            end->cache_[k].combineWith(re->cache_[k]);
            end->cache_[k].cycle_.clear(); // kill cycle as we're complete
          }
         else
          {
            // the FOLLOW cache for other rule has a cycle also.
            // Here is where we bubble up a cycle.  We better recursively
            // wipe out cycles (partial computations).  I'm a little nervous
            // that we might leave a cycle here, however.
            MGrammar_Lookahead refFOLLOW = FOLLOW(k, re);
            end->cache_[k].combineWith( refFOLLOW );
            // all cycles should be gone, but if not, record ref to cycle
            end->cache_[k].cycle_ = refFOLLOW.cycle_;
          }
         if (DEBUG_ANALYZER)
          {
            SYSTEM->out()->println("saving FOLLOW(" +MLANG_String::valueOf(k) + ") for " + rule + ": from "+end->cache_[k].toString(",", *charFormatter_.in(), grammar_));
          }
         // Return the updated cache entry associated
         // with the cycle reference.
         return end->cache_[k];
       }
     }

    end->lock_[k] = true;	// prevent FOLLOW computation cycles

    MGrammar_Lookahead p;

    MGrammar_RuleSymbol_var rs = MGrammar_RuleSymbol::_narrow(grammar_->getSymbol(rule));

    // Walk list of references to this rule to compute FOLLOW
    for (size_t i = 0; i < rs->numReferences(); i++)
     {
        MGrammar_RuleRefElement_var rr = rs->getReference(i);
        if (DEBUG_ANALYZER) 
           SYSTEM->out()->println("next[" + rule + "] is " + 
                                  rr->next_->toString());
        MGrammar_Lookahead q = rr->next_->look(k);
        if (DEBUG_ANALYZER) 
           SYSTEM->out()->println("FIRST of next[" + rule + "] ptr is " + q.toString());
        /* If there is a cycle then if the cycle is to the rule for
         * this end block, you have a cycle to yourself.  Remove the
         * cycle indication--the lookahead is complete.
         */
        if (q.cycle_.length() != 0
        &&  q.cycle_.equals(rule))
         {
           q.cycle_.clear();	// don't want cycle to yourself!
         }
        // add the lookahead into the current FOLLOW computation set
        p.combineWith(q);
        if (DEBUG_ANALYZER) 
           SYSTEM->out()->println("combined FOLLOW[" + rule + "] is " + p.toString());
     }

    end->lock_[k] = false; // we're not doing FOLLOW anymore

    // if no rules follow this, it can be a start symbol or called by a start sym.
    // set the follow to be end of file.
    if (p.fset().nil() && p.cycle_.length() == 0)
     {
       if (grammar_->isTreeWalker())
        {
          // Tree grammars don't see EOF, they see end of sibling list or
          // "NULL TREE LOOKAHEAD".
          p.fset_.add(MParser_Token::NULL_TREE_LOOKAHEAD);
        }
       else if (grammar_->isLexer())
        {
          // Lexical grammars use Epsilon to indicate that the end of rule has been hit
          // EOF would be misleading; any character can follow a token rule not just EOF
          // as in a grammar (where a start symbol is followed by EOF).  There is no
          // sequence info in a lexer between tokens to indicate what is the last token
          // to be seen.
          // p.fset.add(EPSILON_TYPE);
          p.setEpsilon();
        }
       else
        {
          p.fset_.add(MParser_Token::EOF_TYPE);
        }
     }

    // Cache the result of the FOLLOW computation
    if (DEBUG_ANALYZER)
     {
       SYSTEM->out()->println("saving FOLLOW(" +MLANG_String::valueOf(k) + ") for " + 
                              rule + ": " + p.toString(",", *charFormatter_.in(), grammar_));
     }
    end->cache_[k] = p;

    return p;
}

MGrammar_Lookahead MGrammar_LLkAnalyzer::getAltLookahead (
   MGrammar_AlternativeBlock_ptr blk, 
   int alt, 
   int k
)

{
   MGrammar_Lookahead p;
   MGrammar_Alternative_ptr a = blk->getAlternativeAt(alt);
   MGrammar_AlternativeElement_var e = a->head_;
   //SYSTEM->out()->println("getAltLookahead("+k+","+e+"), cache size is "+a.cache.length);
   if (a->cache_[k].is_null())
    {
      p = e->look(k);
      a->cache_[k] = p;
    }
   else
    {
      p = a->cache_[k];
    }
   return p;
}

/**Actions are ignored */
MGrammar_Lookahead MGrammar_LLkAnalyzer::look (
   int k, 
   MGrammar_ActionElement_ptr action
)

{
   if (DEBUG_ANALYZER) 
      SYSTEM->out()->println("lookAction(" +MLANG_String::valueOf(k) + "," + action->toString() + ")");
   return action->next_->look(k);
}

/**Combine the lookahead computed for each alternative */
MGrammar_Lookahead MGrammar_LLkAnalyzer::look (
   int k, 
   MGrammar_AlternativeBlock_ptr blk
)

{
   if (DEBUG_ANALYZER) SYSTEM->out()->println("enter lookAltBlk(" +MLANG_String::valueOf(k) + "," + blk->toString() + ")");
   MGrammar_AlternativeBlock_var saveCurrentBlock = currentBlock_;
   currentBlock_ = MGrammar_AlternativeBlock::_duplicate(blk);
   MGrammar_Lookahead p;
   for (size_t i = 0; i < blk->alternatives_.size(); i++)
    {
       if (DEBUG_ANALYZER) SYSTEM->out()->println("alt " +MLANG_String::valueOf(i) + " of " + blk->toString());
       // must set analysis alt
       currentBlock_->analysisAlt = i;
       MGrammar_Alternative_ptr alt = blk->getAlternativeAt(i);
       MGrammar_AlternativeElement_var elem = alt->head_;
       if (DEBUG_ANALYZER)
        {
          if (alt->head_.in() == alt->tail_.in())
           {
             SYSTEM->out()->println("alt " +MLANG_String::valueOf(i) + " is empty");
           }
        }
       MGrammar_Lookahead q = elem->look(k);
       p.combineWith(q);
    }
   if (k == 1 
   &&  blk->no_
   &&  subruleCanBeInverted(blk, lexicalAnalysis_))
    {
      // Invert the lookahead set
      if (lexicalAnalysis_)
       {
         MParser_BitSet b = ((MGrammar_LexerGrammar_ptr)grammar_)->charVocabulary_;
         ACE_Vector<size_t> elems = p.fset().toArray();
         for (size_t j = 0; j < elems.size(); j++)
          {
             b.remove(elems[j]);
          }
         p.fset(b);
       }
      else
       {
         p.fset_.notInPlace(MParser_Token::MIN_USER_TYPE, grammar_->tokenManager_->maxTokenType());
       }
    }
   if (DEBUG_ANALYZER) SYSTEM->out()->println("leave lookAltBlk(" +MLANG_String::valueOf(k) + "," + blk->toString() + ")=" + p.toString());
   currentBlock_ = saveCurrentBlock;
   return p;
}

/**Compute what follows this place-holder node and possibly
 * what begins the associated loop unless the
 * node is locked.
 * <p>
 * if we hit the end of a loop, we have to include
 * what tokens can begin the loop as well.  If the start
 * node is locked, then we simply found an empty path
 * through this subrule while analyzing it.  If the
 * start node is not locked, then this node was hit
 * during a FOLLOW operation and the FIRST of this
 * block must be included in that lookahead computation.
 */
MGrammar_Lookahead MGrammar_LLkAnalyzer::look (
   int k,
   MGrammar_BlockEndElement_ptr end
)

{
   assert(k<end->lock_.size());
   if (DEBUG_ANALYZER) 
      SYSTEM->out()->println("lookBlockEnd(" +MLANG_String::valueOf(k) + ", " + end->block_->toString() + "); lock is " +MLANG_String::valueOf(end->lock_[k]));
   if (end->lock_[k])
    {
      // computation in progress => the tokens we would have
      // computed (had we not been locked) will be included
      // in the set by that computation with the lock on this
      // node.
      return MGrammar_Lookahead();
    }

   MGrammar_Lookahead p;

   /* Hitting the end of a loop means you can see what begins the loop */
   if (instanceof<MGrammar_ZeroOrMoreBlock>(end->block_)
   ||  instanceof<MGrammar_OneOrMoreBlock>(end->block_))
    {
      // compute what can start the block,
      // but lock end node so we don't do it twice in same
      // computation.
      end->lock_[k] = true;
      p = look(k, end->block_);
      end->lock_[k] = false;
    }
   else
    {
      p = MGrammar_Lookahead();
    }

   /* Tree blocks do not have any follow because they are children
    * of what surrounds them.  For example, A #(B C) D results in
    * a look() for the TreeElement end of NULL_TREE_LOOKAHEAD, which
    * indicates that nothing can follow the last node of tree #(B C)
    */
   if (instanceof<MGrammar_TreeElement>(end->block_))
    {
      p.combineWith(MGrammar_Lookahead::of(MParser_Token::NULL_TREE_LOOKAHEAD));
    }

   /* Syntactic predicates such as ( (A)? )=> have no follow per se.
    * We cannot accurately say what would be matched following a
    * syntactic predicate (you MIGHT be ok if you said it was whatever
    * followed the alternative predicted by the predicate).  Hence,
    * (like end-of-token) we return Epsilon to indicate "unknown
    * lookahead."
    */
   else if (instanceof<MGrammar_SynPredBlock>(end->block_))
    {
      p.setEpsilon();
    }

   // compute what can follow the block
   else
    {
      MGrammar_Lookahead q = end->block_->next_->look(k);
      p.combineWith(q);
    }

   return p;
}

/**Return this char as the lookahead if k=1.
 * <p>### Doesn't work for ( 'a' 'b' | 'a' ~'b' ) yet!!!
 * <p>
 * If the atom has the <tt>not</tt> flag on, then
 * create the set complement of the tokenType
 * which is the set of all characters referenced
 * in the grammar with this char turned off.
 * Also remove characters from the set that
 * are currently allocated for predicting
 * previous alternatives.  This avoids ambiguity
 * messages and is more properly what is meant.
 * ( 'a' | ~'a' ) implies that the ~'a' is the
 * "else" clause.
 * <p>
 * NOTE: we do <b>NOT</b> include exit path in
 * the exclusion set. E.g.,
 * ( 'a' | ~'a' )* 'b'
 * should exit upon seeing a 'b' during the loop.
 */
MGrammar_Lookahead MGrammar_LLkAnalyzer::look (
    int k, 
   MGrammar_CharLiteralElement_ptr atom
)

{
   if (DEBUG_ANALYZER) 
      SYSTEM->out()->println("lookCharLiteral(" +MLANG_String::valueOf(k) + "," + atom->toString() + ")");
   // Skip until analysis hits k==1
   if (k > 1)
    {
      return atom->next_->look(k - 1);
    }
   if (lexicalAnalysis_)
    {
      if (atom->getNo())
       {
         MParser_BitSet b = ((MGrammar_LexerGrammar_ptr)grammar_)->charVocabulary_;
         if (DEBUG_ANALYZER) SYSTEM->out()->println("charVocab is " + b.toString());
         // remove stuff predicted by preceding alts and follow of block
         removeCompetingPredictionSets(b, atom);
         if (DEBUG_ANALYZER) SYSTEM->out()->println("charVocab after removal of prior alt lookahead " + b.toString());
         // now remove element that is stated not to be in the set
         b.clear(atom->getType());
         return MGrammar_Lookahead(b);
       }
      else
       {
         return MGrammar_Lookahead::of(atom->getType());
       }
    }
   else
    {
      // Should have been avoided by MakeGrammar
      tool_->panic("Character literal reference found in parser");
            // ... so we make the compiler happy
      return MGrammar_Lookahead::of(atom->getType());
    }
}

MGrammar_Lookahead MGrammar_LLkAnalyzer::look (
   int k,
   MGrammar_CharRangeElement_ptr r
)

{
   if (DEBUG_ANALYZER) 
      SYSTEM->out()->println("lookCharRange(" +MLANG_String::valueOf(k) + "," + r->toString() + ")");
   // Skip until analysis hits k==1
   if (k > 1)
    {
      return r->next_->look(k - 1);
    }
   MParser_BitSet p = MParser_BitSet::of(r->begin_);
   for (int i = r->begin_ + 1; i <= r->end_; i++)
    {
       p.add(i);
    }
   return MGrammar_Lookahead(p);
}

MGrammar_Lookahead MGrammar_LLkAnalyzer::look (
   int k,
   MGrammar_GrammarAtom_ptr atom
)

{
   if (DEBUG_ANALYZER) 
      SYSTEM->out()->println("look(" +MLANG_String::valueOf(k) + "," + atom->toString() + "[" +MLANG_String::valueOf(atom->getType()) + "])");

   if (lexicalAnalysis_)
    {
      // MakeGrammar should have created a rule reference instead
      tool_->panic("token reference found in lexer");
    }
   // Skip until analysis hits k==1
   if (k > 1)
    {
      return atom->next_->look(k - 1);
    }
   MGrammar_Lookahead l = MGrammar_Lookahead::of(atom->getType());
   if (atom->getNo())
    {
      // Invert the lookahead set against the token vocabulary
      int maxToken = grammar_->tokenManager_->maxTokenType();
      l.fset_.notInPlace(MParser_Token::MIN_USER_TYPE, maxToken);
      // remove stuff predicted by preceding alts and follow of block
      removeCompetingPredictionSets(l.fset_, atom);
    }
   return l;
}

/**The lookahead of a (...)+ block is the combined lookahead of
 * all alternatives and, if an empty path is found, the lookahead
 * of what follows the block.
 */
MGrammar_Lookahead MGrammar_LLkAnalyzer::look (
   int k,
   MGrammar_OneOrMoreBlock_ptr blk
)

{
   if (DEBUG_ANALYZER) SYSTEM->out()->println("look+" +MLANG_String::valueOf(k) + "," + blk->toString() + ")");
   MGrammar_Lookahead p = look(k, (MGrammar_AlternativeBlock_ptr)blk);
   return p;
}

/**Combine the lookahead computed for each alternative.
 * Lock the node so that no other computation may come back
 * on itself--infinite loop.  This also implies infinite left-recursion
 * in the grammar (or an error in this algorithm ;)).
 */
MGrammar_Lookahead MGrammar_LLkAnalyzer::look (
   int k,
   MGrammar_RuleBlock_ptr blk
)

{
   if (DEBUG_ANALYZER) SYSTEM->out()->println("enter lookRuleBlk(" +MLANG_String::valueOf(k) + "," + blk->toString() + ")");
   MGrammar_Lookahead p = look(k, (MGrammar_AlternativeBlock_ptr)blk);
   if (DEBUG_ANALYZER) SYSTEM->out()->println("leave lookRuleBlk(" +MLANG_String::valueOf(k) + "," + blk->toString() + ")=" + p.toString());
   return p;
}

/**If not locked or noFOLLOW set, compute FOLLOW of a rule.
 * <p>
 * TJP says 8/12/99: not true anymore:
 * Lexical rules never compute follow.  They set epsilon and
 * the code generator gens code to check for any character.
 * The code generator must remove the tokens used to predict
 * any previous alts in the same block.
 * <p>
 * When the last node of a rule is reached and noFOLLOW,
 * it implies that a "local" FOLLOW will be computed
 * after this call.  I.e.,
 * <pre>
 *		a : b A;
 *		b : B | ;
 *		c : b C;
 * </pre>
 * Here, when computing the look of rule b from rule a,
 * we want only {B,EPSILON_TYPE} so that look(b A) will
 * be {B,A} not {B,A,C}.
 * <p>
 * if the end block is not locked and the FOLLOW is
 * wanted, the algorithm must compute the lookahead
 * of what follows references to this rule.  If
 * end block is locked, FOLLOW will return an empty set
 * with a cycle to the rule associated with this end block.
 */
MGrammar_Lookahead MGrammar_LLkAnalyzer::look (
   int k,
   MGrammar_RuleEndElement_ptr end
)

{
   if (DEBUG_ANALYZER)
      SYSTEM->out()->println("lookRuleBlockEnd(" +MLANG_String::valueOf(k) + "); noFOLLOW=" +
                   MLANG_String::valueOf(end->noFOLLOW_) + "; lock is " +MLANG_String::valueOf(end->lock_[k]));
   if (/*lexicalAnalysis ||*/ end->noFOLLOW_)
    {
      MGrammar_Lookahead p;
      p.setEpsilon();
      p.hasEpsilonDepth_ = true;
      p.epsilonDepth_ = MParser_BitSet::of(k);
      return p;
    }
   MGrammar_Lookahead p = FOLLOW(k, end);
   return p;
}

/**Compute the lookahead contributed by a rule reference.
 *
 * <p>
 * When computing ruleref lookahead, we don't want the FOLLOW
 * computation done if an empty path exists for the rule.
 * The FOLLOW is too loose of a set...we want only to
 * include the "local" FOLLOW or what can follow this
 * particular ref to the node.  In other words, we use
 * context information to reduce the complexity of the
 * analysis and strengthen the parser.
 *
 * The noFOLLOW flag is used as a means of restricting
 * the FOLLOW to a "local" FOLLOW.  This variable is
 * orthogonal to the <tt>lock</tt> variable that prevents
 * infinite recursion.  noFOLLOW does not care about what k is.
 */
MGrammar_Lookahead MGrammar_LLkAnalyzer::look (
   int k,
   MGrammar_RuleRefElement_ptr rr
)

{
   if (DEBUG_ANALYZER) SYSTEM->out()->println("lookRuleRef(" +MLANG_String::valueOf(k) + "," + rr->toString() + ")");
   MGrammar_RuleSymbol_var rs = Dynamic_Cast<MGrammar_RuleSymbol>(grammar_->getSymbol(rr->targetRule()));
   if (MGrammar_RuleSymbol::is_nil(rs) || !rs->isDefined())
    {
      tool_->error("no definition of rule " + rr->targetRule(), grammar_->getFilename(), rr->getLine(), rr->getColumn());
      return MGrammar_Lookahead();
    }
   MGrammar_RuleBlock_var rb = rs->getBlock();
   MGrammar_RuleEndElement_var end = rb->endNode_;
   bool saveEnd = end->noFOLLOW();
   end->noFOLLOW(true);
   // go off to the rule and get the lookahead (w/o FOLLOW)
   MGrammar_Lookahead p = look(k, rr->targetRule());
   if (DEBUG_ANALYZER) SYSTEM->out()->println("back from rule ref to " + rr->targetRule());
   // restore state of end block
   end->noFOLLOW(saveEnd);

   // check for infinite recursion.  If a cycle is returned: trouble!
   if (p.cycle().length() != 0)
    {
      tool_->error("infinite recursion to rule " + p.cycle() + " from rule " +
                 rr->enclosingRuleName(), grammar_->getFilename(), rr->getLine(), rr->getColumn());
    }

   // is the local FOLLOW required?
   if (p.containsEpsilon())
    {
      if (DEBUG_ANALYZER)
         SYSTEM->out()->println("rule ref to " +
                                rr->targetRule() + " has eps, depth: " + p.epsilonDepth_.toString());

      // remove epsilon
      p.resetEpsilon();
      // fset.clear(EPSILON_TYPE);

      // for each lookahead depth that saw epsilon
      ACE_Vector<size_t> depths = p.epsilonDepth_.toArray();
      assert(depths.size()>0);
      p.hasEpsilonDepth_ = false;
      p.epsilonDepth_ = MParser_BitSet();
      for (size_t i = 0; i < depths.size(); i++)
       {
          int rk = k - (k - depths[i]);
          fprintf(stderr,"epsilonDepth: rule=%s rk=%d,k=%d, depths[%d]=%d\n",
                          rr->targetRule().c_str(), rk, k, i, depths[i]);
          MGrammar_Lookahead q = rr->next_->look(rk);	// see comments in Lookahead
          p.combineWith(q);
       }
      // note: any of these look() computations for local follow can
      // set EPSILON in the set again if the end of this rule is found.
    }

   return p;
}

MGrammar_Lookahead MGrammar_LLkAnalyzer::look (
   int k,
   MGrammar_StringLiteralElement_ptr atom
)

{
   if (DEBUG_ANALYZER) SYSTEM->out()->println("lookStringLiteral(" +MLANG_String::valueOf(k) + "," + atom->toString() + ")");
   if (lexicalAnalysis_)
    {
      // need more lookahead than string can provide?
      if (k > (int)atom->processedAtomText().length())
       {
         return atom->next_->look(k - atom->processedAtomText().length());
       }
      else
       {
         // get char at lookahead depth k, from the processed literal text
         return MGrammar_Lookahead::of(atom->processedAtomText().charAt(k - 1));
       }
    }
   else
    {
      // Skip until analysis hits k==1
      if (k > 1)
       {
         return atom->next_->look(k - 1);
       }
      MGrammar_Lookahead l = MGrammar_Lookahead::of(atom->getType());
      if (atom->getNo())
       {
         // Invert the lookahead set against the token vocabulary
         int maxToken = grammar_->tokenManager_->maxTokenType();
         l.fset_.notInPlace(MParser_Token::MIN_USER_TYPE, maxToken);
       }
      return l;
    }
}

/**The lookahead of a (...)=> block is the lookahead of
 * what follows the block.  By definition, the syntactic
 * predicate block defies static analysis (you want to try it
 * out at run-time).  The LOOK of (a)=>A B is A for LL(1)
 * ### is this even called?
 */
MGrammar_Lookahead MGrammar_LLkAnalyzer::look (
   int k,
   MGrammar_SynPredBlock_ptr blk
)

{
   if (DEBUG_ANALYZER) 
      SYSTEM->out()->println("look=>(" +MLANG_String::valueOf(k) + "," + blk->toString() + ")");
   return blk->next_->look(k);
}

MGrammar_Lookahead MGrammar_LLkAnalyzer::look (
   int k,
   MGrammar_TokenRangeElement_ptr r
)

{
   if (DEBUG_ANALYZER) 
      SYSTEM->out()->println("lookTokenRange(" +MLANG_String::valueOf(k) + "," + r->toString() + ")");
   // Skip until analysis hits k==1
   if (k > 1)
    {
      return r->next_->look(k - 1);
    }
   MParser_BitSet p = MParser_BitSet::of(r->begin());
   for (int i = r->begin() + 1; i <= r->end(); i++)
    {
       p.add(i);
    }
   return MGrammar_Lookahead(p);
}

MGrammar_Lookahead MGrammar_LLkAnalyzer::look (
   int k,
   MGrammar_TreeElement_ptr t
)

{
   if (DEBUG_ANALYZER)
      SYSTEM->out()->println("look(" +MLANG_String::valueOf(k) + "," + t->root_->toString() + "[" +MLANG_String::valueOf(t->root_->getType()) + "])");
   if (k > 1)
    {
      return t->next_->look(k - 1);
    }
   MGrammar_Lookahead l;
   l.set_null();
   if (instanceof<MGrammar_WildcardElement>(t->root_))
    {
      l = t->root_->look(1); // compute FIRST set minus previous rows
    }
   else
    {
      l = MGrammar_Lookahead::of(t->root_->getType());
      if (t->root_->getNo())
       {
         // Invert the lookahead set against the token vocabulary
         int maxToken = grammar_->tokenManager_->maxTokenType();
         l.fset_.notInPlace(MParser_Token::MIN_USER_TYPE, maxToken);
       }
    }
   return l;
}

MGrammar_Lookahead MGrammar_LLkAnalyzer::look (
   int k,
   MGrammar_WildcardElement_ptr wc
)

{
   if (DEBUG_ANALYZER) SYSTEM->out()->println("look(" +MLANG_String::valueOf(k) + "," + wc->toString() + ")");

   // Skip until analysis hits k==1
   if (k > 1)
    {
      return wc->next_->look(k - 1);
    }

   MParser_BitSet b;
   if (lexicalAnalysis_)
    {
      // Copy the character vocabulary
      b = ((MGrammar_LexerGrammar_ptr)grammar_)->charVocabulary_;
    }
   else
    {
      b = MParser_BitSet(1);
      // Invert the lookahead set against the token vocabulary
      int maxToken = grammar_->tokenManager_->maxTokenType();
      b.notInPlace(MParser_Token::MIN_USER_TYPE, maxToken);
      if (DEBUG_ANALYZER) 
         SYSTEM->out()->println("look(" +MLANG_String::valueOf(k) + "," + wc->toString() + ") after not: " + b.toString());
    }

   // Remove prediction sets from competing alternatives
   // removeCompetingPredictionSets(b, wc);

   return MGrammar_Lookahead(b);
}

/** The (...)* element is the combined lookahead of the alternatives and what can
 *  follow the loop.
 */
MGrammar_Lookahead MGrammar_LLkAnalyzer::look (
   int k,
   MGrammar_ZeroOrMoreBlock_ptr blk
)

{
   if (DEBUG_ANALYZER) 
      SYSTEM->out()->println("enter look*(" +MLANG_String::valueOf(k) + "," + blk->toString() + ")");
   MGrammar_Lookahead p = look(k, (MGrammar_AlternativeBlock_ptr)blk);
   MGrammar_Lookahead q = blk->next_->look(k);
   p.combineWith(q);
   if (DEBUG_ANALYZER) 
      SYSTEM->out()->println("leave look*(" +MLANG_String::valueOf(k) + "," + blk->toString() + ")=" + p.toString());
   return p;
}

/**Compute the combined lookahead for all productions of a rule.
 * If the lookahead returns with epsilon, at least one epsilon
 * path exists (one that consumes no tokens).  The noFOLLOW
 * flag being set for this endruleblk, indicates that the
 * a rule ref invoked this rule.
 *
 * Currently only look(RuleRef) calls this.  There is no need
 * for the code generator to call this.
 */
MGrammar_Lookahead MGrammar_LLkAnalyzer::look (
   int k,
   const MLANG_String& rule
)

{
   if (DEBUG_ANALYZER) 
      SYSTEM->out()->println("lookRuleName(" +MLANG_String::valueOf(k) + "," + rule + ")");
   MGrammar_RuleSymbol_var rs = Dynamic_Cast<MGrammar_RuleSymbol>(grammar_->getSymbol(rule));
   MGrammar_RuleBlock_var rb = rs->getBlock();

   if (rb->lock_[k])
    {
      if (DEBUG_ANALYZER)
         SYSTEM->out()->println("infinite recursion to rule " + rb->getRuleName());
      return MGrammar_Lookahead(rule);
    }

   // have we computed it before?
   if (!rb->cache_[k].is_null())
    {
      if (DEBUG_ANALYZER)
       {
         SYSTEM->out()->println("found depth " +MLANG_String::valueOf(k) + " result in FIRST " + rule + " cache: " +
                               rb->cache_[k].toString(",", *charFormatter_.in(), grammar_));
       }
      return rb->cache_[k];
    }

   rb->lock_[k] = true;
   MGrammar_Lookahead p = look(k, (MGrammar_RuleBlock_ptr)rb.in());
   rb->lock_[k] = false;

   // cache results
   rb->cache_[k] = p;
   if (DEBUG_ANALYZER)
    {
      SYSTEM->out()->println("saving depth " +MLANG_String::valueOf(k) + " result in FIRST " + rule + " cache: " +
                            rb->cache_[k].toString(",", *charFormatter_.in(), grammar_));
    }
   return p;
}

/** If the first k-1 sets are singleton sets, the appoximate
 *  lookahead analysis is equivalent to full lookahead analysis.
 */
bool MGrammar_LLkAnalyzer::lookaheadEquivForApproxAndFullAnalysis (
   const ACE_Vector<MGrammar_Lookahead>& bset,
   int k
)

{
   // first k-1 sets degree 1?
   for (int i = 1; i <= k - 1; i++)
    {
       MParser_BitSet look = bset[i].fset();
       if (look.degree() > 1)
        {
          return false;
        }
    }
   return true;
}

/** Remove the prediction sets from preceding alternatives
 * and follow set, but *only* if this element is the first element
 * of the alternative.  The class members currenBlock and
 * currentBlock.analysisAlt must be set correctly.
 * @param b The prediction bitset to be modified
 * @el The element of interest
 */
void MGrammar_LLkAnalyzer::removeCompetingPredictionSets (
   MParser_BitSet& b,
   MGrammar_AlternativeElement_ptr el
)

{
   // Only do this if the element is the first element of the alt,
   // because we are making an implicit assumption that k==1.
   MGrammar_GrammarElement_var head = 
        MGrammar_GrammarElement::_duplicate(currentBlock_->getAlternativeAt(currentBlock_->analysisAlt)->head_);
   // if element is #(. blah) then check to see if el is root
   if (instanceof<MGrammar_TreeElement>(head))
    {
      if (Dynamic_Cast<MGrammar_TreeElement>(head)->root_.in() != el)
       {
         return;
       }
    }
   else if (el != head.in())
    {
      return;
    }
   for (int i = 0; i < currentBlock_->analysisAlt; i++)
    {
       MGrammar_AlternativeElement_var e = currentBlock_->getAlternativeAt(i)->head_;
       b.subtractInPlace(e->look(1).fset());
    }
}

/** Remove the prediction sets from preceding alternatives
 * The class members currenBlock must be set correctly.
 * Remove prediction sets from 1..k.
 * @param look The prediction lookahead to be modified
 * @el The element of interest
 * @k  How deep into lookahead to modify
 */
void MGrammar_LLkAnalyzer::removeCompetingPredictionSetsFromWildcard (
   ACE_Vector<MGrammar_Lookahead>& look,
   MGrammar_AlternativeElement_ptr el,
   int k
)

{
   ACE_UNUSED_ARG(el);
   for (int d = 1; d <= k; d++)
    {
       for (int i = 0; i < currentBlock_->analysisAlt; i++)
        {
           MGrammar_AlternativeElement_var e = currentBlock_->getAlternativeAt(i)->head_;
           look[d].fset_.subtractInPlace(e->look(d).fset());
        }
    }
}

/** reset the analyzer so it looks like a new one */
void MGrammar_LLkAnalyzer::reset()

{
   grammar_ = 0;
   DEBUG_ANALYZER = false;
   currentBlock_ = 0;
   lexicalAnalysis_ = false;
}

    /** Set the grammar for the analyzer */
void MGrammar_LLkAnalyzer::setGrammar(MGrammar_Grammar_ptr g)

{
   if (grammar_ != 0)
    {
      reset();
    }
   grammar_ = g;

   // Is this lexical?
   lexicalAnalysis_ = grammar_->isLexer();
   DEBUG_ANALYZER = grammar_->analyzerDebug_;
   DEBUG_ANALYZER = true;
}

bool MGrammar_LLkAnalyzer::subruleCanBeInverted (
   MGrammar_AlternativeBlock_ptr blk,
   bool forLexer
)

{
   if (instanceof<MGrammar_ZeroOrMoreBlock>(blk)
   ||  instanceof<MGrammar_OneOrMoreBlock>(blk)
   ||  instanceof<MGrammar_SynPredBlock>(blk))
    {
       return false;
    }
   // Cannot invert an empty subrule
   if (blk->alternatives_.size() == 0)
    {
      return false;
    }
   // The block must only contain alternatives with a single element,
   // where each element is a char, token, char range, or token range.
   for (size_t i = 0; i < blk->alternatives_.size(); i++)
    {
       MGrammar_Alternative_ptr alt = blk->getAlternativeAt(i);
       // Cannot have anything interesting in the alternative ...
       if (!MGrammar_SynPredBlock::is_nil(alt->synPred_)
       ||  alt->semPred_.length() != 0 
       ||  !MGrammar_ExceptionSpec::is_nil(alt->exceptionSpec_)) 
        {
          return false;
        }
       // ... and there must be one simple element
       MGrammar_AlternativeElement_var elt = alt->head_;
       if (
           !(
             instanceof<MGrammar_CharLiteralElement>(elt) ||
             instanceof<MGrammar_TokenRefElement>(elt) ||
             instanceof<MGrammar_CharRangeElement>(elt) ||
             instanceof<MGrammar_TokenRangeElement>(elt) ||
             (instanceof<MGrammar_StringLiteralElement>(elt) && !forLexer)
             )
       ||
            !(instanceof<MGrammar_BlockEndElement>(elt->next_)) ||
              elt->getAutoGenType() != MGrammar_GrammarElement::AUTO_GEN_NONE
          )
         {
            return false;
         }
    }
   return true;
}

MGrammar_LLkAnalyzer::~MGrammar_LLkAnalyzer()

{
}
