#ifndef __MGRAMMAR_LLKANALYZER_H__
#define __MGRAMMAR_LLKANALYZER_H__
#include "MLANG/SmartPtr.h"
#include "MParser/BitSet.h"
#include "MGrammar/Lookahead.h"
#include "MParser/CharFormatter.h"

class MGrammar_LLkAnalyzer;
typedef class MGrammar_LLkAnalyzer *MGrammar_LLkAnalyzer_ptr;

#include "MGrammar/AlternativeBlock.h"
#include "MGrammar/BlockWithImpliedExitPath.h"
#include "MGrammar/LLkGrammarAnalyzer.h"
#include "MGrammar/Tool.h"

/* ANTLR Translator Generator
 * Project led by Terence Parr at http://www.jGuru.com
 * Software rights: http://www.antlr.org/license.html
 *
 * $Id: //depot/code/org.antlr/release/antlr-2.7.4/antlr/LLkAnalyzer.java#1 $
 */

/**A linear-approximate LL(k) grammar analzyer.
 *
 * All lookahead elements are sets of token types.
 *
 * @author  Terence Parr, John Lilley
 * @see     antlr.Grammar
 * @see     antlr.Lookahead
 */
class MGrammar_LLkAnalyzer : public MGrammar_LLkGrammarAnalyzer
{
public:
   // Set "analyzerDebug" to true
   bool DEBUG_ANALYZER;
private:
    MGrammar_AlternativeBlock_var currentBlock_;
protected:
    MGrammar_Tool_ptr tool_;
    MGrammar_Grammar_ptr grammar_;
    // True if analyzing a lexical grammar
    bool lexicalAnalysis_;
    // Used for formatting bit sets in default (Java) format
    MParser_CharFormatter_var charFormatter_;

public:
    /** Create an LLk analyzer */
    MGrammar_LLkAnalyzer(MGrammar_Tool_ptr tool);

    virtual ~MGrammar_LLkAnalyzer();

public:

   static bool is_nil (MGrammar_LLkAnalyzer_ptr a)
   { return a == 0; }

   static MGrammar_LLkAnalyzer_ptr _nil()
   { return 0; }

   static MGrammar_LLkAnalyzer_ptr _duplicate(MGrammar_LLkAnalyzer_ptr a)
   { return MLANG_acquire(a); }

protected:
    /** Return true if someone used the '.' wildcard default idiom.
     *  Either #(. children) or '.' as an alt by itself.
     */
    bool altUsesWildcardDefault(MGrammar_Alternative_ptr alt);

    /**Is this block of alternatives LL(k)?  Fill in alternative cache for this block.
     * @return true if the block is deterministic
     */
    bool deterministic(MGrammar_AlternativeBlock_ptr blk);

    /**Is (...)+ block LL(1)?  Fill in alternative cache for this block.
     * @return true if the block is deterministic
     */
    bool deterministic(MGrammar_OneOrMoreBlock_ptr blk);

    /**Is (...)* block LL(1)?  Fill in alternative cache for this block.
     * @return true if the block is deterministic
     */
    bool deterministic(MGrammar_ZeroOrMoreBlock_ptr blk);

    /**Is this (...)* or (...)+ block LL(k)?
     * @return true if the block is deterministic
     */
    bool deterministicImpliedPath(MGrammar_BlockWithImpliedExitPath_ptr blk);

public:
    /**Compute the lookahead set of whatever follows references to
     * the rule associated witht the FOLLOW block.
     */
    MGrammar_Lookahead FOLLOW(int k, MGrammar_RuleEndElement_ptr end);

private:

    MGrammar_Lookahead getAltLookahead(MGrammar_AlternativeBlock_ptr blk, int alt, int k);

public:

    /**Actions are ignored */
    MGrammar_Lookahead look(int k, MGrammar_ActionElement_ptr action);

    /**Combine the lookahead computed for each alternative */
    MGrammar_Lookahead look(int k, MGrammar_AlternativeBlock_ptr blk);

    /**Compute what follows this place-holder node and possibly
     * what begins the associated loop unless the
     * node is locked.
     * <p>
     * if we hit the end of a loop, we have to include
     * what tokens can begin the loop as well.  If the start
     * node is locked, then we simply found an empty path
     * through this subrule while analyzing it.  If the
     * start node is not locked, then this node was hit
     * during a FOLLOW operation and the FIRST of this
     * block must be included in that lookahead computation.
     */
    MGrammar_Lookahead look(int k, MGrammar_BlockEndElement_ptr end);

    /**Return this char as the lookahead if k=1.
     * <p>### Doesn't work for ( 'a' 'b' | 'a' ~'b' ) yet!!!
     * <p>
     * If the atom has the <tt>not</tt> flag on, then
     * create the set complement of the tokenType
     * which is the set of all characters referenced
     * in the grammar with this char turned off.
     * Also remove characters from the set that
     * are currently allocated for predicting
     * previous alternatives.  This avoids ambiguity
     * messages and is more properly what is meant.
     * ( 'a' | ~'a' ) implies that the ~'a' is the
     * "else" clause.
     * <p>
     * NOTE: we do <b>NOT</b> include exit path in
     * the exclusion set. E.g.,
     * ( 'a' | ~'a' )* 'b'
     * should exit upon seeing a 'b' during the loop.
     */
    MGrammar_Lookahead look(int k, MGrammar_CharLiteralElement_ptr atom);

    MGrammar_Lookahead look(int k, MGrammar_CharRangeElement_ptr r);

    MGrammar_Lookahead look(int k, MGrammar_GrammarAtom_ptr atom);

    /**The lookahead of a (...)+ block is the combined lookahead of
     * all alternatives and, if an empty path is found, the lookahead
     * of what follows the block.
     */
    MGrammar_Lookahead look(int k, MGrammar_OneOrMoreBlock_ptr blk);

    /**Combine the lookahead computed for each alternative.
     * Lock the node so that no other computation may come back
     * on itself--infinite loop.  This also implies infinite left-recursion
     * in the grammar (or an error in this algorithm ;)).
     */
    MGrammar_Lookahead look(int k, MGrammar_RuleBlock_ptr blk);

    /**If not locked or noFOLLOW set, compute FOLLOW of a rule.
     * <p>
     * TJP says 8/12/99: not true anymore:
     * Lexical rules never compute follow.  They set epsilon and
     * the code generator gens code to check for any character.
     * The code generator must remove the tokens used to predict
     * any previous alts in the same block.
     * <p>
     * When the last node of a rule is reached and noFOLLOW,
     * it implies that a "local" FOLLOW will be computed
     * after this call.  I.e.,
     * <pre>
     *		a : b A;
     *		b : B | ;
     *		c : b C;
     * </pre>
     * Here, when computing the look of rule b from rule a,
     * we want only {B,EPSILON_TYPE} so that look(b A) will
     * be {B,A} not {B,A,C}.
     * <p>
     * if the end block is not locked and the FOLLOW is
     * wanted, the algorithm must compute the lookahead
     * of what follows references to this rule.  If
     * end block is locked, FOLLOW will return an empty set
     * with a cycle to the rule associated with this end block.
     */
    MGrammar_Lookahead look(int k, MGrammar_RuleEndElement_ptr end);

    /**Compute the lookahead contributed by a rule reference.
     *
     * <p>
     * When computing ruleref lookahead, we don't want the FOLLOW
     * computation done if an empty path exists for the rule.
     * The FOLLOW is too loose of a set...we want only to
     * include the "local" FOLLOW or what can follow this
     * particular ref to the node.  In other words, we use
     * context information to reduce the complexity of the
     * analysis and strengthen the parser.
     *
     * The noFOLLOW flag is used as a means of restricting
     * the FOLLOW to a "local" FOLLOW.  This variable is
     * orthogonal to the <tt>lock</tt> variable that prevents
     * infinite recursion.  noFOLLOW does not care about what k is.
     */
    MGrammar_Lookahead look(int k, MGrammar_RuleRefElement_ptr rr);

    MGrammar_Lookahead look(int k, MGrammar_StringLiteralElement_ptr atom);

    /**The lookahead of a (...)=> block is the lookahead of
     * what follows the block.  By definition, the syntactic
     * predicate block defies static analysis (you want to try it
     * out at run-time).  The LOOK of (a)=>A B is A for LL(1)
     * ### is this even called?
     */
    MGrammar_Lookahead look(int k, MGrammar_SynPredBlock_ptr blk);

    MGrammar_Lookahead look(int k, MGrammar_TokenRangeElement_ptr r);

    MGrammar_Lookahead look(int k, MGrammar_TreeElement_ptr t);

    MGrammar_Lookahead look(int k, MGrammar_WildcardElement_ptr wc);

    /** The (...)* element is the combined lookahead of the alternatives and what can
     *  follow the loop.
     */
    MGrammar_Lookahead look(int k, MGrammar_ZeroOrMoreBlock_ptr blk);

    /**Compute the combined lookahead for all productions of a rule.
     * If the lookahead returns with epsilon, at least one epsilon
     * path exists (one that consumes no tokens).  The noFOLLOW
     * flag being set for this endruleblk, indicates that the
     * a rule ref invoked this rule.
     *
     * Currently only look(RuleRef) calls this.  There is no need
     * for the code generator to call this.
     */
    MGrammar_Lookahead look(int k, const MLANG_String& rule);

    /** If the first k-1 sets are singleton sets, the appoximate
     *  lookahead analysis is equivalent to full lookahead analysis.
     */
    static bool lookaheadEquivForApproxAndFullAnalysis(const ACE_Vector<MGrammar_Lookahead>& bset, int k);

private:

    /** Remove the prediction sets from preceding alternatives
     * and follow set, but *only* if this element is the first element
     * of the alternative.  The class members currenBlock and
     * currentBlock.analysisAlt must be set correctly.
     * @param b The prediction bitset to be modified
     * @el The element of interest
     */
    void removeCompetingPredictionSets(MParser_BitSet& b, MGrammar_AlternativeElement_ptr el);

private:

    /** Remove the prediction sets from preceding alternatives
     * The class members currenBlock must be set correctly.
     * Remove prediction sets from 1..k.
     * @param look The prediction lookahead to be modified
     * @el The element of interest
     * @k  How deep into lookahead to modify
     */
    void removeCompetingPredictionSetsFromWildcard(ACE_Vector<MGrammar_Lookahead>& look, MGrammar_AlternativeElement_ptr el, int k);

    /** reset the analyzer so it looks like a new one */
    void reset();

public:

    /** Set the grammar for the analyzer */
    void setGrammar(MGrammar_Grammar_ptr g);

    bool subruleCanBeInverted(MGrammar_AlternativeBlock_ptr blk, bool forLexer);
};

typedef MLANG_SmartPtr<MGrammar_LLkAnalyzer> MGrammar_LLkAnalyzer_var;

#endif /* __MGRAMMAR_LLKANALYZER_H__ */
