#include "MGrammar/Grammar.h"
#include "MGrammar/Lookahead.h"

/* ANTLR Translator Generator
 * Project led by Terence Parr at http://www.jGuru.com
 * Software rights: http://www.antlr.org/license.html
 *
 * $Id: //depot/code/org.antlr/release/antlr-2.7.4/antlr/Lookahead.java#1 $
 */

/**This object holds all information needed to represent
 * the lookahead for any particular lookahead computation
 * for a <b>single</b> lookahead depth.  Final lookahead
 * information is a simple bit set, but intermediate
 * stages need computation cycle and FOLLOW information.
 *
 * <p>
 * Concerning the <tt>cycle</tt> variable.
 * If lookahead is computed for a RuleEnd node, then
 * computation is part of a FOLLOW cycle for this rule.
 * If lookahead is computed for a RuleBlock node, the
 * computation is part of a FIRST cycle to this rule.
 *
 * <p>
 * Concerning the <tt>epsilonDepth</tt> variable.
 * This is not the depth relative to the rule reference
 * that epsilon was encountered.  That value is
 * <pre>
 * 		initial_k - epsilonDepth + 1
 * </pre>
 * Also, lookahead depths past rule ref for local follow are:
 * <pre>
 * 		initial_k - (initial_k - epsilonDepth)
 * </pre>
 * Used for rule references.  If we try
 * to compute look(k, ruleref) and there are fewer
 * than k lookahead terminals before the end of the
 * the rule, epsilon will be returned (don't want to
 * pass the end of the rule).  We must track when the
 * the lookahead got stuck.  For example,
 * <pre>
 * 		a : b A B E F G;
 * 		b : C ;
 * </pre>
 * LOOK(5, ref-to(b)) is {<EPSILON>} with depth = 4, which
 * indicates that at 2 (5-4+1) tokens ahead, end of rule was reached.
 * Therefore, the token at 4=5-(5-4) past rule ref b must be
 * included in the set == F.
 * The situation is complicated by the fact that a computation
 * may hit the end of a rule at many different depths.  For example,
 * <pre>
 * 		a : b A B C ;
 * 		b : E F		// epsilon depth of 1 relative to initial k=3
 * 		  | G		// epsilon depth of 2
 * 		  ;
 * </pre>
 * Here, LOOK(3,ref-to(b)) returns epsilon, but the depths are
 * {1, 2}; i.e., 3-(3-1) and 3-(3-2).  Those are the lookahead depths
 * past the rule ref needed for the local follow.
 *
 * <p>
 * This is null unless an epsilon is created.
 *
 * @see antlr.Lookahead#combineWith(Lookahead)
 */

MGrammar_Lookahead::MGrammar_Lookahead()
   : hasEpsilonDepth_(false),
     hasEpsilon_(false),
     is_null_(false)

{
}

/** create a new lookahead set with the LL(1) set to the parameter */
MGrammar_Lookahead::MGrammar_Lookahead(const MParser_BitSet& fset)
   : fset_(fset),
     hasEpsilonDepth_(false),
     hasEpsilon_(false),
     is_null_(false)

{
}

/** create an empty lookahead set, but with cycle */
MGrammar_Lookahead::MGrammar_Lookahead(const MLANG_String& cycle)
   : cycle_(cycle),
     hasEpsilonDepth_(false),
     hasEpsilon_(false),
     is_null_(false)

{
   assert(cycle_.length()!=0);
}

/** Make a deep copy of everything in this object */
MGrammar_Lookahead::MGrammar_Lookahead (const MGrammar_Lookahead& b)
   : fset_(b.fset_),
     cycle_(b.cycle_),
     hasEpsilonDepth_(b.hasEpsilonDepth_),
     epsilonDepth_(b.epsilonDepth_),
     hasEpsilon_(b.hasEpsilon_),
     is_null_(b.is_null_)

{
}

/** Make a deep copy of everything in this object */
MGrammar_Lookahead&
MGrammar_Lookahead::operator=(const MGrammar_Lookahead& b)

{
   if (this != &b)
    {
      fset_ = b.fset_;
      cycle_ = b.cycle_;
      hasEpsilonDepth_ = b.hasEpsilonDepth_;
      epsilonDepth_ = b.epsilonDepth_;
      hasEpsilon_ = b.hasEpsilon_;
      is_null_ = b.is_null_;
    }
   return *this;
}

void MGrammar_Lookahead::combineWith(const MGrammar_Lookahead& q)

{
   if (cycle_.length()==0)
    {
      cycle_ = q.cycle_;
    }

   if (q.containsEpsilon())
    {
      hasEpsilon_ = true;
    }

   // combine epsilon depths
   if (hasEpsilonDepth_)
    {
      if (q.hasEpsilonDepth_)
       {
         epsilonDepth_ |= q.epsilonDepth_;
       }
    }
   else if (q.hasEpsilonDepth_)
    {
      hasEpsilonDepth_ = q.hasEpsilonDepth_;
      epsilonDepth_ = q.epsilonDepth_;
    }

   fset_ |= q.fset_;
}

/** What is the intersection of two lookahead depths?
 *  Only the Epsilon "bit" and bitset are considered.
 */
MGrammar_Lookahead MGrammar_Lookahead::intersection(const MGrammar_Lookahead& q)

{
    MGrammar_Lookahead p (fset_ & q.fset_);
    if (hasEpsilon_ && q.hasEpsilon_)
     {
       p.setEpsilon();
     }
#if 0
    if (hasEpsilonDepth_ && q.hasEpsilonDepth_)
     {
       p.hasEpsilonDepth_ = true;
       p.epsilonDepth_ = epsilonDepth_ & q.epsilonDepth_;
     }
#endif
   return p;
}

MGrammar_Lookahead MGrammar_Lookahead::of(int el)

{
    MGrammar_Lookahead look;
    look.fset_.add(el);
    return look;
}

MLANG_String MGrammar_Lookahead::toString() const

{
   MLANG_String e = "",b,f = "",d = "";
   b = fset_.toString(",");
   if (containsEpsilon())
    {
      e = "+<epsilon>";
    }
   if (cycle_.length())
    {
      f += "; FOLLOW(";
      f += cycle_;
      f += ")";
    }
   if (hasEpsilonDepth_)
    {
      d = "; depths=";
      d += epsilonDepth_.toString(",");
    }
   return b + e + f + d;
}

MLANG_String MGrammar_Lookahead::toString (
   const MLANG_String& separator, 
   const MParser_CharFormatter& formatter
) const

{

   MLANG_String e = "",b,f = "",d = "";
   b = fset_.toString(separator, formatter);
   if (containsEpsilon())
    {
      e = "+<epsilon>";
    }
   if (cycle_.length())
    {
      f = "; FOLLOW(";
      f += cycle_;
      f += ")";
    }
   if (hasEpsilonDepth_)
    {
      d = "; depths=";
      d += epsilonDepth_.toString(",");
    }
   return b + e + f + d;
}

MLANG_String MGrammar_Lookahead::toString (
   const MLANG_String& separator,
   const MParser_CharFormatter& formatter,
   const MGrammar_Grammar_ptr g
) const

{
   if (g->isLexer())
    {
       return toString(separator, formatter);
    }
   else
    {
       return toString(separator, g->tokenManager_->getVocabulary());
    }
}

MLANG_String MGrammar_Lookahead::toString (
    const MLANG_String& separator, 
    const VOCABULARY& vocab
) const

{
    MLANG_String b,f = "",d = "";
    b = fset_.toString(separator, vocab);
    if (cycle_.length())
     {
       f = "; FOLLOW(";
       f += cycle_;
       f += ")";
     }
    if (hasEpsilonDepth_)
     {
       d = "; depths=";
       d += epsilonDepth_.toString(",");
     }
    return b + f + d;
}

MGrammar_Lookahead::~MGrammar_Lookahead()

{
}
