#ifndef __MGRAMMAR_LOOKAHEAD_H__
#define __MGRAMMAR_LOOKAHEAD_H__
#include <map>
#include "ace/SString.h"
#include "ace/Vector_T.h"
#include "MLANG/Vector_T.h"
#include "MLANG/String.h"
#include "MParser/CharFormatter.h"
#include "MParser/BitSet.h"

/* ANTLR Translator Generator
 * Project led by Terence Parr at http://www.jGuru.com
 * Software rights: http://www.antlr.org/license.html
 *
 * $Id: //depot/code/org.antlr/release/antlr-2.7.4/antlr/Lookahead.java#1 $
 */

class MGrammar_Grammar;
typedef class MGrammar_Grammar *MGrammar_Grammar_ptr;
typedef MLANG_SmartPtr<MGrammar_Grammar> MGrammar_Grammar_var;

/**This object holds all information needed to represent
 * the lookahead for any particular lookahead computation
 * for a <b>single</b> lookahead depth.  Final lookahead
 * information is a simple bit set, but intermediate
 * stages need computation cycle and FOLLOW information.
 *
 * <p>
 * Concerning the <tt>cycle</tt> variable.
 * If lookahead is computed for a RuleEnd node, then
 * computation is part of a FOLLOW cycle for this rule.
 * If lookahead is computed for a RuleBlock node, the
 * computation is part of a FIRST cycle to this rule.
 *
 * <p>
 * Concerning the <tt>epsilonDepth</tt> variable.
 * This is not the depth relative to the rule reference
 * that epsilon was encountered.  That value is
 * <pre>
 * 		initial_k - epsilonDepth + 1
 * </pre>
 * Also, lookahead depths past rule ref for local follow are:
 * <pre>
 * 		initial_k - (initial_k - epsilonDepth)
 * </pre>
 * Used for rule references.  If we try
 * to compute look(k, ruleref) and there are fewer
 * than k lookahead terminals before the end of the
 * the rule, epsilon will be returned (don't want to
 * pass the end of the rule).  We must track when the
 * the lookahead got stuck.  For example,
 * <pre>
 * 		a : b A B E F G;
 * 		b : C ;
 * </pre>
 * LOOK(5, ref-to(b)) is {<EPSILON>} with depth = 4, which
 * indicates that at 2 (5-4+1) tokens ahead, end of rule was reached.
 * Therefore, the token at 4=5-(5-4) past rule ref b must be
 * included in the set == F.
 * The situation is complicated by the fact that a computation
 * may hit the end of a rule at many different depths.  For example,
 * <pre>
 * 		a : b A B C ;
 * 		b : E F		// epsilon depth of 1 relative to initial k=3
 * 		  | G		// epsilon depth of 2
 * 		  ;
 * </pre>
 * Here, LOOK(3,ref-to(b)) returns epsilon, but the depths are
 * {1, 2}; i.e., 3-(3-1) and 3-(3-2).  Those are the lookahead depths
 * past the rule ref needed for the local follow.
 *
 * <p>
 * This is null unless an epsilon is created.
 *
 * @see antlr.Lookahead#combineWith(Lookahead)
 */
class MGrammar_Lookahead
{
public:
    typedef MLANG_Vector<MLANG_String> VOCABULARY; 
public:
    /** actual bitset of the lookahead */
    MParser_BitSet fset_;
    /** is this computation part of a computation cycle? */
    MLANG_String cycle_;
    /** What k values were being computed when end of rule hit? */
    bool hasEpsilonDepth_;
    MParser_BitSet epsilonDepth_;
    /** Does this lookahead depth include Epsilon token type? This
     *  is used to avoid having a bit in the set for Epsilon as it
     *  conflicts with parsing binary files.
     */
    bool hasEpsilon_;

    bool is_null_;

public:

    MGrammar_Lookahead();

    /** create a new lookahead set with the LL(1) set to the parameter */
    MGrammar_Lookahead(const MParser_BitSet& p);

    /** create an empty lookahead set, but with cycle */
    MGrammar_Lookahead(const MLANG_String& c);

    /** Make a deep copy of everything in this object */
    MGrammar_Lookahead(const MGrammar_Lookahead& b); 

    ~MGrammar_Lookahead ();

public:

    MGrammar_Lookahead& operator=(const MGrammar_Lookahead& b);

public:

    const MParser_BitSet& fset() const
    { return fset_; }

    MGrammar_Lookahead& fset(const MParser_BitSet& new_set)
    { fset_ = new_set; return *this; }

    void combineWith(const MGrammar_Lookahead& q);

    bool containsEpsilon() const
    {
        return hasEpsilon_;
    }

    /** What is the intersection of two lookahead depths?
     *  Only the Epsilon "bit" and bitset are considered.
     */
    MGrammar_Lookahead intersection(const MGrammar_Lookahead& q);

    bool nil() const
    {
        return fset_.nil() && !hasEpsilon_;
    }

    static MGrammar_Lookahead of(int el);

    void resetEpsilon()
    {
        hasEpsilon_ = false;
    }

    void setEpsilon()
    {
        hasEpsilon_ = true;
    }

    const MLANG_String& cycle() const
    { return cycle_; }

    MGrammar_Lookahead& cycle(const MLANG_String& new_cycle)
    { cycle_ = new_cycle; return *this; }

    bool is_null() const
    { return is_null_; }

    MGrammar_Lookahead& set_null()
    { is_null_ = true; return *this; }

    MLANG_String toString() const;

    MLANG_String toString (
       const MLANG_String& separator, 
       const MParser_CharFormatter& formatter
      ) const;

    MLANG_String toString (
       const MLANG_String& separator,
       const MParser_CharFormatter& formatter,
       const MGrammar_Grammar_ptr g
     ) const;

    MLANG_String toString (
       const MLANG_String& separator, 
       const VOCABULARY& vocab) const;
};

#endif /* __MGRAMMAR_LOOKAHEAD_H__ */
