#include "MGrammar/ANTLRTokenTypes.h"
#include "MGrammar/ANTLRLexer.h"
#include "MParser/StringUtils.h"
#include "MGrammar/ActionElement.h"
#include "MParser/System.h"
#include "MParser/CommonToken.h"
#include "MParser/SemanticException.h"
#include "MGrammar/TreeElement.h"
#include "MGrammar/TreeBlockContext.h"
#include "MGrammar/ExceptionSpec.h"
#include "MGrammar/ZeroOrMoreBlock.h"
#include "MGrammar/OneOrMoreBlock.h"
#include "MGrammar/CharLiteralElement.h"
#include "MGrammar/CharRangeElement.h"
#include "MGrammar/StringLiteralElement.h"
#include "MGrammar/TokenRangeElement.h"
#include "MGrammar/WildcardElement.h"
#include "MGrammar/Grammar.h"
#include "MGrammar/Tool.h"
#include "MGrammar/CodeGenerator.h"
#include "MGrammar/MakeGrammar.h"

/* ANTLR Translator Generator
 * Project led by Terence Parr at http://www.jGuru.com
 * Software rights: http://www.antlr.org/license.html
 *
 * $Id: //depot/code/org.antlr/release/antlr-2.7.4/antlr/MakeGrammar.java#1 $
 */

MGrammar_MakeGrammar::MGrammar_MakeGrammar (
   MGrammar_Tool_ptr tool,
   ACE_ARGV& args,
   MGrammar_LLkAnalyzer_ptr analyzer
) : MGrammar_DefineGrammarSymbols(tool, args, analyzer),
    nested_(0),
    grammarError_(false)

{
}

/** Abort the processing of a grammar (due to syntax errors) */
void MGrammar_MakeGrammar::abortGrammar() 

{
   MLANG_String s = "unknown grammar";
   if (grammar_ != 0)
    {
      s = grammar_->getClassName();
    }
   tool_->error("aborting grammar '" + s + "' due to errors");
   super::abortGrammar();
}

void MGrammar_MakeGrammar::addElementToCurrentAlt (
   const MGrammar_AlternativeElement_var& e
)

{
   e->enclosingRuleName (ruleBlock_->getRuleName());
   context()->addAlternativeElement(e);
}

void MGrammar_MakeGrammar::beginAlt(bool doAutoGen)

{
   super::beginAlt(doAutoGen);
   MGrammar_Alternative_var alt = new MGrammar_Alternative();
   alt->setAutoGen(doAutoGen);
   context()->block_->addAlternative(alt);
}

void MGrammar_MakeGrammar::beginChildList()

{
   super::beginChildList();
   context()->block_->addAlternative(new MGrammar_Alternative());
}

/** Add an exception group to a rule (currently a no-op) */
void MGrammar_MakeGrammar::beginExceptionGroup()

{
   super::beginExceptionGroup();
   if (!(instanceof<MGrammar_RuleBlock>(context()->block_)))
    {
      tool_->panic("beginExceptionGroup called outside of rule block");
    }
}

/** Add an exception spec to an exception group or rule block */
void MGrammar_MakeGrammar::beginExceptionSpec (
   const MParser_Token_var& label
)

{
   // Hack the label string a bit to remove leading/trailing space.
//   if (!MParser_Token::is_nil(label))
   if (label)
    {
      label->setText(MParser_StringUtils::stripFront(MParser_StringUtils::stripBack(label->getText(), " \n\r\t"), " \n\r\t"));
    }
   super::beginExceptionSpec(label);
   // Don't check for currentExceptionSpec!=null because syntax errors
   // may leave it set to something.
   currentExceptionSpec_ = new MGrammar_ExceptionSpec(label);
}

void MGrammar_MakeGrammar::beginSubRule (
   const MParser_Token_var& label,
   const MParser_Token_var& start,
   bool no
)

{
   super::beginSubRule(label, start, no);
   // we don't know what kind of subrule it is yet.
   // push a dummy one that will allow us to collect the
   // alternatives.  Later, we'll switch to real object.
   blocks_.push(new MGrammar_BlockContext());
   context()->block_ = new MGrammar_AlternativeBlock(grammar_, start, no);
   context()->altNum_ = 0; // reset alternative number
   nested_++;
   // create a final node to which the last elememt of each
   // alternative will point.
   context()->blockEnd_ = new MGrammar_BlockEndElement(grammar_);
   // make sure end node points to start of block
   context()->blockEnd_->block_ = context()->block_;
   labelElement(context()->block_, label);
}

void MGrammar_MakeGrammar::beginTree (
   const MParser_Token_var& tok
)

{
   if (!(grammar_->isTreeWalker()))
    {
      tool_->error("Trees only allowed in TreeParser", grammar_->getFilename(), tok->getLine(), tok->getColumn());
      throw new MParser_SemanticException("Trees only allowed in TreeParser");
    }
   super::beginTree(tok);
   blocks_.push(new MGrammar_TreeBlockContext());
   context()->block_ = new MGrammar_TreeElement(grammar_, tok);
   context()->altNum_ = 0; // reset alternative number
}

MGrammar_BlockContext_var MGrammar_MakeGrammar::context() 

{
   if (blocks_.size() == 0)
    {
      return MGrammar_BlockContext::_nil();
    }
   else
    {
      return blocks_.top();
    }
}

/**Used to build nextToken() for the lexer.
 * This builds a rule which has every "public" rule in the given Vector of
 * rules as it's alternate.  Each rule ref generates a Token object.
 * @param g  The Grammar that is being processed
 * @param lexRules A vector of lexer rules that will be used to create an alternate block.
 * @param rname The name of the resulting rule.
 */
MGrammar_RuleBlock_var MGrammar_MakeGrammar::createNextTokenRule (
   const MGrammar_Grammar_var& g,
   const MLANG_Vector<MGrammar_RuleSymbol_var>& lexRules,
   const MLANG_String& rname
)

{
   // create actual rule data structure
   MGrammar_RuleBlock_var rb = new MGrammar_RuleBlock(g, rname);
   rb->setDefaultErrorHandler(g->getDefaultErrorHandler());
   MGrammar_RuleEndElement_var ruleEnd = new MGrammar_RuleEndElement(g);
   rb->setEndElement(ruleEnd);
   ruleEnd->block_ = rb;
   // Add an alternative for each element of the rules vector.
   for (size_t i = 0; i < lexRules.size(); i++)
    {
       MGrammar_RuleSymbol_var r = lexRules[i];
       if (!r->isDefined())
        {
          g->antlrTool_->error("Lexer rule " + r->getId().substring(1) + " is not defined");
        }
       else
        {
          if (r->access().equals("public"))
           {
             MGrammar_Alternative_var alt = new MGrammar_Alternative(); // create alt we'll add to ref rule
             MGrammar_RuleBlock_var targetRuleBlock = r->getBlock();
             ACE_Vector<MGrammar_Alternative_var> targetRuleAlts = targetRuleBlock->getAlternatives();
             // collect a sem pred if only one alt and it's at the start;
             // simple, but faster to implement until real hoisting
             if (targetRuleAlts.size()!=0 && targetRuleAlts.size()==1 )
              {
                MGrammar_Alternative_var onlyAlt = targetRuleAlts[0];
                if (onlyAlt->semPred_.length()!=0)
                 {
                   // ok, has sem pred, make this rule ref alt have a pred
                   alt->semPred_ = onlyAlt->semPred_;
// REMOVE predicate from target rule???  NOPE, another
// rule other than nextToken() might invoke it.
                 }
              }

             // create a rule ref to lexer rule
             // the Token is a RULE_REF not a TOKEN_REF since the
             // conversion to mRulename has already taken place
             MGrammar_RuleRefElement_var rr =
                new MGrammar_RuleRefElement (
                          g,
                          new MParser_CommonToken(ANTLRTokenTypes::RULE_REF, r->getId()),
                          MGrammar_GrammarElement::AUTO_GEN_NONE);
             rr->setLabel("theRetToken");
             rr->enclosingRuleName ("nextToken");
             rr->next_ = ruleEnd;
             alt->addElement(rr);  		// add rule ref to alt
             alt->setAutoGen(true);		// keep text of elements
             rb->addAlternative(alt);		// add alt to rule block
             r->addReference(rr);			// track ref to this rule in rule blk
           }
        }
    }

   rb->setAutoGen(true);		// keep text of elements
   rb->prepareForAnalysis();
   //System.out.println(rb);
   return rb;
}

/** Return block as if they had typed: "( rule )?" */
MGrammar_AlternativeBlock_var MGrammar_MakeGrammar::createOptionalRuleRef (
   const MLANG_String& rule,
   const MParser_Token_var& start
)

{
   // Make the subrule
   MGrammar_AlternativeBlock_var blk = 
      new MGrammar_AlternativeBlock(grammar_, start, false);

   // Make sure rule is defined
   MLANG_String mrule = MGrammar_CodeGenerator::encodeLexerRuleName(rule); // can only be a lexer rule!
   if (!grammar_->isDefined(mrule))
    {
      grammar_->define(new MGrammar_RuleSymbol(mrule));
    }

    // Make the rule ref element
    // RK: fixme probably easier to abuse start token..
    MParser_Token_var t = new MParser_CommonToken(ANTLRTokenTypes::TOKEN_REF, rule);
    t->setLine(start->getLine());
    t->setLine(start->getColumn());
    MGrammar_RuleRefElement_var rref =
        new MGrammar_RuleRefElement(grammar_, t, MGrammar_GrammarElement::AUTO_GEN_NONE);

    rref->enclosingRuleName (ruleBlock_->getRuleName());

    // Make the end of block element
    MGrammar_BlockEndElement_var end = new MGrammar_BlockEndElement(grammar_);
    end->block_ = blk;		// end block points back to start of blk

    // Make an alternative, putting the rule ref into it
    MGrammar_Alternative_var alt = new MGrammar_Alternative(rref);
    alt->addElement(end); // last element in alt points to end of block

    // Add the alternative to this block
    blk->addAlternative(alt);

    // create an empty (optional) alt and add to blk
    MGrammar_Alternative_var optAlt = new MGrammar_Alternative();
    optAlt->addElement(end);	// points immediately to end of block

    blk->addAlternative(optAlt);

    blk->prepareForAnalysis();
    return blk;
}

void MGrammar_MakeGrammar::defineRuleName (
   const MParser_Token_var& r,
   const MLANG_String& access,
   bool ruleAutoGen,
   const MLANG_String& docComment

)

{
   //		if ( Character.isUpperCase(r.getText().charAt(0)) ) {
   if (r->getType() == ANTLRTokenTypes::TOKEN_REF)
    {
      if (!(grammar_->isLexer()))
       {
         tool_->error("Lexical rule " + r->getText() +
                    " defined outside of lexer",
                    grammar_->getFilename(), r->getLine(), r->getColumn());
         r->setText(r->getText().toLowerCase());
       }
    }
   else
    {
       if (grammar_->isLexer())
        {
          tool_->error("Lexical rule names must be upper case, '" + r->getText() +
                       "' is not",
                       grammar_->getFilename(), r->getLine(), r->getColumn());
          r->setText(r->getText().toUpperCase());
        }
    }

   super::defineRuleName(r, access, ruleAutoGen, docComment);
   MLANG_String id = r->getText();
   //		if ( Character.isUpperCase(id.charAt(0)) ) { // lexer rule?
   if (r->getType() == ANTLRTokenTypes::TOKEN_REF)
    { // lexer rule?
       id = MGrammar_CodeGenerator::encodeLexerRuleName(id);
    }
   MGrammar_RuleSymbol_var rs = Dynamic_Cast<MGrammar_RuleSymbol>(grammar_->getSymbol(id));
   MGrammar_RuleBlock_var rb = new MGrammar_RuleBlock(grammar_, r->getText(), r->getLine(), ruleAutoGen);

   // Lexer rules do not generate default error handling
   rb->setDefaultErrorHandler(grammar_->getDefaultErrorHandler());

   ruleBlock_ = rb;
   blocks_.push(new MGrammar_BlockContext()); // enter new context
   context()->block_ = rb;
   rs->setBlock(rb);
   ruleEnd_ = new MGrammar_RuleEndElement(grammar_);
   rb->setEndElement(ruleEnd_);
   nested_ = 0;
}

void MGrammar_MakeGrammar::endAlt()

{
   super::endAlt();
   if (nested_ == 0)
    {	// all rule-level alts link to ruleEnd node
      addElementToCurrentAlt(ruleEnd_);
    }
   else
    {
      addElementToCurrentAlt(context()->blockEnd_);
    }
   context()->altNum_++;
}

void MGrammar_MakeGrammar::endChildList()

{
   super::endChildList();
   // create a final node to which the last elememt of the single
   // alternative will point.  Done for compatibility with analyzer.
   // Does NOT point to any block like alternative blocks because the
   // TreeElement is not a block.  This is used only as a placeholder.
   MGrammar_BlockEndElement_var be = new MGrammar_BlockEndElement(grammar_);
   be->block_ = context()->block_;
   addElementToCurrentAlt(be);
}

void MGrammar_MakeGrammar::endExceptionGroup()

{
   super::endExceptionGroup();
}

void MGrammar_MakeGrammar::endExceptionSpec()

{
   super::endExceptionSpec();
   if (currentExceptionSpec_.is_null())
    {
      tool_->panic("exception processing internal error -- no active exception spec");
    }
   if (instanceof<MGrammar_RuleBlock>(context()->block_))
    {
      // Named rule
      Dynamic_Cast<MGrammar_RuleBlock>(context()->block_)->addExceptionSpec(currentExceptionSpec_);
      
    }
   else
    {
      // It must be a plain-old alternative block
      if (!context()->currentAlt()->exceptionSpec_.is_null())
       {
         tool_->error("Alternative already has an exception specification", grammar_->getFilename(), context()->block_->getLine(), context()->block_->getColumn());
       }
      else
       {
         context()->currentAlt()->exceptionSpec_ = currentExceptionSpec_;
       }
    }
   currentExceptionSpec_ = 0;
}

/** Called at the end of processing a grammar */
void MGrammar_MakeGrammar::endGrammar()

{
   if (grammarError_)
    {
      abortGrammar();
    }
   else
    {
      super::endGrammar();
    }
}

void MGrammar_MakeGrammar::endRule (
   const MLANG_String& rule
)

{
   super::endRule(rule);
   MGrammar_BlockContext_var ctx = blocks_.top(); 
   blocks_.pop();	// remove scope
   // record the start of this block in the ending node
   ruleEnd_->block_ = ctx->block_;
   ruleEnd_->block_->prepareForAnalysis();
   //System.out.println(ctx.block);
}

void MGrammar_MakeGrammar::endSubRule() 

{
   super::endSubRule();
   nested_--;
   // remove subrule context from scope stack
   MGrammar_BlockContext_var ctx = blocks_.top();
   blocks_.pop();
   MGrammar_AlternativeBlock_var block = ctx->block_;

   // If the subrule is marked with ~, check that it is
   // a valid candidate for analysis
   if (
       block->no_ &&
       !(instanceof<MGrammar_SynPredBlock>(block)) &&
       !(instanceof<MGrammar_ZeroOrMoreBlock>(block)) &&
       !(instanceof<MGrammar_OneOrMoreBlock>(block))
      )
    {
      if (!analyzer_->subruleCanBeInverted(block, grammar_->isLexer()))
       {
         MLANG_String newline = SYSTEM->getProperty("line.separator");
         tool_->error(
                    "This subrule cannot be inverted.  Only subrules of the form:" + newline +
                    "    (T1|T2|T3...) or" + newline +
                    "    ('c1'|'c2'|'c3'...)" + newline +
                    "may be inverted (ranges are also allowed).",
                    grammar_->getFilename(),
                    block->getLine(), block->getColumn()
                );
        }
    }

   // add the subrule as element if not a syn pred
   if (instanceof<MGrammar_SynPredBlock>(block))
    {
      // record a reference to the recently-recognized syn pred in the
      // enclosing block.
      MGrammar_SynPredBlock_var synpred = Dynamic_Cast<MGrammar_SynPredBlock>(block);
      context()->block_->hasASynPred_ = true;
      context()->currentAlt()->synPred_ = synpred;
      grammar_->hasSyntacticPredicate_ = true;
      synpred->removeTrackingOfRuleRefs(grammar_);
    }
   else
    {
      addElementToCurrentAlt(block);
    }
   ctx->blockEnd_->block_->prepareForAnalysis();
}

void MGrammar_MakeGrammar::endTree()

{
   super::endTree();
   MGrammar_BlockContext_var ctx = blocks_.top();
   blocks_.pop();
   addElementToCurrentAlt(ctx->block_);		// add new TreeElement to enclosing alt.
}

void MGrammar_MakeGrammar::labelElement (
   const MGrammar_AlternativeElement_var& el,
   const MParser_Token_var& label
)

{
   if (!label.is_null())
    {
      // Does this label already exist?
      for (size_t i = 0; i < ruleBlock_->labeledElements_.size(); i++)
       {
          MGrammar_AlternativeElement_var altEl = ruleBlock_->labeledElements_[i];
          MLANG_String l = altEl->getLabel();
          if (l.length() != 0 && l.equals(label->getText())) 
           {
             tool_->error("Label '" + label->getText() + "' has already been defined", grammar_->getFilename(), label->getLine(), label->getColumn());
             return;
           }
       }
      // add this node to the list of labeled elements
      el->setLabel(label->getText());
      ruleBlock_->labeledElements_.push_back(el);
    }
}

void MGrammar_MakeGrammar::noAutoGenSubRule()

{
   context()->block_->setAutoGen(false);
}

void MGrammar_MakeGrammar::oneOrMoreSubRule()

{
   if (context()->block_->no_) 
    {
      tool_->error("'~' cannot be applied to (...)* subrule", grammar_->getFilename(), context()->block_->getLine(), context()->block_->getColumn());
    }
   // create the right kind of object now that we know what that is
   // and switch the list of alternatives.  Adjust the stack of blocks.
   // copy any init action also.
   MGrammar_OneOrMoreBlock_var b = new MGrammar_OneOrMoreBlock(grammar_);
   setBlock(b, context()->block_);
   MGrammar_BlockContext_var old = blocks_.top(); // remove old scope; we want new type of subrule
   blocks_.pop();
   blocks_.push(new MGrammar_BlockContext());
   context()->block_ = b;
   context()->blockEnd_ = old->blockEnd_;
   context()->blockEnd_->block_ = b;
}

void MGrammar_MakeGrammar::optionalSubRule()

{
   if (context()->block_->no_)
    {
      tool_->error("'~' cannot be applied to (...)? subrule", grammar_->getFilename(), context()->block_->getLine(), context()->block_->getColumn());
    }
   // convert (X)? -> (X|) so that we can ignore optional blocks altogether!
   // It already thinks that we have a simple subrule, just add option block.
   beginAlt(false);
   endAlt();
}

void MGrammar_MakeGrammar::refAction (
   const MParser_Token_var& action
)

{
   super::refAction(action);
   context()->block_->hasAnAction_ = true;
   addElementToCurrentAlt(new MGrammar_ActionElement(grammar_, action));
}

void MGrammar_MakeGrammar::setUserExceptions(const MLANG_String& thr)

{
   Dynamic_Cast<MGrammar_RuleBlock>(context()->block_)->throwsSpec_ = thr;
}

// Only called for rule blocks
void MGrammar_MakeGrammar::refArgAction (
   const MParser_Token_var& action
)

{
   Dynamic_Cast<MGrammar_RuleBlock>(context()->block_)->argAction_ = action->getText();
}

void MGrammar_MakeGrammar::refCharLiteral (
   const MParser_Token_var& lit,
   const MParser_Token_var& label,
   bool inverted,
   int autoGenType,
   bool lastInRule
)

{
   if (!(grammar_->isLexer()))
    {
       tool_->error("Character literal only valid in lexer", grammar_->getFilename(), lit->getLine(), lit->getColumn());
       return;
    }
   super::refCharLiteral(lit, label, inverted, autoGenType, lastInRule);
   MGrammar_CharLiteralElement_var cl = new MGrammar_CharLiteralElement(
       Dynamic_Cast<MGrammar_LexerGrammar>(grammar_), lit, inverted, autoGenType);

   // Generate a warning for non-lowercase ASCII when case-insensitive
   if (
       !Dynamic_Cast<MGrammar_LexerGrammar>(grammar_)->caseSensitive_ && cl->getType() < 128 &&
            tolower((char)cl->getType()) != (char)cl->getType()
        )
    {
       tool_->warning("Character literal must be lowercase when caseSensitive=false", grammar_->getFilename(), lit->getLine(), lit->getColumn());
    }

   addElementToCurrentAlt(cl);
   labelElement(cl, label);

   // if ignore option is set, must add an optional call to the specified rule.
   MLANG_String ignore = ruleBlock_->getIgnoreRule();
   if (!lastInRule && ignore.length() != 0)
    {
      addElementToCurrentAlt(
           createOptionalRuleRef(ignore, lit));
    }
}

void MGrammar_MakeGrammar::refCharRange (
   const MParser_Token_var& t1,
   const MParser_Token_var& t2,
   const MParser_Token_var& label,
   int autoGenType,
   bool lastInRule
)

{
   if (!(grammar_->isLexer()))
    {
      tool_->error("Character range only valid in lexer", grammar_->getFilename(), t1->getLine(), t1->getColumn());
      return;
    }
   int rangeMin = ANTLRLexer::tokenTypeForCharLiteral(t1->getText());
   int rangeMax = ANTLRLexer::tokenTypeForCharLiteral(t2->getText());
   if (rangeMax < rangeMin)
    {
      tool_->error("Malformed range.", grammar_->getFilename(), t1->getLine(), t1->getColumn());
      return;
    }

   // Generate a warning for non-lowercase ASCII when case-insensitive
   if (!(Dynamic_Cast<MGrammar_LexerGrammar>(grammar_))->caseSensitive_)
    {
      if (rangeMin < 128 && tolower((char)rangeMin) != (char)rangeMin)
       {
         tool_->warning("Character literal must be lowercase when caseSensitive=false", grammar_->getFilename(), t1->getLine(), t1->getColumn());
       }
      if (rangeMax < 128 && tolower((char)rangeMax) != (char)rangeMax)
       {
         tool_->warning("Character literal must be lowercase when caseSensitive=false", grammar_->getFilename(), t2->getLine(), t2->getColumn());
       }
    }

   super::refCharRange(t1, t2, label, autoGenType, lastInRule);
   MGrammar_CharRangeElement_var cr = new MGrammar_CharRangeElement(Dynamic_Cast<MGrammar_LexerGrammar>(grammar_), t1, t2, autoGenType);
   addElementToCurrentAlt(cr);
   labelElement(cr, label);

   // if ignore option is set, must add an optional call to the specified rule.
   MLANG_String ignore = ruleBlock_->getIgnoreRule();
   if (!lastInRule && ignore.length() != 0)
    {
      addElementToCurrentAlt(createOptionalRuleRef(ignore, t1));
    }
}

void MGrammar_MakeGrammar::refTokensSpecElementOption (
   const MParser_Token_var& tok,
   const MParser_Token_var& option,
   const MParser_Token_var& value
)

{
   /*
      System.out.println("setting tokens spec option for "+tok.getText());
      System.out.println(option.getText()+","+value.getText());
    */
   MGrammar_TokenSymbol_var ts = 
          grammar_->tokenManager_->getTokenSymbol(tok->getText());
   if (ts.is_null())
    {
      tool_->panic("cannot find " + tok->getText() + "in tokens {...}");
    }
   if (option->getText().equals("AST"))
    {
      ts->setASTNodeType(value->getText());
    }
   else
    {
      grammar_->antlrTool_->error("invalid tokens {...} element option:" +
                               option->getText(),
                               grammar_->getFilename(),
                               option->getLine(), option->getColumn());
    }
}

void MGrammar_MakeGrammar::refElementOption (
   const MParser_Token_var& option,
   const MParser_Token_var& value
)

{
   /*
      System.out.println("setting option for "+context().currentElement());
      System.out.println(option.getText()+","+value.getText());
    */
   MGrammar_AlternativeElement_var e = context()->currentElement();
   if (instanceof<MGrammar_StringLiteralElement>(e) 
   ||  instanceof<MGrammar_TokenRefElement>(e) 
   ||  instanceof<MGrammar_WildcardElement>(e))
    {
       Dynamic_Cast<MGrammar_GrammarAtom>(e)->setOption(option, value);
    }
   else
    {
       tool_->error("cannot use element option (" + option->getText() +
                    ") for this kind of element",
                  grammar_->getFilename(), option->getLine(), option->getColumn());
    }
}

/** Add an exception handler to an exception spec */
void MGrammar_MakeGrammar::refExceptionHandler (
   const MParser_Token_var& exTypeAndName,
   const MParser_Token_var& action
)

{
   super::refExceptionHandler(exTypeAndName, action);
   if (currentExceptionSpec_.is_null())
    {
      tool_->panic("exception handler processing internal error");
    }
   currentExceptionSpec_->addHandler(new MGrammar_ExceptionHandler(exTypeAndName, action));
}

void MGrammar_MakeGrammar::refInitAction(const MParser_Token_var& action)

{
   super::refAction(action);
   context()->block_->setInitAction(action->getText());
}

void MGrammar_MakeGrammar::refMemberAction(const MParser_Token_var& act)

{
   grammar_->classMemberAction_ = act;
}

void MGrammar_MakeGrammar::refPreambleAction(const MParser_Token_var& act)

{
   super::refPreambleAction(act);
}

// Only called for rule blocks
void MGrammar_MakeGrammar::refReturnAction (
   const MParser_Token_var& returnAction
)

{
   if (grammar_->isLexer())
    {
      MLANG_String name = MGrammar_CodeGenerator::encodeLexerRuleName (
           Dynamic_Cast<MGrammar_RuleBlock>(context()->block_)->getRuleName());
      MGrammar_RuleSymbol_var rs = Dynamic_Cast<MGrammar_RuleSymbol>(grammar_->getSymbol(name));
      if (rs->access().equals("public"))
       {
         tool_->warning("public Lexical rules cannot specify return type", grammar_->getFilename(), returnAction->getLine(), returnAction->getColumn());
         return;
       }
    }
   Dynamic_Cast<MGrammar_RuleBlock>(context()->block_)->returnAction_ = returnAction->getText();
}

void MGrammar_MakeGrammar::refRule (
   const MParser_Token_var& idAssign,
   const MParser_Token_var& r,
   const MParser_Token_var& label,
   const MParser_Token_var& args,
   int autoGenType
)

{
   // Disallow parser rule references in the lexer
   if (grammar_->isLexer())
    {
      //			if (!Character.isUpperCase(r.getText().charAt(0))) {
      if (r->getType() != ANTLRTokenTypes::TOKEN_REF)
       {
         tool_->error("Parser rule " + r->getText() + " referenced in lexer");
         return;
       }
      if (autoGenType == MGrammar_GrammarElement::AUTO_GEN_CARET)
       {
         tool_->error("AST specification ^ not allowed in lexer", grammar_->getFilename(), r->getLine(), r->getColumn());
       }
    }

   super::refRule(idAssign, r, label, args, autoGenType);
   lastRuleRef_ = new MGrammar_RuleRefElement(grammar_, r, autoGenType);
   if (!MParser_Token::is_nil(args))
    {
      lastRuleRef_->setArgs(args->getText());
    }
   if (!idAssign.is_null())
    {
      lastRuleRef_->setIdAssign(idAssign->getText());
    }
   addElementToCurrentAlt(lastRuleRef_);

   MLANG_String id = r->getText();
   //		if ( Character.isUpperCase(id.charAt(0)) ) { // lexer rule?
   if (r->getType() == ANTLRTokenTypes::TOKEN_REF)
    { // lexer rule?
      id = MGrammar_CodeGenerator::encodeLexerRuleName(id);
    }
   // update symbol table so it knows what nodes reference the rule.
   MGrammar_RuleSymbol_var rs = Dynamic_Cast<MGrammar_RuleSymbol>(grammar_->getSymbol(id));
   rs->addReference(lastRuleRef_);
   labelElement(lastRuleRef_, label);
}

void MGrammar_MakeGrammar::refSemPred (
   const MParser_Token_var& pred
)

{
   //System.out.println("refSemPred "+pred.getText());
   super::refSemPred(pred);
   //System.out.println("context().block: "+context().block);
   if (context()->currentAlt()->atStart())
    {
      context()->currentAlt()->semPred_ = pred->getText();
    }
   else
    {
      MGrammar_ActionElement_var a = new MGrammar_ActionElement(grammar_, pred);
      a->isSemPred (true);
      addElementToCurrentAlt(a);
    }
   //System.out.println("DONE refSemPred "+pred.getText());
}

void MGrammar_MakeGrammar::refStringLiteral (
   const MParser_Token_var& lit,
   const MParser_Token_var& label,
   int autoGenType,
   bool lastInRule
)

{
   super::refStringLiteral(lit, label, autoGenType, lastInRule);
   if (grammar_->isTreeWalker()
   &&  autoGenType == MGrammar_GrammarElement::AUTO_GEN_CARET)
    {
      tool_->error("^ not allowed in here for tree-walker", grammar_->getFilename(), lit->getLine(), lit->getColumn());
    }
   MGrammar_StringLiteralElement_var sl =
       new MGrammar_StringLiteralElement(grammar_, lit, autoGenType);

   // If case-insensitive, then check each char of the stirng literal
   if (grammar_->isLexer() && !(Dynamic_Cast<MGrammar_LexerGrammar>(grammar_))->caseSensitive_)
    {
      for (size_t i = 1; i < lit->getText().length() - 1; i++)
       {
          char c = lit->getText().charAt(i);
          if ((unsigned char)c < 128 && tolower(c) != c)
           {
             tool_->warning("Characters of string literal must be lowercase when caseSensitive=false", grammar_->getFilename(), lit->getLine(), lit->getColumn());
             break;
           }
       }
    }

   addElementToCurrentAlt(sl);
   labelElement(sl, label);

   // if ignore option is set, must add an optional call to the specified rule.
   MLANG_String ignore = ruleBlock_->getIgnoreRule();
   if (!lastInRule && ignore.length() != 0)
    {
      addElementToCurrentAlt(createOptionalRuleRef(ignore, lit));
    }
}

void MGrammar_MakeGrammar::refToken (
   const MParser_Token_var& idAssign,
   const MParser_Token_var& t,
   const MParser_Token_var& label,
   const MParser_Token_var& args,
   bool inverted,
   int autoGenType,
   bool lastInRule
)

{
   if (grammar_->isLexer())
    {
      // In lexer, token references are really rule references
      if (autoGenType == MGrammar_GrammarElement::AUTO_GEN_CARET)
       {
         tool_->error("AST specification ^ not allowed in lexer", grammar_->getFilename(), t->getLine(), t->getColumn());
       }
      if (inverted)
       {
         tool_->error("~TOKEN is not allowed in lexer", grammar_->getFilename(), t->getLine(), t->getColumn());
       }
      refRule(idAssign, t, label, args, autoGenType);

      // if ignore option is set, must add an optional call to the specified token rule.
      MLANG_String ignore = ruleBlock_->getIgnoreRule();
      if (!lastInRule && ignore.length() != 0)
       {
         addElementToCurrentAlt(createOptionalRuleRef(ignore, t));
       }
    }
   else
    {
       // Cannot have token ref args or assignment outside of lexer
       if (!idAssign.is_null())
        {
          tool_->error("Assignment from token reference only allowed in lexer", grammar_->getFilename(), idAssign->getLine(), idAssign->getColumn());
        }
       if (!MParser_Token::is_nil(args)) 
        {
          tool_->error("Token reference arguments only allowed in lexer", grammar_->getFilename(), args->getLine(), args->getColumn());
        }
       super::refToken(idAssign, t, label, args, inverted, autoGenType, lastInRule);
       MGrammar_TokenRefElement_var te = new MGrammar_TokenRefElement(grammar_, t, inverted, autoGenType);
       addElementToCurrentAlt(te);
       labelElement(te, label);
    }
}

void MGrammar_MakeGrammar::refTokenRange (
   const MParser_Token_var& t1,
   const MParser_Token_var& t2,
   const MParser_Token_var& label,
   int autoGenType,
   bool lastInRule
)

{
   if (grammar_->isLexer())
    {
      tool_->error("Token range not allowed in lexer", grammar_->getFilename(), t1->getLine(), t1->getColumn());
      return;
    }
   super::refTokenRange(t1, t2, label, autoGenType, lastInRule);
   MGrammar_TokenRangeElement_var tr = new MGrammar_TokenRangeElement(grammar_, t1, t2, autoGenType);
   if (tr->end() < tr->begin())
    {
      tool_->error("Malformed range.", grammar_->getFilename(), t1->getLine(), t1->getColumn());
      return;
    }
   addElementToCurrentAlt(tr);
   labelElement(tr, label);
}

void MGrammar_MakeGrammar::refTreeSpecifier (
   const MParser_Token_var& treeSpec
)

{
   context()->currentAlt()->treeSpecifier_ = treeSpec;
}

void MGrammar_MakeGrammar::refWildcard (
   const MParser_Token_var& t,
   const MParser_Token_var& label,
   int autoGenType
)

{
   super::refWildcard(t, label, autoGenType);
   MGrammar_WildcardElement_var wc = new MGrammar_WildcardElement(grammar_, t, autoGenType);
   addElementToCurrentAlt(wc);
   labelElement(wc, label);
}

/** Get ready to process a new grammar */
void MGrammar_MakeGrammar::reset()
{
   super::reset();
   blocks_ = std::stack<MGrammar_BlockContext_var>();
   lastRuleRef_ = MGrammar_RuleRefElement::_nil();
   ruleEnd_ = MGrammar_RuleEndElement::_nil();
   ruleBlock_ = MGrammar_RuleBlock::_nil();
   nested_ = 0;
   currentExceptionSpec_ = MGrammar_ExceptionSpec::_nil();
   grammarError_ = false;
}

void MGrammar_MakeGrammar::setArgOfRuleRef (
   const MParser_Token_var& argAction
)

{
   super::setArgOfRuleRef(argAction);
   lastRuleRef_->setArgs(argAction->getText());
}

void MGrammar_MakeGrammar::setBlock (
   const MGrammar_AlternativeBlock_var& b,
   const MGrammar_AlternativeBlock_var& src
)

{
   b->setAlternatives(src->getAlternatives());
   b->initAction_ = src->initAction_;
   //b.lookaheadDepth = src.lookaheadDepth;
   b->label_ = src->label_;
   b->hasASynPred_ = src->hasASynPred_;
   b->hasAnAction_ = src->hasAnAction_;
   b->warnWhenFollowAmbig_ = src->warnWhenFollowAmbig_;
   b->generateAmbigWarnings_ = src->generateAmbigWarnings_;
   b->setLine (src->getLine());
   b->greedy_ = src->greedy_;
   b->greedySet_ = src->greedySet_;
}

void MGrammar_MakeGrammar::setRuleOption (
   const MParser_Token_var& key,
   const MParser_Token_var& value
)

{
   //((RuleBlock)context().block).setOption(key, value);
   ruleBlock_->setOption(key, value);
}

void MGrammar_MakeGrammar::setSubruleOption (
   const MParser_Token_var& key,
   const MParser_Token_var& value
)

{
   Dynamic_Cast<MGrammar_AlternativeBlock>(context()->block_)->setOption(key, value);
}

void MGrammar_MakeGrammar::synPred()

{
    if (context()->block_->no_)
     {
       tool_->error("'~' cannot be applied to syntactic predicate", grammar_->getFilename(), context()->block_->getLine(), context()->block_->getColumn());
     }
    // create the right kind of object now that we know what that is
    // and switch the list of alternatives.  Adjust the stack of blocks.
    // copy any init action also.
    MGrammar_SynPredBlock_var b = new MGrammar_SynPredBlock(grammar_);
    setBlock(b, context()->block_);
    MGrammar_BlockContext_var old = blocks_.top(); // remove old scope; we want new type of subrule
    blocks_.pop();
    blocks_.push(new MGrammar_BlockContext());
    context()->block_ = b;
    context()->blockEnd_ = old->blockEnd_;
    context()->blockEnd_->block_ = b;
}

void MGrammar_MakeGrammar::zeroOrMoreSubRule()

{
    if (context()->block_->no_)
     {
       tool_->error("'~' cannot be applied to (...)+ subrule", grammar_->getFilename(), context()->block_->getLine(), context()->block_->getColumn());
     }
    // create the right kind of object now that we know what that is
    // and switch the list of alternatives.  Adjust the stack of blocks.
    // copy any init action also.
    MGrammar_ZeroOrMoreBlock_var b = new MGrammar_ZeroOrMoreBlock(grammar_);
    setBlock(b, context()->block_);
    MGrammar_BlockContext_var old = blocks_.top(); // remove old scope; we want new type of subrule
    blocks_.pop();
    blocks_.push(new MGrammar_BlockContext());
    context()->block_ = b;
    context()->blockEnd_ = old->blockEnd_;
    context()->blockEnd_->block_ = b;
}

MGrammar_MakeGrammar::~MGrammar_MakeGrammar ()

{
}
