#include "MLANG/StringBuffer.h"
#include "MIO/IOException.h"
#include "MParser/System.h"
#include "MGrammar/Tool.h"
#include "MGrammar/CodeGenerator.h"
#include "MGrammar/Preprocessor/Option.h"
#include "MGrammar/PreprocessorGrammar.h"

/* ANTLR Translator Generator
 * Project led by Terence Parr at http://www.jGuru.com
 * Software rights: http://www.antlr.org/license.html
 *
 * $Id: //depot/code/org.antlr/release/antlr-2.7.4/antlr/preprocessor/Grammar.java#1 $
 */


MGrammar_PreprocessorGrammar::MGrammar_PreprocessorGrammar (
   MGrammar_Tool_ptr tool,
   const MLANG_String& name,
   const MLANG_String& superGrammar,
   const RULES& rules
)
   : name_(name),
     superGrammar_(superGrammar),
     rules_(rules),
     predefined_(false),
     alreadyExpanded_(false),
     specifiedVocabulary_(false),
     antlrTool_(tool)

{
}

void MGrammar_PreprocessorGrammar::addOption (
   const MGrammar_PreprocessorOption_var& o
)

{
   options_[o->getName()] = o;
}

void MGrammar_PreprocessorGrammar::addRule (
   const MGrammar_PreprocessorRule_var& r
)

{
   rules_[r->getName()] = r;
}

/** Copy all nonoverridden rules, vocabulary, and options into this grammar from
 *  supergrammar chain.  The change is made in place; e.g., this grammar's vector
 *  of rules gets bigger.  This has side-effects: all grammars on path to
 *  root of hierarchy are expanded also.
 */
void MGrammar_PreprocessorGrammar::expandInPlace()

{
   // if this grammar already expanded, just return
   if (alreadyExpanded_)
    {
      return;
    }

    // Expand super grammar first (unless it's a predefined or subgrammar of predefined)
    MGrammar_PreprocessorGrammar_var superG = getSuperGrammar();
    if (MGrammar_PreprocessorGrammar::is_nil(superG))
       return; // error (didn't provide superclass)
    if (exportVocab_.length() == 0)
     {
       // if no exportVocab for this grammar, make it same as grammar name
       exportVocab_ = getName();
     }
    if (superG->isPredefined())
       return; // can't expand Lexer, Parser, ...
    superG->expandInPlace();

    // expand current grammar now.
    alreadyExpanded_ = true;
    // track whether a grammar file needed to have a grammar expanded
    MGrammar_GrammarFile_var gf = hier_->getFile(getFileName());
    gf->setExpanded(true);

    // Copy rules from supergrammar into this grammar
    RULES inhRules = superG->getRules();
    for (RULES::const_iterator iter = inhRules.begin(); iter!=inhRules.end(); iter++)
     {
        inherit(iter->second, superG);
     }

    // Copy options from supergrammar into this grammar
    // Modify tokdef options so that they point to dir of enclosing grammar
    OPTIONS inhOptions = superG->getOptions();
    if (inhOptions.size() != 0)
     {
       for (OPTIONS::const_iterator iter = inhOptions.begin(); iter!=inhOptions.end(); iter++)
        {
           inherit(iter->second, superG);
        }
     }

     // add an option to load the superGrammar's output vocab
    if ((options_.size() != 0
    &&  MGrammar_PreprocessorOption::is_nil(options_.get("importVocab"))))
     {
        // no importVocab found, add one that grabs superG's output vocab
        MGrammar_PreprocessorOption_var inputV = 
             new MGrammar_PreprocessorOption("importVocab", superG->exportVocab() + ";", this);
        addOption(inputV);
        // copy output vocab file to current dir
        MLANG_String originatingGrFileName = superG->getFileName();
        MLANG_String path = antlrTool_->pathToFile(originatingGrFileName);
        MLANG_String superExportVocabFileName = path + superG->exportVocab() +
                MGrammar_CodeGenerator::TokenTypesFileSuffix_ +
                MGrammar_CodeGenerator::TokenTypesFileExt_;
        MLANG_String newImportVocabFileName = antlrTool_->fileMinusPath(superExportVocabFileName);
        if (path.equals("." + SYSTEM->getProperty("file.separator")))
         {
           // don't copy tokdef file onto itself (must be current directory)
           // System.out.println("importVocab file same dir; leaving as " + superExportVocabFileName);
         }
        else
         {
           try
            {
               antlrTool_->copyFile(superExportVocabFileName, newImportVocabFileName);
            }
           catch (MIO_IOException& io)
            {
               antlrTool_->toolError("cannot find/copy importVocab file " + superExportVocabFileName);
               return;
            }
         }
     }

    // copy member action from supergrammar into this grammar
    inherit(superG->memberAction_, superG);
}

MGrammar_PreprocessorGrammar_var MGrammar_PreprocessorGrammar::getSuperGrammar()

{
   if (superGrammar_.length() == 0) return MGrammar_PreprocessorGrammar::_nil();
   MGrammar_PreprocessorGrammar_var g = hier_->getGrammar(superGrammar_);
   return g;
}

void MGrammar_PreprocessorGrammar::inherit (
   const MGrammar_PreprocessorOption_var& o, 
   const MGrammar_PreprocessorGrammar_var& superG
)

{
   ACE_UNUSED_ARG (superG);
   // do NOT inherit importVocab/exportVocab options under any circumstances
   if (o->getName().equals("importVocab")
   ||  o->getName().equals("exportVocab"))
    {
       return;
    }

   MGrammar_PreprocessorOption_var overriddenOption;
   overriddenOption = options_.get(o->getName());
   // if overridden, do not add to this grammar
   if (MGrammar_PreprocessorOption::is_nil(overriddenOption))
    { 
      // not overridden
      addOption(o);	// copy option into this grammar--not overridden
    }
}

void MGrammar_PreprocessorGrammar::inherit (
   const MGrammar_PreprocessorRule_var& r, 
   const MGrammar_PreprocessorGrammar_var& superG
)

{
   // if overridden, do not add to this grammar
   MGrammar_PreprocessorRule_var overriddenRule = 
       rules_.get(r->getName());
   if (!MGrammar_PreprocessorRule::is_nil(overriddenRule))
    {
      // rule is overridden in this grammar.
      if (!overriddenRule->sameSignature(r))
       {
         // warn if different sig
         antlrTool_->warning("rule " + getName() + "." + overriddenRule->getName() +
                             " has different signature than " +
                             superG->getName() + "." + overriddenRule->getName());
       }
    }
   else
    {  // not overridden, copy rule into this
      addRule(r);
    }
}

void MGrammar_PreprocessorGrammar::inherit (
   const MLANG_String& memberAction,
   const MGrammar_PreprocessorGrammar_var& superG
)

{
   ACE_UNUSED_ARG (superG);
   if (memberAction_.length() != 0) return;	// do nothing if already have member action
   if (memberAction.length() != 0) 
    { // don't have one here, use supergrammar's action
      memberAction_ = memberAction;
    }
}

void MGrammar_PreprocessorGrammar::setHierarchy (
   const MGrammar_Hierarchy_var& hier
)

{
   hier_ = hier;
}

MLANG_String MGrammar_PreprocessorGrammar::toString() const

{
   MLANG_StringBuffer s;
   if (preambleAction_.length() != 0)
    {
      s.append(preambleAction_);
    }
   if (superGrammar_.length() == 0)
    {
      return "class " + name_ + ";";
    }
   if ( superClass_.length()!=0)
    {
       // replace with specified superclass not actual grammar
       // user must make sure that the superclass derives from super grammar class
       s.append("class " + name_ + " extends " + superClass_ + ";");
    }
   else
    {
       s.append("class " + name_ + " extends " + type_ + ";");
    }
   s.append(
      SYSTEM->getProperty("line.separator") +
      SYSTEM->getProperty("line.separator"));
   if (options_.size() != 0)
    {
      s.append(MGrammar_Hierarchy::optionsToString(options_));
    }
   if (tokenSection_.length() != 0)
    {
      s.append(tokenSection_ + "\n");
    }
   if (memberAction_.length() != 0)
    {
      s.append(memberAction_ + SYSTEM->getProperty("line.separator"));
    }
   for (RULES::const_iterator iter = rules_.begin(); iter!=rules_.end(); iter++)
    {
       MGrammar_PreprocessorRule_var r = iter->second;
       if (!getName().equals(r->enclosingGrammar()->getName()))
        {
          s.append("// inherited from grammar " + r->enclosingGrammar()->getName() + SYSTEM->getProperty("line.separator"));
        }
       s.append(r->toString() +
              SYSTEM->getProperty("line.separator") +
              SYSTEM->getProperty("line.separator"));
    }
   return s.toString();
}

MGrammar_PreprocessorGrammar::~MGrammar_PreprocessorGrammar ()

{
}
