#include "MIO/Reader.h"
#include "MIO/FileReader.h"
#include "MIO/BufferedReader.h"
#include "MParser/System.h"
#include "MParser/RecognitionException.h"
#include "MParser/TokenStreamException.h"
#include "MGrammar/Preprocessor/Grammar.h"
#include "MGrammar/Preprocessor/GrammarFile.h"
#include "MGrammar/Preprocessor/Hierarchy.h"
#include "MGrammar/Preprocessor/Lexer.h"
#include "MGrammar/Preprocessor/Preprocessor.h"

/* ANTLR Translator Generator
 * Project led by Terence Parr at http://www.jGuru.com
 * Software rights: http://www.antlr.org/license.html
 *
 * $Id: //depot/code/org.antlr/release/antlr-2.7.4/antlr/preprocessor/Hierarchy.java#1 $
 */

MGrammar_Hierarchy::MGrammar_Hierarchy (
   MGrammar_Tool_ptr tool
) :
   LexerRoot_(new MGrammar_PreprocessorGrammar(tool, "Lexer")),
   ParserRoot_(new MGrammar_PreprocessorGrammar(tool, "Parser")),
   TreeParserRoot_(new MGrammar_PreprocessorGrammar(tool, "TreeParser")),
   antlrTool_(tool)

{

   LexerRoot_->setPredefined(true);
   ParserRoot_->setPredefined(true);
   TreeParserRoot_->setPredefined(true);

   symbols_.put(LexerRoot_->getName(), LexerRoot_);
   symbols_.put(ParserRoot_->getName(), ParserRoot_);
   symbols_.put(TreeParserRoot_->getName(), TreeParserRoot_);
}

void MGrammar_Hierarchy::addGrammar (
   const MGrammar_PreprocessorGrammar_var& gr
)

{
   gr->setHierarchy(MGrammar_Hierarchy::_duplicate(this));
   // add grammar to hierarchy
   symbols_.put(gr->getName(), gr);
   // add grammar to file.
   MGrammar_GrammarFile_var f = getFile(gr->getFileName());
   f->addGrammar(gr);
}

void MGrammar_Hierarchy::addGrammarFile (
   const MGrammar_GrammarFile_var& gf
)

{
   files_.put(gf->getName(), gf);
}

void MGrammar_Hierarchy::expandGrammarsInFile(const MLANG_String& fileName)

{
   MGrammar_GrammarFile_var f = getFile(fileName);
   for (MGrammar_GrammarFile::GRAMMARS::const_iterator iter = f->getGrammars().begin();
        iter!=f->getGrammars().end();
        iter++)
    {
       MGrammar_PreprocessorGrammar_var g = iter->second;
       g->expandInPlace();
    }
}

MGrammar_PreprocessorGrammar_var MGrammar_Hierarchy::findRoot (
   const MGrammar_PreprocessorGrammar_var& g
)

{
    if (g->getSuperGrammarName().length() == 0)
     {		// at root
       return g;
     }
    // return root of super.
    MGrammar_PreprocessorGrammar_var sg = g->getSuperGrammar();
    if (MGrammar_PreprocessorGrammar::is_nil(sg)) 
       return g;		// return this grammar if super missing
    return findRoot(sg);
}

MGrammar_GrammarFile_var MGrammar_Hierarchy::getFile (
   const MLANG_String& fileName
)

{
   MGrammar_GrammarFile_var temp = files_.get(fileName);
   return temp;
}

MGrammar_PreprocessorGrammar_var MGrammar_Hierarchy::getGrammar (
   const MLANG_String& gr
)

{
   return symbols_.get(gr);
}

MLANG_String MGrammar_Hierarchy::optionsToString (
   const PREOPTIONS& options)

{
   MLANG_String s = "options {" + SYSTEM->getProperty("line.separator");
   for (PREOPTIONS::const_iterator iter = options.begin(); 
        iter!=options.end();
        iter++)
    {
       s += iter->second->toString() + SYSTEM->getProperty("line.separator");
    }
   s += "}" +
        SYSTEM->getProperty("line.separator") +
        SYSTEM->getProperty("line.separator");
   return s;
}

void MGrammar_Hierarchy::readGrammarFile(const MLANG_String& file)

{
   MIO_Reader_var grStream = new MIO_BufferedReader(new MIO_FileReader(file));
   addGrammarFile(new MGrammar_GrammarFile(antlrTool_, file));

   // Create the simplified grammar lexer/parser
   MGrammar_PreprocessorLexer_var ppLexer = new MGrammar_PreprocessorLexer(grStream);
   ppLexer->setFilename(file);
   MGrammar_Preprocessor_var pp = new MGrammar_Preprocessor (
         MParser_TokenStream::_duplicate(ppLexer));
   pp->setTool(antlrTool_);
   pp->setFilename(file);

   // populate the hierarchy with class(es) read in
   try
    {
      pp->grammarFile(this, file);
    }
   catch (MParser_TokenStreamException& io)
    {
      antlrTool_->toolError(MLANG_String("Token stream error reading grammar(s):\n") + io.toString().c_str());
    }
   catch (MParser_Exception& se)
    {
      antlrTool_->toolError(MLANG_String("error reading grammar(s):\n") + se.toString().c_str());
    }
}

/** Return true if hierarchy is complete, false if not */
bool MGrammar_Hierarchy::verifyThatHierarchyIsComplete()

{
   fprintf(stderr,"MGrammar_Hierarchy::verifyThatHierachyIsComplete: enter\n");
   bool complete = true;
   // Make a pass to ensure all grammars are defined
   for (SYMBOLS::const_iterator iter = symbols_.begin();
        iter!=symbols_.end();
        iter++)
    {
       MGrammar_PreprocessorGrammar_var c = iter->second;
       if (c->getSuperGrammarName().length() == 0)
        {
          continue;		// at root: ignore predefined roots
        }
       MGrammar_PreprocessorGrammar_var superG = c->getSuperGrammar();
       if (MGrammar_PreprocessorGrammar::is_nil(superG))
        {
          antlrTool_->toolError("grammar " + c->getSuperGrammarName() + " not defined");
          complete = false;
#if 0
          symbols_.remove(c->getName()); // super not defined, kill sub
#else
          assert(0);
#endif
        }
    }

   if (!complete) 
    {
      fprintf(stderr,"MGrammar_Hierarchy::verifyThatHierachyIsComplete: leave\n");
      return false;
    }

   // Make another pass to set the 'type' field of each grammar
   // This makes it easy later to ask a grammar what its type
   // is w/o having to search hierarchy.
   for (SYMBOLS::const_iterator iter = symbols_.begin();
        iter!= symbols_.end();
        iter++)
    {
       MGrammar_PreprocessorGrammar_var c = iter->second;
       if (c->getSuperGrammarName().length() == 0)
        {
          continue;		// ignore predefined roots
        }
       c->setType(findRoot(c)->getName());
     }

   fprintf(stderr,"MGrammar_Hierarchy::verifyThatHierachyIsComplete: leave\n");
    return true;
}

MGrammar_Hierarchy::~MGrammar_Hierarchy ()

{
}
