#include "MGrammar/Tool.h"
#include "MGrammar/Grammar.h"
#include "MGrammar/CodeGenerator.h"
#include "MGrammar/ExceptionSpec.h"
#include "MGrammar/AlternativeBlock.h"
#include "MGrammar/RuleBlock.h"

/* ANTLR Translator Generator
 * Project led by Terence Parr at http://www.jGuru.com
 * Software rights: http://www.antlr.org/license.html
 *
 * $Id: //depot/code/org.antlr/release/antlr-2.7.4/antlr/RuleBlock.java#1 $
 */

/**A list of alternatives and info contained in
 * the rule definition.
 */
    // grammar-settable options

    /** Construct a named rule. */
MGrammar_RuleBlock::MGrammar_RuleBlock (
   MGrammar_Grammar_ptr g,
   const MLANG_String& r
) : MGrammar_AlternativeBlock(g),
    ruleName_ (r),
    testLiterals_(false),
    cache_(g->maxk_ + 1),
    defaultErrorHandler_(true)

{
   for (int j=0; j<g->maxk_ + 1; j++)
      cache_[j].set_null();
   setAutoGen(g->isParser());
}

/** Construct a named rule with line number information */
MGrammar_RuleBlock::MGrammar_RuleBlock (
   MGrammar_Grammar_ptr g, 
   const MLANG_String& r,
   int line,
   bool doAutoGen
) : MGrammar_AlternativeBlock (g),
    ruleName_ (r),
    testLiterals_(false),
    cache_(g->maxk_ + 1),
    defaultErrorHandler_(true)

{
   for (int j=0; j<g->maxk_ + 1; j++)
      cache_[j].set_null();
   line_ = line;
   setAutoGen(doAutoGen);
}

void MGrammar_RuleBlock::addExceptionSpec(const MGrammar_ExceptionSpec_var& ex)

{
   if (!MGrammar_ExceptionSpec::is_nil(findExceptionSpec(ex->getLabel())))
    {
      if (!MParser_Token::is_nil(ex->getLabel()))
       {
         grammar_->antlrTool_->error("Rule '" + ruleName_ + "' already has an exception handler for label: " + ex->getLabel()->toString());
       }
      else
       {
         grammar_->antlrTool_->error("Rule '" + ruleName_ + "' already has an exception handler");
       }
    }
   else
    {
       exceptionSpecs_[MParser_Token::is_nil(ex->getLabel()) ? "" : ex->getLabel()->getText()] = ex;
    }
}

MGrammar_ExceptionSpec_var MGrammar_RuleBlock::findExceptionSpec (
   const MParser_Token_var& label
) const

{
   MGrammar_ExceptionSpec_var ex;
   EXCEPTIONSPECS::const_iterator iter = exceptionSpecs_.find (
       MParser_Token::is_nil(label) ? MLANG_String("") : label->getText());
   if (iter != exceptionSpecs_.end())
      ex = iter->second; 
   return ex;
}

MGrammar_ExceptionSpec_var MGrammar_RuleBlock::findExceptionSpec(const MLANG_String& label) const

{
   MGrammar_ExceptionSpec_var ex;
   EXCEPTIONSPECS::const_iterator iter = exceptionSpecs_.find (label);
   if (iter != exceptionSpecs_.end())
      ex = iter->second; 
   return ex;
}

void MGrammar_RuleBlock::generate()

{
   grammar_->generator_->gen(this);
}

bool MGrammar_RuleBlock::isLexerAutoGenRule() const

{
   return ruleName_.equals("nextToken");
}

MGrammar_Lookahead MGrammar_RuleBlock::look(int k)

{
   return grammar_->theLLkAnalyzer_->look(k, this);
}

void MGrammar_RuleBlock::prepareForAnalysis()

{
   MGrammar_AlternativeBlock::prepareForAnalysis();
   lock_.resize (0, false);
   lock_.resize (grammar_->maxk_ + 1, false);
}

void MGrammar_RuleBlock::setEndElement(const MGrammar_RuleEndElement_var& re)

{
   endNode_ = re;
}

void MGrammar_RuleBlock::setOption (
   const MParser_Token_var& key,
   const MParser_Token_var& value
)

{
   if (key->getText().equals("defaultErrorHandler"))
    {
      if (value->getText().equals("true"))
       {
         defaultErrorHandler_ = true;
       }
      else if (value->getText().equals("false"))
       {
         defaultErrorHandler_ = false;
       }
      else
       {
         grammar_->antlrTool_->error("Value for defaultErrorHandler must be true or false", grammar_->getFilename(), key->getLine(), key->getColumn());
       }
    }
   else if (key->getText().equals("testLiterals"))
    {
      if (!grammar_->isLexer())
       {
         grammar_->antlrTool_->error("testLiterals option only valid for lexer rules", grammar_->getFilename(), key->getLine(), key->getColumn());
       }
      else
       {
         if (value->getText().equals("true"))
          {
            testLiterals_ = true;
          }
         else if (value->getText().equals("false"))
          {
            testLiterals_ = false;
          }
         else
          {
            grammar_->antlrTool_->error("Value for testLiterals must be true or false", grammar_->getFilename(), key->getLine(), key->getColumn());
          }
       }
     }
    else if (key->getText().equals("ignore"))
     {
       if (!grammar_->isLexer())
        {
          grammar_->antlrTool_->error("ignore option only valid for lexer rules", grammar_->getFilename(), key->getLine(), key->getColumn());
        }
       else
        {
          ignoreRule_ = value->getText();
        }
      }
     else if (key->getText().equals("paraphrase"))
      {
        if (!grammar_->isLexer())
         {
           grammar_->antlrTool_->error("paraphrase option only valid for lexer rules", grammar_->getFilename(), key->getLine(), key->getColumn());
         }
        else
         {
           // find token def associated with this rule
           MGrammar_TokenSymbol_var ts = grammar_->tokenManager_->getTokenSymbol(ruleName_);
           if (MGrammar_TokenSymbol::is_nil(ts))
            {
               grammar_->antlrTool_->panic("cannot find token associated with rule " + ruleName_);
            }
           ts->setParaphrase(value->getText());
         }
      }
     else if (key->getText().equals("generateAmbigWarnings"))
      {
        if (value->getText().equals("true"))
         {
           generateAmbigWarnings_ = true;
         }
        else if (value->getText().equals("false"))
         {
           generateAmbigWarnings_ = false;
         }
        else
         {
           grammar_->antlrTool_->error("Value for generateAmbigWarnings must be true or false", grammar_->getFilename(), key->getLine(), key->getColumn());
         }
      }
     else
      {
        grammar_->antlrTool_->error("Invalid rule option: " + key->getText(), grammar_->getFilename(), key->getLine(), key->getColumn());
      }
}

MLANG_String MGrammar_RuleBlock::toString() const

{
   MLANG_String s = " FOLLOW={";
   ACE_Vector<MGrammar_Lookahead> cache = endNode_->getCache();
   int k = grammar_->maxk_;
   bool allNull = true;
   for (int j = 1; j <= k; j++)
    {
      if (cache[j].nil()) continue;
      s += cache[j].toString(",", grammar_->tokenManager_->getVocabulary());
      allNull = false;
      if (j < k && !cache[j + 1].nil()) s += ";";
    }
   s += "}";
   if (allNull) s = "";
   return ruleName_ + ": " + MGrammar_AlternativeBlock::toString() + " ;" + s;
}

MGrammar_RuleBlock::~MGrammar_RuleBlock ()

{
}
