#include "MLANG/Class.h"
#include "MIO/File.h"
#include "MIO/IOException.h"
#include "MIO/Reader.h"
#include "MIO/FileReader.h"
#include "MIO/InputStreamReader.h"
#include "MIO/BufferedReader.h"
#include "MIO/Writer.h"
#include "MIO/BufferedWriter.h"
#include "MIO/PrintWriter.h"
#include "MParser/System.h"
#include "MParser/StringUtils.h"
#include "MParser/TokenStreamException.h"
#include "MParser/TokenBuffer.h"
#include "MGrammar/PreservingFileWriter.h"
#include "MGrammar/FileCopyException.h"
#include "MGrammar/LLkAnalyzer.h"
#include "MGrammar/DefaultToolErrorHandler.h"
#include "MGrammar/MakeGrammar.h"
#include "MGrammar/Preprocessor/Tool.h"
#include "MGrammar/ANTLRLexer.h"
#include "MGrammar/ANTLRParser.h"
#include "MGrammar/CodeGenerator.h"
#include "MGrammar/Tool.h"

/* ANTLR Translator Generator
 * Project led by Terence Parr at http://www.jGuru.com
 * Software rights: http://www.antlr.org/license.html
 *
 * $Id: //depot/code/org.antlr/release/antlr-2.7.4/antlr/Tool.java#1 $
 */

MLANG_String MGrammar_Tool::version_;

MLANG_String MGrammar_Tool::outputDir_ = ".";

MLANG_String MGrammar_Tool::literalsPrefix_ = "LITERAL_";

bool MGrammar_Tool::upperCaseMangledLiterals_ = false;

/** Construct a new Tool. */
MGrammar_Tool::MGrammar_Tool()
  : errorHandler_(new MGrammar_DefaultToolErrorHandler(this)),
    hasError_( false),
    genDiagnostics_(false),
    genDocBook_(false),
    genHTML_(false),
    f_(new MIO_InputStreamReader(SYSTEM->in())),
    genHashLines_(true),
    noConstructors_(false)

{
}

void MGrammar_Tool::setFileLineFormatter(MIO_FileLineFormatter_ptr formatter) 

{
   MIO_FileLineFormatter::setFormatter(formatter);
}

void MGrammar_Tool::checkForInvalidArguments (
   ACE_ARGV& args,
   const MParser_BitSet& cmdLineArgValid) 

{
   // check for invalid command line args
   for (int a = 1; a < args.argc(); a++)
    {
       if (!cmdLineArgValid.member(a))
        {
          warning(MLANG_String("invalid command-line argument: ") + args[a] + "; ignored");
        }
    }
}

/** This example is from the book _Java in a Nutshell_ by David
 * Flanagan.  Written by David Flanagan.  Copyright (c) 1996
 * O'Reilly & Associates.  You may study, use, modify, and
 * distribute this example for any purpose.  This example is
 * provided WITHOUT WARRANTY either expressed or implied.  */
void MGrammar_Tool::copyFile (
   const MLANG_String& source_name,
   const MLANG_String& dest_name
)

{
   MIO_File_var source_file = new MIO_File(source_name);
   MIO_File_var destination_file = new MIO_File(dest_name);
   MIO_Reader_var source;
   MIO_Writer_var destination;
   ACE_Vector<char> buffer;
   int bytes_read;

   // First make sure the specified source file
   // exists, is a file, and is readable.
   if (!source_file->exists() || !source_file->isFile())
      throw MParser_FileCopyException("FileCopy: no such source file: " +
                                      source_name);
   if (!source_file->canRead())
      throw MParser_FileCopyException("FileCopy: source file "
                                      "is unreadable: " + source_name);

   // If the destination exists, make sure it is a writeable file
   // and ask before overwriting it.  If the destination doesn't
   // exist, make sure the directory exists and is writeable.
   if (destination_file->exists())
    {
      if (destination_file->isFile())
       {
#if 0
         MParser_DataInputStream_var in = new MParser_DataInputStream(SYSTEM->in());
         MLANG_String response;
#endif

         if (!destination_file->canWrite())
            throw MParser_FileCopyException("FileCopy: destination "
                                            "file is unwriteable: " + dest_name);
         /*
            SYSTEM->out()->print("File " + dest_name +
            " already exists.  Overwrite? (Y/N): ");
            SYSTEM->out()->flush();
            response = in.readLine();
            if (!response.equals("Y") && !response.equals("y"))
               throw new FileCopyException("FileCopy: copy cancelled.");
          */
       }
      else
       {
         throw MParser_FileCopyException("FileCopy: destination "
                                        "is not a file: " + dest_name);
       }
    }
   else
    {
      MIO_File_var parentdir = parent(destination_file);
      if (!parentdir->exists())
         throw MParser_FileCopyException("FileCopy: destination "
                                        "directory doesn't exist: " + dest_name);
      if (!parentdir->canWrite())
         throw MParser_FileCopyException("FileCopy: destination "
                                        "directory is unwriteable: " + dest_name);
    }

   // If we've gotten this far, then everything is okay; we can
   // copy the file.
   source = new MIO_BufferedReader(new MIO_FileReader(source_file));
   destination = new MIO_BufferedWriter(new MIO_FileWriter(destination_file));

   buffer.resize (1024, char());
   while (true)
    {
      bytes_read = source->read(&buffer[0], buffer.size());
      if (bytes_read == -1) break;
      destination->write(&buffer[0], bytes_read);
    }
}

/** Perform processing on the grammar file.  Can only be called
 * from main() @param args The command-line arguments passed to
 * main().  This wrapper does the System.exit for use with command-line.
 */
void MGrammar_Tool::doEverythingWrapper(ACE_ARGV& args)

{
   int exitCode = doEverything(args);
   ACE_OS::exit(exitCode);
}

/** Process args and have ANTLR do it's stuff without calling System.exit.
 *  Just return the result code.  Makes it easy for ANT build tool.
 */
int MGrammar_Tool::doEverything(ACE_ARGV& args)

{
   // run the preprocessor to handle inheritance first.

   // Start preprocessor. This strips generates an argument list
   // without -glib options (inside preTool)
   MGrammar_PreprocessorTool_var preTool = new MGrammar_PreprocessorTool(this, args);

   bool preprocess_ok = preTool->preprocess();
   ACE_ARGV modifiedArgs = preTool->preprocessedArgList();

   // process arguments for the Tool
   processArguments(modifiedArgs);
   if (!preprocess_ok)
    {
      return 1;
    }

   f_ = getGrammarReader();

//   ANTLRLexer_var lexer = new ANTLRLexer(f_);
   MParser_TokenStream_var lexer = new ANTLRLexer(f_);
   MParser_TokenBuffer_var tokenBuf = new MParser_TokenBuffer(lexer);
   MGrammar_LLkAnalyzer_var analyzer = new MGrammar_LLkAnalyzer(this);
   MGrammar_MakeGrammar_var behavior = new MGrammar_MakeGrammar(this, args, analyzer);

   try 
    {
       ANTLRParser_var p = new ANTLRParser (
               tokenBuf, 
               MGrammar_GrammarParseBehavior::_duplicate(behavior),
               this);
       p->setFilename(grammarFile_);
       p->grammar();
       if (hasError())
        {
          fatalError("Exiting due to errors.");
        }
       checkForInvalidArguments(modifiedArgs, cmdLineArgValid_);

       // Create the right code generator according to the "language" option
       MGrammar_CodeGenerator_var codeGen;

       // SAS: created getLanguage() method so subclass can override
       //      (necessary for VAJ interface)
       MLANG_String codeGenClassName = "MGrammar_" + getLanguage(behavior) + "CodeGenerator";
       try
        {
           MLANG_Class_var codeGenClass = MLANG_Class::forName(codeGenClassName);
           codeGen = (MGrammar_CodeGenerator_ptr)codeGenClass->newInstance();
           codeGen->setBehavior(behavior);
           codeGen->setAnalyzer(analyzer);
           codeGen->setTool(this);
           codeGen->gen();
        }
       catch (MLANG_ClassNotFoundException& cnfe) 
        {
           panic("Cannot instantiate code-generator: " + codeGenClassName);
        }
       catch (MLANG_InstantiationException& ie) 
        {
           panic("Cannot instantiate code-generator: " + codeGenClassName);
        }
       catch (MLANG_IllegalArgumentException& ie) 
        {
           panic("Cannot instantiate code-generator: " + codeGenClassName);
        }
       catch (MLANG_IllegalAccessException& iae) 
        {
           panic("code-generator class '" + codeGenClassName + "' is not accessible");
        }
     }
    catch (MParser_RecognitionException& pe) 
     {
       fatalError(MLANG_String("Unhandled parser error: ") + pe.getMessage().c_str());
     }
    catch (MParser_TokenStreamException& io) 
     {
       fatalError(MLANG_String("TokenStreamException: ") + io.getMessage().c_str());
     }
    return 0;
}

/** Issue an error
 * @param s The message
 */
void MGrammar_Tool::error(const MLANG_String& s)

{
   hasError_ = true;
   SYSTEM->err()->println("error: " + s);
}

/** Issue an error with line number information
 * @param s The message
 * @param file The file that has the error (or null)
 * @param line The grammar file line number on which the error occured (or -1)
 * @param column The grammar file column number on which the error occured (or -1)
 */
void MGrammar_Tool::error (
   const MLANG_String& s, 
   const MLANG_String& file, 
   int line, 
   int column
)

{
   hasError_ = true;
   SYSTEM->err()->println(MIO_FileLineFormatter::getFormatter()->
                      getFormatString(file, line, column) + s);
}

/** When we are 1.1 compatible...
public static Object factory2 (String p, Object[] initargs) {
     Class c;
     Object o = null;
     try {
     int argslen = initargs.length;
     Class cl[] = new Class[argslen];
     for (int i=0;i&lt;argslen;i++) {
     cl[i] = Class.forName(initargs[i].getClass().getName());
     }
     c = Class.forName (p);
     Constructor con = c.getConstructor (cl);
     o = con.newInstance (initargs);
     } catch (Exception e) {
     SYSTEM->err()->println ("Can't make a " + p);
     }
     return o;
     }
     */
#if 0
public Object factory(String p)
{
  Class c;
  Object o = null;
  try
   {
     c = Class.forName(p);// get class def
     o = c.newInstance(); // make a new one
   }
  catch (Exception e)
   {
     // either class not found,
     // class is interface/abstract, or
     // class or initializer is not accessible.
     warning("Can't create an object of type " + p);
     return null;
   }
  return o;
}
#endif

MLANG_String MGrammar_Tool::fileMinusPath (
   const MLANG_String& f
)

{
   MLANG_String separator = SYSTEM->getProperty("file.separator");
   int endOfPath = f.lastIndexOf(separator);
   if (endOfPath == -1)
    {
      return f;   // no path found
    }
   return f.substring(endOfPath + 1);
}

/** Determine the language used for this run of ANTLR
 *  This was made a method so the subclass can override it
 */
MLANG_String MGrammar_Tool::getLanguage (
   MGrammar_MakeGrammar_ptr behavior
) const

{
   if (genDiagnostics_)
    {
      return "Diagnostic";
    }
   if (genHTML_)
    {
      return "HTML";
    }
   if (genDocBook_)
    {
      return "DocBook";
    }
   return behavior->language();
}

void MGrammar_Tool::help()

{
   SYSTEM->err()->println("usage: java antlr.Tool [args] file.g");
   SYSTEM->err()->println("  -o outputDir       specify output directory where all output generated.");
   SYSTEM->err()->println("  -glib superGrammar specify location of supergrammar file.");
   SYSTEM->err()->println("  -debug             launch the ParseView debugger upon parser invocation.");
   SYSTEM->err()->println("  -html              generate a html file from your grammar.");
   SYSTEM->err()->println("  -docbook           generate a docbook sgml file from your grammar.");
   SYSTEM->err()->println("  -diagnostic        generate a textfile with diagnostics.");
   SYSTEM->err()->println("  -trace             have all rules call traceIn/traceOut.");
   SYSTEM->err()->println("  -traceLexer        have lexer rules call traceIn/traceOut.");
   SYSTEM->err()->println("  -traceParser       have parser rules call traceIn/traceOut.");
   SYSTEM->err()->println("  -traceTreeParser   have tree parser rules call traceIn/traceOut.");
   SYSTEM->err()->println("  -h|-help|--help    this message");
}

void MGrammar_Tool::main (ACE_ARGV& args)

{
#if 0
   SYSTEM->err()->println("ANTLR Parser Generator   Version " +
                           Version.project_version + "   1989-2004 jGuru.com");
   version = Version.project_version;
#else
   SYSTEM->err()->println("ANTLR Parser Generator   Version "
                           "   1989-2004 jGuru.com");
#endif

   try
    {
      if (args.argc() == 1)
       {
         help();
         SYSTEM->exit(1);
       }
      for (int i = 1; i < args.argc(); ++i)
       {
          if (MLANG_String(args[i]).equals("-h")
          ||  MLANG_String(args[i]).equals("-help")
          ||  MLANG_String(args[i]).equals("--help")
            )
           {
             help();
             SYSTEM->exit(1);
           }
       }

      MGrammar_Tool_ptr theTool = new MGrammar_Tool();
      theTool->doEverything(args);
      theTool = 0;
    }
   catch (MParser_Exception& e)
    {
      SYSTEM->err()->println(SYSTEM->getProperty("line.separator") +
                               SYSTEM->getProperty("line.separator"));
      SYSTEM->err()->println("#$%%*&@# internal error: " + MLANG_String(e.toString().c_str()));
      SYSTEM->err()->println("[complain to nearest government official");
      SYSTEM->err()->println(" or send hate-mail to parrt@jguru.com;");
      SYSTEM->err()->println(" please send stack trace with report.]" +
                            SYSTEM->getProperty("line.separator"));
//      e.printStackTrace();
    }
   SYSTEM->exit(0);
}

/** This method is used by all code generators to create new output
 * files. If the outputDir set by -o is not present it will be created here.
 */
MIO_PrintWriter_ptr MGrammar_Tool::openOutputFile (
   const MLANG_String& f
)

{
   if (outputDir_ != ".")
    {
      MIO_File_var out_dir = new MIO_File(outputDir_);
      if (!out_dir->exists())
         out_dir->mkdirs();
    }
   return new MIO_PrintWriter(new MGrammar_PreservingFileWriter(outputDir_ + SYSTEM->getProperty("file.separator") + f));
}

MIO_Reader_ptr MGrammar_Tool::getGrammarReader()

{
   MIO_Reader_var f;
   try
    {
      if (grammarFile_.length()!=0)
       {
         f = new MIO_BufferedReader(new MIO_FileReader(grammarFile_));
       }
    }
   catch (MIO_IOException& e)
    {
      fatalError("cannot open grammar file " + grammarFile_);
    }
   return f._retn();
}

/** @since 2.7.2
 */
void MGrammar_Tool::reportException (
   MLANG_Exception& e,
   const MLANG_String& message
)

{
   SYSTEM->err()->println(message.length() == 0 ? e.getMessage().c_str()
                          : (message + ": " + e.getMessage().c_str()));
}

/** @since 2.7.2
 */
void MGrammar_Tool::reportProgress (
   const MLANG_String& message
)

{
   SYSTEM->out()->println(message);
}

/** An error occured that should stop the Tool from doing any work.
 *  The default implementation currently exits (via
 *  {@link java.lang.System.exit(int)} after printing an error message to
 *  <var>stderr</var>. However, the tools should expect that a subclass
 *  will override this to throw an unchecked exception such as
 *  {@link java.lang.IllegalStateException} or another subclass of
 *  {@link java.lang.RuntimeException}. <em>If this method is overriden,
 *  <strong>it must never return normally</strong>; i.e. it must always
 *  throw an exception or call System.exit</em>.
 *  @since 2.7.2
 *  @param s The message
 */
void MGrammar_Tool::fatalError(const MLANG_String& message)

{
   SYSTEM->err()->println(message);
   SYSTEM->exit(1);
}

/** Issue an unknown fatal error. <em>If this method is overriden,
 *  <strong>it must never return normally</strong>; i.e. it must always
 *  throw an exception or call System.exit</em>.
 *  @deprecated as of 2.7.2 use {@link #fatalError(String)}. By default
 *              this method executes <code>fatalError("panic");</code>.
 */
void MGrammar_Tool::panic()

{
   fatalError("panic");
}

/** Issue a fatal error message. <em>If this method is overriden,
 *  <strong>it must never return normally</strong>; i.e. it must always
 *  throw an exception or call System.exit</em>.
 *  @deprecated as of 2.7.2 use {@link #fatalError(String)}. By defaykt
 *              this method executes <code>fatalError("panic: " + s);</code>.
 * @param s The message
 */
void MGrammar_Tool::panic(const MLANG_String& s)

{
   fatalError("panic: " + s);
}

// File.getParent() can return null when the file is specified without
// a directory or is in the root directory.
// This method handles those cases.
MIO_File_ptr MGrammar_Tool::parent (
   MIO_File_ptr f
)

{
   MLANG_String dirname = f->getParent();
   if (dirname.length() == 0)
    {
      if (f->isAbsolute())
        return new MIO_File(MIO_File::separator());
      else
        return new MIO_File(SYSTEM->getProperty("user.dir"));
    }
   return new MIO_File(dirname);
}

/** Parse a list such as "f1.g;f2.g;..." and return a Vector
 *  of the elements.
 */
ACE_Vector<MLANG_String> MGrammar_Tool::parseSeparatedList (
   const MLANG_String& list,
   char separator
)

{
   ACE_Vector<MLANG_String> v;
   int pos = 0;
   int nextpos;
   while ((nextpos=list.indexOf (separator, pos))!=-1)
    {
       v.push_back (list.substring (pos, nextpos - pos));
       pos = nextpos + 1;
    }
   if (pos < (int)list.length())
      v.push_back (list.substring (pos, list.length() - pos));
   return v;
}

/** given a filename, strip off the directory prefix (if any)
 *  and return it.  Return "./" if f has no dir prefix.
 */
MLANG_String MGrammar_Tool::pathToFile (
   const MLANG_String& f
)

{
   MLANG_String separator = SYSTEM->getProperty("file.separator");
   int endOfPath = f.lastIndexOf(separator);
   if (endOfPath == -1)
    {
      // no path, use current directory
      return "." + SYSTEM->getProperty("file.separator");
    }
   return f.substring(0, endOfPath + 1);
}

/** <p>Process the command-line arguments.  Can only be called by Tool.
 * A bitset is collected of all correct arguments via setArgOk.</p>
 * @param args The command-line arguments passed to main()
 *
 */
void MGrammar_Tool::processArguments(ACE_ARGV& args)

{
    for (int i = 1; i < args.argc(); i++)
     {
        if (MLANG_String(args[i]).equals("-diagnostic"))
         {
           genDiagnostics_ = true;
           genHTML_ = false;
           setArgOK(i);
         }
        else if (MLANG_String(args[i]).equals("-o")) 
         {
           setArgOK(i);
           if (i + 1 >= args.argc()) 
            {
              error("missing output directory with -o option; ignoring");
            }
           else 
            {
              i++;
              setOutputDirectory(args[i]);
              setArgOK(i);
            }
         }
        else if (MLANG_String(args[i]).equals("-html")) 
         {
           genHTML_ = true;
           genDiagnostics_ = false;
           setArgOK(i);
         }
        else if (MLANG_String(args[i]).equals("-docbook")) 
         {
           genDocBook_ = true;
           genDiagnostics_ = false;
           setArgOK(i);
         }
        else
         {
           if (MLANG_String(args[i]).charAt(0) != '-')
            {
              // Must be the grammar file
              grammarFile_ = args[i];
              setArgOK(i);
            }
         }
     }
}


/** Issue an error; used for general tool errors not for grammar stuff
 * @param s The message
 */
void MGrammar_Tool::toolError (const MLANG_String& s)

{
   SYSTEM->err()->println("error: " + s);
}

/** Issue a warning
 * @param s the message
 */
void MGrammar_Tool::warning (const MLANG_String& s)

{
   SYSTEM->err()->println("warning: " + s);
}

/** Issue a warning with line number information
 * @param s The message
 * @param file The file that has the warning (or null)
 * @param line The grammar file line number on which the warning occured (or -1)
 * @param column The grammar file line number on which the warning occured (or -1)
 */
void MGrammar_Tool::warning (
   const MLANG_String& s,
   const MLANG_String& file,
   int line,
   int column)

{
   SYSTEM->err()->println(MIO_FileLineFormatter::getFormatter()->
                          getFormatString(file, line, column) + "warning:" + s);
}

/** Issue a warning with line number information
 * @param s The lines of the message
 * @param file The file that has the warning
 * @param line The grammar file line number on which the warning occured
 */
void MGrammar_Tool::warning (
   const ACE_Vector<MLANG_String>& s,
   const MLANG_String& file,
   int line,
   int column
)

{
   if (s.size() == 0)
    {
      panic("bad multi-line message to Tool.warning");
    }
   SYSTEM->err()->println(MIO_FileLineFormatter::getFormatter()->
                          getFormatString(file, line, column) + "warning:" + s[0]);
   for (size_t i = 1; i < s.size(); i++)
    {
       SYSTEM->err()->println(MIO_FileLineFormatter::getFormatter()->
                          getFormatString(file, line, column) + "    " + s[i]);
    }
}

/**
 * Support C++ & C# namespaces (for now).
 * C++: Add a nested namespace name to the current namespace.
 * C# : Specify an enclosing namespace for the generated code.
 * DAW: David Wagner -- C# support by kunle odutola
 */
void MGrammar_Tool::setNameSpace(const MLANG_String& name)

{
   if (!nameSpace)
      nameSpace = new MGrammar_NameSpace(MParser_StringUtils::stripFrontBack(name, "\"", "\""));
}

MGrammar_Tool::~MGrammar_Tool()

{
}
