#ifndef __MGRAMMAR_TOOL_H__
#define __MGRAMMAR_TOOL_H__
#include "ace/ARGV.h"
#include "ace/Vector_T.h"
#include "MLANG/String.h"
#include "MIO/File.h"
#include "MIO/Reader.h"
#include "MIO/PrintWriter.h"
#include "MParser/BitSet.h"
#include "MParser/Exception.h"
#include "MParser/FileLineFormatter.h"
#include "MGrammar/NameSpace.h"
#include "MGrammar/ToolErrorHandler.h"

class MGrammar_Tool;
typedef class MGrammar_Tool *MGrammar_Tool_ptr;
typedef class MGrammar_Tool *MGrammar_Tool_var;

#include "MGrammar/MakeGrammar.h"

/* ANTLR Translator Generator
 * Project led by Terence Parr at http://www.jGuru.com
 * Software rights: http://www.antlr.org/license.html
 *
 * $Id: //depot/code/org.antlr/release/antlr-2.7.4/antlr/Tool.java#1 $
 */

class MGrammar_Tool

{
public:
    static MLANG_String version_;

    /** Object that handles analysis errors */
    MGrammar_ToolErrorHandler_var errorHandler_;

    /** Was there an error during parsing or analysis? */
    bool hasError_;

    /** Generate diagnostics? (vs code) */
    bool genDiagnostics_;

    /** Generate DocBook vs code? */
    bool genDocBook_;

    /** Generate HTML vs code? */
    bool genHTML_;

    /** Current output directory for generated files */
protected:
    static MLANG_String outputDir_;

    // Grammar input
    MLANG_String grammarFile_;
    MIO_Reader_var f_;
    // SAS: changed for proper text io
    //  transient DataInputStream in = null;

    static MLANG_String literalsPrefix_;
    static bool upperCaseMangledLiterals_;

    /** C++ file level options */
protected:
    MGrammar_NameSpace_var nameSpace;
    MLANG_String namespaceAntlr_;
    MLANG_String namespaceStd_;
    bool genHashLines_;
    bool noConstructors_;

    MParser_BitSet cmdLineArgValid_;

public:

    /** Construct a new Tool. */
    MGrammar_Tool();

    virtual ~MGrammar_Tool();

public:

    const MLANG_String& version() const
    { return version_; }

    bool noConstructors() const
    { return noConstructors_; }

    MGrammar_Tool& noConstructors(bool new_val)
    { noConstructors_ = new_val; return *this; }

    const MLANG_String& grammarFile() const
    { return grammarFile_; }

    MGrammar_Tool& grammarFile(const MLANG_String& new_grammarFile)
    { grammarFile_ = new_grammarFile; return *this; }

    bool hasError() const
    {
        return hasError_;
    }

    const MGrammar_NameSpace_var& getNameSpace() const
    {
        return nameSpace;
    }

    void setNameSpace(const MLANG_String& ns);

    const MLANG_String& getNamespaceStd() const
    {
        return namespaceStd_;
    }

    void setNamespaceStd(const MLANG_String& new_val)
    {
        namespaceStd_ = new_val;
    }

    const MLANG_String& getNamespaceAntlr() const
    {
        return namespaceAntlr_;
    }

    void setNamespaceAntlr(const MLANG_String& new_val)
    {
        namespaceAntlr_ = new_val;
    }

    bool getGenHashLines() const
    {
        return genHashLines_;
    }

    void setGenHashLines(bool new_val)
    {
        genHashLines_ = new_val;
    }

    const MLANG_String& getLiteralsPrefix() const
    {
        return literalsPrefix_;
    }

    void setLiteralsPrefix (const MLANG_String& prefix)
    { literalsPrefix_ = prefix; }

    bool getUpperCaseMangledLiterals() const
    {
        return upperCaseMangledLiterals_;
    }

    void setUpperCaseMangledLiterals(bool new_upper)
    {
        upperCaseMangledLiterals_ = new_upper;
    }

    void setFileLineFormatter(MIO_FileLineFormatter_ptr formatter);

protected:

    void checkForInvalidArguments (
       ACE_ARGV& args, 
       const MParser_BitSet& cmdLineArgValid
    );

public:

    /** This example is from the book _Java in a Nutshell_ by David
     * Flanagan.  Written by David Flanagan.  Copyright (c) 1996
     * O'Reilly & Associates.  You may study, use, modify, and
     * distribute this example for any purpose.  This example is
     * provided WITHOUT WARRANTY either expressed or implied.  */
    void copyFile (
       const MLANG_String& source_name, 
       const MLANG_String& dest_name
     );

    /** Perform processing on the grammar file.  Can only be called
     * from main() @param args The command-line arguments passed to
     * main().  This wrapper does the System.exit for use with command-line.
     */
    void doEverythingWrapper(ACE_ARGV& args);

    /** Process args and have ANTLR do it's stuff without calling System.exit.
     *  Just return the result code.  Makes it easy for ANT build tool.
     */
    int doEverything(ACE_ARGV& args);

    /** Issue an error
     * @param s The message
     */
    void error(const MLANG_String& s);

    /** Issue an error with line number information
     * @param s The message
     * @param file The file that has the error (or null)
     * @param line The grammar file line number on which the error occured (or -1)
     * @param column The grammar file column number on which the error occured (or -1)
     */
    void error (
       const MLANG_String& s, 
       const MLANG_String& file,
       int line,
       int column
     );

#if 0
    Object factory(const MLANG_String& p);
#endif

    MLANG_String fileMinusPath(const MLANG_String& f);

    /** Determine the language used for this run of ANTLR
     *  This was made a method so the subclass can override it
     */
    MLANG_String getLanguage(MGrammar_MakeGrammar_ptr behavior) const;

    const MLANG_String& getOutputDirectory() const
    {
        return outputDir_;
    }

private:

    static void help();

public:

   static void main(ACE_ARGV& args);

   MIO_PrintWriter_ptr openOutputFile(const MLANG_String& f);

   MIO_Reader_ptr getGrammarReader();

    /** @since 2.7.2
     */
    void reportException(MLANG_Exception& e, const MLANG_String& message);

    /** @since 2.7.2
     */
    void reportProgress(const MLANG_String& message);

    /** An error occured that should stop the Tool from doing any work.
     *  The default implementation currently exits (via
     *  {@link java.lang.System.exit(int)} after printing an error message to
     *  <var>stderr</var>. However, the tools should expect that a subclass
     *  will override this to throw an unchecked exception such as
     *  {@link java.lang.IllegalStateException} or another subclass of
     *  {@link java.lang.RuntimeException}. <em>If this method is overriden,
     *  <strong>it must never return normally</strong>; i.e. it must always
     *  throw an exception or call System.exit</em>.
     *  @since 2.7.2
     *  @param s The message
     */
    void fatalError(const MLANG_String& message);

    /** Issue an unknown fatal error. <em>If this method is overriden,
     *  <strong>it must never return normally</strong>; i.e. it must always
     *  throw an exception or call System.exit</em>.
     *  @deprecated as of 2.7.2 use {@link #fatalError(String)}. By default
     *              this method executes <code>fatalError("panic");</code>.
     */
    void panic();

    /** Issue a fatal error message. <em>If this method is overriden,
     *  <strong>it must never return normally</strong>; i.e. it must always
     *  throw an exception or call System.exit</em>.
     *  @deprecated as of 2.7.2 use {@link #fatalError(String)}. By defaykt
     *              this method executes <code>fatalError("panic: " + s);</code>.
     * @param s The message
     */
    void panic(const MLANG_String& s);

    // File.getParent() can return null when the file is specified without
    // a directory or is in the root directory.
    // This method handles those cases.
    MIO_File_ptr parent(MIO_File_ptr f);

    /** Parse a list such as "f1.g;f2.g;..." and return a Vector
     *  of the elements.
     */
    static ACE_Vector<MLANG_String> parseSeparatedList(const MLANG_String& list, char separator);

    /** given a filename, strip off the directory prefix (if any)
     *  and return it.  Return "./" if f has no dir prefix.
     */
    MLANG_String pathToFile(const MLANG_String& f);

    /** <p>Process the command-line arguments.  Can only be called by Tool.
     * A bitset is collected of all correct arguments via setArgOk.</p>
     * @param args The command-line arguments passed to main()
     *
     */
    void processArguments(ACE_ARGV& args);

public:

    void setArgOK(int i)
    {
        cmdLineArgValid_.add(i);
    }

    void setOutputDirectory(const MLANG_String& o)
    {
        outputDir_ = o;
    }

    /** Issue an error; used for general tool errors not for grammar stuff
     * @param s The message
     */
    void toolError(const MLANG_String& s);

    /** Issue a warning
     * @param s the message
     */
    void warning(const MLANG_String& s);

    /** Issue a warning with line number information
     * @param s The message
     * @param file The file that has the warning (or null)
     * @param line The grammar file line number on which the warning occured (or -1)
     * @param column The grammar file line number on which the warning occured (or -1)
     */
    void warning(const MLANG_String& s, const MLANG_String& file, int line, int column);

    /** Issue a warning with line number information
     * @param s The lines of the message
     * @param file The file that has the warning
     * @param line The grammar file line number on which the warning occured
     */
    void warning(const ACE_Vector<MLANG_String>& s, const MLANG_String& file, int line, int column);

    /**
     * Support C++ & C# namespaces (for now).
     * C++: Add a nested namespace name to the current namespace.
     * C# : Specify an enclosing namespace for the generated code.
     * DAW: David Wagner -- C# support by kunle odutola
     */
#if 0
    void setNameSpace(String name);
#endif
};

#endif /* __MGRAMMAR_TOOL_H__ */
