#include "ace/OS.h"
#include "MLANG/StringBuffer.h"
#include "MIO/BufferedReader.h"

MIO_BufferedReader::MIO_BufferedReader (
    const MIO_Reader_var& in,
    size_t sz
) : input_(in),
    buffer_((char*) ACE_OS::malloc(sizeof(char)*sz)),
    size_(sz),
    off_(0),
    count_(0)

{
}

/// Indica si el flujo de entrada se encuentra disponible para
/// ser leido.
/// @retval false, subclases deben implementar este metodo
bool MIO_BufferedReader::ready() 

{
   return count_>0 || input_->ready();
}

/// Marca la posicion actual en el flujo
///
/// @param readAheadLimit Limita el numero de caracteres que pueden ser
/// leidos mientras se preserva la marca. Despues de leer esta cantidad
/// de caracteres reintentar reiniciar el flujo podria falla. 
void MIO_BufferedReader::mark (int readAheadLimit)

{
   ACE_UNUSED_ARG(readAheadLimit);
}

void MIO_BufferedReader::fill()

{
   off_ = 0;
   count_ = input_->read (buffer_, size_);
}

/// Lee una caracter del flujo de entrada
///
/// @retval el caracter leído
/// @retval -1, si se encontro el fin de archivo
/// @throw MIO_IOException Si hubo algun error de entrada/salida
ssize_t MIO_BufferedReader::read ()

{
    if (!(count_>0))
     {
       this->fill();
     }
    if (!(count_>0))
     {
       return -1;
     }
    int c = buffer_[off_++];
    count_--;
    return (char)c;
}

/// Lee un buffer de caracteres
/// 
/// @param buf apunta al buffer de caracters donde se colocaran los caracteres
/// leidos.
/// @param len es el maximo numero de caracteres a ser leidos.
/// @retval el numero de caracteres leidos, o
/// @retval -1 si se ha encontrado el fin del flujo.
/// @throw MIO_IOException si hubo algun error de entrada/salida
ssize_t MIO_BufferedReader::read (char buf[], ssize_t len)

{
   ssize_t tot = 0;
   if (count_>0)
    {
      ssize_t n = len;
      if (count_<n)
         n = count_;
      ACE_OS::memcpy (buf, &buffer_[off_], n);
      count_ -= n;
      tot += n;
      off_ += n;
    }
   if (len>tot)
    {
      int result = input_->read (buf, len - tot);
      if (result != -1)
         tot += result;   
      else if (tot == 0)
         return -1;      
    }
   return tot;
}

/// Cierra el flujo de lectura
///
/// @throws MIO_IOException si hubo algun error de entrada
/// salida
void MIO_BufferedReader::close()

{
   input_->close();
}

/// Reinicia el flujo a la marca mas reciente
///
void MIO_BufferedReader::reset()

{
}

MLANG_String MIO_BufferedReader::readLine()

{
   int c;
   MLANG_StringBuffer sb;
   while ((c=read())!=-1)
    {
       if (c!='\n')
        {
          return sb.toString();
        }
       sb.append ((char) c);
    }
   return sb.toString();
}

MIO_BufferedReader::~MIO_BufferedReader ()

{
   this->close();
   if (buffer_!=0)
    {
      ACE_OS::free(buffer_);
      buffer_ = 0;
    }
}

