//$Id: Class.cpp,v 1.12 2004/12/01 16:42:37 roiser Exp $

#include "MLANG/Reflection/Class.h"
#include "MLANG/Reflection/DataMember.h"
#include "MLANG/Reflection/FunctionMember.h"
#include "MLANG/Reflection/Base.h"
#include "MLANG/Reflection/Tools.h"
#include "MLANG/Reflection/Object.h"
#include "MLANG/Reflection/Type.h"
#include <typeinfo>
#include <iostream>
#include <algorithm>
#if defined (__linux) || defined (__APPLE__)
#include <cxxabi.h>
#endif

//-------------------------------------------------------------------------------
MLANG_Reflection::Class::Class(  const char *           type, 
                             size_t                 size,
                             const std::type_info & typeInfo,
                             unsigned int           modifiers,
                             TYPE                   classType )
//-------------------------------------------------------------------------------
  : TypeBase( type, size, classType, typeInfo ),
    ScopeBase( type, classType ),
    m_modifiers( modifiers ),
    m_allBases( 0 ),
    m_completeType( false ),
    m_constructors( std::vector< Member >()),
    m_destructor( Member()),
    m_pathsToBase( PathsToBase()) {}
    

//-------------------------------------------------------------------------------
MLANG_Reflection::Class & MLANG_Reflection::Class::addBase( const Type &   btype,
                                                    OffsetFunction offsetFP,
                                                    unsigned int   modifiers ) {
//-------------------------------------------------------------------------------
  Base b = Base( btype, offsetFP, modifiers );
  m_bases.push_back( b );
  return * this;
}

    
//-------------------------------------------------------------------------------
void * MLANG_Reflection::Class::castObject( const Type & to, 
                                        void * obj ) const {
//-------------------------------------------------------------------------------
  std::vector< Base > path = std::vector< Base >();
  if ( hasBase( to, path )) { // up cast 
    // in case of up cast the offset has to be calculated by Reflex
    size_t obj2 = (size_t)obj;
    for( std::vector< Base >::reverse_iterator bIter = path.rbegin();
         bIter != path.rend(); ++bIter ) {
      obj2 += bIter->offset((void*)obj2);
    }
    return (void*)obj2;
  }
  path.clear();
  Type t = *this;
  if ( ((Class)to).hasBase( t, path ) ) {  // down cast
  // use the internal dynamic casting of the compiler (e.g. libstdc++.so)
#if defined (__linux) || defined (__APPLE__)
    return abi::__dynamic_cast(obj,
                               (const abi::__class_type_info*)&this->typeInfo(),
                               (const abi::__class_type_info*)&to.typeInfo(),
                               -1); 
#elif defined (_WIN32)
    return __RTDynamicCast(obj,
                           0,
                           (void*)&this->typeInfo(),
                           (void*)&to.typeInfo(),
                           0);
#endif
  }
  // fixme cross cast missing ?? internal cast possible ??
  return 0;
}

    
//-------------------------------------------------------------------------------
MLANG_Reflection::Object MLANG_Reflection::Class::construct( const Type & signature, 
                                                     std::vector < void * > args, 
                                                     void * mem ) const {
//------------------------------------------------------------------------------- 
  static Type defSignature = Type::byName("void (void)");
  Type signature2 = signature;
  
  Member constructor = Member();
  if ( !signature &&  m_constructors.size() > 1 ) 
   {
     signature2 = defSignature; 
   }
  
  for (size_t i = 0; i < m_constructors.size(); ++ i) {
    if ( !signature2 || m_constructors[i].type().id() == signature2.id()) {
      constructor = m_constructors[i];
      break;
    }
  }
  
  if ( constructor.type() ) {
    // no memory address passed -> allocate memory for class
    if ( mem == 0 ) mem = allocate();
    Object obj = Object( type(), mem );
    constructor.invoke( obj, args );
    return obj;
  }
  else {
    throw RuntimeError("No suitable constructor found");
  }
}

    
//-------------------------------------------------------------------------------
void MLANG_Reflection::Class::destruct( void * instance, 
                                    bool dealloc ) const {
//-------------------------------------------------------------------------------
  if ( ! m_destructor.type() ) {
    // destructor for this class not yet revealed
    for ( size_t i = 0; i < functionMemberCount(); ++i ) {
      Member fm = functionMember( i );
      // constructor found set the cache pointer
      if ( fm.isDestructor() ) {
        m_destructor = fm; 
        break;
      }
    }
  }
  if ( m_destructor.type()) {
    // we found a destructor -> invoke it
    Object dummy = Object(Type(), instance);
    m_destructor.invoke( dummy );
    // deallocation of memory wanted 
    if ( dealloc ) { deallocate( instance ); }
  }
  else {
    // this class has no destructor defined (????)
    std::cerr << Reflex::argv0 << ": ERROR: No destructor defined for class "
              << fullName() << std::endl;
  }
}


//-------------------------------------------------------------------------------
struct DynamicType {
//-------------------------------------------------------------------------------
  virtual ~DynamicType() {}
};

    
//-------------------------------------------------------------------------------
MLANG_Reflection::Class MLANG_Reflection::Class::dynamicType( const void * mem ) const {
//-------------------------------------------------------------------------------
  // If no virtual_function_table return itself
  if ( isVirtual() ) {
    // Avoid the case that the first word is a virtual_base_offset_table instead of
    // a virtual_function_table  
    long offset = **(long**)mem;
    if ( offset == 0 ) return * this;
    else {
      Type dytype = Type::byTypeInfo(typeid(*(DynamicType*)mem));
      if ( dytype && dytype.isClass() ) return dytype;
      else                                           return * this;
    }
  }
  else {
    return * this; 
  }
}


//-------------------------------------------------------------------------------
bool MLANG_Reflection::Class::hasBase( const Type & cl ) const {
//-------------------------------------------------------------------------------
  std::vector<Base> v = std::vector<Base>();
  return hasBase(cl, v);
}


//-------------------------------------------------------------------------------
bool MLANG_Reflection::Class::hasBase( const Type & cl,  
                                   std::vector< Base > & path ) const {
//-------------------------------------------------------------------------------
  for ( size_t i = 0; i < baseCount(); ++i ) {
    // is the final base class one of the current class ?
    if ( base( i ).toType().id() == cl.id() ) { 
      // remember the path to this class
      path.push_back( base( i )); 
      return true; 
    }
    // if searched base class is not direct base look in the bases of this one
    else if ( base( i ).toType() && (((Class)(base( i ).toType())).hasBase( cl, path ))) {
      // if successfull remember path
      path.push_back( base( i )); 
      return true; 
    }
  }
  return false;
}


//-------------------------------------------------------------------------------
bool MLANG_Reflection::Class::isComplete() const {
//-------------------------------------------------------------------------------
  if ( ! m_completeType ) m_completeType = isComplete2(); 
  return m_completeType;
}


//-------------------------------------------------------------------------------
bool MLANG_Reflection::Class::isComplete2() const {
//-------------------------------------------------------------------------------
  for (size_t i = 0; i < baseCount(); ++i) {
    Type baseType = base( i ).toType();
    if ( ! baseType )  return false; 
    Class baseClass = baseType;
    if ( ! baseClass.isComplete2()) return false; 
  }
  return true;
}

    
//-------------------------------------------------------------------------------
size_t MLANG_Reflection::Class::allBases() const {
//-------------------------------------------------------------------------------
  size_t allBases = 0;
  for ( size_t i = 0; i < baseCount(); ++i ) {
    ++allBases;
    Type baseType = base( i ).toType();
    if ( baseType ) { 
      allBases += ((Class)baseType).allBases();
    }
  }
  return allBases;
}


//-------------------------------------------------------------------------------
bool MLANG_Reflection::Class::newBases() const {
//-------------------------------------------------------------------------------
  if ( ! m_completeType ) {
    size_t numBases = allBases();
    if ( m_allBases != numBases ) {
      m_completeType = isComplete2();
      m_allBases = numBases;
      return true;
    }
  }
  return false;
}


//-------------------------------------------------------------------------------
void MLANG_Reflection::Class::updateMembers() const {
//-------------------------------------------------------------------------------
  if ( newBases()) {
    std::vector < OffsetFunction > basePath = std::vector < OffsetFunction >();
    updateMembers2( m_members, 
                    m_dataMembers, 
                    m_functionMembers,
                    m_pathsToBase,
                    basePath );
  }
}

    
//-------------------------------------------------------------------------------
const std::vector < MLANG_Reflection::OffsetFunction > & 
MLANG_Reflection::Class::pathToBase( const Scope & base ) const {
//-------------------------------------------------------------------------------
  std::vector < OffsetFunction > * pathToBase = m_pathsToBase[ base.id() ];
  if ( ! pathToBase ) {
    updateMembers();
    pathToBase = m_pathsToBase[ base.id() ];
    /* fixme can get rid of updateMembers() ?
    std::cerr << Reflex::argv0 << ": WARNING: No path found from " 
              << this->fullName() << " to " << base.fullName() << std::endl;
    if ( newBases()) {
      std::cerr << Reflex::argv0 << ": INFO: Not all base classes have resolved, "
                << "do Class::updateMembers() and try again " << std::endl; 
    }
    */
  }
  return * pathToBase;
}


//-------------------------------------------------------------------------------
void MLANG_Reflection::Class::updateMembers2( Members & members,
                                          Members & dataMembers,
                                          Members & functionMembers,
                                          PathsToBase & pathsToBase,
                                          std::vector < OffsetFunction > & basePath ) const {
//-------------------------------------------------------------------------------
  std::vector < Base >::const_iterator bIter;
  for ( bIter = m_bases.begin(); bIter != m_bases.end(); ++bIter ) {
    Type bType = bIter->toType();
    Class bClass = bType;
    basePath.push_back( bIter->offsetFP());
    pathsToBase[ bClass.scope().id() ] = new std::vector < OffsetFunction >( basePath );
    if ( bType ) {
      size_t i = 0;
      for ( i = 0; i < bClass.dataMemberCount(); ++i ) {
        Member dm = bClass.dataMember(i);
        if ( find( dataMembers.begin(),
                   dataMembers.end(),
                   dm ) == dataMembers.end()) {
          members.push_back( dm );
          dataMembers.push_back( dm );
        }
      }
      for ( i = 0; i < bClass.functionMemberCount(); ++i ) {
        Member fm = bClass.functionMember( i );
        if ( find( functionMembers.begin(), 
                   functionMembers.end(),
                   fm ) == functionMembers.end()) {
          members.push_back( fm );
          functionMembers.push_back( fm );
        }
      }
      bClass.updateMembers2( members,
                             dataMembers, 
                             functionMembers,
                             pathsToBase,
                             basePath );
    }
    basePath.pop_back();
  }
  /*
  // breath first search to find the "lowest" members in the hierarchy
  for ( bIter = m_bases.begin(); bIter != m_bases.end(); ++bIter ) {
  const Class * bClass = (*bIter)->toClass();
  if ( bClass ) {  bClass->updateMembers2( members,
  dataMembers, 
  functionMembers,
  pathsToBase,
  basePath );
  }
  }
  */
}


//-------------------------------------------------------------------------------
void MLANG_Reflection::Class::addDataMember( const Member & dm ) const {
//-------------------------------------------------------------------------------
  ScopeBase::addDataMember( dm );
}

    
//-------------------------------------------------------------------------------
void MLANG_Reflection::Class::removeDataMember( const Member & dm ) const {
//-------------------------------------------------------------------------------
  ScopeBase::removeDataMember( dm );
}

    
//-------------------------------------------------------------------------------
void MLANG_Reflection::Class::addFunctionMember( const Member & fm ) const {
//-------------------------------------------------------------------------------
  ScopeBase::addFunctionMember( fm );
  if ( fm.isConstructor() )    m_constructors.push_back( fm );
  else if ( fm.isDestructor() ) m_destructor = fm;
}

    
//-------------------------------------------------------------------------------
void MLANG_Reflection::Class::removeFunctionMember( const Member & fm ) const {
//-------------------------------------------------------------------------------
  ScopeBase::removeFunctionMember( fm );
  if ( fm.isConstructor() ) {
    std::vector< Member >::iterator it;
    for ( it = m_constructors.begin(); it != m_constructors.end(); ++it) {
      if ( it->type().id() == fm.type().id() ) {
        m_constructors.erase(it); 
        break;
      }
    }
  }
  else if ( fm.isDestructor() ) m_destructor = Member();
}

