//$Id: Class.h,v 1.15 2004/12/01 16:42:37 roiser Exp $
#ifndef MLANG_REFLECTION_CLASS_H
#define MLANG_REFLECTION_CLASS_H

// Include files
#include "MLANG/Export.h"
#include "MLANG/Reflection/TypeBase.h"
#include "MLANG/Reflection/ScopeBase.h"
#include "MLANG/Reflection/Member.h"
#include <map>

namespace MLANG_Reflection
{

    // forward declarations
    class MLANG_Export Base;

    /**
     * @class Class Class.h MLANG/Reflection/Class.h
     * @author Stefan Roiser
     * @date 24/11/2003
     * @ingroup Ref
     */
    class MLANG_Export Class : public TypeBase, 
                  public ScopeBase {

    public:

      /** constructor */
      Class( const char *           type, 
             size_t                 size, 
             const std::type_info & typeInfo, 
             unsigned int           modifiers = 0,
             TYPE                   classType = CLASS );


      /** destructor */
      virtual ~Class() {}

      
      /** 
       * the bool operator returns true if the Class is resolved (implemented)
       * @return true if Class is implemented
       */
      operator bool () const;


      /**
       * nthBase will return the nth base class information
       * @param  nth nth base class
       * @return pointer to base class information
       */
      Base base( size_t nth ) const;


      /**
       * baseCount will return the number of base classes
       * @return number of base classes
       */
      size_t baseCount() const;


      /**
       * castObject an object from this class type to another one
       * @param  to is the class type to cast into
       * @param  obj the memory address of the object to be casted
       */
      void * castObject( const Type & to, 
                         void * obj ) const;


      /**
       * construct will call the constructor of a given type and allocate the
       * memory for it
       * @param  signature of the constructor
       * @param  values for parameters of the constructor
       * @param  mem place in memory for implicit construction
       * @return pointer to new instance
       */
      virtual Object construct( const Type & signature = Type(),
                                std::vector < void * > values = std::vector<void*>(),
                                void * mem = 0 ) const;


      /**
       * destruct will call the destructor of a type and remove its memory
       * allocation if desired
       * @param  instance of the type in memory
       * @param  dealloc for also deallacoting the memory
       */
      void destruct( void * instance, 
                     bool dealloc = true ) const;


      /**
       * dynamicType is used to discover whether an object represents the
       * current class type or not
       * @param  mem is the memory address of the object to checked
       * @return the actual class of the object
       */
      Class dynamicType( const void * mem ) const;


      /**
       * fullName returns the fully qualified name of the class
       * @param  typedefexp expand typedefs or not
       * @return fully qualified name of class
       */
      MLANG_String fullName( bool typedefexp = false ) const;


      /**
       * hasBase will check whether this class has a base class given
       * as argument
       * @param  cl the base-class to check for
       * @param  path optionally the path to the base can be retrieved
       * @return true if this class has a base-class cl, false otherwise
       */
      bool hasBase( const Type & cl ) const;
      bool hasBase( const Type & cl,
                    std::vector< Base > & path ) const;


      /**
       * isAbstract will return true if the the class is abstract
       * @return true if the class is abstract
       */
      bool isAbstract() const;


      /** 
       * isClass will return true as this is an implementation of a class
       * @return true
       */
      bool isClass() const;


      /** 
       * isComplete will return true if all classes and base classes of this 
       * class are resolved and fully known in the system
       */
      bool isComplete() const;


      /**
       * isStruct will return true if the type represents a struct (not a class)
       * @return true if type represents a struct
       */
      bool isStruct() const;


      /**
       * isTemplated will return true if the the class is templated
       * @return true if the class is templated
       */
      bool isTemplated() const;


      /**
       * isVirtual will return true if the class contains a virtual table
       * @return true if the class contains a virtual table
       */
      bool isVirtual() const;


      /**
       * name will return the name of the class
       * @return name of class
       */
      const MLANG_String & name() const;


      /** 
       * pathToBase will return a vector of function pointers to the base class
       * ( !!! Attention !!! the most derived class comes first )
       * @param base the base type 
       * @return vector of function pointers to calculate base offset
       */
      const std::vector < OffsetFunction > & pathToBase( const Scope & base ) const;


      /**
       * propertyList will return a pointer to the property list attached
       * to this item
       * @return pointer to property list
       */
      virtual PropertyList & propertyList() const;


      /**
       * scope will return the scope corresponding to the class 
       * @return scope representation of this class
       */
      Scope scope() const;


      /** 
       * updateMembers2 will update the list of Function/Data/Members with all
       * member of base classes currently availabe in the system
       */
      void updateMembers() const;

    public:

      /** 
       * addBase will add the information about a base class
       * @param  base type of the base class
       * @param  offsetFP the pointer to the stub function for calculating the offset
       * @return this
       */
      Class & addBase( const Type &   base,
                       OffsetFunction offsetFP,
                       unsigned int   modifiers = 0 );


      /** 
       * addBase will add the information about a base class
       * @param b the pointer to the base class info
       */
      void addBase( const Base & b );


      /**
       * addDataMember will add the information about a data member
       * @param dm pointer to data member
       */
      virtual void addDataMember( const Member & dm ) const;


      /**
       * removeDataMember will remove the information about a data member
       * @param dm pointer to data member
       */
      virtual void removeDataMember( const Member & dm ) const;


      /**
       * addFunctionMember will add the information about a function member
       * @param fm pointer to function member
       */
      virtual void addFunctionMember( const Member & fm ) const;


      /**
       * removeFunctionMember will remove the information about a function member
       * @param fm pointer to function member
       */
      virtual void removeFunctionMember( const Member & fm ) const;

    private:

      /** map with the class as a key and the path to it as the value
          the key (void*) is a pointer to the unique ScopeName */
      typedef std::map < void *, std::vector < OffsetFunction > * > PathsToBase;

     
      /** 
       * updateMembers2 will update the list of Function/Data/Members with all
       * member of base classes currently availabe in the system
       * @param members the list of members
       * @param dataMembers the list of data members
       * @param functionMembers the list of function members 
       * @param pathsToBase the cache storing pathes to all known bases
       * @param basePath the current path to the base class
       */
      void updateMembers2( Members & members,
                           Members & dataMembers,
                           Members & functionMembers,
                           PathsToBase & pathsToBase,
                           std::vector < OffsetFunction > & basePath ) const;


      /** 
       * newBases will return true if new base classes have been discovered
       * since the last time it was called
       * @return true if new base classes were resolved
       */
      bool newBases() const;


      /** 
       * internal recursive checking for completeness 
       * @return true if class is complete (all bases are resolved)
       */
      bool isComplete2() const;


      /**
       * allBases will return the number of all base classes 
       * (double count even in case of virtual inheritance)
       * @return number of all base classes
       */
      size_t allBases() const;

    private:

      /**
       * pointer to base class info
       * @label class bases
       * @link aggregationByValue
       */
      std::vector < Base > m_bases;


      /** modifiers of class */
      unsigned int m_modifiers;


      /** caches */
      /** all currently known base classes */
      mutable
      size_t m_allBases;


      /** boolean is true if the whole object is resolved */
      mutable
      bool m_completeType;

      
      /**
       * short cut to constructors
       * @label constructors
       * @link aggregationByValue
       */
      mutable
      std::vector < Member > m_constructors;


      /**
       * short cut to destructor
       * @label destructor
       * @link aggregationByValue
       */
      mutable
      Member m_destructor;


      /** map to all inherited datamembers and their inheritance path */
      mutable
      PathsToBase m_pathsToBase;

    }; // class Class
} //namespace MLANG_Reflection

#include "MLANG/Reflection/Base.h"

//-------------------------------------------------------------------------------
inline MLANG_Reflection::Class::operator bool () const {
//-------------------------------------------------------------------------------
  if ( type()) return true;
  return false;
}


//-------------------------------------------------------------------------------
inline void MLANG_Reflection::Class::addBase( const Base & b ) {
//-------------------------------------------------------------------------------
  m_bases.push_back( b );
}


//-------------------------------------------------------------------------------
inline MLANG_Reflection::Base MLANG_Reflection::Class::base( size_t nth ) const {
//-------------------------------------------------------------------------------
  if ( nth < m_bases.size() ) { return m_bases[ nth ]; }
  return Base();
}


//-------------------------------------------------------------------------------
inline size_t MLANG_Reflection::Class::baseCount() const {
//-------------------------------------------------------------------------------
  return m_bases.size();
}


//-------------------------------------------------------------------------------
inline MLANG_String MLANG_Reflection::Class::fullName( bool typedefexp ) const {
//-------------------------------------------------------------------------------
  return ScopeBase::fullName( typedefexp );
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::Class::isAbstract() const {
//-------------------------------------------------------------------------------
  return 0 != (m_modifiers & ABSTRACT);
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::Class::isClass() const {
//-------------------------------------------------------------------------------
  return true;
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::Class::isStruct() const {
//-------------------------------------------------------------------------------
  return 0 != (m_modifiers & STRUCT);
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::Class::isTemplated() const {
//-------------------------------------------------------------------------------
  return 0 != (m_modifiers & TEMPLATED);
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::Class::isVirtual() const {
//-------------------------------------------------------------------------------
  return 0 != (m_modifiers & VIRTUAL);
}


//-------------------------------------------------------------------------------
inline const MLANG_String & MLANG_Reflection::Class::name() const {
//-------------------------------------------------------------------------------
  return ScopeBase::name();
}


//-------------------------------------------------------------------------------
inline MLANG_Reflection::PropertyList & MLANG_Reflection::Class::propertyList() const {
//-------------------------------------------------------------------------------
  return ScopeBase::propertyList();
}


//-------------------------------------------------------------------------------
inline MLANG_Reflection::Scope MLANG_Reflection::Class::scope() const {
//-------------------------------------------------------------------------------
  return ScopeBase::scope();
}


#endif // MLANG_REFLECTION_CLASS_H

