//$Id: ClassBuilder.h,v 1.10 2004/12/01 14:49:54 mato Exp $

#ifndef SEAL_REFLECT_CLASSBUILDER_H
#define SEAL_REFLECT_CLASSBUILDER_H

// Include files
#include "MLANG/Reflection/Reflection.h"
#include "MLANG/Reflection/Tools.h"
#include "MLANG/Reflection/TypeBuilder.h"

namespace MLANG_Reflection
{

    /** 
     * @class ClassBuilderImpl ClassBuilder.h MLANG/Reflection/ClassBuilder.h
     * @author Stefan Roiser
     * @date 30/3/2004
     * @ingroup RefBld
     */
    class ClassBuilderImpl {
    
    public:
      
      /** constructor */
      ClassBuilderImpl( const char * name, 
                        const std::type_info &, 
                        size_t size, 
                        unsigned int modifiers = 0 );


      /** destructor */
      virtual ~ClassBuilderImpl();

      
      /** 
       * addBase will add the information about one base class
       * @param  name of the base class
       * @param  offsetFP function pointer for offset calculation
       * @param  modifiers the modifiers of the class
       */
      void addBase( const Type & base,
                    OffsetFunction offsetFP,
                    unsigned int modifiers = 0 );


      /** addDataMember will add the information about one data
       * member of the class
       *
       * @param  name of the data member
       * @param  type of the data member
       * @param  offset of the data member
       * @param  modifiers the modifiers of the data member
       */ 
      void  addDataMember( const char * name,
                           const Type & type,
                           size_t offset,
                           unsigned int modifiers = 0 );


      /** addFunctionMember will add the information about one
       * function member of the class
       *
       * @param  name of the function member
       * @param  type of the function member
       * @param  stubFP Stub function pointer for the function
       * @param  stubCxt Stub user context for the stub function
       * @param  params pamater names and default values (semi-colon separated)
       * @param  modifiers the modifiers of the data member
       */
      void addFunctionMember( const char * name,
                              const Type & type,
                              StubFunction stubFP, 
                              void *  stubCtx = 0, 
                              const char * params = 0,
                              unsigned int modifiers = 0 );


      /** addProperty will add a property to the property stack
       * which will be emtpied with the next call of a builder
       * class and attached to the item built with this call
       *
       * @param  key the property key
       * @param  value the value of the property
       */
      void  addProperty( const char * key, 
                         Any value );
      void  addProperty( const char * key, 
                         const char * value );
      
    private:

      /** current class being built */
      Class * m_class;

      /** last added member */
      Member m_lastMember;    

    };
    

    /** 
     * @class ClassBuilder ClassBuilder.h MLANG/Reflection/ClassBuilder.h
     * @author Stefan Roiser
     * @date 30/3/2004
     * @ingroup RefBld
     */
    template < class C >
      class ClassBuilder {

    public:
      
      /** constructor */
      ClassBuilder( unsigned int modifiers = 0 );


      /** constructor */
      ClassBuilder( const char* name, 
                    unsigned int modifiers = 0 );


      /** 
       * addBase will add the information about one base class
       * @param  name of the base class
       * @param  offsetFP function pointer for offset calculation
       * @param  modifiers the modifiers of the class
       */
      template < class B >
        ClassBuilder &  addBase( unsigned int modifiers = 0 );


      /** addDataMember will add the information about one data
       * member of the class
       *
       * @param  name of the data member
       * @param  offset of data member
       * @param  modifiers the modifiers of the data member
       * @return a reference to the ClassBuilder
       */
      template < class T > 
        ClassBuilder & addDataMember( const char * name,
                                      size_t offset,
                                      unsigned int modifiers = 0 );
      ClassBuilder & addDataMember( const Type & type,
                                    const char * name, 
                                    size_t offset,
                                    unsigned int modifiers = 0 );


      /** addFunctionMember will add the information about one
       * function member of the class
       *
       * @param  name of the function member
       * @param  function templated function member to extract type information
       * @param  stubFP Stub function pointer for the function
       * @param  stubCxt Stub user context for the stub function
       * @param  params pamater names and default values (semi-colon separated)
       * @param  modifiers the modifiers of the data member
       * @return a reference to the ClassBuilder
       */
      template < class F > 
        ClassBuilder & addFunctionMember( const char * name,
                                          StubFunction stubFP,
                                          void *  stubCtx = 0, 
                                          const char * params = 0,
                                          unsigned int modifiers  = 0 );
      ClassBuilder & addFunctionMember( const Type & type,
                                        const char * name,
                                        StubFunction stubFP,
                                        void *  stubCtx = 0, 
                                        const char * params = 0,
                                        unsigned int modifiers  = 0 );


      /** addProperty will add a property to the last defined
       * data member, method or class.
       * @param  key the property key
       * @param  value the value of the property
       * @return a reference to the building class
       */
      template < typename P >
        ClassBuilder & addProperty( const char * key, 
                                    P value );

    private:
      
      ClassBuilderImpl m_classBuilderImpl;

    }; // class ClassBuilder

} // namespace MLANG_Reflection


//-------------------------------------------------------------------------------
template <class C>
inline MLANG_Reflection::ClassBuilder<C>::ClassBuilder( unsigned int modifiers ) 
//-------------------------------------------------------------------------------
  : m_classBuilderImpl( Tools::demangle(typeid(C)).c_str(),
                        typeid(C),
                        sizeof(C),
                        modifiers) { }
    

//-------------------------------------------------------------------------------
template <class C>
inline MLANG_Reflection::ClassBuilder<C>::ClassBuilder( const char * name, 
                                                    unsigned int modifiers )
//-------------------------------------------------------------------------------
  : m_classBuilderImpl( name, typeid(C), sizeof(C), modifiers ) { }

    
//-------------------------------------------------------------------------------
template <class C> template< class B > 
inline MLANG_Reflection::ClassBuilder<C> & 
MLANG_Reflection::ClassBuilder<C>::addBase( unsigned int modifiers ) {
//-------------------------------------------------------------------------------
  m_classBuilderImpl.addBase( getType<B>(), 
                              baseOffset<C,B>::get(),
                              modifiers );
  return * this;
}


//-------------------------------------------------------------------------------
template <class C> template< class T > 
inline MLANG_Reflection::ClassBuilder<C> & 
MLANG_Reflection::ClassBuilder<C>::addDataMember( const char *  name,
                                              size_t        offset,
                                              unsigned int modifiers )
//-------------------------------------------------------------------------------
{
  m_classBuilderImpl.addDataMember( name,
                                    TypeDistiller<T>::get(),
                                    offset,
                                    modifiers );
  return * this;
}


//-------------------------------------------------------------------------------
template <class C>
inline MLANG_Reflection::ClassBuilder<C> & 
MLANG_Reflection::ClassBuilder<C>::addDataMember( const Type &  type,
                                              const char *  name,
                                              size_t        offset,
                                              unsigned int modifiers )
//-------------------------------------------------------------------------------
{
  m_classBuilderImpl.addDataMember( name,
                                    type,
                                    offset,
                                    modifiers );
  return * this;
}
    
    
//-------------------------------------------------------------------------------
template < class C > template < class F >
inline MLANG_Reflection::ClassBuilder<C> & 
MLANG_Reflection::ClassBuilder<C>::addFunctionMember( const char * name,
                                                  StubFunction stubFP,
                                                  void * stubCtx,
                                                  const char * params, 
                                                  unsigned int modifiers )
//-------------------------------------------------------------------------------
{
  m_classBuilderImpl.addFunctionMember( name,
                                        FunctionDistiller<F>::get(),
                                        stubFP,
                                        stubCtx,
                                        params,
                                        modifiers );
  return * this;
}
    

//-------------------------------------------------------------------------------
template < class C >
inline MLANG_Reflection::ClassBuilder<C> & 
MLANG_Reflection::ClassBuilder<C>::addFunctionMember( const Type & type,
                                                  const char * name,
                                                  StubFunction stubFP,
                                                  void * stubCtx,
                                                  const char * params, 
                                                  unsigned int modifiers ) 
//-------------------------------------------------------------------------------
{
  m_classBuilderImpl.addFunctionMember( name,
                                        type,
                                        stubFP,
                                        stubCtx,
                                        params,
                                        modifiers );
  return * this;
}


//-------------------------------------------------------------------------------
template < class C > template < class P >
inline MLANG_Reflection::ClassBuilder<C> & 
MLANG_Reflection::ClassBuilder<C>::addProperty( const char * key, 
                                            P value )
//-------------------------------------------------------------------------------
{
  m_classBuilderImpl.addProperty(key , value);
  return * this;
}

#endif // SEAL_REFLECT_CLASSBUILDER_H
