//$Id: Member.h,v 1.9 2004/11/30 22:57:58 roiser Exp $

#ifndef MLANG_REFLECTION_MEMBER_H
#define MLANG_REFLECTION_MEMBER_H

// Include files
#include "MLANG/Export.h"
#include "MLANG/Reflection/Kernel.h"

namespace MLANG_Reflection
{

    // forward declarations
    class MLANG_Export MemberBase;
    class MLANG_Export Object;
    class MLANG_Export Type;
    class MLANG_Export Scope;
    class MLANG_Export PropertyList;

    /**
     * @class Member Member.h MLANG/Reflection/Member.h
     * @author Stefan Roiser
     * @date 24/11/2003
     * @ingroup Ref
     */
    class MLANG_Export Member {

    public:

      /** default constructor */
      Member( MemberBase * memberBase = 0 );


      /** Copy Constructor */
      Member( const Member & rh );


      /**
       * lesser than operator 
       */
      bool operator < ( const Member & member ) const;
      

      /** 
       * equal operator 
       */
      bool operator == ( const Member & member ) const;


      /** 
       * assignment operator 
       */
      Member & operator = ( const Member & member );
      

      /** 
       * operator bool will return true if the member is valid
       * @return true if member is implemented
       */
      operator bool () const;


      /** destructor */
      ~Member();


      /** return full name of member */
      MLANG_String fullName( bool typedefexp = false ) const;


      /** get the member value (as void*) */
      void * get( const Object & obj ) const;


      /** invoke the function (if return type as void*) */
      void * invoke( const Object & obj, 
                     const std::vector < void * > & paramList = 
                     std::vector < void * >()) const;


      /** invoke the function (for static functions) */
      void * invoke( const std::vector < void * > & paramList = 
                     std::vector < void * >()) const;


      /** check whether auto is set for the data member */
      bool isAuto() const;


      /** check whether the function member is a constructor */
      bool isConstructor() const;


      /** check whether the function member is a user defined conversion function */
      bool isConverter() const;


      /** check whether the function member is a copy constructor */
      bool isCopyConstructor() const;


      /** return true if this is a data member */
      bool isDataMember() const;


      /** check whether the function member is a destructor */
      bool isDestructor() const;


      /** check whether explicit is set for the function member */
      bool isExplicit() const;


      /** check whether extern is set for the data member */
      bool isExtern() const;


      /** return true if this is a function member */
      bool isFunctionMember() const;


      /** check whether inline is set for the function member */
      bool isInline() const;

      
      /** check whether Mutable is set for the data member */
      bool isMutable() const;


      /** check whether the function member is an operator */
      bool isOperator() const;


      /** check whether the function member is private */
      bool isPrivate() const;


      /** check whether the function member is protected */
      bool isProtected() const;


      /** check whether the function member is public */
      bool isPublic() const;


      /** check whether register is set for the data member */
      bool isRegister() const;


      /** check whether static is set for the data member */
      bool isStatic() const;


      /** check whether virtual is set for the function member */
      bool isVirtual() const;


      /** return the type of the member (function or data member) */
      TYPE memberType() const;


      /** returns the string representation of the member species */
      MLANG_String memberTypeAsString() const;


      /** return the name of the member */
      MLANG_String name() const;


      /** return the offset of the member */
      size_t offset() const;


      /** number of parameters */
      size_t parameterCount() const;


      /** parameter nth default value if declared*/
      MLANG_String parameterDefault( size_t nth ) const;


      /** parameter nth name if declared*/
      MLANG_String parameterName( size_t nth ) const;


      /**
       * propertyList will return a pointer to the property list attached
       * to this item
       * @return pointer to property list
       */
      PropertyList & propertyList() const;


      /** return the scope of the member */
      Scope scope() const;


      /** set the member value */
      void set( Object & obj,
                const void * value ) const;


      /** set the scope of the member */
      void setScope( const Scope & scope ) const;


      /** return the pointer to the stub function */
      StubFunction stubfunction() const;


      /** return pointer to member type */
      Type type() const;

    private:

      /** the pointer to the member implementation */
      MemberBase * m_memberBase;

    }; // class Member

} //namespace MLANG_Reflection

#include "MLANG/Reflection/MemberBase.h"
#include "MLANG/Reflection/Object.h"
#include "MLANG/Reflection/Scope.h"
#include "MLANG/Reflection/PropertyList.h"
#include "MLANG/Reflection/Type.h"


#if defined (__ACE_INLINE__)
#include "MLANG/Reflection/Member.i"
#endif /* __ACE_INLINE__ */

#endif // MLANG_REFLECTION_MEMBER_H



