//$Id: MemberBase.h,v 1.3 2004/12/02 08:35:09 mato Exp $

#ifndef MLANG_REFLECTION_MEMBERBASE_H
#define MLANG_REFLECTION_MEMBERBASE_H

// Include files
#include "MLANG/Reflection/Kernel.h"
#include "MLANG/Reflection/PropertyList.h"
#include "MLANG/Reflection/Type.h"
#include "MLANG/Reflection/Scope.h"

namespace MLANG_Reflection
{

    // forward declarations
    class Object;

    /**
     * @class Member Member.h MLANG/Reflection/Member.h
     * @author Stefan Roiser
     * @date 24/11/2003
     * @ingroup Ref
     */
    class MemberBase {

    public:

      /** default constructor */
      MemberBase( const char *   name,
                  const Type &   type,
                  TYPE           memberType,
                  unsigned int   modifiers );


      /** destructor */
      virtual ~MemberBase();


      /** return full name of member */
      virtual MLANG_String fullName( bool typedefexp = false ) const;


      /** get the member value (as void*) */
      virtual void * get( const Object & obj ) const;


      /** invoke the function (if return type as void*) */
      virtual void * invoke( const Object & obj, 
                             const std::vector < void * > & paramList = 
                             std::vector < void * >()) const;


      /** invoke the function (for static functions) */
      virtual void * invoke( const std::vector < void * > & paramList = 
                             std::vector < void * >()) const;


      /** check whether auto is set for the data member */
      bool isAuto() const;


      /** check whether the function member is a constructor */
      bool isConstructor() const;


      /** check whether the function member is a user defined conversion function */
      bool isConverter() const;


      /** check whether the function member is a copy constructor */
      bool isCopyConstructor() const;


      /** return true if this is a data member */
      bool isDataMember() const;


      /** check whether the function member is a destructor */
      bool isDestructor() const;


      /** check whether explicit is set for the function member */
      bool isExplicit() const;


      /** check whether extern is set for the data member */
      bool isExtern() const;


      /** return true if this is a function member */
      bool isFunctionMember() const;


      /** check whether inline is set for the function member */
      bool isInline() const;

      
      /** check whether Mutable is set for the data member */
      bool isMutable() const;


      /** check whether the function member is an operator */
      bool isOperator() const;


      /** check whether the function member is private */
      bool isPrivate() const;


      /** check whether the function member is protected */
      bool isProtected() const;


      /** check whether the function member is public */
      bool isPublic() const;


      /** check whether register is set for the data member */
      bool isRegister() const;


      /** check whether static is set for the data member */
      bool isStatic() const;


      /** check whether virtual is set for the function member */
      bool isVirtual() const;


      /** return the type of the member (function or data member) */
      TYPE memberType() const;


      /** returns the string representation of the member species */
      MLANG_String memberTypeAsString() const;


      /** return the name of the member */
      const MLANG_String & name() const;


      /** return the offset of the member */
      virtual size_t offset() const;


      /** number of parameters */
      virtual size_t parameterCount() const;


      /** parameter nth default value if declared*/
      virtual MLANG_String parameterDefault( size_t nth ) const;


      /** parameter nth name if declared*/
      virtual MLANG_String parameterName( size_t nth ) const;


      /**
       * propertyList will return a pointer to the property list attached
       * to this item
       * @return pointer to property list
       */
      PropertyList & propertyList() const;


      /** return the scope of the member */
      Scope scope() const;


      /** set the member value */
      virtual void set( const Object & obj,
                        const void * value ) const;


      /** set the scope of the member */
      void setScope( const Scope & scope ) const;


      /** return the pointer to the stub function */
      virtual StubFunction stubfunction() const;


      /** return pointer to member type */
      Type type() const;

    protected:

      /** 
       * calculateBaseObject will calculate the inheritance between an object
       * and the local type if necessary
       * @param obj the object from which the calculation should start
       * @return memory address of new local object relative to obj
       */
      void * calculateBaseObject( const Object & obj ) const;

    protected:

      /**
       * characteristics of the member type
       * @label member type
       * @link aggregation
       * @supplierCardinality 1
       */
      Type m_type;

      
      /** all modifiers of the member */
      unsigned int m_modifiers;

    private:

      /** name of member */
      MLANG_String m_name;


      /**
       * scope of the member
       * @label member scope
       * @link aggregation
       * @supplierCardinality 1
       */
      mutable
      Scope m_scope;


      /** 
       * member type (data or function member) 
       * @label memberType
       */
      TYPE m_memberType;


      /**
       * pointer to the property list
       * @label propertylist
       * @link aggregationByValue
       */
      mutable
      PropertyList m_propertyList;

    }; // class Member

} //namespace MLANG_Reflection

#include "MLANG/Reflection/DataMember.h"
#include "MLANG/Reflection/FunctionMember.h"
#include "MLANG/Reflection/Object.h"

//-------------------------------------------------------------------------------
inline void * MLANG_Reflection::MemberBase::get( const Object & /* obj */ ) const {
//-------------------------------------------------------------------------------
  return 0;
}


//-------------------------------------------------------------------------------
inline void * MLANG_Reflection::MemberBase::invoke( const Object & /* obj */ ,
                                                const std::vector < void * > & /* paramList */ ) const {
//-------------------------------------------------------------------------------
  return 0;
}


//-------------------------------------------------------------------------------
inline void * MLANG_Reflection::MemberBase::invoke( const std::vector < void * > & /* paramList */ ) const {
//-------------------------------------------------------------------------------
  return 0;
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::MemberBase::isAuto() const {
//-------------------------------------------------------------------------------
  return 0 != (m_modifiers & AUTO);
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::MemberBase::isConstructor() const {
//-------------------------------------------------------------------------------
  return 0 != (m_modifiers & CONSTRUCTOR);
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::MemberBase::isConverter() const {
//-------------------------------------------------------------------------------
  return 0 != (m_modifiers & CONVERTER);
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::MemberBase::isCopyConstructor() const {
//-------------------------------------------------------------------------------
  return 0 != (m_modifiers & COPYCONSTRUCTOR);
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::MemberBase::isDataMember() const {
//-------------------------------------------------------------------------------
  return 0 != (m_memberType & DATAMEMBER);
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::MemberBase::isDestructor() const {
//-------------------------------------------------------------------------------
  return 0 != (m_modifiers & DESTRUCTOR);
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::MemberBase::isExplicit() const {
//-------------------------------------------------------------------------------
  return 0 != (m_modifiers & EXPLICIT);
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::MemberBase::isExtern() const {
//-------------------------------------------------------------------------------
  return 0 != (m_modifiers & EXTERN);
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::MemberBase::isFunctionMember() const {
//-------------------------------------------------------------------------------
  return 0 != (m_memberType & FUNCTIONMEMBER);
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::MemberBase::isInline() const {
//-------------------------------------------------------------------------------
  return 0 != (m_modifiers & INLINE);
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::MemberBase::isMutable() const {
//-------------------------------------------------------------------------------
  return 0 != (m_modifiers & MUTABLE);
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::MemberBase::isOperator() const {
//-------------------------------------------------------------------------------
  return 0 != (m_modifiers & OPERATOR);
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::MemberBase::isPrivate() const {
//-------------------------------------------------------------------------------
  return 0 != (m_modifiers & PRIVATE);
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::MemberBase::isProtected() const {
//-------------------------------------------------------------------------------
  return 0 != (m_modifiers & PROTECTED);
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::MemberBase::isPublic() const {
//-------------------------------------------------------------------------------
  return 0 != (m_modifiers & PUBLIC);
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::MemberBase::isRegister() const {
//-------------------------------------------------------------------------------
  return 0 != (m_modifiers & REGISTER);
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::MemberBase::isStatic() const {
//-------------------------------------------------------------------------------
  return 0 != (m_modifiers & STATIC);
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::MemberBase::isVirtual() const {
//-------------------------------------------------------------------------------
  return 0 != (m_modifiers & VIRTUAL);
}


//-------------------------------------------------------------------------------
inline MLANG_Reflection::TYPE MLANG_Reflection::MemberBase::memberType() const {
//-------------------------------------------------------------------------------
  return m_memberType;
}


//-------------------------------------------------------------------------------
inline const MLANG_String & MLANG_Reflection::MemberBase::name() const {
//-------------------------------------------------------------------------------
  return m_name;
}


//-------------------------------------------------------------------------------
inline size_t MLANG_Reflection::MemberBase::offset() const {
//-------------------------------------------------------------------------------
  return 0;
}


//-------------------------------------------------------------------------------
inline size_t MLANG_Reflection::MemberBase::parameterCount() const {
//-------------------------------------------------------------------------------
  return 0; 
}


//-------------------------------------------------------------------------------
inline MLANG_String MLANG_Reflection::MemberBase::parameterDefault( size_t /* nth */ ) const {
//-------------------------------------------------------------------------------
  return "";
}


//-------------------------------------------------------------------------------
inline MLANG_String MLANG_Reflection::MemberBase::parameterName( size_t /* nth */ ) const {
//-------------------------------------------------------------------------------
  return "";
}


//-------------------------------------------------------------------------------
inline void MLANG_Reflection::MemberBase::set( const Object & /* obj */,
                                           const void * /* value */ ) const {}
//-------------------------------------------------------------------------------


//-------------------------------------------------------------------------------
inline void MLANG_Reflection::MemberBase::setScope( const Scope & scope ) const {
//-------------------------------------------------------------------------------
  m_scope = scope;
}


//-------------------------------------------------------------------------------
inline MLANG_Reflection::StubFunction MLANG_Reflection::MemberBase::stubfunction() const {
//-------------------------------------------------------------------------------
  return 0;
}


//-------------------------------------------------------------------------------
inline MLANG_Reflection::Type MLANG_Reflection::MemberBase::type() const {
//-------------------------------------------------------------------------------
  return m_type;
}



#endif // MLANG_REFLECTION_MEMBERBASE_H



