//$Id: Scope.h,v 1.9 2004/11/30 22:57:58 roiser Exp $

#ifndef MLANG_REFLECTION_SCOPE_H
#define MLANG_REFLECTION_SCOPE_H


// Include files
#include "MLANG/Reflection/Kernel.h"
#include <string>

namespace MLANG_Reflection
{

    // forward declarations
    class Namespace;
    class Class;
    class TypeBase;
    class Type;
    class FunctionMember;
    class DataMember;
    class Member;
    class ScopeName;
    class InstantiatedTemplateClass;
    class PropertyList;

    /**
     * @class Scope Scope.h MLANG/Reflection/Scope.h
     * @author Stefan Roiser
     * @date 24/11/2003
     * @ingroup Ref
     */
    class Scope {

    public:

      /** constructor */
      Scope( const ScopeName * scopeName = 0 );


      /** copy constructor */
      Scope( const Scope & rh );


      /** destructor */
      ~Scope();


      /**
       * operator Class will return the corresponding class object of Type
       * @return Class corresponding to Type
       */
      operator Class () const;


      /**
       * the bool operator will return true if the Scope is resolved (implemented)
       * @return true if Scope is implemented
       */
      operator bool () const;


      /**
       * byName will return a pointer to a scope which is given as an argument
       * or 0 if none is found
       * @param  name fully qualified name of scope
       * @return pointer to scope or 0 if none is found
       */
      static Scope byName( const MLANG_String & name );


      /**
       * dataMember will return the nth data member of the scope
       * @param  nth data member
       * @return pointer to data member
       */
      Member dataMember( size_t nth ) const;


      /**
       * dataMemberCount will return the number of data members of this scope
       * @return number of data members
       */
      size_t dataMemberCount() const;


      /**
       * declaringScope will return a pointer to the scope of this one
       * @return pointer to declaring scope
       */
      Scope declaringScope() const;


      /**
       * functionMember will return the nth function member of the scope
       * @param  nth function member
       * @return pointer to function member
       */
      Member functionMember( size_t nth ) const;


      /**
       * functionMemberCount will return the number of function members of
       * this scope
       * @return number of function members
       */
      size_t functionMemberCount() const;


      /**
       * fullName returns the full name of the scope
       * @param  typedefexp expand typedefs or not
       * @return full name of scope
       */
      MLANG_String fullName( bool typedefexp = false ) const;


      /**
       * id returns a unique identifier of the type in the system
       * @return unique identifier
       */
      void * id() const;


      /** 
       * isClass returns true if the type represents a Class
       * @return true if type represents a Class
       */
      bool isClass() const;


      /** 
       * isInstantiatedTemplateClass returns true if the type represents a 
       * IntantiatedTemplateClass
       * @return true if type represents a InstantiatedTemplateClass
       */
      bool isInstantiatedTemplateClass() const;


      /** 
       * isNamespace returns true if the type represents a Namespace
       * @return true if type represents a Namespace
       */
      bool isNamespace() const;


      /**
       * isTopScope will return true if the current scope is the top
       * (empty) namespace
       * @return true if current sope is top namespace
       */
      bool isTopScope() const;

 
      /**
       * member will return the first member with a given name
       * @param  member name
       * @return pointer to member
       */
      Member member( const MLANG_String & name ) const;


      /**
       * member will return the nth member of the scope
       * @param  nth member
       * @return pointer to nth member
       */
      Member member( size_t nth ) const;


      /**
       * memberCount will return the number of members
       * @return number of members
       */
      size_t memberCount() const;


      /**
       * name will return a string representation of name of the scope (NOT
       * fully qualified, see fullName)
       * @return string representation of scope
       */
      const MLANG_String & name() const;


      /**
       * propertyList will return a pointer to the property list attached
       * to this item
       * @return pointer to property list
       */
      PropertyList & propertyList() const;

      
      /**
       * findAll will return a vector of all scopes currently available
       * resolvable scopes
       * @param  nth scope defined in the system
       * @return vector of all available scopes
       */
      static Scope scope( size_t nth );


      /**
       * scopeCount will return the number of currently defined scopes
       * (resolved and unresolved ones)
       * @return number of currently defined scopes
       */
      static size_t scopeCount();


      /**
       * scopeType will return which kind of scope is represented
       * @return type of scope
       */
      TYPE scopeType() const;


      /**
       * scopeTypeAsString will return the string representation of the enum
       * representing the current Scope (e.g. "CLASS")
       * @return string representation of enum for Scope
       */
      MLANG_String scopeTypeAsString() const;


      /**
       * subScope will return a pointer to a sub-scopes
       * @param  nth sub-scope
       * @return pointer to nth sub-scope
       */
      Scope subScope( size_t nth ) const;


      /**
       * scopeCount will return the number of sub-scopes
       * @return number of sub-scopes
       */
      size_t subScopeCount() const;


      /**
       * nthType will return a pointer to the nth sub-type
       * @param  nth sub-type
       * @return pointer to nth sub-type
       */
      Type subType( size_t nth ) const;


      /**
       * typeCount will returnt he number of sub-types
       * @return number of sub-types
       */
      size_t subTypeCount() const;

    public:
 
      /**
       * addDataMember will add the information about a data member
       * @param dm pointer to data member
       */
      void addDataMember( const Member & dm );

      /**
       * removeDataMember will remove the information about a data member
       * @param dm pointer to data member
       */
      void removeDataMember( const Member & dm );

      /**
       * addFunctionMember will add the information about a function member
       * @param fm pointer to function member
       */
      void addFunctionMember( const Member & fm );

      /**
       * removeFunctionMember will remove the information about a function member
       * @param fm pointer to function member
       */
      void removeFunctionMember( const Member & fm );

      /**
       * addSubScope will add a sub-scope to this one
       * @param sc pointer to Scope
       */
      void addSubScope( const Scope & sc );

      /**
       * removeSubScope will remove a sub-scope to this one
       * @param sc pointer to Scope
       */
      void removeSubScope( const Scope & sc );

      /**
       * addSubType will add a sub-type to this scope
       * @param sc pointer to Type
       */
      void addSubType( const Type & ty );

      /**
       * removeSubType will remove a sub-type to this scope
       * @param sc pointer to Type
       */
      void removeSubType( const Type & ty );     

    public:
      
      static Scope __NIRVANA__;

    private:

      /**
       * pointer to the resolved Scope
       * @label scope base
       * @link aggregationByValue
       */
      const ScopeName * m_scopeName;

    }; // class Scope

    bool operator < ( const Scope & lh, 
                      const Scope & rh); 

    bool operator == ( const Scope & lh,
                       const Scope & rh );

} // namespace MLANG_Reflection

#include "MLANG/Reflection/ScopeBase.h"
#include "MLANG/Reflection/ScopeName.h"
#include "MLANG/Reflection/PropertyList.h"

#if defined (__ACE_INLINE__)
#include "MLANG/Reflection/Scope.i"
#endif /* __ACE_INLINE__ */


#endif //MLANG_REFLECTION_SCOPE_H
