//$Id: ScopeBase.h,v 1.13 2004/12/01 14:49:54 mato Exp $

#ifndef MLANG_REFLECTION_SCOPEBASE_H
#define MLANG_REFLECTION_SCOPEBASE_H

// Include files
#include "MLANG/Reflection/Kernel.h"
#include "MLANG/Reflection/Scope.h"
#include "MLANG/Reflection/PropertyList.h"
#include <vector>

namespace MLANG_Reflection
{

    // forward declarations
    class ScopeName;
    class Namespace;
    class Class;
    class InstantiatedTemplateClass;
    class Member;

    /**
     * @class ScopeBase ScopeBase.h MLANG/Reflection/ScopeBase.h
     * @author Stefan Roiser
     * @date 24/11/2003
     * @ingroup Ref
     */
    class ScopeBase {

    public:

      /** constructor within a scope*/
      ScopeBase( const char * scope, 
                 TYPE scopeType );


      /** destructor */
      virtual ~ScopeBase();


      /** 
       * operator Scope will return the corresponding Scope object
       * @return Scope corresponding to this ScopeBase
       */
      operator Scope () const;


      /**
       * nthDataMember will return the nth data member of the scope
       * @param  nth data member
       * @return pointer to data member
       */
      Member dataMember( size_t nth ) const;


      /**
       * dataMemberCount will return the number of data members of this scope
       * @return number of data members
       */
      size_t dataMemberCount() const;


      /**
       * declaringScope will return a pointer to the scope of this one
       * @return pointer to declaring scope
       */
      Scope declaringScope() const;


      /**
       * fullName returns the full name of the scope
       * @param  typedefexp expand typedefs or not
       * @return full name of scope
       */
      virtual MLANG_String fullName( bool typedefexp = false ) const;


      /**
       * nthFunctionMember will return the nth function member of the scope
       * @param  nth function member
       * @return pointer to function member
       */
      Member functionMember( size_t nth ) const;

 
      /**
       * functionMemberCount will return the number of function members of
       * this scope
       * @return number of function members
       */
      size_t functionMemberCount() const;


      /** 
       * isClass returns true if the scope represents a Class
       * @return true if scope represents a Class
       */
      bool isClass() const;


      /** 
       * isInstantiatedTemplateClass returns true if the scope represents a 
       * IntantiatedTemplateClass
       * @return true if scope represents a InstantiatedTemplateClass
       */
      bool isInstantiatedTemplateClass() const;


      /** 
       * isNamespace returns true if the scope represents a Namespace
       * @return true if scope represents a Namespace
       */
      bool isNamespace() const;


      /**
       * isTopScope will return true if the current scope is the top
       * (empty) namespace
       * @return true if current sope is top namespace
       */
      bool isTopScope() const;


      /**
       * member will return the first member with a given name
       * @param name  member name
       * @return pointer to member
       */
      Member member( const MLANG_String & name ) const;


      /**
       * member will return the nth member of the scope
       * @param  nth member
       * @return pointer to nth member
       */
      Member member( size_t nth ) const;


      /**
       * memberCount will return the number of members
       * @return number of members
       */
      size_t memberCount() const;


      /**
       * name will return the name of the scope
       * @return name of scope
       */
      const MLANG_String & name() const;


      /**
       * propertyList will return a pointer to the property list attached
       * to this item
       * @return pointer to property list
       */
      virtual PropertyList & propertyList() const;

      
      /** 
       * scope will return the scope Object of this ScopeBase
       * @return corresponding Scope
       */
      Scope scope() const;

      
      /**
       * scopeType will return which kind of scope is represented
       * @return type of scope
       */
      TYPE scopeType() const;


      /**
       * scopeTypeAsString will return the string representation of the enum
       * representing the current Scope (e.g. "CLASS")
       * @return string representation of enum for Scope
       */
      MLANG_String scopeTypeAsString() const;


      /**
       * subScope will return a pointer to a sub-scopes
       * @param  nth sub-scope
       * @return pointer to nth sub-scope
       */
      Scope subScope( size_t nth ) const;


      /**
       * scopeCount will return the number of sub-scopes
       * @return number of sub-scopes
       */
      size_t subScopeCount() const;


      /**
       * type will return a pointer to the nth sub-type
       * @param  nth sub-type
       * @return pointer to nth sub-type
       */
      Type subType( size_t nth ) const;


      /**
       * typeCount will returnt he number of sub-types
       * @return number of sub-types
       */
      size_t subTypeCount() const;

    protected:

      /** protected constructor for initialisation of the global namespace */
      ScopeBase();

    public:

      /**
       * addDataMember will add the information about a data member
       * @param dm pointer to data member
       */
      virtual void addDataMember( const Member & dm ) const;


      /**
       * removeDataMember will remove the information about a data member
       * @param dm pointer to data member
       */
      virtual void removeDataMember( const Member & dm ) const;


      /**
       * addFunctionMember will add the information about a function member
       * @param fm pointer to function member
       */
      virtual void addFunctionMember( const Member & fm ) const;


      /**
       * removeFunctionMember will remove the information about a function member
       * @param fm pointer to function member
       */
      virtual void removeFunctionMember( const Member & fm ) const;


      /**
       * addSubScope will add a sub-scope to this one
       * @param sc pointer to Scope
       */
      void addSubScope( const Scope & sc );


      /**
       * removeSubScope will remove a sub-scope to this one
       * @param sc pointer to Scope
       */
      void removeSubScope( const Scope & sc );


      /**
       * addSubType will add a sub-type to this scope
       * @param sc pointer to Type
       */
      void addSubType( const Type & ty );


      /**
       * removeSubType will remove a sub-type to this scope
       * @param sc pointer to Type
       */
      void removeSubType( const Type & ty );

    protected:
      
      /** container for all members of the scope */
      typedef std::vector < Member > Members;

      /**
       * pointers to members
       * @label scope members
       * @link aggregationByValue
       */
      mutable
      Members m_members;


      /**
       * container with pointers to all data members in this scope
       * @label scope datamembers
       * @link aggregationByValue
       */
      mutable
      Members m_dataMembers;


      /**
       * container with pointers to all function members in this scope
       * @label scope functionmembers
       * @link aggregationByValue
       */
      mutable
      Members m_functionMembers;

    private:

      /**
       * pointer to the scope name
       * @label scope name
       * @link aggregation
       */
      ScopeName * m_scopeName;


      /**
       * the unresolved name of the scope
       */
      MLANG_String m_name;


      /**
       * type of scope
       * @link aggregationByValue
       * @label scopeType
       */
      TYPE m_scopeType;


      /**
       * pointer to declaring scope
       * @label declaring scope
       * @link aggregation
       */
      Scope m_declaringScope;


      /**
       * pointers to sub-scopes
       * @label sub scopes
       * @link aggregationByValue
       */
      std::vector< Scope > m_subScopes;
 

      /**
       * pointer to types
       * @label scope types
       * @link aggregation
       */
      std::vector < Type > m_types;
 

      /**
       * pointer to the property list
       * @label propertylist
       * @link aggregationByValue
       */
      mutable
      PropertyList m_propertyList;

    }; // class ScopeBase
} //namespace MLANG_Reflection



//-------------------------------------------------------------------------------
inline MLANG_Reflection::Scope MLANG_Reflection::ScopeBase::declaringScope() const {
//-------------------------------------------------------------------------------
  return m_declaringScope;
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::ScopeBase::isClass() const {
//-------------------------------------------------------------------------------
  return ( m_scopeType == CLASS );
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::ScopeBase::isInstantiatedTemplateClass() const {
//-------------------------------------------------------------------------------
  return ( m_scopeType == INSTANTIATEDTEMPLATECLASS );
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::ScopeBase::isNamespace() const {
//-------------------------------------------------------------------------------
  return ( m_scopeType == NAMESPACE );
}


//-------------------------------------------------------------------------------
inline MLANG_Reflection::TYPE MLANG_Reflection::ScopeBase::scopeType() const {
//-------------------------------------------------------------------------------
  return m_scopeType;
}


//-------------------------------------------------------------------------------
inline MLANG_Reflection::Scope MLANG_Reflection::ScopeBase::subScope( size_t nth ) const {
//-------------------------------------------------------------------------------
  if ( nth < m_subScopes.size() ) { return m_subScopes[ nth ]; }
  return Scope( 0 );
}


//-------------------------------------------------------------------------------
inline size_t MLANG_Reflection::ScopeBase::subScopeCount() const {
//-------------------------------------------------------------------------------
  return m_subScopes.size();
}

#endif // MLANG_REFLECTION_SCOPEBASE_H


