// $Id: Type.h,v 1.14 2004/12/02 08:35:09 mato Exp $

#ifndef MLANG_REFLECTION_TYPE_H
#define MLANG_REFLECTION_TYPE_H

// Include files
#include "MLANG/Export.h"
#include "MLANG/Reflection/Kernel.h"
#include <vector>
#include <string>
#include <typeinfo>
#if !defined(WIN32)
#include <utility>
#endif /* !WIN32 */

namespace MLANG_Reflection
{

    // forward declarations
    class Typedef;
    class InstantiatedTemplateFunction;
    class InstantiatedTemplateClass;
    class Pointer;
    class Fundamental;
    class Function;
    class Enum;
    class Class;
    class Array;
    class TypeName;
    class Scope;
    class Object;
    class EnumInfo;
    class UnionInfo;
    class PropertyList;
    


    /**
     * @class Type Type.h MLANG/Reflection/Type.h
     * @author Stefan Roiser
     * @date 05/11/2004
     * @ingroup Ref
     */
    class MLANG_Export Type {

    public:

      /** default constructor */
      Type( const TypeName * typeName = 0,
            unsigned int modifiers = 0 );

      
      /** copy constructor */
      Type( const Type & rh );


      /** copy constructor applying qualification */
      Type( const Type & rh,
            unsigned int modifiers );
      

      /** destructor */
      ~Type();

      
      /** 
       * assignment operator
       */
      Type & operator = ( const Type & type );


      /** 
       * equal operator 
       */
      bool operator == ( const Type & type ) const;

      
      /**
       * lesser than operator
       */
      bool operator < ( const Type & type) const; 


      /**
       * operator Class will return the corresponding class object of Type
       * @return Class corresponding to Type
       */
      operator Class () const;


      /**
       * the bool operator returns true if the Type is resolved (implemented)
       * @return true if Type is implemented 
       */
      operator bool () const;


      /**
       * allocate will reserve memory for the size of the object
       * @return pointer to allocated memory
       */
      void * allocate() const;
      

      /**
       * byName will look for a type given as a string and return a pointer to
       * its reflexion type
       * @param  key fully qualified name of the type as string
       * @return pointer to type or 0 if none is found
       */
      static Type byName( const MLANG_String & key );
      
      
      /**
       * byTypeId will look for a type given as a string representation of a
       * type_info and return a pointer to its reflexion type
       * @param  tid string representation of the type_info type
       * @return pointer to type or 0 if none is found
       */
      static Type byTypeInfo( const std::type_info & tid );


      /**
       * construct will call the constructor of a given type and allocate
       * the memory for it
       * @param  signature of the constructor
       * @param  values for parameters of the constructor
       * @param  mem place in memory for implicit construction
       * @return new object 
       */
      Object construct( const Type & signature = Type( 0 ),
                        const std::vector < void * >& values = std::vector < void * > (),
                        void * mem = 0 ) const;


      /**
       * deallocate will deallocate the memory for a given object
       * @param  instance of the type in memory
       */
      void deallocate( void * instance ) const;


      /**
       * destruct will call the destructor of a type and remove its memory
       * allocation if desired
       * @param  instance of the type in memory
       * @param  dealloc for also deallacoting the memory
       */
      void destruct( void * instance, 
                     bool dealloc = true ) const;


      /**
       * enumInfo will return the nth enum information
       * @param  nth enum
       * @return nth enum info object
       */
      EnumInfo enumInfo( size_t nth ) const;


      /**
       * enumInfoCount will return the number of enums
       * @return number of enums
       */
      size_t enumInfoCount() const;


      /**
       * fullName returns the full name of the type
       * @param  typedefexp epxand typedefs or not
       * @return full name of type
       */
      MLANG_String fullName( bool typedefexp = false ) const;


      /**
       * id returns a unique identifier of the type in the system
       * @return unique identifier
       */
      void * id() const;


      /** 
       * isArray returns true if the type represents a Array
       * @return true if type represents a Array
       */
      bool isArray() const;


      /** 
       * isClass returns true if the type represents a Class
       * @return true if type represents a Class
       */
      bool isClass() const;


      /** 
       * isConst returns true if the type represents a const type
       * @return true if type represents a const type
       */
      bool isConst() const;


      /** 
       * isConstVolatile returns true if the type represents a const volatile type
       * @return true if type represents a const volatile type
       */
      bool isConstVolatile() const;


      /** 
       * isEnum returns true if the type represents a Enum
       * @return true if type represents a Enum
       */
      bool isEnum() const;


      /** 
       * isFunction returns true if the type represents a Function
       * @return true if type represents a Function
       */
      bool isFunction() const;


      /** 
       * isFundamental returns true if the type represents a Fundamental
       * @return true if type represents a Fundamental
       */
      bool isFundamental() const;


      /** 
       * isInstantiatedTemplateClass returns true if the type represents a 
       * IntantiatedTemplateClass
       * @return true if type represents a InstantiatedTemplateClass
       */
      bool isInstantiatedTemplateClass() const;


      /** 
       * isInstantiatedTemplateFunction returns true if the type represents a 
       * InstantiatedTemplateFunction
       * @return true if type represents a InstantiatedTemplateFunction
       */
      bool isInstantiatedTemplateFunction() const;


      /** 
       * isPointer returns true if the type represents a Pointer
       * @return true if type represents a Pointer
       */
      bool isPointer() const;


      /** 
       * isPointerToMember returns true if the type represents a PointerToMember
       * @return true if type represents a PointerToMember
       */
      bool isPointerToMember() const;


      /** 
       * isReference returns true if the type represents a Reference
       * @return true if type represents a Reference
       */
      bool isReference() const;


      /**
       * isTemplatedClass returns true if the type is a templated class
       * @return true if Type is templated class
       */
      bool isTemplatedClass() const;


      /**
       * isTemplatedFunction returns true if the type is a templated function
       * @return true if Type is templated function
       */
      bool isTemplatedFunction() const;


      /** 
       * isTypedef returns true if the type represents a Typedef
       * @return true if type represents a Typedef
       */
      bool isTypedef() const;


      /** 
       * isUnion returns true if the type represents a Union
       * @return true if type represents a 
       */
      bool isUnion() const;


      /** 
       * isUnqualified returns true if the type represents an unqualified type
       * @return true if type represents an unqualified type
       */
      bool isUnqualified() const;


      /** 
       * isVolatile returns true if the type represents a volatile type
       * @return true if type represents a volatile type
       */
      bool isVolatile() const;

 
      /** 
       * size returns the size of the array
       * @return size of array
       */
      size_t length() const;

 
      /**
       * name returns the name of the type
       * @return name of the type
       */
      MLANG_String  name() const;


      /**
       * parameter returns the nth parameter
       * @param  nth nth parameter
       * @return pointer to nth parameter type
       */
      Type parameter( size_t nth ) const;


      /**
       * parameterCount will return the number of parameters of this function
       * @return number of parameters
       */
      size_t parameterCount() const;


      /**
       * propertyList will return a pointer to the property list attached
       * to this item
       * @return pointer to property list
       */
      PropertyList & propertyList() const;


      /**
       * returnType will return a pointer to the type of the return type.
       * @return pointer to Type of return type
       */
      Type returnType() const;
      

      /** scope will return the scope of the Type if any 
       * @return pointer to scope of type
       */
      Scope scope() const;


      /**
       * sizeof will return the size of the type
       * @return size of the type as int
       */
      size_t sizeOf() const;


      /**
       * arrayType will return a pointer to the type of the array.
       * @return pointer to Type of member et. al.
       */
      Type toType() const;


      /**
       * type will return a pointer to the nth Type in the system
       * @param  nth number of type to return
       * @return pointer to nth Type in the system
       */
      static Type type( size_t nth );


      /**
       * typeCount will return the number of currently defined types in
       * the system
       * @return number of currently defined types
       */
      static size_t typeCount();


      /**
       * typeId will return the c++ type_info object of the type
       * @return type_info object of type
       */
      const std::type_info & typeInfo() const;


      /**
       * typeType will return the real type
       * @return real type
       */
      TYPE typeType() const;


      /**
       * typeTypeAsString will return the string representation of the ENUM
       * representing the real type of the Type (e.g. "CLASS")
       * @return string representation of the TYPE enum of the Type
       */
      MLANG_String typeTypeAsString() const;


      /**
       * unionInfo will return the nth union info
       * @param  nth union info
       * @return nth union info
       */
      UnionInfo unionInfo( size_t nth ) const;


      /**
       * unionInfoCount will return the number of union types
       * @return number of union types
       */
      size_t unionInfoCount() const;

    private:

      /** pointer to the TypeName */
      const TypeName * m_typeName;


      /** modifiers */
      unsigned int m_modifiers;
        
    }; // class Type

} //namespace MLANG_Reflection

#include "MLANG/Reflection/TypeName.h"
#include "MLANG/Reflection/TypeBase.h"
#include "MLANG/Reflection/Object.h"
#include "MLANG/Reflection/PropertyList.h"

//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::Type::operator == ( const Type & type ) const {
//-------------------------------------------------------------------------------
  return ( m_typeName == type.m_typeName ); 
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::Type::operator < ( const Type & type ) const {
//-------------------------------------------------------------------------------
  return id() < type.id();
}


//-------------------------------------------------------------------------------
inline MLANG_Reflection::Type::operator bool () const {
//-------------------------------------------------------------------------------
  if ( this->m_typeName && this->m_typeName->m_typeBase ) return true;
  return false;
}


//-------------------------------------------------------------------------------
inline MLANG_Reflection::Type & MLANG_Reflection::Type::operator = ( const Type & type ) {
//-------------------------------------------------------------------------------
  m_typeName = type.m_typeName;
  m_modifiers = type.m_modifiers;
  return * this;
}


//-------------------------------------------------------------------------------
inline MLANG_Reflection::Type::Type( const TypeName * typeName,
                                 unsigned int modifiers ) 
//-------------------------------------------------------------------------------
  : m_typeName( typeName ),
    m_modifiers( modifiers ) {}


//-------------------------------------------------------------------------------
inline MLANG_Reflection::Type::Type( const Type & rh )
//-------------------------------------------------------------------------------
  : m_typeName ( rh.m_typeName ),
    m_modifiers ( rh.m_modifiers ) {}


//-------------------------------------------------------------------------------
inline MLANG_Reflection::Type::Type( const Type & rh, 
                                 unsigned int modifiers ) 
//-------------------------------------------------------------------------------
  : m_typeName( rh.m_typeName ),
    m_modifiers( modifiers ) {}


//-------------------------------------------------------------------------------
inline MLANG_Reflection::Type::~Type() {
//-------------------------------------------------------------------------------
}


//-------------------------------------------------------------------------------
inline void * MLANG_Reflection::Type::allocate() const {
//-------------------------------------------------------------------------------
  if ( * this ) return m_typeName->m_typeBase->allocate();
  return 0;
}


//-------------------------------------------------------------------------------
inline void MLANG_Reflection::Type::deallocate( void * instance ) const {
//-------------------------------------------------------------------------------
  if ( * this ) m_typeName->m_typeBase->deallocate( instance ); 
}


//-------------------------------------------------------------------------------
inline void MLANG_Reflection::Type::destruct( void * instance, 
                                          bool dealloc ) const {
//-------------------------------------------------------------------------------
  if ( * this ) m_typeName->m_typeBase->destruct( instance, dealloc ); 
}


//-------------------------------------------------------------------------------
inline MLANG_String MLANG_Reflection::Type::fullName( bool typedefexp ) const {
//-------------------------------------------------------------------------------
  MLANG_String s = "";
  if ( isConst()) s += "const ";
  if ( isVolatile()) s += "volatile ";
  if ( * this ) s =  m_typeName->m_typeBase->fullName( typedefexp );
  else if ( id() ) s =  m_typeName->name();
  if ( isReference()) s += "&";
  return s;
}


//-------------------------------------------------------------------------------
inline void * MLANG_Reflection::Type::id() const {
//-------------------------------------------------------------------------------
  return (void*)m_typeName;
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::Type::isArray() const {
//-------------------------------------------------------------------------------
  if ( * this ) return m_typeName->m_typeBase->isArray();
  return false;
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::Type::isClass() const {
//-------------------------------------------------------------------------------
  if ( * this ) return m_typeName->m_typeBase->isClass();
  return false;
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::Type::isConst() const {
//-------------------------------------------------------------------------------
  return 0 != (m_modifiers & MLANG_CONST);
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::Type::isConstVolatile() const {
//-------------------------------------------------------------------------------
  return 0 != (m_modifiers & MLANG_CONST & VOLATILE);
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::Type::isEnum() const {
//-------------------------------------------------------------------------------
  if ( * this ) return m_typeName->m_typeBase->isEnum();
  return false;
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::Type::isFunction() const {
//-------------------------------------------------------------------------------
  if ( * this ) return m_typeName->m_typeBase->isFunction();
  return false;
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::Type::isFundamental() const {
//-------------------------------------------------------------------------------
  if ( * this ) return m_typeName->m_typeBase->isFundamental();
  return false;
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::Type::isInstantiatedTemplateClass() const {
//-------------------------------------------------------------------------------
  if ( * this ) return m_typeName->m_typeBase->isInstantiatedTemplateClass();
  return false;
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::Type::isInstantiatedTemplateFunction() const {
//-------------------------------------------------------------------------------
  if ( * this ) return m_typeName->m_typeBase->isInstantiatedTemplateFunction(); 
  return false;
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::Type::isPointer() const {
//-------------------------------------------------------------------------------
  if ( * this ) return m_typeName->m_typeBase->isPointer();
  return false;
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::Type::isPointerToMember() const {
//-------------------------------------------------------------------------------
  if ( * this ) return m_typeName->m_typeBase->isPointerToMember();
  return false;
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::Type::isReference() const {
//-------------------------------------------------------------------------------
  return 0 != ( m_modifiers & REFERENCE );
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::Type::isTemplatedClass() const {
//-------------------------------------------------------------------------------
  if ( * this ) return m_typeName->m_typeBase->isTemplatedClass();
  return false;
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::Type::isTemplatedFunction() const {
//-------------------------------------------------------------------------------
  if ( * this ) return m_typeName->m_typeBase->isTemplatedFunction();
  return false;
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::Type::isTypedef() const {
//-------------------------------------------------------------------------------
  if ( * this ) return m_typeName->m_typeBase->isTypedef();
  return false;
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::Type::isUnion() const {
//-------------------------------------------------------------------------------
  if ( * this ) return m_typeName->m_typeBase->isUnion();
  return false;
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::Type::isUnqualified() const {
//-------------------------------------------------------------------------------
  return 0 == m_modifiers;
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::Type::isVolatile() const {
//-------------------------------------------------------------------------------
  return 0 != ( m_modifiers & VOLATILE );
}


//-------------------------------------------------------------------------------
inline size_t MLANG_Reflection::Type::length() const {
//-------------------------------------------------------------------------------
  if ( * this ) return m_typeName->m_typeBase->length();
  return 0;
}


//-------------------------------------------------------------------------------
inline MLANG_String MLANG_Reflection::Type::name() const {
//-------------------------------------------------------------------------------
  return m_typeName->name();
}


//-------------------------------------------------------------------------------
inline MLANG_Reflection::Type MLANG_Reflection::Type::parameter( size_t nth ) const {
//-------------------------------------------------------------------------------
  if ( * this ) return m_typeName->m_typeBase->parameter( nth );
  return Type();
}


//-------------------------------------------------------------------------------
inline size_t MLANG_Reflection::Type::parameterCount() const {
//-------------------------------------------------------------------------------
  if ( * this ) return m_typeName->m_typeBase->parameterCount();
  return 0;
}


//-------------------------------------------------------------------------------
inline MLANG_Reflection::PropertyList & MLANG_Reflection::Type::propertyList() const {
//-------------------------------------------------------------------------------
  if ( * this ) return m_typeName->m_typeBase->propertyList();
  return PropertyList::__NOPROPERTYLIST__;
}


//-------------------------------------------------------------------------------
inline MLANG_Reflection::Type MLANG_Reflection::Type::returnType() const {
//-------------------------------------------------------------------------------
  if ( * this ) return m_typeName->m_typeBase->returnType();
  return Type();
}


//-------------------------------------------------------------------------------
inline size_t MLANG_Reflection::Type::sizeOf() const {
//-------------------------------------------------------------------------------
  if ( * this ) return m_typeName->m_typeBase->sizeOf();
  return 0;
}


//-------------------------------------------------------------------------------
inline MLANG_Reflection::Type MLANG_Reflection::Type::toType() const {
//-------------------------------------------------------------------------------
  if ( * this ) return m_typeName->m_typeBase->toType();
  return Type();
}


//-------------------------------------------------------------------------------
inline const std::type_info & MLANG_Reflection::Type::typeInfo() const {
//-------------------------------------------------------------------------------
  if ( * this ) return m_typeName->m_typeBase->typeInfo(); 
  return typeid(void);
}


//-------------------------------------------------------------------------------
inline MLANG_Reflection::TYPE MLANG_Reflection::Type::typeType() const {
//-------------------------------------------------------------------------------
  if ( * this ) return m_typeName->m_typeBase->typeType();
  return UNRESOLVED;
}


//-------------------------------------------------------------------------------
inline MLANG_String MLANG_Reflection::Type::typeTypeAsString() const {
//-------------------------------------------------------------------------------
  if ( * this ) return m_typeName->m_typeBase->typeTypeAsString(); 
  return "UNRESOLVED";
}


#endif //MLANG_REFLECTION_TYPE_H
