//$Id: TypeBase.h,v 1.11 2004/11/30 22:57:58 roiser Exp $

#ifndef MLANG_REFLECTION_TYPEBASE_H
#define MLANG_REFLECTION_TYPEBASE_H

// Include files
#include "MLANG/Reflection/Kernel.h"
#include "MLANG/Reflection/Scope.h"
#include "MLANG/Reflection/PropertyList.h"
#include <vector>
#include <typeinfo>

namespace MLANG_Reflection
{

    // forward declarations
    class Fundamental;
    class Function;
    class Array;
    class Class;
    class Enum;
    class Typedef;
    class Pointer;
    class InstantiatedTemplateClass;
    class InstantiatedTemplateFunction;
    class Type;
    class TypeName;
    class Object;
    class EnumInfo;
    class UnionInfo;


    /**
     * @class TypeBase TypeBase.h MLANG/Reflection/TypeBase.h
     * @author Stefan Roiser
     * @date 24/11/2003
     * @ingroup Ref
     */
    class TypeBase {

    public:

      /** default constructor */
      TypeBase( const char *           type, 
                size_t                 size, 
                TYPE                   typeType,
                const std::type_info & typeInfo);


      /** destructor */
      virtual ~TypeBase();


      /**
       * operator Type will return the corresponding Type object
       * @return Type corresponding to this TypeBase
       */
      operator Type() const;


      /**
       * allocate will reserve memory for the size of the object
       * @return pointer to allocated memory
       */
      void * allocate() const;


      /**
       * construct will call the constructor of a given type and allocate the
       * memory for it
       * @param  signature of the constructor
       * @param  values for parameters of the constructor
       * @param  mem place in memory for implicit construction
       * @return pointer to new instance
       */
      virtual Object construct( const Type & signature,
                                std::vector < void * > values,
                                void * mem) const;
      
      /**
       * deallocate will deallocate the memory for a given object
       * @param  instance of the type in memory
       */
      void deallocate( void * instance ) const;


      /**
       * destruct will call the destructor of a type and remove its memory
       * allocation if desired
       * @param  instance of the type in memory
       * @param  dealloc for also deallacoting the memory
       */
      virtual void destruct( void * instance, 
                             bool dealloc = true ) const;


      /**
       * enumInfo will return the nth enum information
       * @param  nth enum
       * @return nth enum info object
       */
      virtual EnumInfo enumInfo(size_t nth) const;


      /**
       * enumInfoCount will return the number of enums
       * @return number of enums
       */
      virtual size_t enumInfoCount() const;


      /**
       * fullName returns the full name of the type
       * @param  typedefexp expand typedefs or not
       * @return full name of type
       */
      virtual MLANG_String fullName( bool typedefexp = false ) const;


      /** 
       * isArray returns true if the type represents a Array
       * @return true if type represents a Array
       */
      bool isArray() const;


      /** 
       * isClass returns true if the type represents a Class
       * @return true if type represents a Class
       */
      bool isClass() const;


      /** 
       * isEnum returns true if the type represents a Enum
       * @return true if type represents a Enum
       */
      bool isEnum() const;


      /** 
       * isFunction returns true if the type represents a Function
       * @return true if type represents a Function
       */
      bool isFunction() const;


      /** 
       * isFundamental returns true if the type represents a Fundamental
       * @return true if type represents a Fundamental
       */
      bool isFundamental() const;


      /** 
       * isInstantiatedTemplateClass returns true if the type represents a 
       * IntantiatedTemplateClass
       * @return true if type represents a InstantiatedTemplateClass
       */
      bool isInstantiatedTemplateClass() const;


      /** 
       * isInstantiatedTemplateFunction returns true if the type represents a 
       * InstantiatedTemplateFunction
       * @return true if type represents a InstantiatedTemplateFunction
       */
      bool isInstantiatedTemplateFunction() const;


      /** 
       * isPointer returns true if the type represents a Pointer
       * @return true if type represents a Pointer
       */
      bool isPointer() const;


      /** 
       * isPointerToMember returns true if the type represents a PointerToMember
       * @return true if type represents a PointerToMember
       */
      bool isPointerToMember() const;


      /**
       * isTemplatedClass returns true if the type is a templated class
       * @return true if Type is templated class
       */
      bool isTemplatedClass() const;


      /**
       * isTemplatedFunction returns true if the type is a templated function
       * @return true if Type is templated function
       */
      bool isTemplatedFunction() const;


      /** 
       * isTypedef returns true if the type represents a Typedef
       * @return true if type represents a Typedef
       */
      bool isTypedef() const;


      /** 
       * isUnion returns true if the type represents a Union
       * @return true if type represents a 
       */
      bool isUnion() const;


      /** Array
       * size returns the size of the array
       * @return size of array
       */
      virtual size_t length() const;

      
      /**
       * name returns the name of the type
       * @return name of type
       */
      const MLANG_String & name() const;

 
      /**
       * parameter returns the nth parameter
       * @param  nth nth parameter
       * @return pointer to nth parameter type
       */
      virtual Type parameter( size_t nth ) const;


      /**
       * parameterCount will return the number of parameters of this function
       * @return number of parameters
       */
      virtual size_t parameterCount() const;


      /**
       * propertyList will return a pointer to the property list attached
       * to this item
       * @return pointer to property list
       */
      virtual PropertyList & propertyList() const;


      /**
       * returnType will return a pointer to the type of the return type.
       * @return pointer to Type of return type
       */
      virtual Type returnType() const;
      

      /**
       * scope will return the scope of the Type if any 
       * @return pointer to scope of type
       */
      Scope scope() const;


      /**
       * sizeof will return the size of the type
       * @return size of the type as int
       */
      size_t sizeOf() const;


      /**
       * arrayType will return a pointer to the type of the array.
       * @return pointer to Type of member et. al.
       */
      virtual Type toType() const;


      /** 
       * type returns the corresponding unqualified Type object
       * @return corresponding type object
       */
      Type type() const;


      /**
       * typeInfo will return the c++ type_info object of the type
       * @return type_info object of type
       */
      const std::type_info & typeInfo() const;


      /**
       * typeType will return the real type
       * @return real type
       */
      TYPE typeType() const;


      /**
       * typeTypeAsString will return the string representation of the TYPE type
       * @return string representation of TYPE type
       */
      MLANG_String typeTypeAsString() const;


      /**
       * unionInfo will return the nth union info
       * @param  nth union info
       * @return nth union info
       */
      virtual UnionInfo unionInfo( size_t nth ) const;


      /**
       * unionInfoCount will return the number of union types
       * @return number of union types
       */
      virtual size_t unionInfoCount() const;

    private:

      /**
       * The scope of the type
       * @label scope type
       * @link aggregation
       */
      Scope m_scope;


      /** size of the type in int */
      size_t m_size;


      /** C++ type_info object */
      const std::type_info & m_typeInfo;


      /**
       * type of the type
       * @link aggregationByValue
       * @label typeType
       */
      TYPE m_typeType;


      /**
       * pointer to the property list
       * @label propertylist
       * @link aggregationByValue
       */
      mutable
      PropertyList m_propertyList;


      /**
       * Pointer to the TypeName 
       * @label type name
       * @ling aggregation
       */
      TypeName * m_typeName;

    }; // class Type
} //namespace MLANG_Reflection


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::TypeBase::isArray() const {
//-------------------------------------------------------------------------------
  return ( m_typeType == ARRAY );
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::TypeBase::isClass() const {
//-------------------------------------------------------------------------------
  return ( m_typeType == CLASS );
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::TypeBase::isEnum() const {
//-------------------------------------------------------------------------------
  return ( m_typeType == ENUM );
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::TypeBase::isFunction() const {
//-------------------------------------------------------------------------------
  return ( m_typeType == FUNCTION );
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::TypeBase::isFundamental() const {
//-------------------------------------------------------------------------------
  return ( m_typeType == FUNDAMENTAL );
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::TypeBase::isInstantiatedTemplateClass() const {
//-------------------------------------------------------------------------------
  return ( m_typeType == INSTANTIATEDTEMPLATECLASS );
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::TypeBase::isInstantiatedTemplateFunction() const {
//-------------------------------------------------------------------------------
  return ( m_typeType == INSTANTIATEDTEMPLATEFUNCTION );
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::TypeBase::isPointer() const {
//-------------------------------------------------------------------------------
  return ( m_typeType == POINTER );
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::TypeBase::isPointerToMember() const {
//-------------------------------------------------------------------------------
  return ( m_typeType == POINTERTOMEMBER );
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::TypeBase::isTemplatedClass() const {
//-------------------------------------------------------------------------------
  return ( m_typeType == INSTANTIATEDTEMPLATECLASS );
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::TypeBase::isTemplatedFunction() const {
//-------------------------------------------------------------------------------
  return ( m_typeType == INSTANTIATEDTEMPLATEFUNCTION );
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::TypeBase::isTypedef() const {
//-------------------------------------------------------------------------------
  return ( m_typeType == TYPEDEF );
}


//-------------------------------------------------------------------------------
inline bool MLANG_Reflection::TypeBase::isUnion() const {
//-------------------------------------------------------------------------------
  return ( m_typeType == UNION );
}


//-------------------------------------------------------------------------------
inline size_t MLANG_Reflection::TypeBase::sizeOf() const { 
//-------------------------------------------------------------------------------
  return m_size; 
}


//-------------------------------------------------------------------------------
inline const std::type_info & MLANG_Reflection::TypeBase::typeInfo() const {
//-------------------------------------------------------------------------------
  return m_typeInfo;
}


//-------------------------------------------------------------------------------
inline size_t MLANG_Reflection::TypeBase::unionInfoCount() const {
//-------------------------------------------------------------------------------
  return 0; 
}

#endif // MLANG_REFLECTION_TYPEBASE_H



