/* -*- C++ -*- */

//=============================================================================
/**
 *  @file SmartPtr.h
 *
 *  Request.h,v 1.x 2004/06/14
 *
 *  @author Arturo Montes <mitosys@colomsat.net.co>
 */
//=============================================================================

#ifndef __MLANG_SMARTPTR_H__
#define __MLANG_SMARTPTR_H__
#include "ace/Auto_Ptr.h"
#include "ace/Synch.h"
#include "ace/Method_Request.h"
#include "MLANG/Export.h"

template <class T>
ACE_INLINE
void MLANG_release (T* ptr)
{
   if (!T::is_nil(ptr))
    {
      ptr->decrement();
    }
}

template <class T>
ACE_INLINE
T* MLANG_acquire (T* ptr)
{
   if (!T::is_nil(ptr))
    {
      ptr->increment();
    }
   return ptr;
}

/**
 * @class MLANG_SmartPtr
 *
 * @brief Este es un Smart Pointer a un objeto de tipo @ref MLANG_SmartPtr.
 * Utiliza el contador de referencias de @ref MLANG_SmartPtr.
 */
template <class T>
class MLANG_SmartPtr
{
public:
   typedef T Pointee;
   typedef T *Pointer;
   // = Creacion y destruccion
   /// Constructor por defecto
   /// 
   /// @param p apunta a un @ref MLANG_SmartPtr por defecto es @ref MLANG_SmartPtr::_nil.
#if !defined(WIN32)
   MLANG_SmartPtr(ACE_TYPENAME MLANG_SmartPtr<T>::Pointer p=T::_nil()) : ptr_(p) {}
#else /* WIN32 */
   MLANG_SmartPtr(Pointer p=T::_nil()) : ptr_(p) {}
#endif /* WIN32 */

   /// Copy constructor. Incrementa el contador de referencias del apuntador
   /// que contiene el Smart Pointer @a a.
   ///
   /// @param a es un Smart Pointer a una solicitud.
   ///
   MLANG_SmartPtr(const MLANG_SmartPtr<T> &a)
     : ptr_(T::_duplicate(Pointer(a))) {}

#if !defined(WIN32)
   /// Copy constructor. Incrementa el contador de referencias del apuntador
   /// que contiene el Smart Pointer @a a.
   ///
   /// @param a es un Smart Pointer a una solicitud.
   ///
   template <class oldT>
   MLANG_SmartPtr(const MLANG_SmartPtr<oldT> &a)
     : ptr_(T::_duplicate(a.in())) {}
#endif /* !WIN32 */

   /// Destructor. Llama a MLANG_SmartPtr::free.
   ///
   ~MLANG_SmartPtr()
   { free(); }
public:

   /// Operador de asignacion de un @ref MLANG_SmartPtr_ptr. Se supone que el llamador
   /// es el dueño del apuntador y se lo esta asignando al SmartPointer. El
   /// llamador esta delegando la propiedad del apuntador al SmartPointer.
   /// 
   /// @param p apunta a un @ref MLANG_SmartPtr.
   /// el objeto anterior es liberado.
   MLANG_SmartPtr &operator=(ACE_TYPENAME MLANG_SmartPtr<T>::Pointer p)
   { reset(p); return *this; }

   /// Operador de asignacion. Tanto @a a como @a this comparten la propiedad
   /// del objeto al que apunta.
   ///
   /// @param a es un SmartPointer.
   template <class oldT>
   MLANG_SmartPtr &operator=(const MLANG_SmartPtr<oldT>& a) {
     free();
     ptr_ = T::_duplicate(a.in());
     return *this;
   }

   /// Operador de asignacion. Tanto @a a como @a this comparten la propiedad
   /// del objeto al que apunta.
   ///
   /// @param a es un SmartPointer.
   MLANG_SmartPtr &operator=(const MLANG_SmartPtr<T>& a) {
     if (this != &a) {
        free();
        ptr_ = T::_duplicate(Pointer(a));
     }
     return *this;
   }

   /// Operador de asignacion. Tanto @a a como @a this comparten la propiedad
   /// del objeto al que apunta.
   ///
   /// @param a es un SmartPointer.
   bool operator==(const MLANG_SmartPtr<T>& a) {
     return (this.ptr_ == a.ptr_);
   }

public:

#if 0
   template <class newT>
   operator MLANG_SmartPtr<newT>() 
   { return MLANG_SmartPtr<newT> (T::_duplicate(ptr_)); }

   template <class newT>
   operator const MLANG_SmartPtr<newT>() const
   { return MLANG_SmartPtr<newT> (T::_duplicate(ptr_)); }
#endif

   /// Operador de conversion cuando se desea probar si este es el apuntador
   /// al objeto null.
   ///
   operator bool () const
   { return !T::is_nil(ptr_); }

   /// Operador de conversion cuando se desea probar si este es el apuntador
   /// al objeto null.
   ///
   bool is_null() const
   { return T::is_nil(ptr_); }

   /// Retorna una referencia al objeto que apunta. Para ser utilizado
   /// como parametro de entrada a un metodo. @a this sigue siendo el
   /// dueño del objeto al que apunta.
   ///
   typename MLANG_SmartPtr<T>::Pointer in() const 
   { return ptr_; }

   /// Retorna una referencia al area de memoria al cual apunta a un
   /// @ref Proceso. Util cuando se utiliza el SmartPointer como parametro
   /// de entrada salida de un metodo.
   ///
   /// @retval ptr_ es una referencia al objeto apuntado por @a this.
   typename MLANG_SmartPtr<T>::Pointer& inout() 
   { return ptr_; }

   /// Retorna una referencia al objeto apuntado por @a this pero liberando
   /// el apuntador. Este metodo se utiliza cuando se desea pasar un SmartPointer
   /// como parametro de salida.
   ///
   /// @retval una referencia al atributo @a ptr_ de @a *this.
   typename MLANG_SmartPtr<T>::Pointer& out()
   {
     reset(T::_nil());
     return ptr_;
   }

   /// Retorna el apuntador que contiene transfiriendo la propiedad del apuntador
   /// al llamador.
   ///
   /// @retval el valor anterior de ptr_, ptr_ es inicializado a MLANG_SmartPtr::_nil.
   typename MLANG_SmartPtr<T>::Pointer _retn()
   {
     Pointer val = ptr_;
     ptr_ = T::_nil();
     return val;
   }

   /// Operador de conversion a un apuntador
   ///
   operator typename MLANG_SmartPtr<T>::Pointer() const 
   { return ptr_; }

   /// @deprecated
   /// Obtiene una referencia al apuntador. No debe utilizarse mas.
   ///
//   operator typename MLANG_SmartPtr<T>::Pointer&() 
//   { return ptr_; }

   /// Operador de conversion. Cuando se desea manipular la estructura del
   /// objeto al que apunta.
   ///
   typename MLANG_SmartPtr<T>::Pointer operator->() const 
   { return ptr_; }

protected:

   /// Apuntador al objeto real
   ///
   Pointer ptr_;

   /// Decrementa el contador de referencias al objeto que apunta
   ///
   void free() 
   { MLANG_release(ptr_); }

   /// Cambia el valor del atributo @ref ptr_ por @a p. Si ptr_ no es T::_nil
   /// decrementa el contado de referencias de lo apuntado por ptr_.
   ///
   /// @param p es el nuevo objeto que controla este SmartPointer.
   void reset(Pointer p) 
   { free(); ptr_ = p; }
};


template <typename newT, typename oldT>
MLANG_SmartPtr<newT> Dynamic_Cast(const MLANG_SmartPtr<oldT>& a)
{
   return MLANG_SmartPtr<newT>(newT::_narrow(a.in())); 
}

template <typename newT, typename oldT>
newT* Dynamic_Cast(oldT* p)
{
   return newT::_narrow(p); 
}

template <typename newT, typename oldT>
bool instanceof (oldT *p)

{ return !newT::is_nil(ACE_dynamic_cast(newT*,p)); }

template <typename newT, typename oldT>
bool instanceof (const MLANG_SmartPtr<oldT>& a)

{ return !newT::is_nil(ACE_dynamic_cast(newT*,a.in())); }

#endif /* __MLANG_SMARTPTR_H__ */
