/* -*- C++ -*- */

//=============================================================================
/**
 *  @file String.h
 *
 *  String.h,v 1.x 2004/06/14
 *
 *  @author Arturo Montes <mitosys@colomsat.net.co>
 */
//=============================================================================

#if !defined(__MLANG_STRING_H__)
#define __MLANG_STRING_H__
//#include "mbase/config.h"
#include "ace/Basic_Types.h"
#include "MLANG/Export.h"
#include "MLANG/Vector_T.h"

namespace MLANG
{
class MLANG_Export String;

void MLANG_Export upcase(String& s);

void MLANG_Export downcase(String& s);

String MLANG_Export operator+(const String& a, const String& b);

ACE_INLINE
int compare (
   const String& a,
   const String& b
);

/**
 * @class String
 *
 * @brief Esta es la clase que implementa un tipo cadena utilizado por
 * un parser. Esta clase a diferencia de las clases de cadenas normales
 * permite armar una cadena en forma incremental sin incurrir el sobrecosto
 * de estar asignando el caracter \\0 por cada adicion de un caracter.
 */
class MLANG_Export String
{
public:
   // = Creacion y destruccion de una cadena

   /// Constructor a partir de un caracter
   ///
   /// @param c es el caracter que forma la cadena inicialmente.
   String(char c);
   /// Constructor a partir de una cadena termina en \\0
   ///
   /// @param s es el apuntador de donde se creara la cadena
   String(const char* s);
   /// Constructor a partir de un buffer
   ///
   /// @param buf es el apuntador de donde se creara la cadena
   /// @param len es el numero de caracteres en el buffer
   String(const char* buf, size_t len);
   /// Crea una cadena \\0 que va a ser utilizada en un parser.
   ///
   /// @param maxlen es la maxima longitud inicial
   /// @param incrlen es el valor en que se debe incrementar la longitud
   /// cuando se necesita crecer la cadena.
   String(size_t maxlen =80, size_t incrlen = 80);
   /// Copy constructor
   ///
   /// @param ts referencia al objeto que se desea copiar.
   String(const String& ts);
   /// Destructor
   ///
   ~String();
public:
   /// Retorna el caracter @a c como una cadena
   ///
   /// @param c es el caracter que se desea convertir
   static String valueOf(char c)
   { return String(c); }

   /// Retorna el valor booleano @a b como una cadena
   ///
   /// @param b es el valor que se desea convertir
   static String valueOf(bool b);

#ifndef HPUX
   static String valueOf (size_t i, int base = 10);
#endif /* !HPUX */

   static String valueOf (short d);
   
   static String valueOf (int i, int base = 10);

   static String valueOf (long l, int base = 10);

   static String valueOf (ACE_UINT64 i, int base = 10);

   static String valueOf (double d);

   static String valueOf (float d);

public:
   /// Le indica que estamos iniciando una nueva cadena
   ///
   void start()
   { len_ = 0; }
   /// Indicamos que finalizamos de construir una cadena.
   ///
   void end()
   { buf_[len_] = '\0'; }
   /// Adicionamos un caracter al final de la cadena
   ///
   /// @param c es el caracter que deseamos adicionar
   void append(int c);
public:
   // = Manipuladores de atributo

   /// Longitud de la cadena
   ///
   /// @retval len_ es la longitud actual de la cadena.
   size_t length() const;

   /// Conseguimos el apuntador al buffer donde reside la cadena
   ///
   /// @retval buf_ apunta a la cadena de caracteres almacenado en String
   const char* c_str() const
   { return buf_; }
   /// Extraemos el elemento @a i-esimo de la cadena. 
   ///
   /// @param i es el indica del elemento en el token.
   char charAt(size_t i) const
   { return buf_[i]; }
   /// Extraemos el elemento @a i-esimo de la cadena. 
   ///
   /// @param i es el indica del elemento en el token.
   char operator[](size_t i) const
   { return buf_[i]; }
   /// Es el token fin?
   ///
   /// @retval 1, si esta es el token \\0.
   /// @retval 0, en caso contrario.
   int is_fin() const
   { return len_ == 0; }
public:
   /// Indica si la cadena s es igual 
   ///
   bool equals (const char *s) const
   { return strcmp (buf_, s) == 0; }
   bool equals (const String& s) const
   { return equals (s.c_str()); }

   int indexOf (const String& s, int startpos = 0) const;

   int indexOf (char c, int startpos = 0) const;

   int lastIndexOf (const String& s) const;

   bool startsWith (const String& s) const;

   bool startsWith (const char* s) const;

   bool endsWith (const String& s) const;

   bool endsWith (const char* s) const;

   String trim () const;

   String substring (int begin, int end = -1) const;

public:

   int find_first_of (const String& s, int pos = 0) const;

   int find_first_of (const char *s, int pos = 0) const;

   int find_first_of (const char c, int pos = 0) const;

public:

   String toLowerCase() const;

   String toUpperCase() const;

   static int npos;

public:

   int compareTo (
      const String& b
    ) const;

   int compareToIgnoreCase (
      const String& b
    ) const;

public:
   // = Operadores de asignacion
   
   /// Asigna desde un String
   ///
   /// @param ts es una referencia al objeto String a ser copiado.
   /// @retval *this una referencia al mismo objeto.
   String& operator=(const String& ts);
   /// Concatenacion al final de un String
   ///
   /// @param ts es una referencia al objeto String a ser concatenado
   /// @retval *this una referencia al mismo objeto.
   String& operator+=(const String& ts);
   /// Operador de concatenacion utilizando un apuntador a una cadena de
   /// caracteres @a s.
   ///
   /// @param s es un apuntador a la cadena de caracteres a ser copiada.
   /// @retval *this una referencia al mismo objeto.
   String& operator=(const char* s);

   int operator< (
      const String& a
    ) const;

   int operator<= (
      const String& a
   ) const;

   int operator== (
      const String& a
   ) const;

   int operator!= (
      const String& a
   ) const;
   /// Borra el contenido de la cadena
   ///
   void clear();

protected:
   size_t maxlen_; // Tamano maximo de buf_
   size_t incrlen_; // Incremento en caracteres 
   size_t len_; // Longitud actual de la cadena
   char* buf_; // Apuntador al buffer donde se encuentra la cadena
public:
   friend void MLANG_Export MLANG::upcase(String& s);
   friend void MLANG_Export MLANG::downcase(String& s);
   friend int compare(const String& a, const String& b);
   friend String MLANG_Export MLANG::operator+(const String& a, const String& b);
};

ACE_INLINE
int compare (
   const String& a,
   const String& b
);

#if 0
ACE_INLINE
int operator< (
   const String& a,
   const String& b
);

ACE_INLINE
int operator<= (
   const String& a,
   const String& b
);

ACE_INLINE
int operator== (
   const String& a,
   const String& b
);

ACE_INLINE
int operator!= (
   const String& a,
   const String& b
);
#endif

}

#if defined (__ACE_INLINE__)
#include "MLANG/String.i"
#endif /* __ACE_INLINE__ */

typedef MLANG::String MLANG_String;

#endif /* __MLANG_STRING_H__ */
