/* -*- C++ -*- */

//=============================================================================
/**
 *  @file StringBuffer.h
 *
 *  StringBuffer.h,v 1.x 2004/06/14
 *
 *  @author Arturo Montes <mitosys@colomsat.net.co>
 */
//=============================================================================

#if !defined(__MPARSER_STRINGBUFFER_H__)
#define __MPARSER_STRINGBUFFER_H__
#include "mbase/config.h"
#include "MLANG/String.h"

namespace MLANG
{
/**
 * @class StringBuffer
 *
 * @brief Esta es la clase que implementa un tipo cadena utilizado por
 * un parser. Esta clase a diferencia de las clases de cadenas normales
 * permite armar una cadena en forma incremental sin incurrir el sobrecosto
 * de estar asignando el caracter \\0 por cada adicion de un caracter.
 */
class MExport StringBuffer
{
public:
   // = Creacion y destruccion de una cadena

   /// Constructor a partir de un caracter
   ///
   /// @param c es el caracter que forma la cadena inicialmente.
   StringBuffer(char c);
   /// Constructor a partir de una cadena termina en \\0
   ///
   /// @param s es el apuntador de donde se creara la cadena
   StringBuffer(const char* s);
   /// Crea una cadena \\0 que va a ser utilizada en un parser.
   ///
   /// @param maxlen es la maxima longitud inicial
   /// @param incrlen es el valor en que se debe incrementar la longitud
   /// cuando se necesita crecer la cadena.
   StringBuffer(size_t maxlen=80,size_t incrlen=80);
   /// Copy constructor
   ///
   /// @param ts referencia al objeto que se desea copiar.
   StringBuffer(const StringBuffer& ts);
   /// Destructor
   ///
   ~StringBuffer();
public:
   /// Le indica que estamos iniciando una nueva cadena
   ///
   void start()
   { len_ = 0; }
   /// Indicamos que finalizamos de construir una cadena.
   ///
   void end()
   { buf_[len_] = '\0'; }
   /// Adicionamos un caracter al final de la cadena
   ///
   /// @param c es el caracter que deseamos adicionar
   StringBuffer& append(int c);
   /// Adicionamos un cadena al final del buffer
   ///
   /// @param s es la cadena que deseamos adicionar
   StringBuffer& append(const char* s);
   /// Adicionamos la cadena de caracteres al final del buffer
   ///
   /// @param s es la cadena que deseamos adicionar
   StringBuffer& append(const String& s);
public:
   // = Manipuladores de atributo

   /// Longitud de la cadena
   ///
   /// @retval len_ es la longitud actual de la cadena.
   size_t length() const
   { return len_; }
   void setLength(size_t len)
   { len_ = len; }
   /// Conseguimos el apuntador al buffer donde reside la cadena
   ///
   /// @retval buf_ apunta a la cadena de caracteres almacenado en StringBuffer
   const char* c_str() const
   { return buf_; }
   /// Conseguimos el apuntador al buffer donde reside la cadena
   ///
   /// @retval buf_ apunta a la cadena de caracteres almacenado en StringBuffer
   const char* getBuffer() const
   { return buf_; }
   /// Extraemos el elemento @a i-esimo de la cadena. 
   ///
   /// @param i es el indica del elemento en el token.
   char operator[](size_t i) const
   { return buf_[i]; }
   /// Es el token fin?
   ///
   /// @retval 1, si esta es el token \\0.
   /// @retval 0, en caso contrario.
   int is_fin() const
   { return len_ == 0; }

   /// Retorna un String contenido en el buffer
   ///
   String toString() const
   { return String (buf_, len_); }
public:
   // = Operadores de asignacion
   
   /// Asigna desde un StringBuffer
   ///
   /// @param ts es una referencia al objeto StringBuffer a ser copiado.
   /// @retval *this una referencia al mismo objeto.
   StringBuffer& operator=(const StringBuffer& ts);
   /// Concatenacion al final de un StringBuffer
   ///
   /// @param ts es una referencia al objeto StringBuffer a ser concatenado
   /// @retval *this una referencia al mismo objeto.
   StringBuffer& operator+=(const StringBuffer& ts);
   /// Operador de concatenacion utilizando un apuntador a una cadena de
   /// caracteres @a s.
   ///
   /// @param s es un apuntador a la cadena de caracteres a ser copiada.
   /// @retval *this una referencia al mismo objeto.
   StringBuffer& operator=(const char* s);
protected:
   size_t maxlen_; // Tamano maximo de buf_
   size_t incrlen_; // Incremento en caracteres 
   size_t len_; // Longitud actual de la cadena
   char* buf_; // Apuntador al buffer donde se encuentra la cadena
};
}

typedef MLANG::StringBuffer MLANG_StringBuffer;

#endif /* __StringBuffer_H__ */
