#include "MParser/System.h"
#include "MParser/CommonAST.h"
#include "MParser/ASTFactory.h"

/* ANTLR Translator Generator
 * Project led by Terence Parr at http://www.jGuru.com
 * Software rights: http://www.antlr.org/license.html
 *
 * $Id: //depot/code/org.antlr/release/antlr-2.7.4/antlr/ASTFactory.java#1 $
 */

/** AST Support code shared by TreeParser and Parser.
 *  We use delegation to share code (and have only one
 *  bit of code to maintain) rather than subclassing
 *  or superclassing (forces AST support code to be
 *  loaded even when you don't want to do AST stuff).
 *
 *  Typically, setASTNodeType is used to specify the
 *  homogeneous type of node to create, but you can override
 *  create to make heterogeneous nodes etc...
 */

MParser_ASTFactory::MParser_ASTFactory()

{
}

/** Create factory with a specific mapping from token type
 *  to Java AST node type.  Your subclasses of ASTFactory
 *  can override and reuse the map stuff.
 */

MParser_ASTFactory::MParser_ASTFactory (
   const CLASSMAP& tokenTypeToClassMap
)

{
   setTokenTypeToASTClassMap(tokenTypeToClassMap);
}

/** Specify an "override" for the Java AST object created for a
 *  specific token.  It is provided as a convenience so
 *  you can specify node types dynamically.  ANTLR sets
 *  the token type mapping automatically from the tokens{...}
 *  section, but you can change that mapping with this method.
 *  ANTLR does it's best to statically determine the node
 *  type for generating parsers, but it cannot deal with
 *  dynamic values like #[LT(1)].  In this case, it relies
 *  on the mapping.  Beware differences in the tokens{...}
 *  section and what you set via this method.  Make sure
 *  they are the same.
 *
 *  Set className to null to remove the mapping.
 *
 *  @since 2.7.2
 */

void MParser_ASTFactory::setTokenTypeASTNodeType (
   int tokenType,
   const MLANG_String& className
)

{
   if (className.length()==0)

    {
      tokenTypeToASTClassMap_.remove(tokenType);
      return;
    }
   MLANG_Class_var c;
   try

    {
       c = MLANG_Class::forName(className);
       tokenTypeToASTClassMap_.put(tokenType, c);
    }
   catch (MLANG_Exception& e)

    {
    #if 0
       throw new IllegalArgumentException("Invalid class, "+className);
    #else
       assert(0);
    #endif
    }
}

/** For a given token type, what is the AST node object type to create
 *  for it?
 *  @since 2.7.2
 */

MLANG_Class_ptr MParser_ASTFactory::getASTNodeType(int tokenType)

{
   // try node specific class
   if (tokenTypeToASTClassMap_.size()!=0 )
    {
      MLANG_Class_var c = tokenTypeToASTClassMap_.get(tokenType);
      if (!MLANG_Class::is_nil(c))

       {
         return c;
       }
    }

   // try a global specified class
   if (!MLANG_Class::is_nil(theASTNodeTypeClass_))

    {
   	return theASTNodeTypeClass_.in();
    }

   // default to the common type
   return MParser_CommonAST::Class();
}

#if 0
/** Add a child to the current AST */

void MParser_ASTFactory::addASTChild(ASTPair currentAST, AST child)

{
   if (child != null)

    {
      if (currentAST_->root() == null)

       {
         // Make new child the current root
         currentAST.root = child;
       }
      else

       {
         if (currentAST.child == null)

          {
            // Add new child to current root
            currentAST.root.setFirstChild(child);
          }
         else

          {
             currentAST.child.setNextSibling(child);
          }
       }
      // Make new child the current child
      currentAST.child = child;
      currentAST.advanceChildToEnd();
    }
}
#endif

/** Create a new empty AST node; if the user did not specify
 *  an AST node type, then create a default one: CommonAST.
 */

MParser_AST_var MParser_ASTFactory::create()

{
   return create(MParser_Token::INVALID_TYPE);
}


MParser_AST_var MParser_ASTFactory::create(int type)

{
   MLANG_Class_var c = getASTNodeType(type);
   MParser_AST_var t = create(c.in());
   if (MParser_AST::is_nil(t))

    {
      t->initialize(type, "");
    }
   return t;
}


MParser_AST_var MParser_ASTFactory::create (
   int type,
   const MLANG_String& txt
)

{
   MParser_AST_var t = create(type);
   if (!MParser_AST::is_nil(t) )

    {
      t->initialize(type, txt);
    }
   return t;
}

/** Create an AST node with the token type and text passed in, but
 *  with a specific Java object type. Typically called when you
 *  say @[PLUS,"+",PLUSNode] in an antlr action.
 *  @since 2.7.2
 */

MParser_AST_var MParser_ASTFactory::create (
   int type,
   const MLANG_String& txt,
   const MLANG_String& className
)

{
   MParser_AST_var t = create(className);
   if ( !MParser_AST::is_nil(t) )

    {
      t->initialize(type, txt);
    }
   return t;
}

/** Create a new empty AST node; if the user did not specify
 *  an AST node type, then create a default one: CommonAST.
 */

MParser_AST_var MParser_ASTFactory::create (
   const MParser_AST_var& tr
)

{
   if (MParser_AST::is_nil(tr)) return tr;   // create(null) == null
   MParser_AST_var t = create(tr->getType());
   if ( !MParser_AST::is_nil(t) )

    {
      t->initialize(tr);
    }
   return t;
}


MParser_AST_var MParser_ASTFactory::create (
   const MParser_Token_var& tok
)

{
   MParser_AST_var t = create(tok->getType());
   if ( !MParser_AST::is_nil(t) )

    {
      t->initialize(tok);
    }
   return t;
}

/** ANTLR generates reference to this when you reference a token
 *  that has a specified heterogeneous AST node type.  This is
 *  also a special case node creation routine for backward
 *  compatibility.  Before, ANTLR generated "new T(tokenObject)"
 *  and so I must call the appropriate constructor not T().
 *
 * @since 2.7.2
 */

MParser_AST_var MParser_ASTFactory::create (
   const MParser_Token_var& tok,
   const MLANG_String& className
)

{
    MParser_AST_var t = createUsingCtor(tok,className);
    return t;
}

/**
 * @since 2.7.2
 */

MParser_AST_var MParser_ASTFactory::create (
   const MLANG_String& className
)

{
   MLANG_Class_var c;
   try

    {
      c = MLANG_Class::forName(className);
    }
   catch (MLANG_Exception& e)

    {
#if 0
      throw MParser_IllegalArgumentException("Invalid class, "+className);
#else
      assert(0);
#endif
    }
   return create(c);
}

/**
 * @since 2.7.2
 */
MParser_AST_var MParser_ASTFactory::createUsingCtor (
   const MParser_Token_var& token,
   const MLANG_String& className
)

{
   MLANG_Class_var c;
   MParser_AST_var t;
#if 0
   try

    {
      c = MLANG_Class::forName(className);
      Class[] tokenArgType = new Class[]

      { antlr.Token.class };
      try

       {
          Constructor ctor = c.getConstructor(tokenArgType);
          t = (MParser_AST_ptr)ctor->newInstance(new Object[]{token}); // make a new one
       }
      catch (NoSuchMethodException e)
       {
          // just do the regular thing if you can't find the ctor
          // Your AST must have default ctor to use this.
          t = create(c);
          if ( !MParser_AST::is_nil(t) )

           {
      	     t->initialize(token);
   	   }
       }
    }
   catch (Exception e)

    {
      throw IllegalArgumentException("Invalid class or can't make instance, "+className);
    }
#else
   assert(0);
#endif
   return t;
}

/**
 * @since 2.7.2
 */
MParser_AST_var MParser_ASTFactory::create (
   const MLANG_Class_var& c
)

{
   MParser_AST_var t;
   try

    {
      t = (MParser_AST_ptr)c->newInstance(); // make a new one
    }
   catch (MLANG_Exception& e)

    {
      error("Can't create AST Node " + c->getName());
      return t;
    }
   return t;
}

/** Copy a single node with same Java AST objec type.
 *  Ignore the tokenType->Class mapping since you know
 *  the type of the node, t.getClass(), and doing a dup.
 *
 *  clone() is not used because we want all AST creation
 *  to go thru the factory so creation can be
 *  tracked.  Returns null if t is null.
 */

MParser_AST_var MParser_ASTFactory::dup (
   const MParser_AST_var& t
)

{
   if ( MParser_AST::is_nil(t) )

    {
      return t;
    }
   MParser_AST_var dup_t = create(t->getClass());
   dup_t->initialize(t);
   return dup_t;
}

/** Duplicate tree including siblings of root. */

MParser_AST_var MParser_ASTFactory::dupList (
   const MParser_AST_var& t
)

{
   MParser_AST_var iter = t;
   MParser_AST_var result = dupTree(iter);            // if t == null, then result==null
   MParser_AST_var nt = result;
   while (!MParser_AST::is_nil(iter))

    {         // for each sibling of the root
      iter = iter->getNextSibling();
      nt->setNextSibling(dupTree(iter));	// dup each subtree, building new tree
      nt = nt->getNextSibling();
    }
   return result;
}

/**Duplicate a tree, assuming this is a root node of a tree--
 * duplicate that node and what's below; ignore siblings of root node.
 */

MParser_AST_var MParser_ASTFactory::dupTree (
   const MParser_AST_var& t
)

{
   MParser_AST_var result = dup(t);   // make copy of root
   // copy all children of root.
   if (!MParser_AST::is_nil(t))

    {
       result->setFirstChild(dupList(t->getFirstChild()));
    }
   return result;
}

/** Make a tree from a list of nodes.  The first element in the
 *  array is the root.  If the root is null, then the tree is
 *  a simple list not a tree.  Handles null children nodes correctly.
 *  For example, build(a, b, null, c) yields tree (a b c).  build(null,a,b)
 *  yields tree (nil a b).
 */

MParser_AST_var MParser_ASTFactory::make (
   const ACE_Vector<MParser_AST_var>& nodes
)

{
   if (nodes.size() == 0)
      return MParser_AST::_nil();
   MParser_AST_var root = nodes[0];
   MParser_AST_var tail;
   if (!MParser_AST::is_nil(root))

    {
       root->setFirstChild(MParser_AST::_nil());	// don't leave any old pointers set
    }
   // link in children;
   for (size_t i = 1; i < nodes.size(); i++)

    {
       if (MParser_AST::is_nil(nodes[i])) continue;	// ignore null nodes
       if (MParser_AST::is_nil(root))

        {
           // Set the root and set it up for a flat list
           root = tail = nodes[i];
        }
       else if (MParser_AST::is_nil(tail))

       {
           root->setFirstChild(nodes[i]);
           tail = root->getFirstChild();
       }
      else

       {
           tail->setNextSibling(nodes[i]);
           tail = tail->getNextSibling();
       }
      // Chase tail to last sibling
      while (!MParser_AST::is_nil(tail->getNextSibling()))

       {
          tail = tail->getNextSibling();
       }
    }
   return root;
}

#if 0
/** Make a tree from a list of nodes, where the nodes are contained
 * in an ASTArray object
 */

MParser_AST_ptr MParser_ASTFactory::make(ASTArray nodes)

{
   return make(nodes.array);
}
#endif

#if 0
/** Make an AST the root of current AST */

void MParser_ASTFactory::makeASTRoot(ASTPair currentAST, AST root)

{
   if (root != null)

    {
      // Add the current root as a child of new root
      root.addChild(currentAST.root);
      // The new current child is the last sibling of the old root
      currentAST.child = currentAST.root;
      currentAST.advanceChildToEnd();
      // Set the new root
      currentAST.root = root;
    }
}
#endif


void MParser_ASTFactory::setASTNodeClass(const MLANG_String& t)

{
   theASTNodeType_ = t;
   try

    {
        theASTNodeTypeClass_ = MLANG_Class::forName(t); // get class def
    }
   catch (MLANG_Exception& e)

    {
       // either class not found,
       // class is interface/abstract, or
       // class or initializer is not accessible.
       error("Can't find/access AST Node type" + t);
    }
}

/** Specify the type of node to create during tree building.
 * 	@deprecated since 2.7.1
 */

void MParser_ASTFactory::setASTNodeType(const MLANG_String& t)

{
  setASTNodeClass(t);
}


const MParser_ASTFactory::CLASSMAP& MParser_ASTFactory::getTokenTypeToASTClassMap() const

{
   return tokenTypeToASTClassMap_;
}


void MParser_ASTFactory::setTokenTypeToASTClassMap (
   const CLASSMAP& tokenTypeToClassMap
)

{
   tokenTypeToASTClassMap_ = tokenTypeToClassMap;
}

/** To change where error messages go, can subclass/override this method
 *  and then setASTFactory in Parser and TreeParser.  This method removes
 *  a prior dependency on class antlr.Tool.
 */

void MParser_ASTFactory::error(const MLANG_String& e)

{
   SYSTEM->err()->println(e);
}

MParser_ASTFactory::~MParser_ASTFactory()

{
}
