#ifndef __MPARSER_ASTFACTORY_H__
#define __MPARSER_ASTFACTORY_H__
#include "ace/Vector_T.h"
#include "MLANG/Object.h"
#include "MLANG/SmartPtr.h"
#include "MLANG/String.h"
#include "MLANG/HashMap.h"
#include "MLANG/Class.h"
#include "MParser/Token.h"
#include "MParser/AST.h"

/* ANTLR Translator Generator
 * Project led by Terence Parr at http://www.jGuru.com
 * Software rights: http://www.antlr.org/license.html
 *
 * $Id: //depot/code/org.antlr/release/antlr-2.7.4/antlr/ASTFactory.java#1 $
 */

class MParser_ASTFactory;
typedef class MParser_ASTFactory *MParser_ASTFactory_ptr;

/** AST Support code shared by TreeParser and Parser.
 *  We use delegation to share code (and have only one
 *  bit of code to maintain) rather than subclassing
 *  or superclassing (forces AST support code to be
 *  loaded even when you don't want to do AST stuff).
 *
 *  Typically, setASTNodeType is used to specify the
 *  homogeneous type of node to create, but you can override
 *  create to make heterogeneous nodes etc...
 */
class MParser_ASTFactory : public MLANG_Object

{
public:
   typedef MLANG_HashMap<int, MLANG_Class_var> CLASSMAP;
protected:
    /** Name of AST class to create during tree construction.
     *  Null implies that the create method should create
     *  a default AST type such as CommonAST.  This is for
     *  homogeneous nodes.
     */
   MLANG_String theASTNodeType_;
protected:
   MLANG_Class_var theASTNodeTypeClass_;

   /** How to specify the classname to create for a particular
    *  token type.  Note that ANTLR allows you to say, for example,
    *
       tokens;
         PLUS<AST=PLUSNode>;
         ...
        }
    *
    *  and it tracks everything statically.  #[PLUS] will make you
    *  a PLUSNode w/o use of this table.
    *
    *  For tokens that ANTLR cannot track statically like #[i],
    *  you can use this table to map PLUS (Integer) -> PLUSNode (Class)
    *  etc... ANTLR sets the class map from the tokens;...} section
    *  via the ASTFactory(Hashtable) ctor in antlr.Parser.
    */
protected:
   CLASSMAP tokenTypeToASTClassMap_;

public:
   MParser_ASTFactory();

	/** Create factory with a specific mapping from token type
	 *  to Java AST node type.  Your subclasses of ASTFactory
	 *  can override and reuse the map stuff.
	 */
public:
   MParser_ASTFactory (const CLASSMAP& tokenTypeToClassMap);

   virtual ~MParser_ASTFactory ();

public:
   static bool is_nil (MParser_ASTFactory_ptr factory)
   { return factory == 0; }

   static MParser_ASTFactory_ptr _nil()
   { return 0; }

   static MParser_ASTFactory_ptr _duplicate(MParser_ASTFactory_ptr factory)
   { return MLANG_acquire(factory); }

public:
	/** Specify an "override" for the Java AST object created for a
	 *  specific token.  It is provided as a convenience so
	 *  you can specify node types dynamically.  ANTLR sets
	 *  the token type mapping automatically from the tokens{...}
	 *  section, but you can change that mapping with this method.
	 *  ANTLR does it's best to statically determine the node
	 *  type for generating parsers, but it cannot deal with
	 *  dynamic values like #[LT(1)].  In this case, it relies
	 *  on the mapping.  Beware differences in the tokens{...}
	 *  section and what you set via this method.  Make sure
	 *  they are the same.
	 *
	 *  Set className to null to remove the mapping.
	 *
	 *  @since 2.7.2
	 */
   void setTokenTypeASTNodeType (
     int tokenType, const MLANG_String& className
    );

   /** For a given token type, what is the AST node object type to create
    *  for it?
    *  @since 2.7.2
    */
public:
   MLANG_Class_ptr getASTNodeType(int tokenType);

#if 0
    /** Add a child to the current AST */
public:
   void addASTChild(ASTPair currentAST, MParser_AST_ptr child);
#endif

    /** Create a new empty AST node; if the user did not specify
     *  an AST node type, then create a default one: CommonAST.
     */
public:
   MParser_AST_var create();

public:
   MParser_AST_var create(int type);

public:
   MParser_AST_var create(int type, const MLANG_String& txt);

	/** Create an AST node with the token type and text passed in, but
	 *  with a specific Java object type. Typically called when you
	 *  say @[PLUS,"+",PLUSNode] in an antlr action.
	 *  @since 2.7.2
	 */
public:
   MParser_AST_var create (
      int type, 
      const MLANG_String& txt, 
      const MLANG_String& className
    );

    /** Create a new empty AST node; if the user did not specify
     *  an AST node type, then create a default one: CommonAST.
     */
public:
   MParser_AST_var create(const MParser_AST_var& tr);

public:
   MParser_AST_var create(const MParser_Token_var& tok);

	/** ANTLR generates reference to this when you reference a token
	 *  that has a specified heterogeneous AST node type.  This is
	 *  also a special case node creation routine for backward
	 *  compatibility.  Before, ANTLR generated "new T(tokenObject)"
	 *  and so I must call the appropriate constructor not T().
	 *
	 * @since 2.7.2
	 */
public:
   MParser_AST_var create (
      const MParser_Token_var& tok, 
      const MLANG_String& className
    );

   /**
    * @since 2.7.2
    */
public:
   MParser_AST_var create (
      const MLANG_String& className
    );

   /**
    * @since 2.7.2
    */
protected:
   MParser_AST_var createUsingCtor (
      const MParser_Token_var& token,
      const MLANG_String& className
    );

   /**
    * @since 2.7.2
    */
protected:
   MParser_AST_var create(const MLANG_Class_var& c);

   /** Copy a single node with same Java AST objec type.
    *  Ignore the tokenType->Class mapping since you know
    *  the type of the node, t.getClass(), and doing a dup.
    *
    *  clone() is not used because we want all AST creation
    *  to go thru the factory so creation can be
    *  tracked.  Returns null if t is null.
    */
public:
   MParser_AST_var dup(const MParser_AST_var& t);

    /** Duplicate tree including siblings of root. */
public:
   MParser_AST_var dupList(const MParser_AST_var& t);

    /**Duplicate a tree, assuming this is a root node of a tree--
     * duplicate that node and what's below; ignore siblings of root node.
     */
public:
   MParser_AST_var dupTree(const MParser_AST_var& t);

    /** Make a tree from a list of nodes.  The first element in the
     *  array is the root.  If the root is null, then the tree is
     *  a simple list not a tree.  Handles null children nodes correctly.
     *  For example, build(a, b, null, c) yields tree (a b c).  build(null,a,b)
     *  yields tree (nil a b).
     */
public:
   MParser_AST_var make(const ACE_Vector<MParser_AST_var>& nodes);

#if 0
    /** Make a tree from a list of nodes, where the nodes are contained
     * in an ASTArray object
     */
public:
   MParser_AST_ptr make(ASTArray nodes);
#endif

#if 0
    /** Make an AST the root of current AST */
public:
   void makeASTRoot(ASTPair currentAST, MParser_AST_ptr root);
#endif

public:
   void setASTNodeClass(const MLANG_String& t);

    /** Specify the type of node to create during tree building.
     * 	@deprecated since 2.7.1
     */
public:
   void setASTNodeType(const MLANG_String& t);

public:
   const CLASSMAP& getTokenTypeToASTClassMap() const;

public:
   void setTokenTypeToASTClassMap(const CLASSMAP& tokenTypeToClassMap);

    /** To change where error messages go, can subclass/override this method
     *  and then setASTFactory in Parser and TreeParser.  This method removes
     *  a prior dependency on class antlr.Tool.
     */
public:
   void error(const MLANG_String& e);
};

typedef MLANG_SmartPtr<MParser_ASTFactory> MParser_ASTFactory_var;

#endif /* __MPARSER_ASTFACTORY_H__ */
