#include "MLANG/StringBuffer.h"
#include "MParser/BaseAST.h"

/* ANTLR Translator Generator
 * Project led by Terence Parr at http://www.jGuru.com
 * Software rights: http://www.antlr.org/license.html
 *
 * $Id: //depot/code/org.antlr/release/antlr-2.7.4/antlr/BaseAST.java#1 $
 */

/**
 * A Child-Sibling Tree.
 *
 * A tree with PLUS at the root and with two children 3 and 4 is
 * structured as:
 *
 *		PLUS
 *		  |
 *		  3 -- 4
 *
 * and can be specified easily in LISP notation as
 *
 * (PLUS 3 4)
 *
 * where every '(' starts a new subtree.
 *
 * These trees are particular useful for translators because of
 * the flexibility of the children lists.  They are also very easy
 * to walk automatically, whereas trees with specific children
 * reference fields can't easily be walked automatically.
 *
 * This class contains the basic support for an AST.
 * Most people will create ASTs that are subclasses of
 * BaseAST or of CommonAST.
 */

bool MParser_BaseAST::verboseStringConversion_ = false;

ACE_Vector<MLANG_String> MParser_BaseAST::tokenNames_;

MParser_BaseAST::MParser_BaseAST ()

{
}

/**Add a node to the end of the child list for this node */
void MParser_BaseAST:: addChild (
   const MParser_AST_var& node
)

{
   if (MParser_AST::is_nil(node))
      return;
   MParser_BaseAST_var t = this->down();
   if (!MParser_BaseAST::is_nil(t))
    {
      while (!MParser_BaseAST::is_nil(t->right()))
       {
         t = t->right();
       }
      t->right (MParser_BaseAST::_narrow(node));
    }
   else
    {
      this->down (MParser_BaseAST::_narrow (node));
    }
}

/** How many children does this node have? */
int MParser_BaseAST::getNumberOfChildren()

{
   MParser_BaseAST_var t = this->down();
   int n = 0;
   if (!MParser_BaseAST::is_nil(t))
    {
      n = 1;
      while (!MParser_BaseAST::is_nil(t->right()))
       {
         t = t->right();
         n++;
       }
      return n;
    }
   return n;
}

void MParser_BaseAST::doWorkForFindAll (
   ASTEnumeration& v,
   const MParser_AST_var& target,
   bool partialMatch
)

{
   MParser_AST_var sibling;

   // Start walking sibling lists, looking for matches.
   for (sibling = MParser_BaseAST::_duplicate(this);
        !MParser_AST::is_nil(sibling);
        sibling = sibling->getNextSibling())
    {
        if ((partialMatch && sibling->equalsTreePartial(target)) ||
           (!partialMatch && sibling->equalsTree(target)))
	 {
            v.push_back(sibling);
         }
        // regardless of match or not, check any children for matches
        if (!MParser_AST::is_nil(sibling->getFirstChild()))
	 {
	   MParser_BaseAST_var child = 
	       MParser_BaseAST::_narrow (sibling->getFirstChild());
           child->doWorkForFindAll (v, target, partialMatch);
        }
    }
}

/** Is node t equal to this in terms of token type and text? */
bool MParser_BaseAST::equals (
   const MParser_AST_var& t
)

{
   if (MParser_AST::is_nil(t))
      return false;
   return this->getText().equals(t->getText()) &&
          this->getType() == t->getType();
}

/** Is t an exact structural and equals() match of this tree.  The
 *  'this' reference is considered the start of a sibling list.
 */
bool MParser_BaseAST::equalsList (
   const MParser_AST_var& b
)

{
   MParser_AST_var t = b;
   MParser_AST_var sibling;

   // the empty tree is not a match of any non-null tree.
   if (MParser_AST::is_nil(t))
    {
      return false;
    }

   // Otherwise, start walking sibling lists.  First mismatch, return false.
   for (sibling = this;
        !MParser_AST::is_nil(sibling) && !MParser_AST::is_nil(t);
        sibling = sibling->getNextSibling(), t = t->getNextSibling())
    {
       // as a quick optimization, check roots first.
       if (!sibling->equals(t))
	{
          return false;
        }
       // if roots match, do full list match test on children.
       if (!MParser_AST::is_nil(sibling->getFirstChild()))
	{
          if (!sibling->getFirstChild()->equalsList(t->getFirstChild()))
	   {
             return false;
           }
        }
       // sibling has no kids, make sure t doesn't either
       else if (!MParser_AST::is_nil(t->getFirstChild()))
	{
          return false;
        }
    }
   if (MParser_AST::is_nil(sibling) && MParser_AST::is_nil(t))
    {
      return true;
    }
   // one sibling list has more than the other
   return false;
}

/** Is 'sub' a subtree of this list?
 *  The siblings of the root are NOT ignored.
 */
bool MParser_BaseAST::equalsListPartial (
   const MParser_AST_var& b
)

{
   MParser_AST_var sub = b;
   MParser_AST_var sibling;

   // the empty tree is always a subset of any tree.
   if (MParser_AST::is_nil(sub))
    {
      return true;
    }

   // Otherwise, start walking sibling lists.  First mismatch, return false.
   for (sibling = MParser_AST::_duplicate(this);
        !MParser_AST::is_nil(sibling) && !MParser_AST::is_nil(sub);
        sibling = sibling->getNextSibling(), sub = sub->getNextSibling())
    {
       // as a quick optimization, check roots first.
       if (!sibling->equals(sub)) return false;
       // if roots match, do partial list match test on children.
       if (!MParser_AST::is_nil(sibling->getFirstChild()))
	{
          if (!sibling->getFirstChild()->equalsListPartial(sub->getFirstChild()))
	     return false;
        }
    }
   if (MParser_AST::is_nil(sibling) && !MParser_AST::is_nil(sub))
    {
      // nothing left to match in this tree, but subtree has more
      return false;
    }
   // either both are null or sibling has more, but subtree doesn't
   return true;
}

/** Is tree rooted at 'this' equal to 't'?  The siblings
 *  of 'this' are ignored.
 */
bool MParser_BaseAST::equalsTree (
   const MParser_AST_var& t
)

{
   // check roots first.
   if (!this->equals(t))
      return false;

   // if roots match, do full list match test on children.
   if (!MParser_AST::is_nil(this->getFirstChild()))
    {
      if (!this->getFirstChild()->equalsList(t->getFirstChild()))
	 return false;
    }
   // sibling has no kids, make sure t doesn't either
   else if (!MParser_AST::is_nil(t->getFirstChild()))
    {
      return false;
    }
   return true;
}

/** Is 't' a subtree of the tree rooted at 'this'?  The siblings
 *  of 'this' are ignored.
 */
bool MParser_BaseAST::equalsTreePartial (
   const MParser_AST_var& sub
)

{
   // the empty tree is always a subset of any tree.
   if (MParser_AST::is_nil(sub))
    {
      return true;
    }

   // check roots first.
   if (!this->equals(sub))
      return false;

   // if roots match, do full list partial match test on children.
   if (!MParser_AST::is_nil(this->getFirstChild()))
    {
      if (!this->getFirstChild()->equalsListPartial(sub->getFirstChild())) 
	 return false;
    }
   return true;
}

/** Walk the tree looking for all exact subtree matches.  Return
 *  an ASTEnumerator that lets the caller walk the list
 *  of subtree roots found herein.
 */
ASTEnumeration MParser_BaseAST::findAll (
   const MParser_AST_var& target
)

{
   ASTEnumeration roots;
   MParser_AST_var sibling;

   // the empty tree cannot result in an enumeration
   if (MParser_AST::is_nil(target))
    {
      return roots;
    }

   doWorkForFindAll(roots, target, false);  // find all matches recursively
   return roots;
}

/** Walk the tree looking for all subtrees.  Return
 *  an ASTEnumerator that lets the caller walk the list
 *  of subtree roots found herein.
 */
ASTEnumeration MParser_BaseAST::findAllPartial (
   const MParser_AST_var& sub
)

{
   ASTEnumeration roots;
   MParser_AST_var sibling;

   // the empty tree cannot result in an enumeration
   if (MParser_AST::is_nil(sub))
    {
      return roots;
    }

   doWorkForFindAll(roots, sub, true);  // find all matches recursively
   return roots;
}

/** Get the first child of this node; null if not children */
MParser_AST_var MParser_BaseAST::getFirstChild()

{
   return MParser_AST::_duplicate(down_);
}

/** Get the next sibling in line after this one */
MParser_AST_var MParser_BaseAST::getNextSibling()

{
   return MParser_AST::_duplicate(right_);
}

/** Get the token text for this node */
MLANG_String MParser_BaseAST::getText() const

{
   return "";
}

/** Get the token type for this node */
int MParser_BaseAST::getType()

{
   return 0;
}

int MParser_BaseAST::getLine()

{
   return 0;
}

int MParser_BaseAST::getColumn()

{
   return 0;
}

/** Remove all children */
void MParser_BaseAST:: removeChildren()

{
   down_ = MParser_BaseAST::_nil();
}

void MParser_BaseAST:: setFirstChild (
   const MParser_AST_var& c
)

{
   down_ = MParser_BaseAST::_narrow(c);
}

void MParser_BaseAST::setNextSibling (
   const MParser_AST_var& n
)

{
    right_ = MParser_BaseAST::_narrow(n);
}

/** Set the token text for this node */
void MParser_BaseAST::setText (
   const MLANG_String& text
)

{
   ACE_UNUSED_ARG(text);
}

    /** Set the token type for this node */
void MParser_BaseAST:: setType(int ttype)

{
   ACE_UNUSED_ARG(ttype);
}

void MParser_BaseAST::setVerboseStringConversion (
   bool verbose,
   const ACE_Vector<MLANG_String>& names
)

{
   verboseStringConversion_ = verbose;
   tokenNames_ = names;
}

/** Return an array of strings that maps token ID to it's text. @since 2.7.3 */
const ACE_Vector<MLANG_String>& MParser_BaseAST::getTokenNames()

{
    return tokenNames_;
}

MLANG_String MParser_BaseAST::toString() const

{
    MLANG_StringBuffer b;
    // if verbose and type name not same as text (keyword probably)
    #if 0
    if (verboseStringConversion_
    &&  !getText().equalsIgnoreCase(tokenNames_[getType()])
    &&  !getText().equalsIgnoreCase(MParser_StringUtils::stripFrontBack(tokenNames[getType()], "\"", "\"")))
     {
       b.append('[');
       b.append(getText());
       b.append(",<");
       b.append(tokenNames[getType()]);
       b.append(">]");
       return b.toString();
    }
    #else
    if (verboseStringConversion_)
    assert(0);
    #endif
   return getText();
}

/** Print out a child-sibling tree in LISP notation */
MLANG_String MParser_BaseAST::toStringList() const

{
   MParser_AST_var t = MParser_BaseAST::_duplicate(ACE_const_cast(MParser_BaseAST_ptr, this));
   MLANG_String ts = "";
   if (!MParser_AST::is_nil(t->getFirstChild())) ts += " (";
   ts += " ";
   ts += this->toString();
   if (!MParser_AST::is_nil(t->getFirstChild()))
    {
      ts += MParser_BaseAST_var(MParser_BaseAST::_narrow(t->getFirstChild()))->toStringList();
    }
   if (!MParser_AST::is_nil(t->getFirstChild())) ts += " )";
   if (!MParser_AST::is_nil(t->getNextSibling()))
    {
      ts += MParser_BaseAST_var(MParser_BaseAST::_narrow(t->getNextSibling()))->toStringList();
    }
   return ts;
}

MLANG_String MParser_BaseAST::toStringTree() const

{
   MParser_AST_var t = MParser_BaseAST::_duplicate(ACE_const_cast(MParser_BaseAST_ptr, this));
   MLANG_String ts = "";
   if (!MParser_AST::is_nil(t->getFirstChild()))
    {
      ts += " (";
    }
   ts += " " + this->toString();
   if (!MParser_AST::is_nil(t->getFirstChild()))
    {
      MParser_BaseAST_var child = MParser_BaseAST::_narrow (t->getFirstChild());
      ts += child->toStringList();
    }
   if (!MParser_AST::is_nil(t->getFirstChild()))
    {
      ts += " )";
    }
   return ts;
}

MLANG_String MParser_BaseAST::decode (
   const MLANG_String& text
)

{
   char c, c1, c2, c3, c4, c5;
   MLANG_StringBuffer n;
   for (size_t i = 0; i < text.length(); i++)
    {
       c = text.charAt(i);
       if (c == '&')
	{
          c1 = text.charAt(i + 1);
          c2 = text.charAt(i + 2);
          c3 = text.charAt(i + 3);
          c4 = text.charAt(i + 4);
          c5 = text.charAt(i + 5);

          if (c1 == 'a' && c2 == 'm' && c3 == 'p' && c4 == ';')
	   {
             n.append("&");
             i += 5;
           }
          else if (c1 == 'l' && c2 == 't' && c3 == ';')
	   {
             n.append("<");
             i += 4;
           }
          else if (c1 == 'g' && c2 == 't' && c3 == ';')
	   {
             n.append(">");
             i += 4;
           }
          else if (c1 == 'q' && c2 == 'u' && c3 == 'o' &&
                   c4 == 't' && c5 == ';')
           {
             n.append("\"");
             i += 6;
           }
          else if (c1 == 'a' && c2 == 'p' && c3 == 'o' &&
                   c4 == 's' && c5 == ';')
           {
             n.append("'");
             i += 6;
           }
          else
             n.append("&");
        }
       else
          n.append(c);
    }
   return n.toString();
}

MLANG_String MParser_BaseAST::encode (
   const MLANG_String& text
)

{
   char c;
   MLANG_StringBuffer n;
   for (size_t i = 0; i < text.length(); i++)
    {
       c = text.charAt(i);
       switch (c)
	{
        case '&':
              {
                 n.append("&amp;");
                 break;
              }
        case '<':
              {
                 n.append("&lt;");
                 break;
              }
        case '>':
              {
                 n.append("&gt;");
                 break;
              }
        case '"':
              {
                 n.append("&quot;");
                 break;
              }
        case '\'':
              {
                 n.append("&apos;");
                 break;
              }
        default :
              {
                 n.append(c);
                 break;
              }
       }
    }
   return n.toString();
}

void MParser_BaseAST::xmlSerializeNode (
   const MIO_Writer_var& out
)

{
   MLANG_StringBuffer buf;
   buf.append("<");
   buf.append(getClass()->getName() + " ");
   buf.append("text=\"" + encode(getText()) + "\" type=\"" +
             MLANG_String::valueOf(getType()) + "\"/>");
   out->write(buf.toString());
}

void MParser_BaseAST:: xmlSerializeRootOpen (
   const MIO_Writer_var& out
)

{
   MLANG_StringBuffer buf;
   buf.append("<");
   buf.append(getClass()->getName() + " ");
   buf.append("text=\"" + encode(getText()) + "\" type=\"" +
                  MLANG_String::valueOf(getType()) + "\">\n");
   out->write(buf.toString());
}

void MParser_BaseAST::xmlSerializeRootClose (
   const MIO_Writer_var& out
)

{
    out->write("</" + getClass()->getName() + ">\n");
}

void MParser_BaseAST:: xmlSerialize (
   const MIO_Writer_var& out
)

{
   // print out this node and all siblings
   for (MParser_AST_var node = MParser_AST::_duplicate(this);
        !MParser_AST::is_nil(node);
        node = node->getNextSibling())
     {
       if (MParser_AST::is_nil(node->getFirstChild()))
	{
          // print guts (class name, attributes)
          MParser_BaseAST_var ast(MParser_BaseAST::_narrow(node));
	  ast->xmlSerializeNode(out);
        }
       else
	{
          MParser_BaseAST_var ast(MParser_BaseAST::_narrow(node));
	  ast->xmlSerializeRootOpen(out);

          // print children
          ast = MParser_BaseAST::_narrow(node->getFirstChild());
          ast->xmlSerialize(out);

          // print end tag
          ast = MParser_BaseAST::_narrow(node);
          ast->xmlSerializeRootClose(out);
        }
    }
}

MParser_BaseAST::~MParser_BaseAST ()

{
}

