#ifndef __MPARSER_BASEAST_H__
#define __MPARSER_BASEAST_H__
#include "ace/Vector_T.h"
#include "MLANG/SmartPtr.h"
#include "MIO/Writer.h"
#include "MParser/AST.h"

/* ANTLR Translator Generator
 * Project led by Terence Parr at http://www.jGuru.com
 * Software rights: http://www.antlr.org/license.html
 *
 * $Id: //depot/code/org.antlr/release/antlr-2.7.4/antlr/BaseAST.java#1 $
 */
class MParser_BaseAST;
typedef MParser_BaseAST *MParser_BaseAST_ptr;
typedef MLANG_SmartPtr<MParser_BaseAST> MParser_BaseAST_var;

/**
 * A Child-Sibling Tree.
 *
 * A tree with PLUS at the root and with two children 3 and 4 is
 * structured as:
 *
 *		PLUS
 *		  |
 *		  3 -- 4
 *
 * and can be specified easily in LISP notation as
 *
 * (PLUS 3 4)
 *
 * where every '(' starts a new subtree.
 *
 * These trees are particular useful for translators because of
 * the flexibility of the children lists.  They are also very easy
 * to walk automatically, whereas trees with specific children
 * reference fields can't easily be walked automatically.
 *
 * This class contains the basic support for an AST.
 * Most people will create ASTs that are subclasses of
 * BaseAST or of CommonAST.
 */
class MParser_BaseAST : public MParser_AST
{
protected:
   MParser_BaseAST_var down_;

   const MParser_BaseAST_var& down() const
   { return down_; }

   MParser_BaseAST& down (const MParser_BaseAST_var& node)
   { down_ = node; return *this; }

protected:
   MParser_BaseAST_var right_;

   const MParser_BaseAST_var& right() const
   { return right_; }

   MParser_BaseAST& right (const MParser_BaseAST_var& node)
   { right_ = node; return *this; }

private:
   static bool verboseStringConversion_;
private:
   static ACE_Vector<MLANG_String> tokenNames_;

public:
   MParser_BaseAST();

   virtual ~MParser_BaseAST();

public:
   static bool is_nil (MParser_BaseAST_ptr ast)
   { return ast == 0; }

   static MParser_BaseAST_ptr _nil()
   { return 0; }

   static MParser_BaseAST_ptr _duplicate(MParser_BaseAST_ptr ast)
   { return MLANG_acquire(ast); }

   static MParser_BaseAST_ptr _narrow(MParser_AST_ptr ast)
   { return MLANG_acquire(ACE_dynamic_cast(MParser_BaseAST_ptr, ast)); }

public:
    /**Add a node to the end of the child list for this node */
   void addChild(const MParser_AST_var& node);

public:
   /** How many children does this node have? */
   int getNumberOfChildren();

private:
   void doWorkForFindAll (
      ASTEnumeration& v,
      const MParser_AST_var& target,
      bool partialMatch
    );

    /** Is node t equal to this in terms of token type and text? */
public:
   bool equals(const MParser_AST_var& t);

    /** Is t an exact structural and equals() match of this tree.  The
     *  'this' reference is considered the start of a sibling list.
     */
public:
   bool equalsList(const MParser_AST_var& t);

    /** Is 'sub' a subtree of this list?
     *  The siblings of the root are NOT ignored.
     */
public:
   bool equalsListPartial(const MParser_AST_var& sub);

    /** Is tree rooted at 'this' equal to 't'?  The siblings
     *  of 'this' are ignored.
     */
public:
   bool equalsTree(const MParser_AST_var& t);

    /** Is 't' a subtree of the tree rooted at 'this'?  The siblings
     *  of 'this' are ignored.
     */
public:
   bool equalsTreePartial(const MParser_AST_var& sub);

    /** Walk the tree looking for all exact subtree matches.  Return
     *  an ASTEnumerator that lets the caller walk the list
     *  of subtree roots found herein.
     */
public:
   ASTEnumeration findAll(const MParser_AST_var& target);

    /** Walk the tree looking for all subtrees.  Return
     *  an ASTEnumerator that lets the caller walk the list
     *  of subtree roots found herein.
     */
public:
   ASTEnumeration findAllPartial(const MParser_AST_var& sub);

    /** Get the first child of this node; null if not children */
public:
   MParser_AST_var getFirstChild();

    /** Get the next sibling in line after this one */
public:
   MParser_AST_var getNextSibling();

    /** Get the token text for this node */
public:
   MLANG_String getText() const;

    /** Get the token type for this node */
public:
   int getType();

public:
   int getLine();

public:
   int getColumn();

public:
   virtual void initialize(int t, const MLANG_String& txt) = 0;

public:
   virtual void initialize(const MParser_AST& t) = 0;

public:
   virtual void initialize(const MParser_Token& t) = 0;

    /** Remove all children */
public:
   void removeChildren();

public:
   void setFirstChild(const MParser_AST_var& c);

public:
   void setNextSibling(const MParser_AST_var& n);

    /** Set the token text for this node */
public:
   void setText(const MLANG_String& text);

    /** Set the token type for this node */
public:
   void setType(int ttype);

public:
   static void setVerboseStringConversion (
      bool verbose,
      const ACE_Vector<MLANG_String>& names
    );

    /** Return an array of strings that maps token ID to it's text. @since 2.7.3 */
public:
   static const ACE_Vector<MLANG_String>& getTokenNames();

public:
   MLANG_String toString() const;

    /** Print out a child-sibling tree in LISP notation */
public:
   MLANG_String toStringList() const;

public:
   MLANG_String toStringTree() const;

public:
   static MLANG_String decode(const MLANG_String& text);

public:
   static MLANG_String encode(const MLANG_String& text);

public:
   void xmlSerializeNode(const MIO_Writer_var& out);

public:
   void xmlSerializeRootOpen(const MIO_Writer_var& out);

public:
   void xmlSerializeRootClose(const MIO_Writer_var& out);

public:
   void xmlSerialize(const MIO_Writer_var& out);
};

#endif /* __MPARSER_BASEAST_H__ */
