#include "MParser/BitSet.h"

/* ANTLR Translator Generator
 * Project led by Terence Parr at http://www.jGuru.com
 * Software rights: http://www.antlr.org/license.html
 *
 * $Id: //depot/code/org.antlr/release/antlr-2.7.4/antlr/collections/impl/BitSet.java#1 $
 */

/**A BitSet to replace java.util.BitSet.
 * Primary differences are that most set operators return new sets
 * as opposed to oring and anding "in place".  Further, a number of
 * operations were added.  I cannot contain a BitSet because there
 * is no way to access the internal bits (which I need for speed)
 * and, because it is final, I cannot subclass to add functionality.
 * Consider defining set degree.  Without access to the bits, I must
 * call a method n times to test the ith bit...ack!
 *
 * Also seems like or() from util is wrong when size of incoming set is bigger
 * than this.bits.length.
 *
 * @author Terence Parr
 * @author <br><a href="mailto:pete@yamuna.demon.co.uk">Pete Wells</a>
 */

/// Assignment operator
///
MParser_BitSet& MParser_BitSet::operator=(const MParser_BitSet& b)

{
  if (this != &b)
   {
     bits_ = b.bits_;
   }
  return *this;
}

    /** or this element into this set (grow as necessary to accommodate) */
MParser_BitSet& MParser_BitSet::add(int el)

{
   size_t n = wordNumber(el);
   if (n >= bits_.size())
    {
      growToInclude(el);
    }
   bits_[n] |= bitMask(el);
   return *this;
}

MParser_BitSet operator& (
   const MParser_BitSet& a,
   const MParser_BitSet& b
) 

{
   MParser_BitSet s (a);
   s &= b;
   return s;
}

#ifndef MIN
#define MIN(x,y) ((x) < (y) ? (x) : (y))
#endif /* MIN */

MParser_BitSet& MParser_BitSet::operator&= (const MParser_BitSet& a)

{
   size_t n = MIN(bits_.size(), a.bits_.size());
   size_t i;
   for (i = 0; i <n; i++)
    {
       bits_[i] &= a.bits_[i];
    }
   // clear all bits in this not present in a (if this bigger than a).
   for (i = n; i < bits_.size(); i++)
    {
      bits_[i] = 0;
    }
   return *this;
}

ACE_UINT64 MParser_BitSet::bitMask(size_t bitNumber)

{
   ACE_UINT64 bitPosition = ((ACE_UINT64)bitNumber) & MOD_MASK; // bitNumber mod BITS
   return ((ACE_UINT64)1L) << bitPosition;
}

void MParser_BitSet::clear()

{
   for (size_t i = 0; i < bits_.size(); i++)
    {
      bits_[i] = 0;
    }
}

void MParser_BitSet::clear(int el)

{
    size_t n = wordNumber(el);
    if (n >= bits_.size())
     {	// grow as necessary to accommodate
       growToInclude(el);
     }
    bits_[n] &= ~bitMask(el);
}

size_t MParser_BitSet::degree() const

{
    size_t deg = 0;
    for (size_t i = 0; i< bits_.size(); i++)
     {
        ACE_UINT64 word = bits_[i];
        while (word != 0L)
         {
           deg++;
           word &= (word - 1);
         }
      }
     return deg;
}

/** code "inherited" from java.util.BitSet */
bool MParser_BitSet::equals(const MParser_BitSet& set) const

{
   size_t n = MIN(bits_.size(), set.bits_.size());
   for (size_t i = 0; i < n; i++)
    {
       if (bits_[i] != set.bits_[i])
        {
          return false;
        }
    }
   if (bits_.size() > n)
    {
      for (size_t i = n; i < bits_.size(); i++)
       {
          if (bits_[i] != 0)
           {
             return false;
           }
       }
    }
   else if (set.bits_.size() > n)
    {
      for (size_t i = n; i < set.bits_.size(); i++)
       {
          if (set.bits_[i] != 0)
           {
             return false;
           }
       }
    }
   return true;
}

#if 0
/** Find ranges in a set element array.  @param elems The array of
 * elements representing the set, usually from Bit Set.toArray().
 * @return Vector of ranges.
 */
static MLANG_Vector<int> MParser_BitSet::getRanges (const MLANG_Vector<int> &elems)

{
   MLANG_Vector<int> ranges;
   if (elems.size() == 0)
    {
      return ranges;
    }
   int begin = elems[0];
   int end = elems[elems.size() - 1];
   if (elems.size() <= 2)
    {
      // Not enough elements for a range expression
      return ranges;
    }

   ranges.size(5);
   // look for ranges
        for (int i = 0; i < elems.size() - 2; i++) {
            int lastInRange;
            lastInRange = elems.size() - 1;
            for (int j = i + 1; j < elems.size(); j++) {
                if (elems[j] != elems[j - 1] + 1) {
                    lastInRange = j - 1;
                    break;
                }
            }
            // found a range
            if (lastInRange - i > 2) {
                ranges.appendElement(new IntRange(elems[i], elems[lastInRange]));
            }
        }
        return ranges;
    }
#endif

#ifndef MAX
#define MAX(x,y) ((x) > (y) ? (x) : (y))
#endif /* MAX */

/**
 * Grows the set to a larger number of bits.
 * @param bit element that must fit in set
 */
void MParser_BitSet::growToInclude(int bit)

{
    size_t newSize = MAX(bits_.size() << 1, numWordsToHold(bit));
    bits_.resize(newSize, 0);
}

bool MParser_BitSet::member(int el) const

{
   size_t n = wordNumber(el);
   if (n >= bits_.size()) 
      return false;
   return (bits_[n] & bitMask(el)) != 0;
}

bool MParser_BitSet::nil() const

{
   for (size_t i = 0; i<bits_.size(); i++)
    {
       if (bits_[i] != 0) 
          return false;
    }
   return true;
}

#if 0
MParser_BitSet not() {
        BitSet s = (BitSet)this.clone();
        s.notInPlace();
        return s;
    }
#endif

MParser_BitSet& MParser_BitSet::notInPlace()

{
   for (size_t i = 0; i<bits_.size(); i++)
    {
       bits_[i] = ~bits_[i];
    }
   return *this;
}

/** complement bits in the range 0..maxBit. */
MParser_BitSet& MParser_BitSet::notInPlace(size_t maxBit)

{
   return notInPlace(0, maxBit);
}

/** complement bits in the range minBit..maxBit.*/
MParser_BitSet& MParser_BitSet::notInPlace(size_t minBit, size_t maxBit)

{
   // make sure that we have room for maxBit
   growToInclude(maxBit);
   for (size_t i = minBit; i <= maxBit; i++)
    {
       size_t n = wordNumber(i);
       bits_[n] ^= bitMask(i);
    }
   return *this;
}

size_t MParser_BitSet::numWordsToHold(int el)

{
   return (el >> LOG_BITS) + 1;
}

MParser_BitSet MParser_BitSet::of(int el)

{
   MParser_BitSet s (el + 1);
   s.add(el);
   return s;
}

/** return this | a in a new set */
MParser_BitSet operator| (
   const MParser_BitSet& a,
   const MParser_BitSet& b
)

{
    MParser_BitSet s (b);
    s |= a;
    return s;
}

MParser_BitSet& MParser_BitSet::operator|=(const MParser_BitSet& a)

{
    // If this is smaller than a, grow this first
    if (a.bits_.size() > bits_.size())
     {
       setSize(a.bits_.size());
     }
    size_t min = MIN(bits_.size(), a.bits_.size());
    assert (min==a.bits_.size() && a.bits_.size() <= bits_.size());
    for (size_t i = 0; i < min; i++)
     {
        bits_[i] |= a.bits_[i];
     }
    return *this;
}

// remove this element from this set
MParser_BitSet& MParser_BitSet::remove (int el)

{
   size_t n = wordNumber(el);
   if (n >= bits_.size())
    {
      growToInclude(el);
    }
   bits_[n] &= ~bitMask(el);
   return *this;
}

/**Is this contained within a? */
bool MParser_BitSet::subset(const MParser_BitSet& a)

{
   return this->equals(a & *this);
}

/**Subtract the elements of 'a' from 'this' in-place.
 * Basically, just turn off all bits of 'this' that are in 'a'.
 */
MParser_BitSet& MParser_BitSet::subtractInPlace(const MParser_BitSet& a)

{
   // for all words of 'a', turn off corresponding bits of 'this'
   for (size_t i = 0; i < bits_.size() && i < a.bits_.size(); i++)
    {
       bits_[i] &= ~a.bits_[i];
    }
   return *this;
}

MLANG_Vector<size_t> MParser_BitSet::toArray() const

{
   MLANG_Vector<size_t> elems;
   elems.resize (degree(), 0);
   size_t en = 0;
   for (size_t i = 0; i < (bits_.size() << LOG_BITS); i++)
    {
       if (member(i))
        {
          elems[en++] = i;
        }
    }
   assert(en==degree());
   return elems;
}

/** Transform a bit set into a string by formatting each element as an integer
 * @separator The string to put in between elements
 * @return A commma-separated list of values
 */
MLANG_String MParser_BitSet::toString (const MLANG_String& separator) const

{
   MLANG_String str = "";
   for (size_t i = 0; i < (bits_.size() << LOG_BITS); i++)
    {
       if (member(i))
        {
          if (str.length() > 0)
           {
             str += separator;
           }
          str += MLANG_String::valueOf(i);
        }
    }
   return str;
}

/** Transform a bit set into a string of characters.
 * @separator The string to put in between elements
 * @param formatter An object implementing the CharFormatter interface.
 * @return A commma-separated list of character constants.
 */
MLANG_String MParser_BitSet::toString (
   const MLANG_String& separator,
   const MParser_CharFormatter& formatter
) const

{
   MLANG_String str = "";

   for (size_t i = 0; i < (bits_.size() << LOG_BITS); i++)
    {
       if (member(i))
        {
          if (str.length() > 0)
           {
             str += separator;
           }
          str += formatter.literalChar(i);
        }
    }
   return str;
}

/**Create a string representation where instead of integer elements, the
 * ith element of vocabulary is displayed instead.  Vocabulary is a Vector
 * of Strings.
 * @separator The string to put in between elements
 * @return A commma-separated list of character constants.
 */
MLANG_String MParser_BitSet::toString (
   const MLANG_String& separator,
   const VOCABULARY& vocabulary
) const

{
   if (vocabulary.size() == 0)
    {
      return toString(separator);
    }
   MLANG_String str = "";
   for (size_t i = 0; i < (bits_.size() << LOG_BITS); i++)
    {
       if (member(i))
        {
          if (str.length() > 0)
           {
             str += separator;
           }
          if (i >= vocabulary.size())
           {
              str += "<bad element ";
              str += MLANG_String::valueOf(i);
              str += ">";
           }
          else 
           {
              str += vocabulary[i];
           }
        }
    }
   return str;
}

/**
 * Dump a comma-separated list of the words making up the bit set.
 * Split each 64 bit number into two more manageable 32 bit numbers.
 * This generates a comma-separated list of C++-like unsigned long constants.
 */
MLANG_String MParser_BitSet::toStringOfHalfWords() const

{
   MLANG_String s;
   for (size_t i = 0; i < bits_.size(); i++)
    {
       if (i != 0) s += ", ";
       ACE_UINT64 tmp = bits_[i];
       tmp &= 0xFFFFFFFFL;
       s += MLANG_String::valueOf(tmp);
       s += "UL";
       s += ", ";
       tmp = bits_[i] >> 32;
       tmp &= 0xFFFFFFFFL;
       s +=MLANG_String::valueOf(tmp);
       s += "UL";
    }
   return s;
}

/**
 * Dump a comma-separated list of the words making up the bit set.
 * This generates a comma-separated list of Java-like long int constants.
 */
MLANG_String MParser_BitSet::toStringOfWords() const

{
   MLANG_String s;
   for (size_t i = 0; i < bits_.size(); i++)
    {
      if (i != 0) s += ", ";
      s +=MLANG_String::valueOf(bits_[i]);
      s += "UL";
    }
   return s;
}

    /** Print out the bit set but collapse char ranges. */
MLANG_String MParser_BitSet::toStringWithRanges (
   const MLANG_String& separator,
   const MParser_CharFormatter& formatter
) const

{
   MLANG_String str;
   MLANG_Vector<size_t> elems = toArray();
   if (elems.size() == 0)
    {
      return "";
    }
   // look for ranges
   size_t i = 0;
   while (i < elems.size())
    {
      int lastInRange;
      lastInRange = 0;
      for (size_t j = i + 1; j < elems.size(); j++)
       {
          if (elems[j] != elems[j - 1] + 1)
           {
             break;
           }
          lastInRange = j;
       }
      // found a range
      if (str.length() > 0)
       {
         str += separator;
       }
      if (lastInRange - i >= 2)
       {
         str += formatter.literalChar(elems[i]);
         str += "..";
         str += formatter.literalChar(elems[lastInRange]);
         i = lastInRange;	// skip past end of range for next range
       }
      else 
       {	// no range, just print current char and move on
         str += formatter.literalChar(elems[i]);
       }
      i++;
    }
   return str;
}

size_t MParser_BitSet::wordNumber(int bit)

{
   return bit >> LOG_BITS; // bit / BITS
}
