#ifndef __MPARSER_BITSET_H__
#define __MPARSER_BITSET_H__
#include <map>
#include "ace/SString.h"
#include "ace/Basic_Types.h"
#include "MLANG/Vector_T.h"
#include "MLANG/String.h"
#include "MParser/CharFormatter.h"

/* ANTLR Translator Generator
 * Project led by Terence Parr at http://www.jGuru.com
 * Software rights: http://www.antlr.org/license.html
 *
 * $Id: //depot/code/org.antlr/release/antlr-2.7.4/antlr/collections/impl/BitSet.java#1 $
 */

typedef MLANG_Vector<ACE_UINT64> longArray;

/**A BitSet to replace java.util.BitSet.
 * Primary differences are that most set operators return new sets
 * as opposed to oring and anding "in place".  Further, a number of
 * operations were added.  I cannot contain a BitSet because there
 * is no way to access the internal bits (which I need for speed)
 * and, because it is final, I cannot subclass to add functionality.
 * Consider defining set degree.  Without access to the bits, I must
 * call a method n times to test the ith bit...ack!
 *
 * Also seems like or() from util is wrong when size of incoming set is bigger
 * than this.bits.length.
 *
 * @author Terence Parr
 * @author <br><a href="mailto:pete@yamuna.demon.co.uk">Pete Wells</a>
 */
class MParser_BitSet

{
public:
   typedef MLANG_Vector<MLANG_String> VOCABULARY;
public:
    enum
    {
       BITS = 64 ,   // number of bits / long
       NIBBLE = 4,
       LOG_BITS = 6,  // 2^6 == 64
       MOD_MASK = BITS - 1
    };

    /* We will often need to do a mod operator (i mod nbits).  Its
     * turns out that, for powers of two, this mod operation is
     * same as (i & (nbits-1)).  Since mod is slow, we use a
     * precomputed mod mask to do the mod instead.
     */

/** The actual data bits */
protected:

   MLANG_Vector<ACE_UINT64> bits_;

public:
    // = Creacion y destruccion

    /** Construct a bitset given the size
     * @param nbits The size of the bitset in bits
     */
    MParser_BitSet(size_t nbits = BITS)
      : bits_(((nbits - 1) >> LOG_BITS) + 1) {}

    /** Construction from a static array of longs */
    MParser_BitSet(const MLANG_Vector<ACE_UINT64> &bits)
      : bits_(bits) {}

    /// Copy constructor
    ///
    MParser_BitSet(const MParser_BitSet& b)
     : bits_(b.bits_) {}

    /// Destruccion
    ///
    ~MParser_BitSet() {}

public:

    /** or this element into this set (grow as necessary to accommodate) */
    MParser_BitSet& add(int el);

    MParser_BitSet& operator=(const MParser_BitSet& b);

    friend MParser_BitSet operator & (
           const MParser_BitSet& a, 
           const MParser_BitSet& b
      );

    MParser_BitSet& operator &= (const MParser_BitSet& a);

    static ACE_UINT64 bitMask(size_t bitNumber);

    void clear();

    void clear(int el);

    size_t degree() const;

    /** code "inherited" from java.util.BitSet */
    bool equals(const MParser_BitSet& obj) const;

    /** Find ranges in a set element array.  @param elems The array of
     * elements representing the set, usually from Bit Set.toArray().
     * @return Vector of ranges.
     */
    static MLANG_Vector<int> getRanges(const MLANG_Vector<int>& elems);

    /**
     * Grows the set to a larger number of bits.
     * @param bit element that must fit in set
     */
    void growToInclude(int bit);

    bool member(int el) const;

    bool nil() const;

//    friend MParser_BitSet operator!();

    MParser_BitSet& notInPlace();

    /** complement bits in the range 0..maxBit. */
    MParser_BitSet& notInPlace(size_t maxBit);

    /** complement bits in the range minBit..maxBit.*/
    MParser_BitSet& notInPlace(size_t minBit, size_t maxBit);

    static size_t numWordsToHold(int el);

    static MParser_BitSet of(int el);

    /** return this | a in a new set */
    friend MParser_BitSet operator| (
        const MParser_BitSet& a,
        const MParser_BitSet& b
      );

    MParser_BitSet& operator|=(const MParser_BitSet& a);

    // remove this element from this set
    MParser_BitSet& remove(int el);

    /**
     * Sets the size of a set.
     * @param nwords how many words the new set should be
     */
    MParser_BitSet& setSize(int nwords)
    { bits_.resize (nwords, 0); return *this; }


    size_t size() const
    { return bits_.size() << LOG_BITS; // num words * bits per word
    }

    /** return how much space is being used by the bits array not
     *  how many actually have member bits on.
     */
    size_t lengthInLongWords() const
    { return bits_.size(); }

    /**Is this contained within a? */
    bool subset(const MParser_BitSet& a);

    /**Subtract the elements of 'a' from 'this' in-place.
     * Basically, just turn off all bits of 'this' that are in 'a'.
     */
    MParser_BitSet& subtractInPlace(const MParser_BitSet& a);

    MLANG_Vector<size_t> toArray() const;

    MLANG_Vector<ACE_UINT64> toPackedArray() const
    { return bits_; }

    /** Transform a bit set into a string by formatting each element as an integer
     * @separator The string to put in between elements
     * @return A commma-separated list of values
     */
    MLANG_String toString(const MLANG_String& separator = ",") const;

    /** Transform a bit set into a string of characters.
     * @separator The string to put in between elements
     * @param formatter An object implementing the CharFormatter interface.
     * @return A commma-separated list of character constants.
     */
    MLANG_String toString (
        const MLANG_String& separator, 
        const MParser_CharFormatter& formatter
      ) const;

    /**Create a string representation where instead of integer elements, the
     * ith element of vocabulary is displayed instead.  Vocabulary is a Vector
     * of Strings.
     * @separator The string to put in between elements
     * @return A commma-separated list of character constants.
     */
    MLANG_String toString (
      const MLANG_String& separator,
      const VOCABULARY& vocabulary
     ) const;

    /**
     * Dump a comma-separated list of the words making up the bit set.
     * Split each 64 bit number into two more manageable 32 bit numbers.
     * This generates a comma-separated list of C++-like unsigned long constants.
     */
    MLANG_String toStringOfHalfWords() const;

    /**
     * Dump a comma-separated list of the words making up the bit set.
     * This generates a comma-separated list of Java-like long int constants.
     */
    MLANG_String toStringOfWords() const;

    /** Print out the bit set but collapse char ranges. */
    MLANG_String toStringWithRanges (
       const MLANG_String& separator,
       const MParser_CharFormatter& formatter
      ) const;

private:
    static size_t wordNumber(int bit);
};

#endif /* __MPARSER_BITSET_H__ */
