#include "MLANG/Class.h"
#include "MParser/System.h"
#include "MParser/MismatchedCharException.h"
#include "MParser/CharScanner.h"

using namespace MParser;

/* ANTLR Translator Generator
 * Project led by Terence Parr at http://www.jGuru.com
 * Software rights: http://www.antlr.org/license.html
 *
 * $Id: //depot/code/org.antlr/release/antlr-2.7.4/antlr/CharScanner.java#1 $
 */


CharScanner::CharScanner()
   : saveConsumedInput_(true),
     caseSensitive_(true),
     caseSensitiveLiterals_(true),
     tabsize_(8),
     commitToPath_(false),
     traceDepth_(0)

{
//   hashString = new ANTLRHashString(this);
   setTokenObjectClass("MParser_CommonToken");
}

CharScanner::CharScanner (
   const MParser_InputBuffer_var& cb
) : saveConsumedInput_(true),
    caseSensitive_(true),
    caseSensitiveLiterals_(true),
    tabsize_(8),
    inputState_ (new MParser_LexerSharedInputState(cb)),
    commitToPath_(false),
    traceDepth_(0)

{
   setTokenObjectClass("MParser_CommonToken");
}

CharScanner::CharScanner (
   const MParser_LexerSharedInputState_var& sharedState
) : saveConsumedInput_(true),
    caseSensitive_(true),
    caseSensitiveLiterals_(true),
    tabsize_(8),
    inputState_ (sharedState),
    commitToPath_(false),
    traceDepth_(0)

{
   setTokenObjectClass("MParser_CommonToken");
}

void CharScanner::append(char c)

{
   if (saveConsumedInput_)
    {
      text_.append(c);
    }
}

void CharScanner::append(const MLANG_String& s)

{
    if (saveConsumedInput_)
     {
       text_.append(s);
     }
}

void CharScanner::commit()

{
    inputState_->input->commit();
}

void CharScanner::consume()

{ 
   if (inputState_->guessing == 0)
    {
       char c = LA(1);
       if (caseSensitive_)
        {
          append(c);
        }
       else
        {
          // use input.LA(), not LA(), to get original case
          // CharScanner.LA() would toLower it.
          append(inputState_->input->LA(1));
        }
       if (c == '\t')
        {
           tab();
        }
       else
        {
           inputState_->column++;
        }
   }
  inputState_->input->consume();
}

/** Consume chars until one matches the given char */
void CharScanner::consumeUntil(int c)

{
   while (LA(1) != EOF_CHAR && LA(1) != c)
    {
      consume();
    }
}

/** Consume chars until one matches the given set */
void CharScanner::consumeUntil(const MParser_BitSet& set)

{
   while (LA(1) != EOF_CHAR && !set.member(LA(1)))
    {
      consume();
    }
}

MParser_Token_var CharScanner::makeToken(int t)

{
   try 
    {
       MParser_Token_var tok = (MParser_Token_ptr)tokenObjectClass_->newInstance();
       tok->setType(t);
       tok->setColumn(inputState_->tokenStartColumn);
       tok->setLine(inputState_->tokenStartLine);
       // tracking real start line now: tok.setLine(inputState.line);
       return tok._retn();
    }
   catch (MLANG_InstantiationException& ie)
    {
       panic("can't instantiate token: " + tokenObjectClass_->toString());
    }
   catch (MLANG_IllegalAccessException& iae)
    {
       panic("Token class is not accessible" + tokenObjectClass_->toString());
    }
   return MParser_Token::_duplicate(MParser_Token::badToken_);
}

void CharScanner::match(char c)

{
   if (LA(1) != c)
    {
      throw MParser_MismatchedCharException(LA(1), c, false, this);
    }
   consume();
}

void CharScanner::match(const MParser_BitSet& b)

{
   if (!b.member(LA(1)))
    {
      throw MParser_MismatchedCharException(LA(1), b, false, this);
    }
   else
    {
      consume();
    }
}

void CharScanner::match(const MLANG_String& s)

{
   int len = s.length();
   for (int i = 0; i < len; i++)
    {
       if (LA(1) != s.charAt(i)) 
        {
          throw MParser_MismatchedCharException(LA(1), s.charAt(i), false, this);
        }
       consume();
    }
}

void CharScanner::matchNot(char c)

{
   if (LA(1) == c)
    {
      throw MParser_MismatchedCharException(LA(1), c, true, this);
    }
   consume();
}

void CharScanner::matchRange(char c1, char c2)

{
   if (LA(1) < c1 || LA(1) > c2) 
      throw MParser_MismatchedCharException(LA(1), c1, c2, false, this);
   consume();
}

/** advance the current column number by an appropriate amount
 *  according to tab size. This method is called from consume().
 */
void CharScanner::tab()

{
    int c = getColumn();
    int nc = ( ((c-1)/tabsize_) + 1) * tabsize_ + 1;  // calculate tab stop
    setColumn( nc );
}

/** @see #panic(String)
 */
/** This method is executed by ANTLR internally when it detected an illegal
 *  state that cannot be recovered from.
 *  The default implementation of this method calls
 *  {@link java.lang.System.exit(int)} and writes directly to
 *  {@link java.lang.System.err)} , which is usually not appropriate when
 *  a translator is embedded into a larger application. <em>It is highly
 *  recommended that this method be overridden to handle the error in a
 *  way appropriate for your application (e.g. throw an unchecked
 *  exception)</em>.
 */
void CharScanner::panic(const MLANG_String& s)

{
    SYSTEM->err()->println ("CharScanner: panic " + s);
    SYSTEM->exit(1);
}

/** Parser error-reporting function can be overridden in subclass */
void CharScanner::reportError(const MParser_RecognitionException& ex)

{
   fprintf(stderr, "%s\n", ex.getMessage().c_str());
}

/** Parser error-reporting function can be overridden in subclass */
void CharScanner::reportError(const MLANG_String& s)

{
   if (getFilename().length() == 0)
    {
      fprintf(stderr, "error: %s\n", s.c_str());
    }
   else
    {
      fprintf(stderr, "%s: error: %s\n", getFilename().c_str(), s.c_str());
    }
}

    /** Parser warning-reporting function can be overridden in subclass */
void CharScanner::reportWarning(const MLANG_String& s)

{
    if (getFilename().length() == 0) 
     {
       fprintf(stderr, "warning: %s\n", s.c_str());
     }
    else
     {
       fprintf(stderr, "%s: warning: %s\n", getFilename().c_str(), s.c_str());
     }
}

void CharScanner::resetText()

{
    text_.setLength(0);
    inputState_->tokenStartColumn = inputState_->column;
    inputState_->tokenStartLine = inputState_->line;
}

void CharScanner::rewind(int pos)

{
   inputState_->input->rewind(pos);
   // RK: should not be here, it is messing up column calculation
   // setColumn(inputState.tokenStartColumn);
}

void CharScanner::setTokenObjectClass(const MLANG_String& cl)

{
   try
    {
      tokenObjectClass_ = MLANG_Class::forName(cl);
    }
    catch (MLANG_ClassNotFoundException& ce)
    {
      panic(MLANG_String("ClassNotFoundException: ") + cl);
    }
}

// Test the token text against the literals table
// Override this method to perform a different literals test
int CharScanner::testLiteralsTable(int ttype) 

{
   ACE_CString word(text_.getBuffer(), text_.length());
   if (!caseSensitive_)
    {
      MLANG_String s (text_.getBuffer(), text_.length());
      MLANG::downcase(s);
      word = s.c_str();
    }
   MAPLIT::iterator iter = literals_.find(word);
   if (iter != literals_.end())
    {
      ttype = iter->second;
    }
   return ttype;
}

/** Test the text passed in against the literals table
 * Override this method to perform a different literals test
 * This is used primarily when you want to test a portion of
 * a token.
 */
int CharScanner::testLiteralsTable (
   const MLANG_String& text,
   int ttype
)

{
   ACE_CString word(text.c_str());
   MAPLIT::iterator iter = literals_.find(word);
   if (iter != literals_.end())
    {
      ttype = iter->second;
    }
   return ttype;
}

void CharScanner::traceIndent()

{
   for (size_t i = 0; i < traceDepth_; i++)
       fprintf(stdout, " ");
}

void CharScanner::traceIn(const MLANG_String& rname)

{
   traceDepth_ += 1;
   traceIndent();
   fprintf(stdout, "> lexer %s; c==%c\n", rname.c_str(), LA(1));
}

void CharScanner::traceOut(const MLANG_String& rname)

{
    traceIndent();
    fprintf (stdout, "< lexer %s; c==%c\n", rname.c_str(), LA(1));
    traceDepth_ -= 1;
}

CharScanner::~CharScanner()

{
}
