#ifndef __MPARSER_CHARSCANNER_H__
#define __MPARSER_CHARSCANNER_H__
#include "ace/Map.h"
#include "MLANG/SmartPtr.h"
#include "MLANG/HashMap.h"
#include "MLANG/String.h"
#include "MLANG/StringBuffer.h"
#include "MLANG/Class.h"
#include "MParser/BitSet.h"
#include "MParser/TokenStream.h"
#include "MParser/RecognitionException.h"
#include "MParser/LexerSharedInputState.h"

/* ANTLR Translator Generator
 * Project led by Terence Parr at http://www.jGuru.com
 * Software rights: http://www.antlr.org/license.html
 *
 * $Id: //depot/code/org.antlr/release/antlr-2.7.4/antlr/CharScanner.java#1 $
 */

namespace MParser
{

class CharScanner;
typedef class CharScanner *CharScanner_ptr;
typedef MLANG_SmartPtr<CharScanner> CharScanner_var;

class CharScanner : public MParser_TokenStream

{
protected:

    enum {
      NO_CHAR = 0,
      EOF_CHAR = -1
    };
    MLANG_StringBuffer text_; // text of current token

protected:

    bool saveConsumedInput_;
    MLANG_Class_var tokenObjectClass_; // what kind of tokens to create?
    bool caseSensitive_;
    bool caseSensitiveLiterals_;
    typedef MLANG_HashMap<ACE_CString, int> MAPLIT;
    MAPLIT literals_; // set by subclass

    /** Tab chars are handled by tab() according to this value; override
     *  method to do anything weird with tabs.
     */
    size_t tabsize_;

    MParser_Token_var returnToken_;

    // Hash string used so we don't new one every time to check literals table
    ACE_CString hashString_;

    MParser_LexerSharedInputState_var inputState_;

    /** Used during filter mode to indicate that path is desired.
     *  A subsequent scan error will report an error as usual if
     *  acceptPath=true;
     */
    bool commitToPath_;

    /** Used to keep track of indentdepth for traceIn/Out */
    size_t traceDepth_;

public:

    CharScanner();

    CharScanner(const MParser_InputBuffer_var& cb);

    CharScanner(const MParser_LexerSharedInputState_var& sharedState);

    virtual ~CharScanner();

public:

   static bool is_nil (CharScanner_ptr scanner)
   { return scanner == 0; }

   static CharScanner_ptr _nil()
   { return 0; }

   static CharScanner_ptr _duplicate(CharScanner_ptr scanner)
   { return MLANG_acquire(scanner); }

public:

    void append(char c);

    void append(const MLANG_String& s);

    void commit();

    void consume();

    /** Consume chars until one matches the given char */
    void consumeUntil(int c);

    /** Consume chars until one matches the given set */
    void consumeUntil(const MParser_BitSet& set);

    bool getCaseSensitive() const
    { return caseSensitive_; }

    bool getCaseSensitiveLiterals() const
    { return caseSensitiveLiterals_; }

    int getColumn() const
    { return inputState_->column; }

    void setColumn(int c) 
    { inputState_->column = c; }

    bool getCommitToPath() const
    { return commitToPath_; }

    MLANG_String getFilename() const
    { return inputState_->filename; }

    const MParser_InputBuffer_var& getInputBuffer() const
    { return inputState_->input; }

    const MParser_LexerSharedInputState_var& getInputState() const
    { return inputState_; }

    void setInputState(const MParser_LexerSharedInputState_var& state)
    { inputState_ = state; }

    int getLine() const
    { return inputState_->line; }

    /** return a copy of the current text buffer */
    MLANG_String getText() const
    { return text_.toString(); }

    const MParser_Token_var& getTokenObject() const
    { return returnToken_; }

    int LA(int i) const
    {
        if (caseSensitive_)
         {
            return inputState_->input->LA(i);
         }
        else
         {
            return toLower(inputState_->input->LA(i));
         }
    }

    MParser_Token_var makeToken(int t);

    int mark()
    { return inputState_->input->mark(); }

    void match(char c);

    void match(const MParser_BitSet& b);

    void match(const MLANG_String& s);

    void matchNot(char c);

    void matchRange(char c1, char c2);

    void newline() 
    {
        inputState_->line++;
        inputState_->column = 1;
    }

    /** advance the current column number by an appropriate amount
     *  according to tab size. This method is called from consume().
     */
    void tab();

    void setTabSize( size_t size )
    {
	 tabsize_ = size;
    }

    int getTabSize() const
    { return tabsize_; }

    /** This method is executed by ANTLR internally when it detected an illegal
     *  state that cannot be recovered from.
     *  The default implementation of this method calls
     *  {@link java.lang.System.exit(int)} and writes directly to
     *  {@link java.lang.System.err)} , which is usually not appropriate when
     *  a translator is embedded into a larger application. <em>It is highly
     *  recommended that this method be overridden to handle the error in a
     *  way appropriate for your application (e.g. throw an unchecked
     *  exception)</em>.
     */
    void panic(const MLANG_String&  s = "");

    /** Parser error-reporting function can be overridden in subclass */
    void reportError(const MParser_RecognitionException& ex);

    /** Parser error-reporting function can be overridden in subclass */
    void reportError(const MLANG_String& s);

    /** Parser warning-reporting function can be overridden in subclass */
    void reportWarning(const MLANG_String& s);

    void resetText() ;

    void rewind(int pos);

    void setCaseSensitive(bool t)
    {
        caseSensitive_ = t;
    }

    void setCommitToPath(bool commit)
    {
        commitToPath_ = commit;
    }

    void setFilename(const MLANG_String& f)
    {
        inputState_->filename = f;
    }

    void setLine(int line)
    {
        inputState_->line = line;
    }

    void setText(const MLANG_String& s)
    {
        resetText();
        text_.append(s);
    }

    void setTokenObjectClass(const MLANG_String& cl);

    // Test the token text against the literals table
    // Override this method to perform a different literals test
    int testLiteralsTable(int ttype);

    /** Test the text passed in against the literals table
     * Override this method to perform a different literals test
     * This is used primarily when you want to test a portion of
     * a token.
     */
    int testLiteralsTable (
       const MLANG_String& text, 
       int ttype
     );

    // Override this method to get more specific case handling
    virtual char toLower(char c) const
    {
        return tolower(c);
    }

    void traceIndent();

    void traceIn(const MLANG_String& rname);

    void traceOut(const MLANG_String& rname);

    /** This method is called by YourLexer.nextToken() when the lexer has
     *  hit EOF condition.  EOF is NOT a character.
     *  This method is not called if EOF is reached during
     *  syntactic predicate evaluation or during evaluation
     *  of normal lexical rules, which presumably would be
     *  an IOException.  This traps the "normal" EOF condition.
     *
     *  uponEOF() is called after the complete evaluation of
     *  the previous token and only if your parser asks
     *  for another token beyond that last non-EOF token.
     *
     *  You might want to throw token or char stream exceptions
     *  like: "Heh, premature eof" or a retry stream exception
     *  ("I found the end of this file, go back to referencing file").
     */
    virtual void uponEOF() {}
};
}

typedef MParser::CharScanner MParser_CharScanner;
typedef MParser_CharScanner *MParser_CharScanner_ptr;
typedef MLANG_SmartPtr<MParser_CharScanner> MParser_CharScanner_var;

#define LABEL(label) \
   bool _b##label = true; \
label: \
   if (_b##label)

#define BREAK(label) \
{ \
   _b##label = false; \
   goto label; \
}

#define CONTINUE(label) \
{ \
   _b##label = true; \
   goto label; \
}

#endif /* __MPARSER_CHARSCANNER_H__ */
