#include "MParser/InputBuffer.h"

/* ANTLR Translator Generator
 * Project led by Terence Parr at http://www.jGuru.com
 * Software rights: http://www.antlr.org/license.html
 *
 * $Id: //depot/code/org.antlr/release/antlr-2.7.4/antlr/InputBuffer.java#1 $
 */

using namespace MParser;

// SAS: Added this class to genericise the input buffers for scanners
//      This allows a scanner to use a binary (FileInputStream) or
//      text (FileReader) stream of data; the generated scanner
//      subclass will define the input stream
//      There are two subclasses to this: CharBuffer and ByteBuffer

/**A Stream of characters fed to the lexer from a InputStream that can
 * be rewound via mark()/rewind() methods.
 * <p>
 * A dynamic array is used to buffer up all the input characters.  Normally,
 * "k" characters are stored in the buffer.  More characters may be stored during
 * guess mode (testing syntactic predicate), or when LT(i>k) is referenced.
 * Consumption of characters is deferred.  In other words, reading the next
 * character is not done by conume(), but deferred until needed by LA or LT.
 * <p>
 *
 * @see MParser_CharQueue
 */
InputBuffer::InputBuffer()
   : _nMarkers(0),
    _markerOffset(0),
    _numToConsume(0)
{
}

/** This method updates the state of the input buffer so that
 *  the text matched since the most recent mark() is no longer
 *  held by the buffer.  So, you either do a mark/rewind for
 *  failed predicate or mark/commit to keep on parsing without
 *  rewinding the input.
 */
void InputBuffer::commit()

{
   _nMarkers--;
}

/** Mark another character for deferred consumption */
void InputBuffer::consume()

{
   _numToConsume++;
}

MLANG_String InputBuffer::getLAChars()

{
   MLANG_String la;
   la.start();
   for (int i = _markerOffset; i < _queue._nbrEntries; i++)
       la.append(_queue.elementAt(i));
   la.end();
   return la;
}

MLANG_String InputBuffer::getMarkedChars()

{
   MLANG_String marked;
   marked.start();
   for (int i = 0; i < _markerOffset; i++)
       marked.append(_queue.elementAt(i));
   marked.end();
   return marked;
}

bool InputBuffer::isMarked()

{
   return (_nMarkers != 0);
}

/** Get a lookahead character */
int InputBuffer::LA(int i) 

{
   fill(i);
   return _queue.elementAt(_markerOffset + i - 1);
}

/**Return an integer marker that can be used to rewind the buffer to
 * its current state.
 */
int InputBuffer::mark()

{
    syncConsume();
    _nMarkers++;
    return _markerOffset;
}

/**Rewind the character buffer to a marker.
 * @param mark Marker returned previously from mark()
 */
void InputBuffer::rewind(int mark)

{
    syncConsume();
    _markerOffset = mark;
    _nMarkers--;
}

/** Reset the input buffer
 */
void InputBuffer::reset()

{
    _nMarkers = 0;
    _markerOffset = 0;
    _numToConsume = 0;
    _queue.reset();
}

/** Sync up deferred consumption */
void InputBuffer::syncConsume()

{
   while (_numToConsume > 0) 
    {
      if (_nMarkers > 0)
       {
         // guess mode -- leave leading characters and bump offset.
         _markerOffset++;
       }
      else
       {
         // normal mode -- remove first character
         _queue.removeFirst();
       }
      _numToConsume--;
    }
}

InputBuffer::~InputBuffer()

{
}
