#include "MParser/System.h"
#include "MParser/MismatchedTokenException.h"
#include "MParser/Parser.h"

/**A generic ANTLR parser (LL(k) for k>=1) containing a bunch of
 * utility routines useful at any lookahead depth.  We distinguish between
 * the LL(1) and LL(k) parsers because of efficiency.  This may not be
 * necessary in the near future.
 *
 * Each parser object contains the state of the parse including a lookahead
 * cache (the form of which is determined by the subclass), whether or
 * not the parser is in guess mode, where tokens come from, etc...
 *
 * <p>
 * During <b>guess</b> mode, the current lookahead token(s) and token type(s)
 * cache must be saved because the token stream may not have been informed
 * to save the token (via <tt>mark</tt>) before the <tt>try</tt> block.
 * Guessing is started by:
 * <ol>
 * <li>saving the lookahead cache.
 * <li>marking the current position in the TokenBuffer.
 * <li>increasing the guessing level.
 * </ol>
 *
 * After guessing, the parser state is restored by:
 * <ol>
 * <li>restoring the lookahead cache.
 * <li>rewinding the TokenBuffer.
 * <li>decreasing the guessing level.
 * </ol>
 *
 * @see antlr.Token
 * @see antlr.TokenBuffer
 * @see antlr.Tokenizer
 * @see antlr.LL1Parser
 * @see antlr.LLkParser
 */


MParser_Parser::MParser_Parser() 
  : inputState_(new MParser_ParserSharedInputState()),
    ignoreInvalidDebugCalls_(false),
    traceDepth_(0)

{
}

MParser_Parser::MParser_Parser (
   const MParser_ParserSharedInputState_var& state
)
  : inputState_(state),
    ignoreInvalidDebugCalls_(false),
    traceDepth_(0)

{
}

void MParser_Parser::setTokenNames (
   const char *vec[],
   size_t n
)

{
   tokenNames_.clear();
   for (size_t i=0; i<n; i++)
     tokenNames_.push_back (vec[i]);
}

void MParser_Parser::setTokenNames (
   const MLANG_Vector<MLANG_String>& vec
)

{
   tokenNames_ = vec;
}

#if 0
void MParser_Parser::addMessageListener(MessageListener l)

{
   if (!ignoreInvalidDebugCalls)
      throw new IllegalArgumentException("addMessageListener() is only valid if parser built for debugging");
}
#endif

#if 0
void MParser_Parser::addParserListener(ParserListener l) 

{
        if (!ignoreInvalidDebugCalls)
            throw new IllegalArgumentException("addParserListener() is only valid if parser built for debugging");
}
#endif

#if 0
void MParser_Parser::addParserMatchListener(ParserMatchListener l) 

{
        if (!ignoreInvalidDebugCalls)
            throw new IllegalArgumentException("addParserMatchListener() is only valid if parser built for debugging");
}
#endif

#if 0
void MParser_Parser::addParserTokenListener(ParserTokenListener l) 

{
        if (!ignoreInvalidDebugCalls)
            throw new IllegalArgumentException("addParserTokenListener() is only valid if parser built for debugging");
}
#endif

#if 0
void MParser_Parser::addSemanticPredicateListener(SemanticPredicateListener l) 

{
        if (!ignoreInvalidDebugCalls)
            throw new IllegalArgumentException("addSemanticPredicateListener() is only valid if parser built for debugging");
}
#endif

#if 0
void MParser_Parser::addSyntacticPredicateListener(SyntacticPredicateListener l) 

{
        if (!ignoreInvalidDebugCalls)
            throw new IllegalArgumentException("addSyntacticPredicateListener() is only valid if parser built for debugging");
}
#endif

#if 0
void MParser_Parser::addTraceListener(TraceListener l) 

{
        if (!ignoreInvalidDebugCalls)
            throw new IllegalArgumentException("addTraceListener() is only valid if parser built for debugging");
}
#endif

/** Consume tokens until one matches the given token */
void MParser_Parser::consumeUntil(int tokenType)

{
   while (LA(1) != MParser_Token::EOF_TYPE && LA(1) != tokenType) 

    {
      consume();
    }
}

/** Consume tokens until one matches the given token set */
void MParser_Parser::consumeUntil (const MParser_BitSet& set)

{
   while (LA(1) != MParser_Token::EOF_TYPE && !set.member(LA(1))) 
    {
       consume();
    }
}

void MParser_Parser::defaultDebuggingSetup (
   const MParser_TokenStream_var& ,
   const MParser_TokenBuffer_var& 
)

{
   // by default, do nothing -- we're not debugging
}

/** Get the AST return value squirreled away in the parser */
const MParser_AST_var& MParser_Parser::getAST() 

{
   return returnAST_;
}

const MParser_ASTFactory_var& MParser_Parser::getASTFactory() 

{
   return astFactory_;
}

MLANG_String MParser_Parser::getFilename() const

{
   return inputState_->filename();
}

const MParser_ParserSharedInputState_var& MParser_Parser::getInputState() const

{
   return inputState_;
}

void MParser_Parser::setInputState (
   const MParser_ParserSharedInputState_var& state
)

{
   inputState_ = state;
}

MLANG_String MParser_Parser::getTokenName(int num) const

{
   return tokenNames_[num];
}

const ACE_Vector<MLANG_String>& MParser_Parser::getTokenNames() const

{
   return tokenNames_;
}

bool MParser_Parser::isDebugMode() const

{
   return false;
}

    // Forwarded to TokenBuffer
int MParser_Parser::mark() 

{
   return inputState_->input()->mark();
}

    /**Make sure current lookahead symbol matches token type <tt>t</tt>.
     * Throw an exception upon mismatch, which is catch by either the
     * error handler or by the syntactic predicate.
     */
void MParser_Parser::match(int t)

{
   if (LA(1) != t)
      throw MParser_MismatchedTokenException(tokenNames_, LT(1), t, false, getFilename());
   else
      // mark token as consumed -- fetch next token deferred until LA/LT
      consume();
}

/**Make sure current lookahead symbol matches the given set
 * Throw an exception upon mismatch, which is catch by either the
 * error handler or by the syntactic predicate.
 */
void MParser_Parser::match (const MParser_BitSet& b)

{
   if (!b.member(LA(1)))
      throw MParser_MismatchedTokenException(tokenNames_, LT(1), b, false, getFilename());
   else
      // mark token as consumed -- fetch next token deferred until LA/LT
      consume();
}

void MParser_Parser::matchNot(int t)

{
   if (LA(1) == t)
      // Throws inverted-sense exception
      throw MParser_MismatchedTokenException(tokenNames_, LT(1), t, true, getFilename());
   else
      // mark token as consumed -- fetch next token deferred until LA/LT
      consume();
}

/** @deprecated as of 2.7.2. This method calls System.exit() and writes
 *  directly to stderr, which is usually not appropriate when
 *  a parser is embedded into a larger application. Since the method is
 *  <code>static</code>, it cannot be overridden to avoid MParser_Parser::these problems.
 *  ANTLR no longer uses this method internally or in generated code.
 */
void MParser_Parser::panic() 

{
   SYSTEM->err()->println("Parser: panic");
   SYSTEM->exit(1);
}

#if 0
void MParser_Parser::removeMessageListener(MessageListener l) 

{
   if (!ignoreInvalidDebugCalls)
            throw new RuntimeException("removeMessageListener() is only valid if parser built for debugging");
}
#endif

#if 0
void MParser_Parser::removeParserListener(ParserListener l) 

{
        if (!ignoreInvalidDebugCalls)
            throw new RuntimeException("removeParserListener() is only valid if parser built for debugging");
}
#endif

#if 0
void MParser_Parser::removeParserMatchListener(ParserMatchListener l) 

{
        if (!ignoreInvalidDebugCalls)
            throw new RuntimeException("removeParserMatchListener() is only valid if parser built for debugging");
}
#endif

#if 0
void MParser_Parser::removeParserTokenListener(ParserTokenListener l) 

{
        if (!ignoreInvalidDebugCalls)
            throw new RuntimeException("removeParserTokenListener() is only valid if parser built for debugging");
}
#endif

#if 0
void MParser_Parser::removeSemanticPredicateListener(SemanticPredicateListener l) 

{
        if (!ignoreInvalidDebugCalls)
            throw new IllegalArgumentException("removeSemanticPredicateListener() is only valid if parser built for debugging");
}
#endif

#if 0
void MParser_Parser::removeSyntacticPredicateListener(SyntacticPredicateListener l) 

{
        if (!ignoreInvalidDebugCalls)
            throw new IllegalArgumentException("removeSyntacticPredicateListener() is only valid if parser built for debugging");
}
#endif

#if 0
void MParser_Parser::removeTraceListener(TraceListener l) 

{
        if (!ignoreInvalidDebugCalls)
            throw new RuntimeException("removeTraceListener() is only valid if parser built for debugging");
}
#endif

    /** Parser error-reporting function can be overridden in subclass */
void MParser_Parser::reportError(MParser_RecognitionException& ex) 

{
   SYSTEM->err()->println(ex.toString().c_str());
}

    /** Parser error-reporting function can be overridden in subclass */
void MParser_Parser::reportError(const MLANG_String& s) 

{
   if (getFilename().length() == 0)
    {
      SYSTEM->err()->println("error: " + s);
    }
   else
    {
      SYSTEM->err()->println(getFilename() + ": error: " + s);
    }
}

/** Parser warning-reporting function can be overridden in subclass */
void MParser_Parser::reportWarning (const MLANG_String& s) 

{
   if (getFilename().length() == 0)
    {
      SYSTEM->err()->println("warning: " + s);
    }
   else
    {
      SYSTEM->err()->println(getFilename() + ": warning: " + s);
    }
}

void MParser_Parser::rewind(int pos) 

{
   inputState_->input()->rewind(pos);
}

    /** Specify an object with support code (shared by
     *  Parser and TreeParser.  Normally, the programmer
     *  does not play with this, using setASTNodeType instead.
     */
void MParser_Parser::setASTFactory (const MParser_ASTFactory_var& f) 

{
   astFactory_ = f;
}

void MParser_Parser::setASTNodeClass (const MLANG_String& cl) 

{
   astFactory_->setASTNodeType(cl);
}

/** Specify the type of node to create during tree building; use setASTNodeClass now
 *  to be consistent with Token Object Type accessor.
 *  @deprecated since 2.7.1
 */
void MParser_Parser::setASTNodeType (const MLANG_String& nodeType) 

{
   setASTNodeClass(nodeType);
}

void MParser_Parser::setDebugMode (bool debugMode) 

{
   ACE_UNUSED_ARG (debugMode);
#if 0
   if (!ignoreInvalidDebugCalls_)
      throw new RuntimeException("setDebugMode() only valid if parser built for debugging");
#endif
}

void MParser_Parser::setFilename (const MLANG_String& f) 

{
   inputState_->filename (f);
}

void MParser_Parser::setIgnoreInvalidDebugCalls(bool value) 

{
   ignoreInvalidDebugCalls_ = value;
}

/** Set or change the input token buffer */
void MParser_Parser::setTokenBuffer (const MParser_TokenBuffer_var& t) 

{
   inputState_->input(t);
}

void MParser_Parser::traceIndent() 

{
   for (int i = 0; i < traceDepth_; i++)
       SYSTEM->out()->print(" ");
}

void MParser_Parser::traceIn(const MLANG_String& rname)

{
   traceDepth_ += 1;
   traceIndent();
   SYSTEM->out()->println("> " + rname + "; LA(1)==" + LT(1)->getText() +
                   ((inputState_->guessing() > 0)?" [guessing]":""));
}

void MParser_Parser::traceOut(const MLANG_String& rname)

{
   traceIndent();
   SYSTEM->out()->println("< " + rname + "; LA(1)==" + LT(1)->getText() +
                     ((inputState_->guessing() > 0)?" [guessing]":""));
   traceDepth_ -= 1;
}

MParser_Parser::~MParser_Parser() 

{
}
