#ifndef __MPARSER_PARSER_H__
#define __MPARSER_PARSER_H__
#include "MLANG/Vector_T.h"
#include "MLANG/SmartPtr.h"
#include "MLANG/String.h"
#include "MLANG/Object.h"
#include "MLANG/HashMap.h"
#include "MLANG/Class.h"
#include "MParser/Exception.h"
#include "MParser/RecognitionException.h"
#include "MParser/BitSet.h"
#include "MParser/Token.h"
#include "MParser/TokenStream.h"
#include "MParser/TokenBuffer.h"
#include "MParser/ParserSharedInputState.h"
#include "MParser/AST.h"
#include "MParser/ASTFactory.h"

/* ANTLR Translator Generator
 * Project led by Terence Parr at http://www.jGuru.com
 * Software rights: http://www.antlr.org/license.html
 *
 * $Id: //depot/code/org.antlr/release/antlr-2.7.4/antlr/Parser.java#1 $
 */

class MParser_Parser;
typedef class MParser_Parser *MParser_Parser_ptr;

/**A generic ANTLR parser (LL(k) for k>=1) containing a bunch of
 * utility routines useful at any lookahead depth.  We distinguish between
 * the LL(1) and LL(k) parsers because of efficiency.  This may not be
 * necessary in the near future.
 *
 * Each parser object contains the state of the parse including a lookahead
 * cache (the form of which is determined by the subclass), whether or
 * not the parser is in guess mode, where tokens come from, etc...
 *
 * <p>
 * During <b>guess</b> mode, the current lookahead token(s) and token type(s)
 * cache must be saved because the token stream may not have been informed
 * to save the token (via <tt>mark</tt>) before the <tt>try</tt> block.
 * Guessing is started by:
 * <ol>
 * <li>saving the lookahead cache.
 * <li>marking the current position in the TokenBuffer.
 * <li>increasing the guessing level.
 * </ol>
 *
 * After guessing, the parser state is restored by:
 * <ol>
 * <li>restoring the lookahead cache.
 * <li>rewinding the TokenBuffer.
 * <li>decreasing the guessing level.
 * </ol>
 *
 * @see antlr.Token
 * @see antlr.TokenBuffer
 * @see antlr.Tokenizer
 * @see antlr.LL1Parser
 * @see antlr.LLkParser
 */

class MParser_Parser : public MLANG_Object

{
protected:
   MParser_ParserSharedInputState_var inputState_;

    /** Nesting level of registered handlers */
    // protected:
   int exceptionLevel_;

    /** Table of token type to token names */
protected:
   MLANG_Vector<MLANG_String> tokenNames_;

    /** AST return value for a rule is squirreled away here */
protected:
   MParser_AST_var returnAST_;

    /** AST support code; parser delegates to this object.
     *  This is set during parser construction by default
     *  to either "new ASTFactory()" or a ctor that
     *  has a token type to class map for hetero nodes.
     */
protected:
   MParser_ASTFactory_var astFactory_;

   /** Constructed if any AST types specified in tokens{..}.
    *  Maps an Integer->Class object.
    */
protected:
   typedef MLANG_HashMap<int, MLANG_Class_var> CLASSMAP;
   CLASSMAP tokenTypeToASTClassMap_;

private:
   bool ignoreInvalidDebugCalls_;

    /** Used to keep track of indentdepth for traceIn/Out */
protected:
   int traceDepth_;

public:
   MParser_Parser();

   MParser_Parser (
      const MParser_ParserSharedInputState_var& state
    );

   virtual ~MParser_Parser();

public:

   static bool is_nil (MParser_Parser_ptr parser)
   { return parser == 0; }

   static MParser_Parser_ptr _nil()
   { return 0; }

   static MParser_Parser_ptr _duplicate(MParser_Parser_ptr parser)
   { return MLANG_acquire(parser); }

   /** If the user specifies a tokens{} section with heterogeneous
    *  AST node types, then ANTLR generates code to fill
    *  this mapping.
    */
public:
   void setTokenNames (const char *vec[], size_t n);

   void setTokenNames (const MLANG_Vector<MLANG_String>& vec);

   const CLASSMAP& getTokenTypeToASTClassMap() const
   { return tokenTypeToASTClassMap_; }

#if 0
public:
   void addMessageListener (
      MParser_MessageListener_ptr l
    );
#endif

#if 0
public:
   void addParserListener (
      MParser_ParserListener_ptr l
    );
#endif

#if 0
public:
   void addParserMatchListener (
      MParser_ParserMatchListener_ptr l
    );
#endif

#if 0
public:
   void addParserTokenListener (
      MParser_ParserTokenListener_ptr l
    );
#endif

#if 0
public:
   void addSemanticPredicateListener (
      MParser_SemanticPredicateListener_ptr l
    );
#endif

#if 0
public:
   void addSyntacticPredicateListener (
      MParser_SyntacticPredicateListener_ptr l
    );
#endif

#if 0
public:
   void addTraceListener (
      MParser_TraceListener_ptr l
    );
#endif

    /**Get another token object from the token stream */
public:
   virtual void consume () = 0;

    /** Consume tokens until one matches the given token */
public:
   void consumeUntil (
      int tokenType
    );

    /** Consume tokens until one matches the given token set */
public:
   void consumeUntil (
      const MParser_BitSet& set
    );

protected:
   void defaultDebuggingSetup (
      const MParser_TokenStream_var& lexer,
      const MParser_TokenBuffer_var& tokBuf
    );

    /** Get the AST return value squirreled away in the parser */
public:
   virtual const MParser_AST_var& getAST();

public:
   virtual const MParser_ASTFactory_var& getASTFactory();

public:
   virtual MLANG_String getFilename() const;

public:
   virtual const MParser_ParserSharedInputState_var& getInputState() const;

public:
   virtual void setInputState (
      const MParser_ParserSharedInputState_var& state
    );

public:
   virtual MLANG_String getTokenName (int num) const;

public:
   virtual const ACE_Vector<MLANG_String>& getTokenNames() const;

public:
   virtual bool isDebugMode() const;

    /** Return the token type of the ith token of lookahead where i=1
     * is the current token being examined by the parser (i.e., it
     * has not been matched yet).
     */
public:
   virtual int LA(int i) = 0;

    /**Return the ith token of lookahead */
public:
   virtual MParser_Token_var LT(int i) = 0;

    // Forwarded to TokenBuffer
public:
   virtual int mark();

    /**Make sure current lookahead symbol matches token type <tt>t</tt>.
     * Throw an exception upon mismatch, which is catch by either the
     * error handler or by the syntactic predicate.
     */
public:
   virtual void match(int t);

    /**Make sure current lookahead symbol matches the given set
     * Throw an exception upon mismatch, which is catch by either the
     * error handler or by the syntactic predicate.
     */
public:
   virtual void match (
      const MParser_BitSet& b
    );

public:
   virtual void matchNot (
      int t
    );

    /** @deprecated as of 2.7.2. This method calls System.exit () and writes
     *  directly to stderr, which is usually not appropriate when
     *  a parser is embedded into a larger application. Since the method is
     *  <code>static</code>, it cannot be overridden to avoid these problems.
     *  ANTLR no longer uses this method internally or in generated code.
     */
public:
   static void panic();

#if 0
public:
   virtual void removeMessageListener (
      MParser_MessageListener_ptr l
    );
#endif

#if 0
public:
   virtual void removeParserListener (
      MParser_ParserListener_ptr l
    );
#endif

#if 0
public:
   virtual void removeParserMatchListener (
      MParser_ParserMatchListener_ptr l
    );
#endif

#if 0
public:
   virtual void removeParserTokenListener (
      MParser_ParserTokenListener_ptr l
    );
#endif

#if 0
public:
   virtual void removeSemanticPredicateListener (
      MParser_SemanticPredicateListener_ptr l
    );
#endif

#if 0
public:
   virtual void removeSyntacticPredicateListener (
      MParser_SyntacticPredicateListener_ptr l
    );
#endif

#if 0
public:
   virtual void removeTraceListener (
      MParser_TraceListener_ptr l
    );
#endif

    /** Parser error-reporting function can be overridden in subclass */
public:
   virtual void reportError (
      MParser_RecognitionException& ex
    );

    /** Parser error-reporting function can be overridden in subclass */
public:
   virtual void reportError (
      const MLANG_String& s
    );

    /** Parser warning-reporting function can be overridden in subclass */
public:
   virtual void reportWarning (
      const MLANG_String& s
    );

public:
   virtual void rewind (
      int pos
    );

    /** Specify an object with support code (shared by
     *  Parser and TreeParser.  Normally, the programmer
     *  does not play with this, using setASTNodeType instead.
     */
public:
   virtual void setASTFactory (
      const MParser_ASTFactory_var& f
    );

public:
   virtual void setASTNodeClass (
      const MLANG_String& cl
    );

    /** Specify the type of node to create during tree building; use setASTNodeClass now
     *  to be consistent with Token Object Type accessor.
	 *  @deprecated since 2.7.1
     */
public:
   virtual void setASTNodeType (
      const MLANG_String& nodeType
    );

public:
   virtual void setDebugMode (
      bool debugMode
    );

public:
   virtual void setFilename (
      const MLANG_String& f
    );

public:
   virtual void setIgnoreInvalidDebugCalls (
      bool value
    );

    /** Set or change the input token buffer */
public:
   virtual void setTokenBuffer (
      const MParser_TokenBuffer_var& t
    );

public:
   virtual void traceIndent();

public:
   virtual void traceIn (
      const MLANG_String& rname
    );

public:
   virtual void traceOut (
      const MLANG_String& rname
    );
};

typedef MLANG_SmartPtr<MParser_Parser> MParser_Parser_var;

#endif /* __MPARSER_PARSER_H__ */
