#include "MParser/TokenBuffer.h"

/* ANTLR Translator Generator
 * Project led by Terence Parr at http://www.jGuru.com
 * Software rights: http://www.antlr.org/license.html
 *
 * $Id: //depot/code/org.antlr/release/antlr-2.7.4/antlr/TokenBuffer.java#1 $
 */

/**A Stream of Token objects fed to the parser from a Tokenizer that can
 * be rewound via mark()/rewind() methods.
 * <p>
 * A dynamic array is used to buffer up all the input tokens.  Normally,
 * "k" tokens are stored in the buffer.  More tokens may be stored during
 * guess mode (testing syntactic predicate), or when LT(i>k) is referenced.
 * Consumption of tokens is deferred.  In other words, reading the next
 * token is not done by conume(), but deferred until needed by LA or LT.
 * <p>
 *
 * @see antlr.Token
 * @see antlr.Tokenizer
 * @see antlr.TokenQueue
 */


/** Create a token buffer */
MParser_TokenBuffer::MParser_TokenBuffer (
   const MParser_TokenStream_var& input
) : input_(input),
    nMarkers_(0),
    markerOffset_(0),
    numToConsume_(0)

{
}

/** Reset the input buffer to empty state */
void MParser_TokenBuffer::reset()

{
   nMarkers_ = 0;
   markerOffset_ = 0;
   numToConsume_ = 0;
   queue_.reset();
}

/** Mark another token for deferred consumption */
void MParser_TokenBuffer::consume()

{
   numToConsume_++;
}

/** Ensure that the token buffer is sufficiently full */
void MParser_TokenBuffer::fill(int amount)

{
   syncConsume();
   // Fill the buffer sufficiently to hold needed tokens
   while (queue_.nbrEntries() < (amount + markerOffset_)) 
    {
      // Append the next token
      MParser_Token_var t = input_->nextToken();
      queue_.append(t);
    }
}

/** Get a lookahead token value */
int MParser_TokenBuffer::LA(int i)

{
  assert(i>0);
  fill(i);
  assert(markerOffset_ + i - 1>=0 && markerOffset_ + i - 1<queue_.nbrEntries());
  assert(queue_.offset()>=0);

  if(MParser_Token::is_nil(queue_.elementAt(markerOffset_ + i - 1)))
   {
     fprintf(stderr, "MParser_TokenBuffer::LA: i=%d offset=%d markerOffset_=%d markerOffset_ + i - 1=%d\n",
             i, queue_.offset(), markerOffset_, markerOffset_ + i - 1);
     assert(!MParser_Token::is_nil(queue_.elementAt(markerOffset_ + i - 1)));
   }
  return queue_.elementAt(markerOffset_ + i - 1)->getType();
}

/** Get a lookahead token */
MParser_Token_var MParser_TokenBuffer::LT(int i)

{
  fill(i);
  return queue_.elementAt(markerOffset_ + i - 1);
}

/**Return an integer marker that can be used to rewind the buffer to
 * its current state.
 */
int MParser_TokenBuffer::mark()

{
   syncConsume();
//System.out.println("Marking at " + markerOffset);
//try { for (int i = 1; i <= 2; i++) { System.out.println("LA("+i+")=="+LT(i).getText()); } } catch (ScannerException e) {}
   nMarkers_++;
   return markerOffset_;
}

/**Rewind the token buffer to a marker.
 * @param mark Marker returned previously from mark()
 */
void MParser_TokenBuffer::rewind(int mark) 

{
   syncConsume();
   markerOffset_ = mark;
   nMarkers_--;
//System.out.println("Rewinding to " + mark);
//try { for (int i = 1; i <= 2; i++) { System.out.println("LA("+i+")=="+LT(i).getText()); } } catch (ScannerException e) {}
    }

/** Sync up deferred consumption */
void MParser_TokenBuffer::syncConsume() 

{
   while (numToConsume_ > 0)
    {
      if (nMarkers_ > 0)
       {
         // guess mode -- leave leading tokens and bump offset.
         markerOffset_++;
       }
      else 
       {
         // normal mode -- remove first token
         queue_.removeFirst();
       }
      numToConsume_--;
    }
}

MParser_TokenBuffer::~MParser_TokenBuffer ()

{
}

