#include <limits.h>
#include <errno.h>
#include <stdlib.h>
#include <memory.h>
#include <assert.h>
#include "mbase/debug.h"
#include "mbase/sqltypes.h"
#include "mbase/decimal.h"
#include "mbase/byutils.h"

static int packdgts (
   register const char* from,
   register ACE_DECIMAL_BASE_T* to,
   register int pos
)

{
   register int carry = 0;
   while (*from!='\0' || carry!=0)
    {
      if (*from!='\0')
       carry += *from++ - '0';
      // is odd pos?
      if ((pos+1)%ACE_DECIMAL_BASE_DIGITS==0)
       {
         *to++ = (ACE_DECIMAL_BASE_T) carry;
         carry = 0;
       }
      else
       carry *= 10;
      pos++;
    }
   return pos / ACE_DECIMAL_BASE_DIGITS;
}

static char *unpackdgts (
   register const ACE_DECIMAL_BASE_T* dgts,
   register int ndgts,
   register char* to
)

{
   while (--ndgts>=0)
    {
      register int i;
      register int carry = *dgts++;
      for (i=ACE_DECIMAL_BASE_DIGITS-1; i>=0; i--)
       {
        to[i] = carry % 10 + '0';
        carry /= 10;
       }
      to += ACE_DECIMAL_BASE_DIGITS;
    }
   *to = '\0';
   return to;
}

int mdecload (
   mdec_t* d,
   int pos, /* Signo */
   int exp, /* Exponente */
   ACE_DECIMAL_BASE_T* dgts, /* Digitos */
   int ndgts /* Numero de digitos */
)

{
   register ACE_DECIMAL_BASE_T* p; 
   if (pos==-1)
    {
      d->dec_pos = -1;
      return 0;
    }

   // Eliminate zeroes to right.
   for (p = &dgts[ndgts]; ndgts>0  && *--p==0; ndgts--);
   // Eliminate zeroes to left disminishing the exponent
   while (*dgts==0 && ndgts>0)
    {
      --exp;
      --ndgts;
      dgts++;
    }
   // Buffer in digits exceed max number of precision in decimal?
   if (ndgts>ACE_DECSIZE)
    {
      // Round to most significant digits in dgts[ACE_DECSIZE-1] 
      p = &dgts[ACE_DECSIZE];
      if (*p>=ACE_DECIMAL_BASE/2)
       {
         do
          {
            *p-- = 0;
            (*p)++;
          }
         while (*p>=ACE_DECIMAL_BASE);
       }
      // This possible case when 9999999... 9 (ACE_DECSIZE) n where n >= 50
      if (p<dgts)
       {
         ++exp;
         --dgts;
       }
      ndgts = ACE_DECSIZE;
    }
   d->dec_ndgts = ndgts;
   if (ndgts>0)
    {
      memcpy(&d->dec_dgts[0],dgts, ndgts*sizeof(ACE_DECIMAL_BASE_T));
    }
   else
    {
      exp = -63;
      pos = 1;
    }
   d->dec_pos = pos;
   d->dec_exp = exp;
   if (exp<-63)
    {
      d->dec_exp = -63;
      return  -1201;
    }
   if (exp>63)
    {
      d->dec_exp = 63;
      return  -1200;
    }
   return  0;
}

int dec2prec (
   const dec_t* d
)

{
   register int dec;
   register int tot;

   if (d->dec_ndgts==0 || d->dec_pos==-1)    
    return 0;

   dec = ACE_DECIMAL_BASE_DIGITS*(d->dec_ndgts - d->dec_exp);
   tot = ACE_DECIMAL_BASE_DIGITS*d->dec_ndgts;
   if (d->dec_dgts[d->dec_ndgts-1]%(ACE_DECIMAL_BASE/10)==0)
    {
      --dec;
      --tot;
    }
   if (d->dec_dgts[0]/(ACE_DECIMAL_BASE/10)==0)
    --tot;
   if (dec<0)
    {
      tot -= dec;
      dec = 0;
    }

#if 0
   if (tot<dec)
    tot = dec;
#endif

   return ACE_PRECMAKE(tot,dec);
}

int decprec (
   const dec_t* d
)

{
   int prec = dec2prec(d);
   int tot = ACE_PRECTOT(prec);
   int dec = ACE_PRECDEC(prec);
   if (tot<dec)
      tot = dec;

   return ACE_PRECMAKE(tot, dec);
}

int mdectoexp_r (
   const dec_t* d,
   char* p,
   int maxlen,
   int exp,
   int sign,
   char* buf,
   int bufsize
)

{
   register char* start;
   register char* from;
   int length_exp = sign + 4;
   register char* to;

   if (--exp<0)
    ++length_exp;
   if (exp<-(ACE_DECIMAL_BASE-1) || exp>(ACE_DECIMAL_BASE-1))
    ++length_exp;
   if (length_exp>=maxlen)
    return 1;

   start = dececvt_r(d,maxlen-length_exp,&exp,&sign,buf,bufsize);
   from = start + maxlen - length_exp;
   while (--from>start && *from=='0')
    *from = 0;

   from = start;
   to = p;
   if (sign!=0)
    *to++ = '-';
   *to++ = *from++;
   if (*from!='\0')
    *to++ = '.';
   while (*from!='\0')
    *to++ = *from++;
   *to++ = 'e';
   if (--exp<0)
    {
      *to++ = '-';
      exp = -exp;
    }
   if (exp>=100)
    {
      *to++ = (exp / 100) + '0';
      exp = exp % 100;
    }
   *to++ = exp / 10 + '0';
   *to++ = exp % 10 + '0';
   return 0;
}

int mdectoasc (
   const dec_t* d,
   char* buf,
   int maxlen,
   int right
)

{
   register char *p = buf;
   register char* temp;
   int exp;
   int sign;
   int len;
   char tempbuf[5*ACE_DECSIZE*sizeof(ACE_DECIMAL_BASE_T)+1];

   // Is null decimal?
   if (d->dec_pos==-1)
    {
      memset(p, ' ', maxlen);
      return 0;
    }
   if (right<0)
    {
      if (d->dec_ndgts==0)
       {
	 right = 1;
       }
      else
       {
	 right = ACE_DECIMAL_BASE_DIGITS*(d->dec_ndgts - d->dec_exp);
	 if (d->dec_ndgts>0
         &&  ((d->dec_dgts[d->dec_ndgts - 1]%(ACE_DECIMAL_BASE/10))==0))
	  right--;
         if (right<=0)
	  right = 1;
       }
    }
   memset(p, ' ', maxlen);
   for (;;)
    {
      temp = mdecfcvt_r(d, right, &exp, &sign, tempbuf, sizeof(tempbuf));
      len = sign + 1 + right + (right>0?1:0);
      if (exp>0)
       len += exp - 1;
      if (len<=maxlen)
       break;
      if (exp>-4 && exp>-right)
       {
         if (right>=(len - maxlen))
          {
            right -= len - maxlen;
	    continue;
          }
	 if (right==(len - 1 - maxlen) && right!=0)
          {
	    right = 0;
	    continue;
          }
       }
      if (mdectoexp_r(d,p,maxlen,exp,sign,tempbuf,sizeof(tempbuf))==0)
       return 0;
      memset(p, '*', maxlen);
      return -1;
    }
   if (sign!=0)
    *p++ = '-';
   if (exp>0)
    {
      memcpy(p, temp, exp);
      p += exp;
      temp += exp;
      exp = 0;
    }
   else
    *p++ = '0';
   if (right>0)
    *p++ = '.';
   // Copy decimal digits
   while (--right>=0)
    {
      *p++ = (exp++<0)?'0':*temp++;
    }
   return 0;
}

char* mdeccvt_r (
   const dec_t* d,
   int len,
   int* decpt,
   int* sign,
   int use_exp,
   char* buf,
   int maxlen
)

{
   register char* to = buf;
   char* next;
   int ndgts;
   register int i;

   if (len>=maxlen)
    len = maxlen - 1;
   if (len<0)
    len = 0;
   // Is null value?
   if (d->dec_pos==-1)
    {
      *sign = 0;
      *decpt = 0;
      memset(to, ' ', len);
      to[len] = '\0';
      return to;
    }
   // Find sign of number.
   *sign = d->dec_pos==0?1:0;
   // Is zero?
#if 0
   if (d->dec_ndgts<=0 && d->dec_exp==-64)
#else
   if (d->dec_ndgts<=0 || d->dec_exp==-64)
#endif
    {
      *decpt = 0;
    }
   else
    {
      *decpt = ACE_DECIMAL_BASE_DIGITS*d->dec_exp;
    }

   // Unpack the decimal number
   next = unpackdgts(&d->dec_dgts[0], d->dec_ndgts, to);
   // Number of digits is multiply by base digits
   ndgts = ACE_DECIMAL_BASE_DIGITS*d->dec_ndgts;
   // We erase the ACE_DECIMAL_BASE_DIGITS-1 leading zeros.
   for (i=0; i<ACE_DECIMAL_BASE_DIGITS-1 && *to=='0'; i++)
    {
      to++;
      ndgts--;
      (*decpt)--;
    }
   if (use_exp)
    len += *decpt;
   // We exceed the bufsize?
   if (len<0)
    {
      buf[0] = '\0';
      return buf;
    }
   // Buffer exceed the number of digits
   if (len>ndgts)
    {
      memset(next,'0',len - ndgts);
      to[len] = '\0';
      return to;
    }
   if (len>=ndgts)
    {
      to[len] = '\0';
      return to;
    }
   // Round the last digit
   next = &to[len];
   (*next) += 0x5;
   while (*next>'9')
    {
      *next = '0';
      if (next>to)
       {
         next--;
         (*next)++;
	 continue;
       }
      else
       {
         *next = '1';
         (*decpt)++;
         if (use_exp)
	  len++;
	 to[len] = '\0';
         return to;
        }
     }
   to[len] = '\0';
   return to;
}

char* mdececvt_r(
   const dec_t* d,
   int len,
   int* decpt,
   int* sign,
   char* buf,
   int maxlen
)

{
   return mdeccvt_r(d, len, decpt, sign, 0, buf, maxlen);
}

char* mdecfcvt_r(
   const dec_t* d,
   int right,
   int* decpt,
   int* sign,
   char* buf,
   int maxlen
)

{
   return mdeccvt_r(d, right, decpt, sign, 1, buf, maxlen);
}

int mdeccvlong (
   long l,
   dec_t* d
)

{
   register long n = l; // Only for speed
   register int pos = 0; // Suppose negative number
   ACE_DECIMAL_BASE_T dgts[ACE_DECIMAL_DIGITS(ACE_LONG_PREC)];
   register int i;
   
   if (n<0)
    n = -n;
   else
    pos++;

   // Number of digits depend of precision of long
   for (i = sizeof(dgts) - 1; i>=0; i--)
    {
      dgts[i] = (ACE_DECIMAL_BASE_T) (n % ACE_DECIMAL_BASE);
      n = n / ACE_DECIMAL_BASE;
    }
   
   return mdecload(d, pos, sizeof(dgts), &dgts[0], sizeof(dgts));
}

int deccvulong (
   unsigned long l,
   dec_t* d
)

{
   register unsigned long n = l; // Only for speed
   register int pos = 1; // Suppose positive number
   ACE_DECIMAL_BASE_T dgts[ACE_DECIMAL_DIGITS(ACE_ULONG_PREC)];
   register int i;
   
   // Number of digits depend of precision of long
   for (i = sizeof(dgts) - 1; i>=0; i--)
    {
      dgts[i] = (ACE_DECIMAL_BASE_T) (n % ACE_DECIMAL_BASE);
      n = n / ACE_DECIMAL_BASE;
    }
   
   return mdecload(d, pos, sizeof(dgts), &dgts[0], sizeof(dgts));
}

int mdivbyint (
   mdec_t* d,
   int i,
   int* r
)

{
   int ndgts;
   int msn;
   ACE_DECIMAL_BASE_T* p;
   ACE_DECIMAL_BASE_T* q;
   int result;

   result = 0;
   msn = 0;
   p = &d->dec_dgts[0];
   q = p;
   for (ndgts = d->dec_ndgts;
        ndgts > 0;
        ndgts--)
    {
       result = 100 * result + *q++;
       *p = result / i;
       if (*p!=0 || msn!=0)
        {
          p++;
          msn++;
        }
       result = result % i;
    }

   if ((ndgts = q - p)!=0)
    {
      d->dec_ndgts -= ndgts;
      d->dec_exp -= ndgts;
    }

   *r = result;
   return 0;
}

int dectolong (
   const dec_t* d,
   long* pl
)

{
   register long l;
   register int i;

   if (d->dec_pos==-1)
    {
      // Undefined value
      *pl = 0x80000000;
      return 0;
    }

   // Verify we not exceed maximun long precision
   if (d->dec_exp>ACE_DECIMAL_DIGITS(ACE_LONG_PREC))
    return  -1200;

   // Verify we not exceed maximun long precision
   if (d->dec_exp==ACE_DECIMAL_DIGITS(ACE_LONG_PREC))
    {
      dec_t temp;
      deccvlong(ACE_LONG_MAX, &temp);
      if (mdeccmp(&temp, d) < 0)
       return -1200;
      deccvlong(ACE_LONG_MIN, &temp);
      if (mdeccmp(d, &temp) < 0)
       return -1200;
    }

   l = 0; 
   for (i = 0; i<d->dec_exp; i++)
    {
       l *= ACE_DECIMAL_BASE;
       if (i<d->dec_ndgts)
        l += d->dec_dgts[i];
    }
   *pl = (d->dec_pos==0)?-l:l;
   return 0;
}

int dectoshort (
   const dec_t* d,
   short* pl
)

{
   register short l;
   register int i;

   if (d->dec_pos==-1)
    {
      // Undefined value
      rsetnull(CSHORTTYPE, pl);
      return 0;
    }

   // Verify we not exceed maximun long precision
   if (d->dec_exp>ACE_DECIMAL_DIGITS(ACE_SHORT_PREC))
    return  -1200;

   // Verify we not exceed maximun long precision
   if (d->dec_exp==ACE_DECIMAL_DIGITS(ACE_SHORT_PREC))
    {
      dec_t temp;
      deccvlong(ACE_SHRT_MAX, &temp);
      if (mdeccmp(&temp, d) < 0)
       return -1200;
      deccvlong(ACE_SHRT_MIN, &temp);
      if (mdeccmp(d, &temp) < 0)
       return -1200;
    }

   l = 0; 
   for (i = 0; i<d->dec_exp; i++)
    {
       l *= ACE_DECIMAL_BASE;
       if (i<d->dec_ndgts)
        l += d->dec_dgts[i];
    }
   *pl = (d->dec_pos==0)?-l:l;
   return 0;
}

int dectoulong (
   const dec_t* d,
   unsigned long* pl
)

{
   register unsigned long l;
   register int i;

   if (d->dec_pos==-1)
    {
      // Undefined value
      *pl = 0x80000000;
      return 0;
    }

   // Verify we not exceed maximun long precision
   if (d->dec_exp>ACE_DECIMAL_DIGITS(ACE_LONG_PREC))
    return  -1200;

   // Verify we not exceed maximun long precision
   if (d->dec_exp==ACE_DECIMAL_DIGITS(ACE_LONG_PREC))
    {
      dec_t temp;
      deccvulong(ACE_ULONG_MAX, &temp);
      if (mdeccmp(&temp, d) > 0)
       return -1200;
    }

   l = 0; 
   for (i = 0; i<d->dec_exp; i++)
    {
       l *= ACE_DECIMAL_BASE;
       if (i<d->dec_ndgts)
        l += d->dec_dgts[i];
    }
   *pl = l;
   return 0;
}

int mdectoushort (
   const dec_t* d,
   unsigned short* pl
)

{
   register unsigned short l;
   register int i;

   if (d->dec_pos==-1)
    {
      // Undefined value
      rsetnull(CSHORTTYPE, pl);
      return 0;
    }

   // Verify we not exceed maximun long precision
   if (d->dec_exp>ACE_DECIMAL_DIGITS(ACE_USHORT_PREC))
    return  -1200;

   // Verify we not exceed maximun long precision
   if (d->dec_exp==ACE_DECIMAL_DIGITS(ACE_USHORT_PREC))
    {
      dec_t temp;
      deccvulong(ACE_USHRT_MAX, &temp);
      if (mdeccmp(&temp, d) > 0)
       return -1200;
    }

   l = 0; 
   for (i = 0; i<d->dec_exp; i++)
    {
       l *= ACE_DECIMAL_BASE;
       if (i<d->dec_ndgts)
        l += d->dec_dgts[i];
    }
   *pl = l;
   return 0;
}

int mdectoint (
   const dec_t* d,
   int* i
)

{
   long l;
   int retcode = mdectolong (d, &l); 
   if (retcode!=0)
      return retcode;

   if (mrisnull(CLONGTYPE, &l))
      mrsetnull(CINTTYPE, i);
   else
      *i = l;
   return retcode;
}

// Compare a and b ignoring sign.
int mudeccmp (
   const dec_t* a,
   const dec_t* b
)

{
   register int result = 0;

   if (a->dec_pos==-1 || b->dec_pos==-1)
    return -2;
   if (a->dec_ndgts==0)
    result = (b->dec_ndgts==0)?0:-1;
   else if (b->dec_ndgts==0)
    result = 1;
   else if (a->dec_exp<b->dec_exp)
    result = -1;
   else if (a->dec_exp>b->dec_exp)
    result = 1;
   else
    {
      register int i;
      int a_has_less_digits = a->dec_ndgts<b->dec_ndgts;
      register int ndgts = a_has_less_digits?a->dec_ndgts:b->dec_ndgts;
      for (i=0; i<ndgts; i++)
       if ((result=(a->dec_dgts[i] - b->dec_dgts[i]))!=0)
          break;

      if (result==0 && a->dec_ndgts!=b->dec_ndgts)
       {
         if (ndgts<a->dec_ndgts)
          {
            while (ndgts<a->dec_ndgts
            &&     a->dec_dgts[ndgts]==0)
              ndgts++;
            if (ndgts<a->dec_ndgts)
               result = 1;
          }
         else
          {
            while (ndgts<b->dec_ndgts
            &&     b->dec_dgts[ndgts]==0)
              ndgts++;
            if (ndgts<b->dec_ndgts)
               result = -1;
          }
       }
    }
   return result;
}

int mdecadd (
   const mdec_t* a,
   const mdec_t* b,
   mdec_t* r
)

{
   int diff_exp; // Differences between exponent
   int lsd_in_b; // Starting end position in b to add
   int ndgts; // Number of digits in the result
   ACE_DECIMAL_BASE_T dgts[ACE_DECSIZE+2]; // Digits of the result
   register ACE_DECIMAL_BASE_T* to;
   register const ACE_DECIMAL_BASE_T* from;
   int carry;
   int op_sub;

   // We have a null decimal?
   if (a->dec_pos==-1 || b->dec_pos==-1)
    {
      r->dec_pos = -1;
      return 0;
    }
   // If a is less than b (ignoring sign), swap a and b.
   if (mudeccmp(a,b)<0)
    {
      const dec_t* temp = a;
      a = b;
      b = temp;
    }
  
   // Calculate the differences in exponent 
   diff_exp = a->dec_exp - b->dec_exp;
   // Calculate the expecting number of resulting digits
   ndgts = b->dec_ndgts + diff_exp;
   // If number of resulting digits is less than digits in a (exponent of
   // b is too low). Use number of a digits.
   if (ndgts<a->dec_ndgts)
    ndgts = a->dec_ndgts;
   // If number of resulting digits exceed number of digits in ACE_Decimal
   // trunk the remaining result digits.
   if (ndgts>ACE_DECSIZE+1)
    ndgts = ACE_DECSIZE+1;

   ++ndgts;
   // Initialize resulting digits to 0.
   memset(&dgts[0],0,ndgts*sizeof(ACE_DECIMAL_BASE_T));
   // Reserve the first digit in dgts when a + b increment number of signi-
   // ficants digits in the result.
   memcpy(&dgts[1],&a->dec_dgts[0],a->dec_ndgts*sizeof(ACE_DECIMAL_BASE_T));
   lsd_in_b = ndgts - diff_exp - 1;
   if (lsd_in_b>b->dec_ndgts)
    lsd_in_b = b->dec_ndgts;

   to = &dgts[lsd_in_b + diff_exp + 1];
   from = &b->dec_dgts[lsd_in_b];
   carry = 0;
   op_sub = a->dec_pos==b->dec_pos?0:1;
   while(--from>=b->dec_dgts)
    {
      register int result;
      --to;
      if (op_sub)
        result = carry + *to - *from;
      else
        result = carry + *to + *from;
      carry = 0;
      while (result>=ACE_DECIMAL_BASE)
       {
         carry++;
         result -= ACE_DECIMAL_BASE;
       }
      while (result<0)
       {
         carry--;
         result += ACE_DECIMAL_BASE;
       }
      *to = (char)result;
    }
   if (carry!=0)
    {
      --to;
      if (carry>0)
       {
         while (++(*to)>=ACE_DECIMAL_BASE) 
          *to-- = 0;
       }
      else
       {
         while ((*to)--==0)
          *to-- = ACE_DECIMAL_BASE - 1;
       }
    }
   return mdecload(r, a->dec_pos, a->dec_exp + 1,
                &dgts[0], ndgts);
}

int mdeccmp(const mdec_t* a,const mdec_t* b)

{
   register int result = 0;

   if (a->dec_pos==-1 || b->dec_pos==-1)
    return -2;
   if (a->dec_pos!=b->dec_pos)
    result = 1;
   else if (a->dec_ndgts==0)
    result = (b->dec_ndgts==0)?0:-1;
   else if (b->dec_ndgts==0)
    result = 1;
   else if (a->dec_exp<b->dec_exp)
    result = -1;
   else if (a->dec_exp>b->dec_exp)
    result = 1;
   else
    {
      register int i;
      int a_has_less_digits = a->dec_ndgts<b->dec_ndgts;
      register int ndgts = a_has_less_digits?a->dec_ndgts:b->dec_ndgts;
      for (i=0; i<ndgts; i++)
        if ((result=(a->dec_dgts[i] - b->dec_dgts[i]))!=0)
         break;

      if (result==0 && a->dec_ndgts!=b->dec_ndgts)
       {
         if (ndgts<a->dec_ndgts)
          {
            while (ndgts<a->dec_ndgts
            &&     a->dec_dgts[ndgts]==0) 
              ndgts++;
            if (ndgts<a->dec_ndgts)
               result = 1;
          }
         else 
          {
            while (ndgts<b->dec_ndgts
            &&     b->dec_dgts[ndgts]==0) 
              ndgts++;
            if (ndgts<b->dec_ndgts)
               result = -1;
          }
       }
    }
   return a->dec_pos!=0?result:-result;
}

void mdeccopy (
   const dec_t* a,
   mdec_t* b
)

{
   register int i;

   b->dec_pos = a->dec_pos;
   b->dec_exp = a->dec_exp;
   b->dec_ndgts = a->dec_ndgts;
   if (a->dec_pos!=-1)
   for (i=0; i<a->dec_ndgts; i++)
     b->dec_dgts[i] = a->dec_dgts[i];
}

int mdecsub (
   const mdec_t* a,
   const mdec_t* b,
   mdec_t* r
)

{
   mdec_t temp;

   if (a->dec_pos==-1 || b->dec_pos==-1)
    {
      r->dec_pos = -1;
      return  0;
    }

   mdeccopy(b, &temp);
   temp.dec_pos = (temp.dec_pos == 0?1:0);
   return mdecadd(a, &temp, r);
}

int mdecmul (
   const mdec_t* a,
   const mdec_t* b,
   mdec_t* r
)

{
   register int ndgts;
   ACE_DECIMAL_BASE_T dgts[2*ACE_DECSIZE];
   register int i;
   register int j;

   if (a->dec_pos==-1 || b->dec_pos==-1)
    {
      r->dec_pos = -1;
      return 0;
    }

   ndgts = a->dec_ndgts + b->dec_ndgts;

   // The digits of result
   memset(&dgts[0],0,ndgts*sizeof(ACE_DECIMAL_BASE_T));
   // Initialize to zero digits
   // Do multiply
   for (i = a->dec_ndgts - 1; i>=0; i--)
    {
      for (j = b->dec_ndgts - 1; j>=0; j--)
       {
          register int result = a->dec_dgts[i] * b->dec_dgts[j];
          register ACE_DECIMAL_BASE_T* p = &dgts[j + 1 + i];
          while (result)
           {
             result += *p;
             *p-- = (ACE_DECIMAL_BASE_T) (result % ACE_DECIMAL_BASE);
             result = result / ACE_DECIMAL_BASE;
           }
       }
    }
   // Normalize the number
   return mdecload(r, a->dec_pos==b->dec_pos?1:0,
                  a->dec_exp + b->dec_exp,
                  &dgts[0], ndgts);
}

int mdecdiv (
   const mdec_t* a,
   const mdec_t* b,
   mdec_t* r
)

{
   register ACE_DECIMAL_BASE_T* p;
   ACE_DECIMAL_BASE_T bdgts[ACE_DECSIZE+2];
   ACE_DECIMAL_BASE_T dgts[ACE_DECSIZE+2];
   ACE_DECIMAL_BASE_T adgts[2*ACE_DECSIZE+2];
   int dividend;
   int quotient;
   int divisor;
   int i;
   register int ndgts;

   if (a->dec_pos==-1 || b->dec_pos==-1)
    {
      r->dec_pos = -1;
      r->dec_ndgts = 0;
      return 0;
    }

   if (b->dec_ndgts==0)
    {
      deccvlong(0, r);
      return -1202;
    }

   memset(&adgts[0], 0, sizeof(adgts));
   bdgts[0] = 0;
   memcpy(&adgts[1], &a->dec_dgts[0], a->dec_ndgts*sizeof(ACE_DECIMAL_BASE_T));
   memcpy(&bdgts[1], &b->dec_dgts[0], b->dec_ndgts*sizeof(ACE_DECIMAL_BASE_T));
   divisor = b->dec_dgts[0] * ACE_DECIMAL_BASE;
   if (b->dec_ndgts>1)
    divisor += b->dec_dgts[1];

   for (ndgts=0; ndgts<ACE_DECSIZE+2; ndgts++)
    {
       register int carry;

       dividend = adgts[ndgts] * ACE_DECIMAL_BASE + adgts[ndgts + 1];
       dividend = dividend * ACE_DECIMAL_BASE + adgts[ndgts + 2];
       quotient = dividend / divisor;
       dgts[ndgts] = quotient;
       if (quotient==0)
        continue;

       carry = 0;
       p = &adgts[ndgts + b->dec_ndgts];
       for (i = b->dec_ndgts; i>=0; i--)
        {
           carry += *p + (ACE_DECIMAL_BASE*ACE_DECIMAL_BASE) - quotient * bdgts[i];
           *p-- = carry % ACE_DECIMAL_BASE;
           carry = carry / ACE_DECIMAL_BASE;
           carry -= ACE_DECIMAL_BASE;
        }
       if (carry<0)
        {
          dgts[ndgts]--;
          carry = 0;
          p = &adgts[ndgts + b->dec_ndgts];
          for (i = b->dec_ndgts; i>=0; i--)
           {
             carry += *p + bdgts[i]; 
             *p-- = carry % ACE_DECIMAL_BASE;
             carry = carry / ACE_DECIMAL_BASE;
           }
        }
    }
   return mdecload(r, a->dec_pos==b->dec_pos?1:0,
                 a->dec_exp - b->dec_exp + 1,
                 &dgts[0], ndgts); 
}

int mdeclpow (
   const mdec_t* x,
   long n,
   mdec_t* r
)

{
   int sign;
   int err;
   mdec_t t;

   if (x->dec_pos==-1 || risnull(CLONGTYPE, &n))
    {
      r->dec_pos = -1;
      r->dec_ndgts = 0;
      return 0;
    }

   sign = 0;
   if (n<0)
    {
      sign = 1;
      n = - n;
    }
   
   mdeccvint (1, r);
   mdeccvint (1, &t);
   while (n>0)
    {
       if ((err = mdecmul (&t, x, &t)))
        {
          return err;
        }
       /* Es n un numero impar?
        */
       if (n&1
       &&  (err = mdecmul (r, &t, r)))
        {
          break;
        }
       n >>= 1;
    }
   /* Si el exponente era negativo, calcula el inverso
    */
   if (sign)
    {
      mdeccvint (1, &t);
      return mdecdiv (&t, r, r);
    }
   return 0;
}

int mdecpow (
   const mdec_t* x,
   const mdec_t* y,
   mdec_t* r
)

{
   if (x->dec_pos==-1 || y->dec_pos==-1)
    {
      r->dec_pos = -1;
      r->dec_ndgts = 0;
      return 0;
    }
   if (y->dec_exp!=y->dec_ndgts)
    {
      mdecln (x, r);
      mdecmul (y, r, r);
      return mdecexp (r, r);
    }
   else
    {
      int sign;
      int err;
      mdec_t n;
      mdec_t zero;
      mdec_t t;
      mdec_t s;
      char temp [2*ACE_DECSIZE + 1 + 1 + 1];

      sign = 0;
      /* Si n es negativo, entonces es 1 / x ** -n
       */
      mdeccopy (y, &n);
      if (n.dec_pos==0)
       {
         sign = 1;
         n.dec_pos = 1;
       }
      /* Le quita inicialmente las cifras decimales
       */
      mdectrunc (&n, 0); 
      mdeccvint (1, &s);
      mdeccopy (x, &t);
      mdeccvint (0, &zero);
      while (mdeccmp(&n,&zero)>0)
       {
          int res;
          /* n por 2 y obtiene el residuo en res
           */
          mdivbyint (&n, 2, &res);
          /* Era n un numero impar?
           */
          if (res
          &&  (err = mdecmul (&s, &t, &s)))
           {
             return err;
           }
          if ((err = mdecmul (&t, &t, &t)))
           {
             return err;
           }
       }
      mdectoasc (&n, temp, sizeof(temp), 0);
      DPRINTF(("mdecpow: temp=%s\n", temp));
      /* Si el exponente era negativo, calcula el inverso
       */
      mdeccopy (&s, r);
      if (sign)
       {
         mdeccvint (1, &t);
         return mdecdiv (&t, r, r);
       }
      return 0;
    }
}

int mdeccvasc (
   const char* from,
   int len,
   mdec_t* d
)

{
   register const char* p;
   register const char* end;
   int pos; // Is positive
   char dgts[ACE_DECIMAL_BASE_DIGITS*ACE_DECSIZE+1];
   register char* to;
   register int exp;
   register int i;

   if (len == -1) len = strlen(from);

   p = from; 
   end = &from[len];
   
   // Skip leading blanks
   while (p<end && *p==' ') p++;
   if (p==end)
    {
      d->dec_pos = -1;
      return 0;
    }
   pos = *p;
   if (*p=='-' || *p=='+')
    p++;

   if (p==end)
    return -1213;

   if ((*p<'0' || *p>'9') && *p!='.')
    return -1213;

   // Is an hexadecimal number?
   if (*p=='0' && (*p=='X' || *p=='x'))
    {
      register long l = 0;
      for (p+=2;p<end;p++)
       {
          register int d = *p;
          if (d>='0' && d<='9')
           d = d - '0';
          else if (d>='A' && d<='Z')
           d = d - 'A' + 10;
          else if (d>='a' && d<='z')
           d = d - 'a' + 10;
          else
           break;
          l = 16*l + d;
       }
      while (p<end && *p==' ') p++;
      if (pos=='-')
       l = -l;
      deccvlong (l, d);
      return  0;
    }

   to = &dgts[0];
   exp = 0;
   i = 0;
   for (;p<end && ((*p>='0' && *p<='9') || *p=='.' || *p==','); p++)
    {
      if (*p=='.')
       {
         if (i++)
          return -1213;
       }
      else if (*p==',')
       {
         if (i)
          return -1213;
       }
      else if (*p=='0' && to==&dgts[0])
       {
         if (i)
          --exp;
       }
      else
       {
         if (to<&dgts[2*ACE_DECSIZE])
          *to++ = *p;
         if (i==0)
          exp++;
       }
    }
   *to = '\0';
   // Decimal Number in exponential form?
   if (p<end && (*p=='e' || *p=='E'))
    {
      register int pos;

      ++p;
      pos = 0;
      // Skip sign if necessary
      if (p<end && (*p=='-' || *p=='+'))
         pos = *p++;
      // Calculate the exponent
      for (i=0; p<end && *p>='0' && *p<='9'; p++)
       i = i * 10 + *p++ - '0';
      // Exponent is negative?
      if (pos=='-')
       i = -i;
      // Increment exponent
      exp += i;
    }
   // Skip trailing blanks
   while (p<end && *p==' ') p++;
   // Set the decimal number to base ACE_DECIMAL_BASE.
   d->dec_ndgts = packdgts(&dgts[0],
			      &d->dec_dgts[0],
			      exp&1);
   return mdecload(d, pos=='-'?0:1, 
		  ((exp + 1)&~1)/2,
                  &d->dec_dgts[0],d->dec_ndgts);
}

int mdectrunc (
   mdec_t* d,
   int prec
)

{
   int ndgts; /* esi */

   if (d->dec_pos==-1)
      return 0;

   ndgts = ((prec + 1) & ~1) / 2 + d->dec_exp;
   if (ndgts>d->dec_ndgts)
      return -1;

   if (ndgts>0
   &&  (prec&1)
   &&  (d->dec_dgts[ndgts - 1] = (d->dec_dgts[ndgts - 1] / 10) * 10)==0)
      --ndgts;

   if (ndgts<=0)
    {
      d->dec_ndgts = 0;
      d->dec_exp = 0;
    }
   else
    {
      d->dec_ndgts = ndgts;
    }

   return 0;
}

int mdecround (
   mdec_t* d,
   int prec
)

{
   mdec_t t;

   if (d->dec_pos==-1)
      return 0;

   if (prec>=2*(d->dec_ndgts - d->dec_exp))
      return -1;

   t.dec_pos = d->dec_pos;
   t.dec_ndgts = 1;
   t.dec_dgts[0] = (prec&0x1) ? 5 : 50;
   t.dec_exp = - (prec & ~1) / 2;
   mdecadd(d, &t, d);
   return mdectrunc(d, prec);
}

int mdeccvint (
   int i,
   mdec_t *d
)

{
   if (i==(short)0x8000)
    {
      d->dec_ndgts = 0;
      d->dec_exp = 0;
      d->dec_pos = -1;
      return 0;
    }
   return deccvlong(i, d);
}

int mdeccvfd (
   double dbl,
   mdec_t* d,
   short ndigit
)

{
   int decpt;
   int sign;

   const char* from = ecvt(dbl, ndigit, &decpt, &sign);
   d->dec_ndgts = packdgts(from, &d->dec_dgts[0], decpt&1);
   return mdecload(d,
           (sign==0) ? 1 : 0,
           ((decpt + 1) & ~1) / 2,
           &d->dec_dgts[0],
           d->dec_ndgts);
}

int mdeccvflt (
   double dbl,
   mdec_t* d
)

{
   return mdeccvfd(dbl, d, 8);
}

int mdeccvdbl (
   double dbl,
   mdec_t* d
)

{
   return mdeccvfd(dbl, d, 16);
}


int mdectodbl (
   const dec_t* d,
   double* dbl   
)

{
   char tempbuf[0x28];
   int ndgts;
   char* to;
   int exp;
   const ACE_DECIMAL_BASE_T* dgts;

   to = &tempbuf[0];
   if (d->dec_pos==0)
      *to++ = '-';

   exp = d->dec_exp;
   dgts = &d->dec_dgts[0];
   ndgts = d->dec_ndgts;
   if (exp<=0)
    {
      *to++ = '0';
      *to++ = '.';
    }
   else if (exp<ndgts)
    {
      to = unpackdgts(dgts, exp, to);
      *to++ = '.';
      dgts += exp;
      ndgts -= exp;
      exp = 0;
    }
   else
    {
      exp -= ndgts;
    }

   to = unpackdgts(dgts, ndgts, to);
   if (exp!=0)
    {
      exp = 2 * exp;
      *to++ = 'E';
      if (exp<0)
       {
         exp = - exp;
        *to++ = '-';
       }

      *to++ = (exp / 10) + '0';
      *to++ = (exp % 10) + '0';
    }

   *to++ = '\0';
   errno = 0;
   *dbl = atof(&tempbuf[0]);
   if (errno!=0)
      return 0xfffffb50;
   else
      return 0;
}

void deccomplement (
   ACE_DECIMAL_BASE_T* dgts,
   int ndgts
)

{
   int accum;
   ACE_DECIMAL_BASE_T* to;

   for (accum = 100, to = &dgts[ndgts] - 1; to>=dgts; to--)
    {
       if (*to!=0 || accum!=100)
        {
          *to = accum - *to;
          accum = 99;
        }
    }
}

void stdecimal (
   const dec_t* d,
   ACE_DECIMAL_BASE_T* dgts,
   int ndgts
)

{
   if (d->dec_pos==-1)
    {
      byfill(dgts, ndgts, 0);
      return;
    }

   *dgts++ = (d->dec_exp + 64)&0x7f;
   --ndgts;
   bycopy(&d->dec_dgts, dgts, ndgts);
   if (ndgts>d->dec_ndgts)
    {
      byfill(dgts + d->dec_ndgts, ndgts - d->dec_ndgts, 0);
    }
   else if (ndgts<d->dec_ndgts && d->dec_dgts[ndgts]>=50)
    {
      ACE_DECIMAL_BASE_T* to;

      to = &dgts[ndgts];
      for (;;)
       {
         to--;
         *to = *to + 1;
         if (*to<100)
            break;

         if (to==dgts)
          {
            *to = 1;
            dgts[-1]++;
            break;
          }
         *to = 0;
       }
    }

   if (d->dec_pos==0)
    {
      deccomplement(dgts, ndgts);
      dgts[-1] ^= 0x7f;
    }
   else
    {
      dgts[-1] |= 0x80;
    }
}

int mlddecimal (
   const ACE_DECIMAL_BASE_T* from,
   int len,
   mdec_t* d
)

{
   ACE_DECIMAL_BASE_T dgts[ACE_DECSIZE + 1];
   int exp;

   if (from[0]==0 && from[1]==0)
    {
      d->dec_ndgts = 0;
      d->dec_exp = 0;
      d->dec_pos = -1;
      return 0;
    }
   --len;
   bycopy(from+1, &dgts[0], len);
   exp = (unsigned char)*from;
   if (!(exp&0x80))
    {
      deccomplement(&dgts[0], len);
      exp ^= 0x7f;
    }
   exp = (exp&0x7f) - 64;
   return mdecload(d, (*from&0x80) ? 1 : 0, exp, &dgts[0], len);
}

int rdefmtdec (
   const char* s,
   int len,
   const char* sqlformat,
   dec_t* d
)

{
   return mdeccvasc (s, len, d);
}

#define EPS "0.000000000000000000001"

/* Calcula el binomial a de x.
 */

int mdecbinomial (
   const mdec_t* x,
   const mdec_t* a,
   mdec_t* r
)

{
   register int n;
   mdec_t t;
   mdec_t eps;

   mdeccvint (1, &t);
   mdeccvint (0, r);
   mdeccvasc (EPS, strlen(EPS), &eps);
   for (n=1;n<100;n++)
    {
       mdec_t err;
       mdec_t t1, t2;
       char temp[2*ACE_DECSIZE+1+1+1];

       mdeccopy (r, &err);
       mdecadd (r, &t, r);
       mdeccopy (a, &t1);
       mdeccvint (n - 1, &t2);
       mdecsub (&t1, &t2, &t1);
       mdecmul (&t, &t1, &t);
       mdecmul (&t, x, &t);
       mdeccvint (n, &t1);
       mdecdiv (&t, &t1, &t);
       mdecsub (&err, r, &err);
       err.dec_pos = 1;
       if (mdeccmp(&err, &eps)<=0)
        {
           break;
        }
       mdectoasc (r, temp, sizeof(temp), -1);
#if 0
       fprintf(stderr, "err=%s\n", temp);
#endif
    }
   return 0;
}

int mdecsqrt (
   const mdec_t* x0,
   mdec_t* r
)

{
#if 0
   mdec_t a, y;
   mdeccvasc ("0.5", strlen("0.5"), &a);
   mdeccvint (-1, &y);
   mdecadd (x0, &y, &y);
   return mdecbinomial (&y, &a, r);
#else
   mdec_t _2;
   mdec_t x;
   mdec_t eps;

   mdeccvint (2, &_2);
   mdecdiv (x0, &_2, &x);
   mdeccvasc (EPS, strlen(EPS), &eps);
   for (;;)
    {
       mdec_t t, temp;

       mdecmul (&x, &x, &t);
       mdecsub (&t, x0, &t);
       mdecdiv (&t, &_2, &t);
       mdecdiv (&t, &x, &t);
       mdecsub (&x, &t, &t);       
       mdecsub (&x, &t, &temp);
       temp.dec_pos = 1;
       if (mdeccmp (&temp, &eps)<=0)
        {
          break;
        }
       mdeccopy (&t, &x);
    }
   mdeccopy (&x, r);
   return 0;
#endif
}

int mdecexp (
   const mdec_t* x0,
   mdec_t* r
)

{
#if 0
   register int n;
   mdec_t t;
   mdec_t eps;
   mdec_t tx;

   n = 0;
   mdeccopy (x0, &x);
   mdeccvint (1, &t);
   mdeccvint (0, r);
   mdeccvasc (EPS, strlen(EPS), &eps);
   while (1)
    {
       mdec_t err;
       mdec_t temp;

       mdeccopy (r, &err);
       mdecadd (r, &t, r);
       mdecmul (&t, &x, &t);
       n++;
       mdeccvint (n, &temp);
       mdecdiv (&t, &temp, &t);
       mdecsub (&err, r, &err);
       err.dec_pos = 1;
       if (mdeccmp(&err, &eps)<=0)
        {
           break;
        }
    }
   return 0;
#else
/* Uses the fact that e^x = (e^(x/2))^2
   When x is small enough, we use the series:
     e^x = 1 + x + x^2/2! + x^3/3! + ...
*/

  mdec_t x;
  int sign = 0;
  int f = 0;
  int i;
  mdec_t v;
  mdec_t e;
  mdec_t eps;
  mdec_t a;
  mdec_t d;
  mdec_t _1;

  mdeccvint (1, &_1);
  /* a - holds x^y of x^y/y! */
  /* d - holds y! */
  /* e - is the value x^y/y! */
  /* v - is the sum of the e's */
  /* f - number of times x was divided by 2. */
  /* m - is 1 if x was minus. */
  /* i - iteration count. */
  /* n - the scale to compute the sum. */
  /* z - orignal scale. */

  /* Check the sign of x. */
  deccopy (x0, &x);
  if (x.dec_pos==1)
   {
     sign = 1;
     x.dec_pos = 0;
   } 

  /* Precondition x. */
  while (mdeccmp (&x, &_1) > 0)
   {
     mdec_t _2;
     f++;
     mdeccvint (2, &_2);
     mdecdiv (&x, &_2, &x);
   }

  /* Initialize the variables. */
  mdeccvint (1, &v);
  mdecadd (&v, &x, &v);
  mdeccopy (&x, &a);
  mdeccvint (1, &d);
  mdeccvasc (EPS, strlen(EPS), &eps);
  for (i=2; ; i++)
   {
      mdec_t temp;
      mdecmul (&a, &x, &a);
      mdeccvint (i, &temp);
      mdecmul (&d, &temp, &d);
      mdecdiv (&a, &d, &e);
      mdeccopy (&e, &temp);
      temp.dec_pos = 1;
      if (mdeccmp(&temp,&eps) <= 0) {
        if (f>0)
         {
           while (f--) 
            {
              mdecmul (&v, &v, &v);
            }
         }
        if (sign)
         {
           mdeccvint (1, r);
           mdecdiv (r, &v, r);
           return 0;
         }
        mdeccopy (&v, r);
        return 0;
      }
      mdecadd (&v, &e, &v);
   }
#endif
}

int mdecln (
   const mdec_t* x0,
   mdec_t* r
)

{
#if 0
   register int n;
   mdec_t t;
   mdec_t eps;
   mdec_t y;

   mdeccvint (1, &y);
   mdecsub (&y, x0, &y);
   mdeccopy (&y, &t);
   mdeccopy (&y, r);
   mdeccvasc (EPS, strlen(EPS), &eps);
   for (n=1;n<200;n++)
    {
       mdec_t err;
       mdec_t temp;

       mdeccopy (r, &err);
       mdecmul (&t, &y, &t);
       n++;
       mdeccvint (n, &temp);
       mdecdiv (&t, &temp, &temp);
       mdecadd (r, &temp, r);
       mdecsub (&err, r, &err);
       err.dec_pos = 1;
       if (mdeccmp(&err, &eps)<=0)
        {
           break;
        }
    }

   if (r->dec_pos==0)
      r->dec_pos = 1;
   else
      r->dec_pos = 0;
   return 0;
#else
  mdec_t e;
  mdec_t _2;
  mdec_t __5;
  mdec_t f;
  int i;
  mdec_t m, n;
  mdec_t x;
  mdec_t eps;
  mdec_t v;

  /* return something for the special case. */
  mdeccopy (x0, &x);
  if (x.dec_pos==0)
   {
     return -1;
   }

  /* Precondition x to make .5 < x < 2.0. */
  mdeccvint (2, &f);
  i=0;
  mdeccvint (2, &_2);
  while (mdeccmp (&x, &_2) >= 0)
   {  /* for large numbers */
     mdecmul (&f, &_2, &f);
     mdecsqrt (&x, &x);
   }
  mdeccvasc ("0.5", strlen("0.5"), &__5);
  while (mdeccmp (&x, &__5) <= 0)
   {  /* for small numbers */
     mdecmul (&f, &_2, &f);
     mdecsqrt (&x, &x);
   }

  /* Set up the loop. */
  mdeccvint (-1, &n);
  mdecadd (&n, &x, &n);
  mdeccvint (1, &v);
  mdecadd (&v, &x, &v);
  mdecdiv (&n, &v, &n);
  mdeccopy (&n, &v);
  mdecmul (&n, &n, &m);

  /* Sum the series. */
  mdeccvasc (EPS, strlen(EPS), &eps);
  for (i=3; ; i+=2)
   {
      mdec_t temp;
      mdecmul (&n, &m, &n);
      mdeccvint (i, &temp);
      mdecdiv (&n, &temp, &e);
      mdeccopy (&e, &temp);
      if (mdeccmp(&temp, &eps)<=0)
       {
         mdecmul (&v, &f, &v);
         mdeccopy (&v, r);
         return 0;
       }
      mdecadd (&v, &e, &v);
   }
#endif
}
