#include "mbase/debug.h"
#include "mbase/sqltypes.h"
#include "mbase/byutils.h"
#include "mbase/rdate.h"
#include "mbase/rmoney.h"
#include "mbase/datetime.h"
#include "mbase/decimal.h"
#include "mbase/locator.h"
#include "mbase/vstring.h"

static char fltnull[sizeof(float)] = {-1, -1, -1, -1};
static char dblnull[sizeof(double)] = {-1, -1, -1, -1, -1, -1, -1, -1};

static short toctypes[]=
{
  CCHARTYPE,
  CSHORTTYPE,
  CLONGTYPE,
  CDOUBLETYPE,
  CFLOATTYPE,
  CDECIMALTYPE,
  CLONGTYPE,
  CDATETYPE,
  CMONEYTYPE,
  -1,
  CDTIMETYPE,
  CLOCATORTYPE,
  CLOCATORTYPE,
  CVCHARTYPE,
  CINVTYPE,
  CVSTRINGTYPE
};

static short tosqltypes[]=
{
  SQLCHAR,
  SQLSMINT,
  SQLSMINT,
  SQLINT,
  SQLSMFLOAT,
  SQLFLOAT,
  -1,
  SQLDECIMAL,
  SQLCHAR,
  SQLCHAR,
  SQLDATE,
  SQLMONEY,
  SQLDTIME,
  SQLBYTES,
  SQLVCHAR,
  SQLINTERVAL,
  SQLBYTES,
  SQLVSTRING,
  -1,
  -1,
  -1,
  -1
};

int mrtosqltype (
   int ctype
)

{
   if (ctype==CINTTYPE && sizeof(int)==sizeof(long))
      ctype = CLONGTYPE;

   if (ctype>=CCHARTYPE && ctype<=CVSTRINGTYPE)
    {
      return tosqltypes[ctype - CCHARTYPE];
    }
   else if (ISSQLTYPE(ctype))
    {
      return ctype;
    }
   else
    {
      return -1;
    }
}

int mrtoctype (
   int sqltype
)

{
   if (sqltype==CINTTYPE && sizeof(int)==sizeof(long))
    sqltype = CLONGTYPE;

   if (ISSQLTYPE(sqltype))
    {
      return toctypes[sqltype];
    }
   else
    {
      return sqltype;
    }
}


int mrisnull (
   int ctype,
   const void * sqldata
)

{
   if (ctype==CINTTYPE && sizeof(int)==sizeof(long))
      ctype = CLONGTYPE;

   if (ISSQLTYPE(ctype))
      ctype = toctypes[ctype];

   switch(ctype)
    {
    case CSHORTTYPE:
         if (*((const short*) sqldata)==(short) 0x8000)
            return 1;
         break;

    case CINTTYPE:
         if (*((const int*)sqldata)==(short) 0x8000)
            return 1;
         break;

    case CLONGTYPE:
         if (*((const long*)sqldata)==(long)0x80000000)
            return 1;
         break;

    case CDATETYPE:
         if (*((const mjdate*)sqldata)==(long)0x80000000)
            return 1;
         break;

    case CFLOATTYPE:
         if (bycmpr(sqldata, &fltnull[0], sizeof(fltnull))==0)
            return 1;
         break;

    case CDOUBLETYPE:
         if (bycmpr(sqldata, &dblnull[0], sizeof(dblnull))==0)
            return 1;
         break;

    case CCHARTYPE:
    case CFIXCHARTYPE:
    case CSTRINGTYPE:
    case CVCHARTYPE:
    case CFILETYPE:
         if (sqldata==0 || *((const char*)sqldata)=='\0')
            return 1;
         break;

    case CDTIMETYPE:
         if (((const dtime_t*)sqldata)->dt_dec.dec_pos==-1)
            return 1;
         break;

    case CINVTYPE:
         if (((const intrvl_t*)sqldata)->in_dec.dec_pos==-1)
            return 1;
         break;

    case CDECIMALTYPE:
    case CMONEYTYPE:
         if (((const dec_t*)sqldata)->dec_pos==-1)
            return 1;
         break;

   case CLOCATORTYPE:
        if (((const mbase_loc_t*)sqldata)->loc_indicator<0)
          return 1;
        break;

    case CVSTRINGTYPE:
         if (sqldata==0 || ((mvstring)sqldata)->len==0)
            return 1;
         break;

    default:
         assert(0);
    }

   return 0;
}

int mrsetnull (
   int ctype,
   void* sqldata
)

{
   if (ctype==CINTTYPE && sizeof(int)==sizeof(long))
      ctype = CLONGTYPE;

   if (ISSQLTYPE(ctype))
     ctype = toctypes[ctype];

   switch(ctype)
    {
    case CSHORTTYPE:
         *((short*) sqldata) = (short)0x8000;
         break;

    case CINTTYPE:
         *((int*) sqldata) = (short)0x8000;
         break;

    case CLONGTYPE:
         *((long*) sqldata) = 0x80000000;
         break;

    case CDATETYPE:
         *((mjdate*) sqldata) = 0x80000000;
         break;

    case CFLOATTYPE: 
         byfill(sqldata, sizeof(float), -1);
         break;

    case CDOUBLETYPE:
         byfill(sqldata, sizeof(double), -1);
         break;

    case CCHARTYPE:
    case CFIXCHARTYPE:
    case CSTRINGTYPE:
    case CVCHARTYPE:
    case CFILETYPE:
         *((char*)sqldata) = '\0';
         break;

    case CDTIMETYPE:
         ((dtime_t*)sqldata)->dt_dec.dec_pos = -1;
         ((dtime_t*)sqldata)->dt_dec.dec_exp = 0;
         ((dtime_t*)sqldata)->dt_dec.dec_ndgts = 0;
         break;

    case CINVTYPE:
         ((intrvl_t*)sqldata)->in_dec.dec_pos = -1;
         ((intrvl_t*)sqldata)->in_dec.dec_exp = 0;
         ((intrvl_t*)sqldata)->in_dec.dec_ndgts = 0;
         break;

    case CDECIMALTYPE:
    case CMONEYTYPE:
         ((dec_t*)sqldata)->dec_pos = -1;
         ((dec_t*)sqldata)->dec_exp = 0;
         ((dec_t*)sqldata)->dec_ndgts = 0;
         break;

    case CLOCATORTYPE:
         ((mbase_loc_t*)sqldata)->loc_indicator = -1;
         break;

    case CVSTRINGTYPE:
         mvstring_setnull((mvstring)sqldata);
         break;

    default:
         assert(0);
    }

   return 0;
}

/* Estructura para el alineamiento de un short
 */
struct short_align_s
{
   char c;
   short s;
};

/* Estructura para el alineamiento de un int
 */
struct int_align_s
{
   char c;
   int i;
};

/* Estructura para el alineamiento de un long
 */
struct long_align_s
{
   char c;
   long l;
};

/* Estructura para el alineamiento de un date
 */
struct date_align_s
{
   char c;
   mjdate jdate;
};

/* Estructura para el alineamiento de un dec
 */
struct dec_align_s
{
   char c;
   dec_t d;
};

/* Estructura para el alineamiento de un float
 */
struct float_align_s
{
   char c;
   float f;
};

/* Estructura para el alineamiento de un double
 */
struct double_align_s
{
   char c;
   double d;
};

/* Estructura para el alineamiento de un dtime
 */
struct dtime_align_s
{
   char c;
   dtime_t dt;
};

/* Estructura para el alineamiento de un interval
 */
struct intrvl_align_s
{
   char c;
   intrvl_t in;
};

/* Estructura para el alineamiento de un mlocator
 */
struct mloc_align_s
{
   char c;
   mbase_loc_t loc;
};

/* Estructura para el alineamiento de un mlocator
 */
struct mvstring_align_s
{
   char c;
   struct mvstring_s vs;
};

int mrtypalign (
   int pos,
   int sqltype
)

{
   register int alignment = 1;
   switch(sqltype&SQLTYPE)
    {
    case CCHARTYPE:
    case CFIXCHARTYPE:
    case CSTRINGTYPE:
         break;

    case CSHORTTYPE:
         alignment = offsetof(struct short_align_s, s);
         break;

    case CINTTYPE:
         alignment = offsetof(struct int_align_s, i);
         break;

    case CLONGTYPE:
         alignment = offsetof(struct long_align_s, l);
         break;

    case CFLOATTYPE:
         alignment = offsetof(struct float_align_s, f);
         break;

    case CDOUBLETYPE:
         alignment = offsetof(struct double_align_s, d);
         break;

    case CDECIMALTYPE:
         alignment = offsetof(struct dec_align_s, d);
         break;

    case CDATETYPE:
         alignment = offsetof(struct date_align_s, jdate);
         break;

    case CMONEYTYPE:
         alignment = offsetof(struct dec_align_s, d);
         break;

    case CDTIMETYPE:
         alignment = offsetof(struct dtime_align_s, dt);
         break;

    case CLOCATORTYPE:
         alignment = offsetof(struct mloc_align_s, loc);
         break;

    case CINVTYPE:
         alignment = offsetof(struct intrvl_align_s, in);
         break;

    case CVSTRINGTYPE:
         alignment = offsetof(struct mvstring_align_s, vs);
         break;

    case SQLCHAR:
    case SQLVCHAR:
         break;

    case SQLSMINT:
         alignment = offsetof(struct short_align_s, s);
         break;

    case SQLINT:
         alignment = offsetof(struct long_align_s, l);
         break;

    case SQLFLOAT:
         alignment = offsetof(struct double_align_s, d);
         break;

    case SQLSMFLOAT:
         alignment = offsetof(struct float_align_s, f);
         break;

    case SQLDECIMAL:
         alignment = offsetof(struct dec_align_s, d);
         break;

    case SQLSERIAL:
         alignment = offsetof(struct long_align_s, l);
         break;

    case SQLDATE:
         alignment = offsetof(struct date_align_s, jdate);
         break;

    case SQLMONEY:
         alignment = offsetof(struct dec_align_s, d);
         break;

    case SQLDTIME:
         alignment = offsetof(struct dtime_align_s, dt);
         break;

    case SQLBYTES:
    case SQLTEXT:
         alignment = sizeof(short);
         break;

    case SQLINTERVAL:
         alignment = offsetof(struct intrvl_align_s, in);
         break;

    case SQLVSTRING:
         alignment = offsetof(struct mvstring_align_s, vs);
         break;

    default:
         EPRINTF(("rtypalign: %d\n", sqltype));
         assert(0);
    }
   alignment--;
   pos = (pos + alignment) & ~alignment;
   return pos;
}

static int datdigits ()

{
   int ndigits = 10;
   const char* p;

   if ((p=getenv("DBDATE"))==0)
      return ndigits;

   while (*p!='\0')
    {
      if (*p=='y' || *p=='Y')
       {
         ndigits = (p[1] == '2') ? 8 : 10;
         return ndigits;
       }
      p++;
    }

   return ndigits;
}

int mrtypwidth (
   int sqltype,
   int sqlprec
)

{
   register int width;

   switch(sqltype&SQLTYPE)
    {
    case SQLCHAR:
    case SQLBYTES:
    case SQLTEXT:
         width = sqlprec;
         break;

    case SQLVCHAR:
         width = sqlprec&0xff;
         break;

    case SQLSMINT:
         width = 6;
         break;

    case SQLINTERVAL:
         width = ACE_TU_LEN(sqlprec)
             + ((((ACE_TU_END(sqlprec)>ACE_TU_SECOND) ? ACE_TU_FRAC : ACE_TU_END(sqlprec)) - ACE_TU_START(sqlprec)) >> 1)
             + (ACE_TU_START(sqlprec)==ACE_TU_FRAC ? 1 : 0) + 1;
         break;

    case SQLDTIME:
         width = ACE_TU_LEN(sqlprec);
         width += (((ACE_TU_END(sqlprec)>ACE_TU_SECOND) ? ACE_TU_FRAC : ACE_TU_END(sqlprec)) - ACE_TU_START(sqlprec))>>1;
         width += (ACE_TU_END(sqlprec)==ACE_TU_FRAC) ? 1 : 0;
         break;

    case SQLDATE:
         width = datdigits();
         break;

    case SQLINT:
    case SQLSERIAL:
         width = 11;
         break;

    case SQLFLOAT:
    case SQLSMFLOAT:
         width = 14;
         break;

    case SQLMONEY:
         width = ACE_PRECTOT(sqlprec);
         width += mrfrontlen();
         width += mrbacklen();
         width += 2;
         if (ACE_PRECTOT(sqlprec)==ACE_PRECDEC(sqlprec))
            ++width;
         break;

    case SQLDECIMAL:
         width = ACE_PRECTOT(sqlprec);
         width += 1+1;
         /* El punto mas el signo
          */
         if (ACE_PRECTOT(sqlprec)==ACE_PRECDEC(sqlprec))
            ++width;
         break;

   default:
         width = 0;
         break;
    }

   return width;
}

int mrtypmsize (
   int sqltype,
   int sqlprec
)

{
   register int msize;

   switch(sqltype&SQLTYPE)
    {
    case SQLCHAR:
         msize = sqlprec + 1;
         break;

    case SQLVCHAR:
    case CVCHARTYPE:
         msize = (sqlprec&0xff) + 1;
         break;

    case SQLSMINT:
         msize = SIZSMINT;
         break;

    case CSHORTTYPE:
         msize = sizeof(short);
         break;

    case SQLINT:
    case SQLSERIAL:
         msize = SIZINT;
         break;

    case CINTTYPE:
         msize = sizeof(int);
         break;

    case CLONGTYPE:
         msize = sizeof(long);
         break;

    case SQLDATE:
         msize = SIZDATE;
         break;

    case CDATETYPE:
         msize = sizeof(mjdate);
         break;

    case SQLSMFLOAT:
    case CFLOATTYPE:
         msize = SIZSMFLOAT;
         break;

    case SQLFLOAT:
    case CDOUBLETYPE:
         msize = SIZFLOAT;
         break;

    case SQLDECIMAL:
    case SQLMONEY:
    case CDECIMALTYPE:
    case CMONEYTYPE:
         msize = sizeof(dec_t);
         break;

    case SQLDTIME:
    case CDTIMETYPE:
         msize = sizeof(dtime_t);
         break;

    case SQLINTERVAL:
    case CINVTYPE:
         msize = sizeof(intrvl_t);
         break;

    case CCHARTYPE:
    case CSTRINGTYPE:
    case CFIXCHARTYPE:
         msize = sqlprec;
         break;

    case SQLVSTRING:
    case CVSTRINGTYPE:
         msize = sizeof(struct mvstring_s);
         break;

    default:
         msize = 0;
         break;
    }

   return msize;
}

int mrtyplength (
   int sqltype,
   int sqlprec
)

{
   if (ISDECTYPE(sqltype))
      return ACE_DECLENGTH(sqlprec);
   else
      return sqlprec;
}

